// DRGNURBSSurface.h: interface for the CDRGNURBSSurface class.
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
// PURPOSE:
//    
// Declaration of the CDRGNURBSSurface class for rendering NURBS surfaces.  
// Accompanies the article "Rendering NURBS Surfaces in Real-Time".  Please refer
// to the article for an understanding of the methods in this class.
// ------------------------------------------------------------------------------------
//
// Author: Dean Macri - Intel Developer Relations Divison -- Tools and Technology Group
//
////////////////////////////////////////////////////////////////////////////////////////


#if !defined(AFX_DRGNURBSSURFACE_H__0D6D594B_6F3B_11D3_BE36_00902752C5DF__INCLUDED_)
#define AFX_DRGNURBSSURFACE_H__0D6D594B_6F3B_11D3_BE36_00902752C5DF__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define SUPPORT_PENTIUM_III	1

#if SUPPORT_PENTIUM_III
#define SIMD_SIZE 4		// Number of floats per SIMD register -- Just for code
						// clarity.  Changing this will probably cause bugs-a-plenty
#define CPUID	_asm _emit 0x0f _asm _emit 0xA2 // We'll need this to check for a Pentium III
#else
#define SIMD_SIZE 1
#endif

//
// macros for allocating and freeing memory aligned on 16-byte boundaries using the 
// _mm_malloc and _mm_free routines provided with the Intel compiler.  If you're
// not going to use Pentium III optimizations, just change these to
//
// #define ALIGNED_NEW( s, t ) new t[s];
// #define ALIGNED_DELETE( x ) do { if( (x) ) { delete x; x = NULL; } } while (0)

#if SUPPORT_PENTIUM_III
#define ALIGNED_NEW( s, t ) (t *)_mm_malloc( (s) * sizeof( t ), 16 )
#define ALIGNED_DELETE( x ) do { if( (x) ) { _mm_free( (char *)x ); x = NULL; } } while (0)
#else
#define ALIGNED_NEW( s, t ) (t *)malloc( (s) * sizeof( t ) )
#define ALIGNED_DELETE( x ) do { if( (x) ) { free( (char *)x ); x = NULL; } } while (0)
#endif

//
// Some handy-dandy macros
//
#define CLAMP_UPPER(x) if(x>1.0f) { x = 1.0f; }
#define CLAMP_LOWER(x) if(x<0.0f) { x = 0.0f; }

struct Point4D
{
	float x;
	float y;
	float z;
	float w;
};

class CDRGNURBSSurface : public CDRGObject  
{
private:
	Point4D *m_pControlPoints;
	float *m_UBasisCoefficients;
	float *m_VBasisCoefficients;
	float *m_UKnots;
	float *m_VKnots;
	float *m_UBasis, *m_dUBasis;
	float *m_VBasis, *m_dVBasis;
	Point4D *m_UTemp, *m_dUTemp;
	int *m_TessUKnotSpan;
	int *m_TessVKnotSpan;

	int m_iUDegree, m_iVDegree;
	int m_iUOrder, m_iVOrder;
	int m_iUKnots, m_iVKnots;
	int m_iUControlPoints, m_iVControlPoints;

	int m_iUBasisSpans, m_iVBasisSpans;

	int m_iUTessellations, m_iVTessellations;

	bool m_bUseSSE, m_bIsPentiumIII;

	D3DVERTEX *m_pVertices;
	WORD *m_iVertexIndices[2];

	void Cleanup( void );
	float ComputeCoefficient( float *fKnots, int iInterval, int i, int p, int k );
	void ComputeBasisCoefficients( void );
	void EvaluateBasisFunctions( void );

public:
	CDRGNURBSSurface();
	virtual ~CDRGNURBSSurface();

	HRESULT Init(	CDRGWrapper *pDRGWrapper,
					int	uDegree,
					int	vDegree,
					int uControlPoints,
					int vControlPoints,
					Point4D *pControlPoints,
					float *pUKnots,
					float *pVKnots,
					int iDefaultUTessellations = 10,
					int iDefaultVTessellations = 10
				);
				
	void UpdateControlPoints( Point4D *pControlPoints );

	virtual bool SetUseSSE( bool bUseSSE );
	virtual void SetTessellations( int iUTessellations, int iVTessellations );
	virtual void TessellateSurface();
#if SUPPORT_PENTIUM_III
	virtual void TessellateSurfaceSSE();
#endif
	virtual HRESULT Render(DRGMatrix& MxWorld, bool bRenderChildren = true );
	virtual int GetTriangleCount();

	bool IsPentiumIII( void );
};

#endif // !defined(AFX_DRGNURBSSURFACE_H__0D6D594B_6F3B_11D3_BE36_00902752C5DF__INCLUDED_)
