/*
	The Universe Development Kit
	Copyright (C) 2000  Sean O'Neil
	soneil@home.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __Matrix_h__
#define __Matrix_h__

#include "Noise.h"	// Has some useful defines and inline functions

class CVector;
class CMatrix;
class CQuaternion;
class C3DBase;
class C3DObject;

/*******************************************************************************
* Class: CVector
********************************************************************************
* This class implements a simple 3D vector with x, y, and z coordinates. Several
* functions and operators are defined to make working with vectors easier.
*******************************************************************************/
class CVector
{
public:
	float x, y, z;

public:
	// Constructors
	CVector()							{}
	CVector(float f)					{ *this = f; }
	CVector(float *p)					{ *this = p; }
	CVector(CVector &v)					{ *this = v; }
	CVector(float a, float b, float c)	{ x = a; y = b; z = c; }

	// Casting, unary, equal, and comparison operators
	operator float *()					{ return &x; }
	float &operator[](int n)			{ return (&x)[n]; }
	CVector operator-()					{ return CVector(-x, -y, -z); }
	void operator=(float f)				{ x = y = z = f; }
	void operator=(float *p)			{ x = p[0]; y = p[1]; z = p[2]; }
	void operator=(CVector &v)			{ x = v.x; y = v.y; z = v.z; }
	bool operator==(CVector &v)			{ return (Abs(x - v.x) <= DELTA && Abs(y - v.y) <= DELTA && Abs(z - v.z) <= DELTA); }
	bool operator!=(CVector &v)			{ return !(*this == v); }

	// Arithmetic operators (vector with scalar)
	CVector operator+(float f)			{ return CVector(x+f, y+f, z+f); }
	CVector operator-(float f)			{ return CVector(x-f, y-f, z-f); }
	CVector operator*(float f)			{ return CVector(x*f, y*f, z*f); }
	CVector operator/(float f)			{ f = 1/f; return CVector(x*f, y*f, z*f); }
	void operator+=(float f)			{ x += f; y += f; z += f; }
	void operator-=(float f)			{ x -= f; y -= f; z -= f; }
	void operator*=(float f)			{ x *= f; y *= f; z *= f; }
	void operator/=(float f)			{ f = 1/f; x *= f; y *= f; z *= f; }

	// Arithmetic operators (vector with vector)
	CVector operator+(CVector &v)		{ return CVector(x+v.x, y+v.y, z+v.z); }
	CVector operator-(CVector &v)		{ return CVector(x-v.x, y-v.y, z-v.z); }
	CVector operator*(CVector &v)		{ return CVector(x*v.x, y*v.y, z*v.z); }
	CVector operator/(CVector &v)		{ return CVector(x/v.x, y/v.y, z/v.z); }
	void operator+=(CVector &v)			{ x += v.x; y += v.y; z += v.z; }
	void operator-=(CVector &v)			{ x -= v.x; y -= v.y; z -= v.z; }
	void operator*=(CVector &v)			{ x *= v.x; y *= v.y; z *= v.z; }
	void operator/=(CVector &v)			{ x /= v.x; y /= v.y; z /= v.z; }

	// Dot and cross product operators
	float operator|(CVector &v)			{ return x*v.x + y*v.y + z*v.z; }
	CVector operator^(CVector &v)		{ return CVector(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x); }
	void operator^=(CVector &v)			{ *this = *this ^ v; }

	// Magnitude/distance methods
	float MagnitudeSquared()			{ return x*x + y*y + z*z; }
	float Magnitude()					{ return sqrtf(MagnitudeSquared()); }
	void Normalize()					{ *this /= Magnitude(); }
	float DistanceSquared(CVector &v)	{ return (*this - v).MagnitudeSquared(); }
	float Distance(CVector &v)			{ return (*this - v).Magnitude(); }
	CVector Midpoint(CVector &v)		{ return CVector((*this - v) * 0.5f + v); }
	CVector Average(CVector &v)			{ return CVector((*this + v) * 0.5f); }
	float Angle(CVector &v)				{ return acosf(*this | v); }

	// Advanced methods
	CVector Reflect(CVector &n)
	{
		float f = Magnitude();
		CVector v = *this / f;
		return (v - n * (2.0f * (v | n))) * f;
	}
	CVector Rotate(float fAngle, CVector &n)
	{
		float fCos = cosf(fAngle);
		float fSin = sinf(fAngle);
		return CVector(*this * fCos + ((n * *this) * (1.0f - fCos)) * n + (*this ^ n) * fSin);
	}
};

// Returns the normal vector of two vectors (the normalized cross product)
inline CVector NormalVector(CVector &v1, CVector &v2)
{
	CVector v = v1 ^ v2;
	v.Normalize();
	return v;
}

// Returns the normal vector of a triangle or 3 points on a plane (assumes counter-clockwise orientation)
inline CVector NormalVector(CVector &p1, CVector &p2, CVector &p3)
{
	return NormalVector(p2-p1, p3-p1);
}

// Returns the direction vector between two points
inline CVector DirectionVector(CVector &p1, CVector &p2)
{
	CVector v = p2 - p1;
	v.Normalize();
	return v;
}


/*******************************************************************************
* Class: CQuaternion
********************************************************************************
* This class implements a 4D quaternion. Several functions and operators are
* defined to make working with quaternions easier. Quaternions are often used to
* represent rotations, and have a number of advantages over other constructs.
* Their main disadvantage is that they are unintuitive.
*******************************************************************************/
class CQuaternion : public CVector
{
public:
	float w;

public:
	// Constructors
	CQuaternion()							{}
	CQuaternion(float a, float b, float c, float d)	{ x = a; y = b; z = c; w = d; }
	CQuaternion(CVector &v, float f)		{ SetAxisAngle(v, f); }
	CQuaternion(CVector &v)					{ *this = v; }
	CQuaternion(CQuaternion &q)				{ *this = q; }
	CQuaternion(CMatrix &m)					{ *this = m; }
	CQuaternion(float *p)					{ *this = p; }

	// Casting and unary operators
	operator float*()						{ return &x; }
	float &operator[](int n)				{ return (&x)[n]; }
	CQuaternion operator-()					{ return CQuaternion(-x, -y, -z, -w); }

	// Equal and comparison operators
	void operator=(CVector &v)				{ x = v.x; y = v.y; z = v.z; w = 0; }
	void operator=(CQuaternion &q)			{ x = q.x; y = q.y; z = q.z; w = q.w; }
	void operator=(CMatrix &m);
	void operator=(float *p)				{ x = p[0]; y = p[1]; z = p[2]; w = p[3]; }
	bool operator==(CQuaternion &q)			{ return (Abs(x - q.x) <= DELTA && Abs(y - q.y) <= DELTA && Abs(z - q.z) <= DELTA && Abs(w - q.w) <= DELTA); }
	bool operator!=(CQuaternion &q)			{ return !(*this == q); }

	// Arithmetic operators (quaternion and scalar)
	CQuaternion operator+(float f)			{ return CQuaternion(x+f, y+f, z+f, w+f); }
	CQuaternion operator-(float f)			{ return CQuaternion(x-f, y-f, z-f, w-f); }
	CQuaternion operator*(float f)			{ return CQuaternion(x*f, y*f, z*f, w*f); }
	CQuaternion operator/(float f)			{ f=1/f; return CQuaternion(x*f, y*f, z*f, w*f); }
	void operator+=(float f)				{ x+=f; y+=f; z+=f; w+=f; }
	void operator-=(float f)				{ x-=f; y-=f; z-=f; w-=f; }
	void operator*=(float f)				{ x*=f; y*=f; z*=f; w*=f; }
	void operator/=(float f)				{ f=1/f; x*=f; y*=f; z*=f; w*=f; }

	// Arithmetic operators (quaternion and quaternion)
	CQuaternion operator+(CQuaternion &q)	{ return CQuaternion(x+q.x, y+q.y, z+q.z, w+q.w); }
	CQuaternion operator-(CQuaternion &q)	{ return CQuaternion(x-q.x, y-q.y, z-q.z, w-q.w); }
	CQuaternion operator*(CQuaternion &q);
	void operator+=(CQuaternion &q)			{ x+=q.x; y+=q.y; z+=q.z; w+=q.w; }
	void operator-=(CQuaternion &q)			{ x-=q.x; y-=q.y; z-=q.z; w-=q.w; }
	void operator*=(CQuaternion &q)			{ *this = *this * q; }

	float MagnitudeSquared()				{ return x*x + y*y + z*z + w*w; }
	float Magnitude()						{ return sqrtf(MagnitudeSquared()); }
	void Normalize()						{ *this /= Magnitude(); }
	CQuaternion Conjugate()					{ return CQuaternion(-x, -y, -z, w); }
	CQuaternion Inverse()					{ return Conjugate() / MagnitudeSquared(); }
	CQuaternion UnitInverse()				{ return Conjugate(); }
	CVector RotateVector(CVector &v)		{ return (*this * CQuaternion(v) * UnitInverse()); }

	void SetAxisAngle(CVector &vAxis, float fAngle)
	{
		// 4 muls, 2 trig function calls
		fAngle *= 0.5f;
		*this = vAxis * sinf(fAngle);
		w = cosf(fAngle);
	}
	void GetAxisAngle(CVector &vAxis, float &fAngle)
	{
		// 4 muls, 1 div, 2 trig function calls
		fAngle = acosf(w);
		vAxis = *this / sinf(fAngle);
		fAngle *= 2.0f;
	}

	void Rotate(CQuaternion &q)				{ *this = q * *this; }
	void Rotate(CVector &vAxis, float fAngle)
	{
		CQuaternion q;
		q.SetAxisAngle(vAxis, fAngle);
		Rotate(q);
	}

	CVector GetViewAxis()
	{
		// 6 muls, 7 adds
		float x2 = x + x, y2 = y + y, z2 = z + z;
		float xx = x * x2, xz = x * z2;
		float yy = y * y2, yz = y * z2;
		float wx = w * x2, wy = w * y2;
		return -CVector(xz+wy, yz-wx, 1-(xx+yy));
	}
	CVector GetUpAxis()
	{
		// 6 muls, 7 adds
		float x2 = x + x, y2 = y + y, z2 = z + z;
		float xx = x * x2, xy = x * y2;
		float yz = y * z2, zz = z * z2;
		float wx = w * x2, wz = w * z2;
		return CVector(xy-wz, 1-(xx+zz), yz+wx);
	}
	CVector GetRightAxis()
	{
		// 6 muls, 7 adds
		float x2 = x + x, y2 = y + y, z2 = z + z;
		float xy = x * y2, xz = x * z2;
		float yy = y * y2, zz = z * z2;
		float wy = w * y2, wz = w * z2;
		return CVector(1-(yy+zz), xy+wz, xz-wy);
	}
};

extern CQuaternion Slerp(CQuaternion &q1, CQuaternion &q2, float t);

/*******************************************************************************
* Class: CMatrix
********************************************************************************
* This class implements a 4x4 matrix. Several functions and operators are
* defined to make working with matrices easier. The values are kept in column-
* major order to make it easier to use with OpenGL. For performance reasons,
* most of the functions assume that all matrices are orthogonal, which means the
* bottom row is [ 0 0 0 1 ]. Since I plan to use the GL_PROJECTION matrix to
* handle the projection matrix, I should never need to use any other kind of
* matrix, and I get a decent performance boost by ignoring the bottom row.
*******************************************************************************/
class CMatrix
{
public:
	// This class uses column-major order, as used by OpenGL
	// Here are the ways in which the matrix values can be accessed:
	// | f11 f21 f31 f41 |   | f1[0] f1[4] f1[8]  f1[12] |   | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |
	// | f12 f22 f32 f42 |   | f1[1] f1[5] f1[9]  f1[13] |   | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |
	// | f13 f23 f33 f43 | = | f1[2] f1[6] f1[10] f1[14] | = | f2[0][2] f2[1][2] f2[2][2] f2[3][2] |
	// | f14 f24 f34 f44 |   | f1[3] f1[7] f1[11] f1[15] |   | f2[0][3] f2[1][3] f2[2][3] f2[3][3] |
	union
	{
		struct { float f11, f12, f13, f14, f21, f22, f23, f24, f31, f32, f33, f34, f41, f42, f43, f44; };
		float f1[16];
		float f2[4][4];
	};

	CMatrix()						{}
	CMatrix(float f)				{ *this = f; }
	CMatrix(float *pf)				{ *this = pf; }
	CMatrix(CQuaternion &q)			{ *this = q; }

	// Init functions
	void ZeroMatrix()
	{
		f11 = f12 = f13 = f14 = f21 = f22 = f23 = f24 = f31 = f32 = f33 = f34 = f41 = f42 = f43 = f44 = 0;
	}
	void IdentityMatrix()
	{
		f12 = f13 = f14 = f21 = f23 = f24 = f31 = f32 = f34 = f41 = f42 = f43 = 0;
		f11 = f22 = f33 = f44 = 1;
	}

	operator float*()				{ return f1; }
	float &operator[](int n)		{ return f1[n]; }
	float &operator()(int i, int j)	{ return f2[i][j]; }

	void operator=(float f)			{ for(register int i=0; i<16; i++) f1[i] = f; }
	void operator=(float *pf)		{ for(register int i=0; i<16; i++) f1[i] = pf[i]; }
	void operator=(CQuaternion &q);

	CMatrix operator*(CMatrix &m);
	void operator*=(CMatrix &m)		{ *this = *this * m; }
	CVector operator*(CVector &v)	{ return TransformVector(v); }

	CVector TransformVector(CVector &v)
	{
		// 9 muls, 9 adds
		// | f11 f21 f31 f41 |   | v.x |   | f11*v.x+f21*v.y+f31*v.z+f41 |
		// | f12 f22 f32 f42 |   | v.y |   | f12*v.x+f22*v.y+f32*v.z+f42 |
		// | f13 f23 f33 f43 | * | v.z | = | f13*v.x+f23*v.y+f33*v.z+f43 |
		// | 0   0   0   1   |   | 1   |   | 1                           |
		return CVector((f11*v.x+f21*v.y+f31*v.z+f41),
					   (f12*v.x+f22*v.y+f32*v.z+f42),
					   (f13*v.x+f23*v.y+f33*v.z+f43));
	}
	CVector TransformNormal(CVector &v)
	{
		// 9 muls, 6 adds
		// | f11 f21 f31 f41 |   | v.x |   | f11*v.x+f21*v.y+f31*v.z |
		// | f12 f22 f32 f42 |   | v.y |   | f12*v.x+f22*v.y+f32*v.z |
		// | f13 f23 f33 f43 | * | v.z | = | f13*v.x+f23*v.y+f33*v.z |
		// | 0   0   0   1   |   | 1   |   | 1                       |
		return CVector((f11*v.x+f21*v.y+f31*v.z),
					   (f12*v.x+f22*v.y+f32*v.z),
					   (f13*v.x+f23*v.y+f33*v.z));
	}

	// Translate functions
	void TranslateMatrix(float x, float y, float z)
	{
		// | 1  0  0  x |
		// | 0  1  0  y |
		// | 0  0  1  z |
		// | 0  0  0  1 |
		f12 = f13 = f14 = f21 = f23 = f24 = f31 = f32 = f34 = 0;
		f11 = f22 = f33 = f44 = 1;
		f41 = x; f42 = y; f43 = z;
	}
	void TranslateMatrix(float *pf)		{ TranslateMatrix(pf[0], pf[1], pf[2]); }
	void Translate(float x, float y, float z)
	{
		// 9 muls, 9 adds
		// | f11 f21 f31 f41 |   | 1  0  0  x |   | f11 f21 f31 f11*x+f21*y+f31*z+f41 |
		// | f12 f22 f32 f42 |   | 0  1  0  y |   | f12 f22 f32 f12*x+f22*y+f32*z+f42 |
		// | f13 f23 f33 f43 | * | 0  0  1  z | = | f13 f23 f33 f13*x+f23*y+f33*z+f43 |
		// | 0   0   0   1   |   | 0  0  0  1 |   | 0   0   0   1                     |
		f41 = f11*x+f21*y+f31*z+f41;
		f42 = f12*x+f22*y+f32*z+f42;
		f43 = f13*x+f23*y+f33*z+f43;
	}
	void Translate(float *pf)			{ Translate(pf[0], pf[1], pf[2]); }

	// Scale functions
	void ScaleMatrix(float x, float y, float z)
	{
		// | x  0  0  0 |
		// | 0  y  0  0 |
		// | 0  0  z  0 |
		// | 0  0  0  1 |
		f12 = f13 = f14 = f21 = f23 = f24 = f31 = f32 = f34 = f41 = f42 = f43 = 0;
		f11 = x; f22 = y; f33 = z; f44 = 1;
	}
	void ScaleMatrix(float *pf)			{ ScaleMatrix(pf[0], pf[1], pf[2]); }
	void Scale(float x, float y, float z)
	{
		// 9 muls
		// | f11 f21 f31 f41 |   | x  0  0  0 |   | f11*x f21*y f31*z f41 |
		// | f12 f22 f32 f42 |   | 0  y  0  0 |   | f12*x f22*y f32*z f42 |
		// | f13 f23 f33 f43 | * | 0  0  z  0 | = | f13*x f23*y f33*z f43 |
		// | 0   0   0   1   |   | 0  0  0  1 |   | 0     0     0     1   |
		f11 *= x; f21 *= y; f31 *= z;
		f12 *= x; f22 *= y; f32 *= z;
		f13 *= x; f23 *= y; f33 *= z;
	}
	void Scale(float *pf)				{ Scale(pf[0], pf[1], pf[2]); }

	// Rotate functions
	void RotateXMatrix(float fRadians)
	{
		// | 1 0    0     0 |
		// | 0 fCos -fSin 0 |
		// | 0 fSin fCos  0 |
		// | 0 0    0     1 |
		f12 = f13 = f14 = f21 = f24 = f31 = f34 = f41 = f42 = f43 = 0;
		f11 = f44 = 1;

		float fCos = cosf(fRadians);
		float fSin = sinf(fRadians);
		f22 = f33 = fCos;
		f23 = fSin;
		f32 = -fSin;
	}
	void RotateX(float fRadians)
	{
		// 12 muls, 6 adds, 2 trig function calls
		// | f11 f21 f31 f41 |   | 1 0    0     0 |   | f11 f21*fCos+f31*fSin f31*fCos-f21*fSin f41 |
		// | f12 f22 f32 f42 |   | 0 fCos -fSin 0 |   | f12 f22*fCos+f32*fSin f32*fCos-f22*fSin f42 |
		// | f13 f23 f33 f43 | * | 0 fSin fCos  0 | = | f13 f23*fCos+f33*fSin f33*fCos-f23*fSin f43 |
		// | 0   0   0   1   |   | 0 0    0     1 |   | 0   0                 0                 1   |
		float fTemp, fCos, fSin;
		fCos = cosf(fRadians);
		fSin = sinf(fRadians);
		fTemp = f21*fCos+f31*fSin;
		f31 = f31*fCos-f21*fSin;
		f21 = fTemp;
		fTemp = f22*fCos+f32*fSin;
		f32 = f32*fCos-f22*fSin;
		f22 = fTemp;
		fTemp = f23*fCos+f33*fSin;
		f33 = f33*fCos-f23*fSin;
		f23 = fTemp;
	}
	void RotateYMatrix(float fRadians)
	{
		// | fCos  0 fSin  0 |
		// | 0     1 0     0 |
		// | -fSin 0 fCos  0 |
		// | 0     0 0     1 |
		f12 = f14 = f21 = f23 = f24 = f32 = f34 = f41 = f42 = f43 = 0;
		f22 = f44 = 1;

		float fCos = cosf(fRadians);
		float fSin = sinf(fRadians);
		f11 = f33 = fCos;
		f13 = -fSin;
		f31 = fSin;
	}
	void RotateY(float fRadians)
	{
		// 12 muls, 6 adds, 2 trig function calls
		// | f11 f21 f31 f41 |   | fCos  0 fSin  0 |   | f11*fCos-f31*fSin f21 f11*fSin+f31*fCos f41 |
		// | f12 f22 f32 f42 |   | 0     1 0     0 |   | f12*fCos-f32*fSin f22 f12*fSin+f32*fCos f42 |
		// | f13 f23 f33 f43 | * | -fSin 0 fCos  0 | = | f13*fCos-f33*fSin f23 f13*fSin+f33*fCos f43 |
		// | 0   0   0   1   |   | 0     0 0     1 |   | 0                 0   0                 1   |
		float fTemp, fCos, fSin;
		fCos = cosf(fRadians);
		fSin = sinf(fRadians);
		fTemp = f11*fCos-f31*fSin;
		f31 = f11*fSin+f31*fCos;
		f11 = fTemp;
		fTemp = f12*fCos-f32*fSin;
		f32 = f12*fSin+f32*fCos;
		f12 = fTemp;
		fTemp = f13*fCos-f33*fSin;
		f33 = f13*fSin+f33*fCos;
		f13 = fTemp;
	}
	void RotateZMatrix(float fRadians)
	{
		// | fCos -fSin 0 0 |
		// | fSin fCos  0 0 |
		// | 0    0     1 0 |
		// | 0    0     0 1 |
		f13 = f14 = f23 = f24 = f31 = f32 = f34 = f41 = f42 = f43 = 0;
		f33 = f44 = 1;

		float fCos = cosf(fRadians);
		float fSin = sinf(fRadians);
		f11 = f22 = fCos;
		f12 = fSin;
		f21 = -fSin;
	}
	void RotateZ(float fRadians)
	{
		// 12 muls, 6 adds, 2 trig function calls
		// | f11 f21 f31 f41 |   | fCos -fSin 0 0 |   | f11*fCos+f21*fSin f21*fCos-f11*fSin f31 f41 |
		// | f12 f22 f32 f42 |   | fSin fCos  0 0 |   | f12*fCos+f22*fSin f22*fCos-f12*fSin f32 f42 |
		// | f13 f23 f33 f43 | * | 0    0     1 0 | = | f13*fCos+f23*fSin f23*fCos-f13*fSin f33 f43 |
		// | 0   0   0   1   |   | 0    0     0 1 |   | 0                 0                 0   1   |
		float fTemp, fCos, fSin;
		fCos = cosf(fRadians);
		fSin = sinf(fRadians);
		fTemp = f11*fCos+f21*fSin;
		f21 = f21*fCos-f11*fSin;
		f11 = fTemp;
		fTemp = f12*fCos+f22*fSin;
		f22 = f22*fCos-f12*fSin;
		f12 = fTemp;
		fTemp = f13*fCos+f23*fSin;
		f23 = f23*fCos-f13*fSin;
		f13 = fTemp;
	}
	void RotateMatrix(CVector &v, float f)
	{
		// 15 muls, 10 adds, 2 trig function calls
		float fCos = cosf(f);
		CVector vCos = v * (1 - fCos);
		CVector vSin = v * sinf(f);

		f14 = f24 = f34 = f41 = f42 = f43 = 0;
		f44 = 1;

		f11 = (v.x * vCos.x) + fCos;
		f21 = (v.x * vCos.y) - (vSin.z);
		f31 = (v.x * vCos.z) + (vSin.y);
		f12 = (v.y * vCos.x) + (vSin.z);
		f22 = (v.y * vCos.y) + fCos;
		f32 = (v.y * vCos.z) - (vSin.x);
		f13 = (v.z * vCos.x) - (vSin.y);
		f32 = (v.z * vCos.y) + (vSin.x);
		f33 = (v.z * vCos.z) + fCos;
	}
	void Rotate(CVector &v, float f)
	{
		// 51 muls, 37 adds, 2 trig function calls
		CMatrix mat;
		mat.RotateMatrix(v, f);
		*this *= mat;
	}

	void ModelMatrix(CQuaternion &q, CVector &vFrom)
	{
		*this = q;
		f41 = vFrom.x;
		f42 = vFrom.y;
		f43 = vFrom.z;
	}
	void ViewMatrix(CQuaternion &q, CVector &vFrom)
	{
		*this = q;
		f41 = -(vFrom.x*f11 + vFrom.y*f21 + vFrom.z*f31);
		f42 = -(vFrom.x*f12 + vFrom.y*f22 + vFrom.z*f32);
		f43 = -(vFrom.x*f13 + vFrom.y*f23 + vFrom.z*f33);
	}
	void ViewMatrix(CVector &vFrom, CVector &vView, CVector &vUp, CVector &vRight)
	{
		// 9 muls, 9 adds
		f11 = vRight.x;	f21 = vRight.y;	f31 = vRight.z;	f41 = -(vFrom | vRight);
		f12 = vUp.x;	f22 = vUp.y;	f32 = vUp.z;	f42 = -(vFrom | vUp);
		f13 = -vView.x;	f23 = -vView.y;	f33 = -vView.z;	f43 = -(vFrom | -vView);
		f14 = 0;		f24 = 0;		f34 = 0;		f44 = 1;
	}
	void ViewMatrix(CVector &vFrom, CVector &vAt, CVector &vUp)
	{
		CVector vView = vAt - vFrom;
		vView.Normalize();
		CVector vRight = vView ^ vUp;
		vRight.Normalize();
		vUp = vRight ^ vView;
		vUp.Normalize();
		ViewMatrix(vFrom, vView, vUp, vRight);
	}
	void ProjectionMatrix(float fNear, float fFar, float fFOV, float fAspect)
	{
		// 2 muls, 3 divs, 2 adds, 1 trig function call
		float h = 1.0f / tanf(DEGTORAD(fFOV * 0.5f));
		float Q = fFar / (fFar - fNear);
		f12 = f13 = f14 = f21 = f23 = f24 = f31 = f32 = f41 = f42 = f44 = 0;
		f11 = h / fAspect;
		f22 = h;
		f33 = Q;
		f34 = 1;
		f43 = -Q*fNear;
	}

	// For orthogonal matrices, I belive this also gives you the inverse.
	void Transpose()
	{
		float f;
		SWAP(f12, f21, f);
		SWAP(f13, f31, f);
		SWAP(f14, f41, f);
		SWAP(f23, f32, f);
		SWAP(f24, f42, f);
		SWAP(f34, f43, f);
	}
};


/*******************************************************************************
* Class: CLine
********************************************************************************
* This class implements a 3-dimensional line. It is initialized with two points
* on the line. It can be used to find the shortest distance between two lines or
* between a point and a line, which can be useful for things like collision
* detection.
*******************************************************************************/
class CLine
{
public:
	CVector m_vStart;
	CVector m_vDir;

	CLine()		{}
	CLine(CVector &p1, CVector &p2)	{ Init(p1, p2); }
	void Init(CVector &p1, CVector &p2)
	{
		m_vStart = p1;
		m_vDir = p2 - p1;
		m_vDir.Normalize();
	}

	CVector ClosestPoint(CVector &p)	{ return m_vStart + (m_vDir * ((m_vDir | p) - (m_vDir | m_vStart))); }
	float Distance(CVector &p)			{ return p.Distance(ClosestPoint(p)); }
	float Distance(CLine &l)
	{
		CVector v = m_vDir ^ l.m_vDir;
		v.Normalize();
		return Abs((m_vStart - l.m_vStart) | v);
	}
/*
	// Calculate the shortest distance between a point to a line
	(known) point p = p, or (px, py, pz)
	(known) point on line = a, or (ax, ay, az)
	(known) direction vector of line = n, or (nx, ny, nz)
	(unknown) point on line closest to p = P, or (Px, Py, Pz)

	// The vector from p to P must be perpendicular to n, which gives us:
	(px-Px, py-Py, pz-Pz) . (nx, ny, nz) = 0
	// To make sure P is actually on the line, it must also fit this line equation:
	(ax, ay, az) + k * (nx, ny, nz) = (Px, Py, Pz)

	// Expanding those equations gives us this:
	nx * (px - Px) + ny * (py - Py) + nz * (pz - Pz) = 0
	// and these:
	Px = ax + k * nx
	Py = ay + k * ny
	Pz = az + k * nz

	// Substituting and solving for k gives us:
	nx * (px - (ax + k * nx)) + ny * (py - (ay + k * ny)) + nz * (pz - (az + k * nz)) = 0
	nx*px - nx*ax - k*nx2 + ny*py - ny*ay - k*ny2 + nz*pz - nz*az - k*nz2 = 0
	k * (nx2 + ny2 + nz2) = nx*px - nx*ax + ny*py - ny*ay + nz*pz - nz*az
	k = (nx*px + ny*py + nz*pz - nx*ax - ny*ay - nz*az) / (nx2 + ny2 + nz2)
	k = (n.p - n.a) / (n.n)
	
	// If n is normalized then n.n will always equal 1, so...
	P = a + (n.p - n.a) * n
	distance = | P - p |
*/

/*
	// Calculate the shortest distance between two lines (call them line A and line B)
	(known) point on A = AP, or (APx, APy, APz)
	(known) direction vector of A = AN, or (ANx, ANy, ANz)
	(known) point on B = BP, or (BPx, BPy, BPz)
	(known) direction vector of B = BN, or (BNx, BNy, BNz)
	(unknown) point on A closest to B = P1, or (P1x, P1y, P1z)
	(unknown) point on B closest to A = P2, or (P2x, P2y, P2z)

	// The vector from P1 to P2 is perpendicular to both lines:
	(P2x - P1x, P2y - P1y, P2z - P1z) . (ANx, ANy, ANz) = 0
	(P2x - P1x, P2y - P1y, P2z - P1z) . (BNx, BNy, BNz) = 0
	// To find P1 and P2, we need to find values k and l for the line equations:
	(APx, APy, APz) + k * (ANx, ANy, ANz) = (P1x, P1y, P1z)
	(BPx, BPy, BPz) + l * (BNx, BNy, BNz) = (P2x, P2y, P2z)

	// Expanding these equations gives us these:
	ANx * (P2x - P1x) + ANy * (P2y - P1y) + ANz * (P2z - P1z) = 0
	BNx * (P2x - P1x) + BNy * (P2y - P1y) + BNz * (P2z - P1z) = 0
	// And these:
	P1x = APx + k * ANx
	P1y = APy + k * ANy
	P1z = APz + k * ANz
	P2x = BPx + l * BNx
	P2y = BPy + l * BNy
	P2z = BPz + l * BNz

	// Substituting and solving for k and l gives us:
	ANx * ((BPx + l * BNx) - (APx + k * ANx)) + ANy * ((BPy + l * BNy) - (APy + k * ANy)) + ANz * ((BPz + l * BNz) - (APz + k * ANz)) = 0
	BNx * ((BPx + l * BNx) - (APx + k * ANx)) + BNy * ((BPy + l * BNy) - (APy + k * ANy)) + BNz * ((BPz + l * BNz) - (APz + k * ANz)) = 0

	(ANx*BPx + l*ANx*BNx - ANx*APx - k*ANx2) + (ANy*BPy + l*ANy*BNy - ANy*APy - k*ANy2) + (ANz*BPz + l*ANz*BNz - ANz*APz - k*ANz2) = 0
	(BNx*BPx + l*BNx2 - BNx*APx - k*ANx*BNx) + (BNy*BPy + l*BNy2 - BNy*APy - k*ANy*BNy) + (BNz*BPz + l*BNz2 - BNz*APz - k*ANz*BNz) = 0

	k = (ANx*BPx + l*ANx*BNx - ANx*APx + ANy*BPy + l*ANy*BNy - ANy*APy + ANz*BPz + l*ANz*BNz - ANz*APz) / (ANx2 + ANy2 + ANz2)
	k = (BNx*BPx + l*BNx2 - BNx*APx + BNy*BPy + l*BNy2 - BNy*APy + BNz*BPz + l*BNz2 - BNz*APz) / (ANx*BNx + ANy*BNy + ANz*BNz)

	(ANx*BNx + ANy*BNy + ANz*BNz) * (ANx*BPx + l*ANx*BNx - ANx*APx + ANy*BPy + l*ANy*BNy - ANy*APy + ANz*BPz + l*ANz*BNz - ANz*APz)  = (ANx2 + ANy2 + ANz2) * (BNx*BPx + l*BNx2 - BNx*APx + BNy*BPy + l*BNy2 - BNy*APy + BNz*BPz + l*BNz2 - BNz*APz)

	(ANx*BNx*ANy*BPy + ANx*BNx*l*ANy*BNy - ANx*BNx*ANy*APy + ANx*BNx*ANz*BPz + ANx*BNx*l*ANz*BNz - ANx*BNx*ANz*APz) + 
	(ANy*BNy*ANx*BPx + ANy*BNy*l*ANx*BNx - ANy*BNy*ANx*APx + ANy*BNy*ANz*BPz + ANy*BNy*l*ANz*BNz - ANy*BNy*ANz*APz) +
	(ANz*BNz*ANx*BPx + ANz*BNz*l*ANx*BNx - ANz*BNz*ANx*APx + ANz*BNz*ANy*BPy + ANz*BNz*l*ANy*BNy - ANz*BNz*ANy*APy) =
	(ANx2*BNy*BPy + ANx2*l*BNy2 - ANx2*BNy*APy + ANx2*BNz*BPz + ANx2*l*BNz2 - ANx2*BNz*APz) +
	(ANy2*BNx*BPx + ANy2*l*BNx2 - ANy2*BNx*APx + ANy2*BNz*BPz + ANy2*l*BNz2 - ANy2*BNz*APz) +
	(ANz2*BNx*BPx + ANz2*l*BNx2 - ANz2*BNx*APx + ANz2*BNy*BPy + ANz2*l*BNy2 - ANz2*BNy*APy)

	l * (2*ANx*BNx*ANy*BNy + 2*ANz*BNz*ANx*BNx + 2*ANy*BNy*ANz*BNz - ANx2*BNy2 - ANy2*BNx2 - ANz2*BNx2 - ANx2*BNz2 - ANy2*BNz2 - ANz2*BNy2) = 
	(ANx2*BNy*BPy - ANx2*BNy*APy + ANx2*BNz*BPz - ANx2*BNz*APz) +
	(ANy2*BNx*BPx - ANy2*BNx*APx + ANy2*BNz*BPz - ANy2*BNz*APz) +
	(ANz2*BNx*BPx - ANz2*BNx*APx + ANz2*BNy*BPy - ANz2*BNy*APy) +
	(-ANx*BNx*ANy*BPy + ANx*BNx*ANy*APy - ANx*BNx*ANz*BPz + ANx*BNx*ANz*APz) +
	(-ANy*BNy*ANx*BPx + ANy*BNy*ANx*APx - ANy*BNy*ANz*BPz + ANy*BNy*ANz*APz) +
	(-ANz*BNz*ANx*BPx + ANz*BNz*ANx*APx - ANz*BNz*ANy*BPy + ANz*BNz*ANy*APy)

	// Next think about using AN ^ BN to get the direction vector between P1 and P2
	// Normalize that vector and call it n
	// The distance between the two lines might be the absolute value of (P2-P1).n
	// That still doesn't tell us where the closest points on the line are, but it might help find them
*/
};


/*******************************************************************************
* Class: CPlane
********************************************************************************
* This class implements a 3-dimensional plane equation. It can be initialized
* with 3 points on the plane or a normal and a point on the plane, and it can
* be used to find the distance from a point to a plane. I plan to add more to
* this class later to make it more useful.
*******************************************************************************/
class CPlane
{
public:
	CVector m_vNormal;
	float D;

	CPlane()		{}
	void Init(CVector &p1, CVector &p2, CVector &p3)
	{	// Initializes the plane based on three points in the plane
		Init(NormalVector(p1, p2, p3), p1);
	}
	void Init(CVector &vNormal, CVector &p)
	{	// Initializes the plane based on a normal and a point in the plane
		m_vNormal = vNormal;
		D = -(p | m_vNormal);
	}

	bool operator==(CPlane &p)	{ return (Abs(D - p.D) <= DELTA && m_vNormal == p.m_vNormal); }
	bool operator!=(CPlane &p)	{ return !(*this == p); }

	float Distance(CVector &p)
	{	// Returns the distance between the plane and point p
		return (m_vNormal | p) + D;	// A positive, 0, or negative result indicates the point is in front of, on, or behind the plane
	}

	bool Intersection(CVector &vPos, CVector &vDir)
	{	// Returns true if the line intersects the plane and changes vPos to the location of the intersection
		float f = m_vNormal | vDir;
		if(ABS(f) < DELTA)
			return false;
		vPos -= vDir * (Distance(vPos) / f);
		return true;
	}
};

/****************************************************************************
* Class: C3DBase
*****************************************************************************
* This class represents the most basic 3D object in the scene. It has a 3D
* position, an orientation, and a parent which provides its frame of reference
* in the scene. For instance, this could be a gun turret on a ship. It's
* really part of the ship, but can also be treated as a separate object. This
* is really just a base class meant to be derived from to provide some
* standardized functionality.
* Note: This class is derived from CQuaternion so it will inherit useful
*       functions like Rotate(), GetViewAxis(), GetUpAxis(), and GetRightAxis().
****************************************************************************/
class C3DBase : public CQuaternion
{
protected:
	C3DBase *m_pParent;
	CVector m_vPosition;

public:
	C3DBase() : CQuaternion(0.0f, 0.0f, 0.0f, 1.0f), m_vPosition(0.0f)
	{
		m_pParent = NULL;
	}
	void SetParent(C3DBase *p)			{ m_pParent = p; }
	C3DBase *GetParent()				{ return m_pParent; }
	void SetPosition(CVector &v)		{ m_vPosition = v; }
	CVector GetPosition()				{ return m_vPosition; }

	CMatrix GetViewMatrix()
	{
		CMatrix m;
		m.ViewMatrix(*this, m_vPosition);
		return m;
	}
	CMatrix GetModelMatrix()
	{
		CMatrix m;
		m.ModelMatrix(*this, m_vPosition);
		return m;
	}
	float DistanceTo(C3DBase &obj)	{ return m_vPosition.Distance(obj.m_vPosition); }
	CVector VectorTo(C3DBase &obj)	{ return obj.m_vPosition - m_vPosition; }
};

/****************************************************************************
* Class: C3DObject
*****************************************************************************
* Adding to C3DBase, this class represents a movable object in the scene. It
* has a velocity vector and certain methods to make using it easier.
****************************************************************************/
class C3DObject : public C3DBase
{
protected:
	CVector m_vVelocity;

public:
	C3DObject() : C3DBase(), m_vVelocity(0.0f) {}
	void SetVelocity(CVector &v)			{ m_vVelocity = v; }
	CVector GetVelocity()					{ return m_vVelocity; }
	void AddVelocity(CVector &v)			{ m_vVelocity += v; }
	void UpdatePosition(float fSeconds)		{ m_vPosition += m_vVelocity * fSeconds; }
	void UpdatePosition(int nMilliseconds)	{ m_vPosition += m_vVelocity * (nMilliseconds * 0.001f); }
};

#endif //__Matrix_h__
