/*
	The Universe Development Kit
	Copyright (C) 2000  Sean O'Neil
	soneil@home.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __PixelBuffer_h__
#define __PixelBuffer_h__

#include "Noise.h"

/*******************************************************************************
* Class: CPixelBuffer
********************************************************************************
* This class implements a general-purpose pixel buffer to be used for anything.
* It is often used by CTexture to set up OpenGL textures, so many of the
* parameters you use to initialize it look like the parameters you would pass
* to glTexImage1D or glTexImage2D. Some of the standard pixel buffer routines
* call fast MMX functions implemented in PixelBuffer.asm.
*******************************************************************************/
class CPixelBuffer
{
protected:
	int m_nWidth;				// The width of the pixel buffer
	int m_nHeight;				// The height of the pixel buffer
	int m_nChannels;			// The number of channels of data (1-4)
	int m_nFormat;				// The format of the pixel data (i.e. GL_LUMINANCE, GL_RGBA)
	int m_nDataType;			// The data type of the pixel data (i.e. GL_UNSIGNED_BYTE, GL_FLOAT)
	void *m_pBuffer;			// The pointer to the pixel buffer

public:
	CPixelBuffer()				{ m_pBuffer = NULL; }
	CPixelBuffer(int nWidth, int nHeight, int nChannels=3, int nFormat=GL_RGB, int nDataType=GL_UNSIGNED_BYTE)
	{
		m_pBuffer = NULL;
		Init(nWidth, nHeight, nChannels, nFormat, nDataType);
	}
	~CPixelBuffer()				{ Cleanup(); }

	void SwapBuffer(CPixelBuffer *p)
	{
		void *pTemp;
		SWAP(m_pBuffer, p->m_pBuffer, pTemp);
	}

	int GetWidth()				{ return m_nWidth; }
	int GetHeight()				{ return m_nHeight; }
	int GetChannels()			{ return m_nChannels; }
	int GetFormat()				{ return m_nFormat; }
	int GetDataType()			{ return m_nDataType; }
	void *GetBuffer()			{ return m_pBuffer; }
	void *GetPos(int x, int y)
	{
		x = MAX(0, MIN(m_nWidth-1, x));
		y = MAX(0, MIN(m_nHeight-1, y));
		int nIndex = (y*m_nWidth + x) * m_nChannels;
		return (m_nDataType == GL_FLOAT) ? (void *)&((float *)m_pBuffer)[nIndex] : (void *)&((unsigned char *)m_pBuffer)[nIndex];
	}

	void Init(int nWidth, int nHeight, int nChannels=3, int nFormat=GL_RGB, int nDataType=GL_UNSIGNED_BYTE);
	void Cleanup()
	{
		if(m_pBuffer)
		{
			delete m_pBuffer;
			m_pBuffer = NULL;
		}
	}

	// Miscellaneous initalization routines
	void MakeNoise(unsigned long nSeed);
};

#endif // __PixelBuffer_h__
