/*
	The Universe Development Kit
	Copyright (C) 2000  Sean O'Neil
	soneil@home.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __ROAMSphere_h__
#define __ROAMSphere_h__

#include "Matrix.h"
#include "PixelBuffer.h"

// This function was just hacked together to generate a texture map
// that indicates height and lattitude (a little randomized) to give
// an indication of climate from high altitudes. Up close it looks
// terrible, and a better texturing scheme needs to be written.
void BuildTexture(CPixelBuffer *pPixelBuffer, float fVariance);

//#define USE_ALLOCATOR

#ifdef USE_ALLOCATOR
/****************************************************************************
* Class: CMemoryBlock
*****************************************************************************
* This class implements a simple memory allocator. It allows you to allocate
* blocks of memory for a specific structure or class so you don't have to
* make an expensive system call to allocate or free memory every time you
* want to lock or free some memory. Just initialize it with the maximum number
* of elements and the element size, then call GetElement() to get a pointer
* to a free element in the block and ReleaseElement() to free it. A simple
* stack is used to keep track of the free pointers.
* Note: Comment out the USE_ALLOCATOR define above to remove CMemoryBlock.
****************************************************************************/
class CMemoryBlock
{
protected:
	int m_nElements;
	int m_nElementSize;
	void *m_pBuffer;
	void **m_pFreeStack;
	int m_nFreePosition;

public:
	CMemoryBlock()
	{
		m_nElements = 0;
		m_nElementSize = 0;
		m_pBuffer = NULL;
		m_pFreeStack = NULL;
		m_nFreePosition = 0;
	}
	CMemoryBlock(int nElements, int nElementSize)
	{
		m_pBuffer = NULL;
		m_pFreeStack = NULL;
		Init(nElements, nElementSize);
	}
	~CMemoryBlock()
	{
		Cleanup();
	}
	void Cleanup()
	{
		if(m_pFreeStack)
		{
			delete m_pFreeStack;
			m_pFreeStack = NULL;
		}
		if(m_pBuffer)
		{
			delete m_pBuffer;
			m_pBuffer = NULL;
		}
	}
	void Init(int nElements, int nElementSize)
	{
		Cleanup();
		m_nElements = nElements;
		m_nElementSize = nElementSize;
		m_pBuffer = new char[nElements * nElementSize];
		TRACE("CMemoryBlock::Init() allocating %d bytes.\n", nElements * nElementSize);
		m_pFreeStack = new void *[nElements];
		for(m_nFreePosition = 0; m_nFreePosition < nElements; m_nFreePosition++)
			m_pFreeStack[m_nFreePosition] = (void *)((int)m_pBuffer + m_nFreePosition * nElementSize);
	}
	int GetElementSize()			{ return m_nElementSize; }
	int GetUsedElementCount()		{ return m_nElements - m_nFreePosition; }
	int GetFreeElementCount()		{ return m_nFreePosition; }
	void *GetElement()
	{
		void *pBlock = NULL;
		if(m_nFreePosition)
			pBlock = m_pFreeStack[--m_nFreePosition];
		return pBlock;
	}
	void ReleaseElement(void *pBlock)
	{
		m_pFreeStack[m_nFreePosition++] = pBlock;
	}
};
#endif

class CPackedArray
{
protected:
	unsigned short m_nElements;
	unsigned short m_nElementSize;
	unsigned short m_nBlockSize;
	unsigned short m_nFreePosition;
	unsigned short *m_pFreeStack;
	void *m_pBuffer;

public:
	CPackedArray()
	{
		m_pBuffer = NULL;
		m_pFreeStack = NULL;
	}
	CPackedArray(unsigned short nElements, unsigned short nElementSize, unsigned short nBlockSize=1000)
	{
		m_pBuffer = NULL;
		m_pFreeStack = NULL;
		Init(nElements, nElementSize, nBlockSize);
	}
	~CPackedArray()
	{
		Cleanup();
	}
	void Cleanup()
	{
		if(m_pFreeStack)
		{
			delete m_pFreeStack;
			m_pFreeStack = NULL;
		}
		if(m_pBuffer)
		{
			delete m_pBuffer;
			m_pBuffer = NULL;
		}
	}
	void Init(unsigned short nElements, unsigned short nElementSize, unsigned short nBlockSize=1000)
	{
		Cleanup();
		m_nElements = nElements;
		m_nElementSize = nElementSize;
		m_nBlockSize = nBlockSize;
		m_pBuffer = new char [m_nElements * m_nElementSize];
		memset(m_pBuffer, 0, m_nElements * m_nElementSize);
		m_nFreePosition = 0;
		m_pFreeStack = new unsigned short [m_nElements];
		for(unsigned short i = 0; i < m_nElements; i++)
			m_pFreeStack[i] = i;
	}
	unsigned short GetElementCount()		{ return m_nElements; }
	unsigned short GetElementSize()			{ return m_nElementSize; }
	unsigned short GetUsedElementCount()	{ return m_nFreePosition; }
	unsigned short GetFreeElementCount()	{ return m_nElements - m_nFreePosition; }
	unsigned short GetElement()
	{
		if(m_nFreePosition == m_nElements)
		{
			// Need to allocate a larger array
			m_nElements += m_nBlockSize;
			void *pBuffer = new char [m_nElements * m_nElementSize];
			memcpy(pBuffer, m_pBuffer, m_nFreePosition * m_nElementSize);
			delete m_pBuffer;
			m_pBuffer = pBuffer;
			delete m_pFreeStack;
			m_pFreeStack = new unsigned short [m_nElements];
			for(int i=m_nFreePosition; i<m_nElements; i++)
				m_pFreeStack[i] = i;
		}
		return m_pFreeStack[m_nFreePosition++];
	}
	void ReleaseElement(unsigned short nElement)
	{
		m_pFreeStack[--m_nFreePosition] = nElement;
	}
	//void *operator[](unsigned short n)	{ return (void *)((int)m_pBuffer + (int)(m_nElementSize * n)); }
};

class CROAMVertex;
class CROAMVertexArray : public CPackedArray
{
public:
	CROAMVertexArray(unsigned short nElements, unsigned short nElementSize, unsigned short nBlockSize=1000) : CPackedArray(nElements, nElementSize, nBlockSize) {}
	CROAMVertex *operator[](unsigned short n)	{ return (CROAMVertex *)((int)m_pBuffer + (int)(m_nElementSize * n)); }
};

/****************************************************************************
* Class: CROAMVertex
*****************************************************************************
* This class implements a simple vertex with a position, a normal, and a
* set of 2-dimensional texture coordinates. Note that it has a static vertex
* array object, which allows all ROAM objects to be drawn using one vertex
* array.
****************************************************************************/
class CROAMVertex
{
// Attributes
public:
	static CROAMVertexArray Array;
	float m_fTexCoord[2];		// The texture coordinates of the vertex
	CVector m_vNormal;			// The normal of the vertex (used for smooth shading)
	CVector m_vPosition;		// The position of the vertex

// Operations
public:
	CROAMVertex()	{}
	void Init(CVector &v)
	{
		m_vPosition = v;
		m_vNormal = v;
		m_fTexCoord[0] = m_fTexCoord[1] = 0;
	}
};

/****************************************************************************
* Class: CROAMTriangle
*****************************************************************************
* This class implements the ROAM triangle. It is quite large because it
* relies heavily on caching information so it doesn't have to calculate things
* every frame. Some deal with calculating split priority while others deal
* with handling the actual split. Like CROAMVertex, it is derived from CListNode
* and uses CMemoryBlock if USE_ALLOCATOR is defined.
*
* Note that the parent pointer points to the parent of this triangle, but this
* is still not a tree-like structure! I do not keep a binary tree of all
* triangles used to reach the current model, I only keep the triangles that
* are part of the current model. When I split 2 triangles into 4, I modify
* the 2 existing triangles and add 2 new ones, setting the parent pointers of
* the new triangles to point to the existing ones. Later when I merge 4 into
* 2, the two child triangles get deleted and the parents get changed back.
*
* One more important thing to note is vertex and edge order. ROAM will only
* split two right triangles that make up a square, so I think of every
* triangle as a right triangle (even though it's not really a right triangle
* because it's on a sphere). My vertices always run counter-clockwise with
* the middle vertex being in the middle. Edges are handled in a similar
* fashion. Edge 0 indicates vertices 0 - 1, edge 1 indicates vertices 1 - 2,
* and edge 2 indicates vertices 2 - 0.
*
* It looks something like this:
* (square)
* 1----0 /2   Note: m_pEdge[2] is always the edge that is split, and a
* |   /	/ |         triangle can be split only if m_pEdge[2]->m_pEdge[2]
* |  / /  |         points back to itself. If it doesn't, its neighbor must
* | / /   |         be split first (recursively).
* 2/ 0----1
****************************************************************************/
#define CROAMTriangleList CLinkedList<CROAMTriangle>
class CROAMTriangle : public CListNode<CROAMTriangle>
{
// Attributes
public:
	// Triangle members
	unsigned short m_nVertex[3];	// Indices into a CROAMVertexArray
	CVector m_vNormal;				// The normal of the triangle (used for flat shading)
	CROAMTriangle *m_pEdge[3];		// Pointers to the triangles sharing edges with this one
	CROAMTriangle *m_pParent;		// Pointer to the parent triangle that created this one

	// Split priority members
	CVector m_vMidpoint;		// The midpoint of the third edge (the one to split)
	float m_fOffset;			// The offset in height the midpoint will require when split
	float m_fLength;			// The length of the third edge (the one to split)

// Operations
public:
	CROAMTriangle()	{ memset((char *)this, 0, sizeof(CROAMTriangle)); }
	CROAMTriangle(unsigned short n1, unsigned short n2, unsigned short n3)
	{
		memset((char *)this, 0, sizeof(CROAMTriangle));
		SetVertices(n1, n2, n3);
	}

	CROAMVertex *Vertex(unsigned short n)		{ return CROAMVertex::Array[m_nVertex[n]]; }

	void CopyPriorityInfo(CROAMTriangle *pTriangle)
	{
		m_vMidpoint = pTriangle->m_vMidpoint;
		m_fOffset = pTriangle->m_fOffset;
		m_fLength = pTriangle->m_fLength;
	}
	float GetPriority(CVector &vPosition, CVector &vHeading, float fHorizon)
	{
		float fPriority = 0.0f;
		float fDistance = vPosition.DistanceSquared(m_vMidpoint);
		// Only calculate priority for vertices closer than the horizon
		// (Make an exception if the triangle is larger than the camera's distance to it)
		if(fDistance < m_fLength || fDistance < fHorizon)
		{
			// Only calculate priority for vertices inside the angle of view
			// (Make an exception if the triangle is larger than the camera's distance to it)
			if(fDistance < m_fLength || (vHeading | DirectionVector(vPosition, m_vMidpoint)) > 0.75f)
			{
				float fError = m_fOffset*m_fOffset;
				if(fDistance < m_fLength && fDistance > 1.0f)
					fError = Max(fError, m_fLength/fDistance);
				fPriority = fError / sqrtf(fDistance);
			}
		}
		return fPriority;
	}

	void SetVertices(unsigned short n1, unsigned short n2, unsigned short n3)
	{
		m_nVertex[0] = n1; m_nVertex[1] = n2; m_nVertex[2] = n3;
		ComputeNormal();
	}
	void SetEdges(CROAMTriangle *p1, CROAMTriangle *p2, CROAMTriangle *p3)
	{
		m_pEdge[0] = p1; m_pEdge[1] = p2; m_pEdge[2] = p3;
	}
	void ReplaceEdge(CROAMTriangle *pOld, CROAMTriangle *pNew)
	{
		if(m_pEdge[0] == pOld)
			m_pEdge[0] = pNew;
		else if(m_pEdge[1] == pOld)
			m_pEdge[1] = pNew;
		else if(m_pEdge[2] == pOld)
			m_pEdge[2] = pNew;
	}

	void ComputeNormal()	                { m_vNormal = NormalVector(Vertex(0)->m_vPosition, Vertex(1)->m_vPosition, Vertex(2)->m_vPosition); }
	bool HasVertex(unsigned short n)		{ return (n == m_nVertex[0] || n == m_nVertex[1] || n == m_nVertex[2]); }
	unsigned short GetOpposite(unsigned short n1, unsigned short n2)
	{
		if(m_nVertex[0] != n1 && m_nVertex[0] != n2)
			return m_nVertex[0];
		if(m_nVertex[1] != n1 && m_nVertex[1] != n2)
			return m_nVertex[1];
		return m_nVertex[2];
	}
	void Draw(GLenum nMode, unsigned short *nArray, unsigned short &nIndex)
	{
		if(nMode == GL_SMOOTH)
		{
			nArray[nIndex++] = m_nVertex[0];
			nArray[nIndex++] = m_nVertex[1];
			nArray[nIndex++] = m_nVertex[2];
		}
		else
		{
			glNormal3fv(m_vNormal);
			glArrayElement(m_nVertex[0]);
			glArrayElement(m_nVertex[1]);
			glArrayElement(m_nVertex[2]);
		}
	}

#ifdef USE_ALLOCATOR
public:
	static CMemoryBlock m_mbTriangle;
	void *operator new(unsigned int n)
	{
		return m_mbTriangle.GetElement();
	}
	void operator delete(void *p)
	{
		m_mbTriangle.ReleaseElement(p);
	}
#endif
};

/****************************************************************************
* Class: CROAMDiamond
*****************************************************************************
* This class implements the ROAM diamond. A diamond represents a set of 4
* triangles that can be merged. ROAM's constraints say that you can't merge
* triangles that have themselves been split until they are merged. So unless
* you plan to build the entire tree every frame, you need to keep a list of
* diamonds that can currently be merged. Every split causes existing diamonds
* to be destroyed and new ones to be created and so does every merge, so
* keeping track of them properly is important. Most of the code for that is
* in CROAMSphere.
****************************************************************************/
#define CROAMDiamondList CLinkedList<CROAMDiamond>
class CROAMDiamond : public CListNode<CROAMDiamond>
{
public:
	// Triangles that make up the diamond
	CROAMTriangle *m_pParent[2];
	CROAMTriangle *m_pChild[2];

	// Merge priority members
	CVector m_vMidpoint;			// The midpoint of the longest edge of the parent triangle
	float m_fOffset;				// The offset in height of the split vertex
	float m_fLength;				// The length of the longest edge of the parent triangle

public:
	CROAMDiamond()	{ memset((char *)this, 0, sizeof(CROAMDiamond)); }
	CROAMDiamond(CROAMTriangle *p1, CROAMTriangle *p2, CROAMTriangle *p3, CROAMTriangle *p4)
	{
		m_pParent[0] = p1; m_pParent[1] = p2;
		m_pChild[0] = p3; m_pChild[1] = p4;
		m_vMidpoint = m_pParent[0]->Vertex(2)->m_vPosition.Midpoint(m_pChild[0]->Vertex(0)->m_vPosition);
		m_fOffset = m_pParent[0]->Vertex(1)->m_vPosition.Magnitude() - m_vMidpoint.Magnitude();
		m_fLength = m_pParent[0]->Vertex(2)->m_vPosition.DistanceSquared(m_pChild[0]->Vertex(0)->m_vPosition);
	}
	float GetPriority(CVector &vPosition, CVector &vHeading, float fHorizon)
	{
		float fPriority = 0.0f;
		float fDistance = vPosition.DistanceSquared(m_vMidpoint);
		// Only calculate priority for vertices closer than the horizon
		// (Make an exception if the triangle is larger than the camera's distance to it)
		if(fDistance < m_fLength || fDistance < fHorizon)
		{
			// Only calculate priority for vertices inside the angle of view
			// (Make an exception if the triangle is larger than the camera's distance to it)
			if(fDistance < m_fLength || (vHeading | DirectionVector(vPosition, m_vMidpoint)) > 0.75f)
			{
				float fError = m_fOffset*m_fOffset;
				if(fDistance < m_fLength && fDistance > 1.0f)
					fError = Max(fError, m_fLength/fDistance);
				fPriority = 1.1f * fError / sqrtf(fDistance);
			}
		}
		return fPriority;
	}

	bool HasTriangle(CROAMTriangle *p)	{ return (m_pParent[0] == p || m_pParent[1] == p || m_pChild[0] == p || m_pChild[1] == p); }

#ifdef USE_ALLOCATOR
public:
	static CMemoryBlock m_mbDiamond;
	void *operator new(unsigned int n)
	{
		return m_mbDiamond.GetElement();
	}
	void operator delete(void *p)
	{
		m_mbDiamond.ReleaseElement(p);
	}
#endif
};


/****************************************************************************
* Class: CROAMSphere
*****************************************************************************
* This class implements the spherical ROAM algorithm. It keeps lists of the
* vertices, triangles, and diamonds. Because it is currently used to implement
* a fractal-based planet, it also has a member variable of the CFractal class
* and a number of others for defining the planet's basic properties. Its main
* methods are Init(), Update(), and Draw(). Init() initializes the planet
* with a random seed and certain parameters, like radius and roughness.
* Update() updates the priority lists and performs any necessary splits or
* merges in the mesh. Draw() draws the mesh using either smooth or flat shading.
****************************************************************************/
class CROAMSphere : public C3DObject
{
protected:
	CROAMTriangleList m_llTriangle;
	CROAMDiamondList m_llDiamond;
	float m_fRadius;		// The radius of the planet
	float m_fMaxVariance;	// The maximum variance above or below the radius
	float m_fVariance;
	float m_fOffset;
	CFractal m_frTest;		// The fractal object used to determine altitudes for new vertices

	void RebuildEdgeList();
	void BuildCube();
	void UpdateTriangleOffset(CROAMTriangle *pTriangle);
	void Split(CROAMTriangle *pTriangle);
	void Merge(CROAMDiamond *pDiamond);
	void MergeAll()
	{
		while(!m_llDiamond.IsEmpty())
			Merge(m_llDiamond.GetHead());
	}

public:
	CROAMSphere()
	{
	}
	~CROAMSphere()
	{
	}

	int GetTriangleCount()	{ return m_llTriangle.GetCount(); }
	int GetDiamondCount()	{ return m_llDiamond.GetCount(); }

	void Init(int nSeed, float fRadius, float fVariance, float fRoughness);
	bool Update(C3DBase *pCamera, float fMaxError);
	void Draw(C3DBase *pCamera, GLenum nMode=GL_SMOOTH);
};

#endif // __ROAMSphere_h__
