#if WINDOWS

	#include <winsock2.h>

#elif MACOS

	#include <OpenTransport.h>

#endif

const long networkError = 1; // Generic error code
const unsigned long maxPacketDataSize = 256;
const unsigned short portNumber = 28000;

// This structure holds the 32-bit IP address
// and 16-bit port number which makes up a
// complete internet address
struct NetworkAddress
{
	unsigned long		host;
	unsigned short		port;
};

// A generic data packet structure
struct NetworkPacket
{
	NetworkAddress		address;
	unsigned long		size;
	char					data[maxPacketDataSize];
};

// The NetworkEndpoint class encapsulates all of the
// functionality needed to send and receive packets
class NetworkEndpoint
{
	private:
		
		NetworkAddress		address;
		volatile bool		created;
		
		#if WINDOWS
		
			SOCKET			socket;
			WSAEVENT			socketEvent[2];
			HANDLE			socketThread;
			
			static DWORD WINAPI SocketThread(LPVOID data);
		
		#elif MACOS
		
			volatile bool	bound;
			
			EndpointRef		endpoint;
			TEndpointInfo	info;
			
			static pascal void EndpointNotify(void *contextPtr, OTEventCode code,
				OTResult result, void *cookie);
		
		#endif
	
	public:
		
		NetworkEndpoint(const NetworkAddress& addr);
		~NetworkEndpoint();
		
		bool Send(NetworkPacket *packet);
		bool Receive(NetworkPacket *packet);
};

class NetworkMgr
{
	private:
		
		NetworkAddress		localAddress;
		NetworkEndpoint	*endpoint;
	
	public:
		
		NetworkMgr();
		~NetworkMgr();
};

NetworkMgr::NetworkMgr()
{
	#if WINDOWS
	
		WSADATA	data;
		
		// Initialize WinSock
		if (WSAStartup(MAKEWORD(2, 0), &data) == 0)
		{
			try
			{
				char	name[256];
				
				// Get our local address
				gethostname(name, 256);
				const HOSTENT *hostInfo = gethostbyname(name);
				if (!hostInfo) throw networkError;
				localAddress.host =
					ntohl(*(unsigned long *) hostInfo->h_addr_list[0]);
				localAddress.port = portNumber;
				
				// Create the socket
				endpoint = new NetworkEndpoint(localAddress);
			}
			catch(long error)
			{
				WSACleanup();
				throw;
			}
		}
		else throw networkError;
	
	#elif MACOS
	
		// Initialize OpenTransport
		if (InitOpenTransport() == noErr)
		{
			try
			{
				InetInterfaceInfo		info;
				
				// Get our local address
				OSStatus err = OTInetGetInterfaceInfo(&info, kDefaultInetInterface);
				if (err != noErr) throw networkError;
				localAddress.host = info.fAddress;
				localAddress.port = portNumber;
				
				// Create the endpoint
				endpoint = new NetworkEndpoint(localAddress);
			}
			catch(long error)
			{
				CloseOpenTransport();
				throw;
			}
		}
		else throw networkError;
	
	#endif
}

NetworkMgr::~NetworkMgr()
{
	delete endpoint;
	
	#if WINDOWS
	
		WSACleanup();
	
	#elif MACOS
	
		CloseOpenTransport();
	
	#endif
}

NetworkEndpoint::NetworkEndpoint(const NetworkAddress& addr)
{
	address = addr;
	created = false;
	
	#if WINDOWS
	
		// Create a socket for UDP
		socket = WSASocket(AF_INET, SOCK_DGRAM, 0, NULL, 0, 0);
		if (socket != INVALID_SOCKET)
		{
			DWORD				threadID;
			SOCKADDR_IN		sockaddr;
			
			created = true;
			socketEvent[0] = WSACreateEvent();
			socketEvent[1] = WSACreateEvent();
			
			// Create a thread for receiving network events
			socketThread =
				CreateThread(NULL, 0, &SocketThread, this, 0, &threadID);
			
			// Have WinSock tell us about read and write events
			WSAEventSelect(socket, socketEvent[0], FD_READ | FD_WRITE);
			
			// Bind the socket to our address
			sockaddr.sin_family = AF_INET;
			sockaddr.sin_port = htons(address.port);
			sockaddr.sin_addr.s_addr = htonl(address.host);
			bind(socket, (LPSOCKADDR) &sockaddr, sizeof(SOCKADDR_IN));
		}
		else throw networkError;
	
	#elif MACOS
	
		endpoint = kOTInvalidEndpointRef;
		bound = false;
		
		// Begin the initialization of the endpoint
		OSStatus err = OTAsyncOpenEndpoint(OTCreateConfiguration(kUDPName), 0,
			&info, &EndpointNotify, this);
		if (err != noErr) throw networkError;
	
	#endif
}

NetworkEndpoint::~NetworkEndpoint()
{
	if (created)
	{
		#if WINDOWS
		
			WSAEventSelect(socket, socketEvent[0], 0);
			
			// Signal the thread to exit
			WSASetEvent(socketEvent[1]);
			WaitForSingleObject(socketThread, 500);
			
			CloseHandle(socketThread);
			CloseHandle(socketEvent[1]);
			CloseHandle(socketEvent[0]);
			closesocket(socket);
		
		#elif MACOS
		
			OTSetSynchronous(endpoint);
			if (bound) OTUnbind(endpoint);
			OTCloseProvider(endpoint);
		
		#endif
	}
}

#if WINDOWS

	DWORD WINAPI NetworkEndpoint::SocketThread(LPVOID data)
	{
		NetworkEndpoint *networkEndpoint = static_cast<NetworkEndpoint *>(data);
		
		for (;;)
		{
			// Wait for something to happen
			DWORD d = WSAWaitForMultipleEvents(2, networkEndpoint->socketEvent,
				FALSE, WSA_INFINITE, FALSE);
			if (d == WSA_WAIT_EVENT_0 + 1) break;
			
			if (d == WSA_WAIT_EVENT_0)
			{
				WSANETWORKEVENTS	events;
				
				if (WSAEnumNetworkEvents(networkEndpoint->socket,
					networkEndpoint->socketEvent[0], &events) == 0)
				{
					if (events.lNetworkEvents & FD_READ)
					{
						// Receive packets here
					}
					if (events.lNetworkEvents & FD_WRITE)
					{
						// Send packets here
					}
				}
			}
		}
		return (0);
	}
	
#elif MACOS

	pascal void NetworkEndpoint::EndpointNotify(void *contextPtr,
		OTEventCode code, OTResult result, void *cookie)
	{
		NetworkEndpoint *networkEndpoint =
			static_cast<NetworkEndpoint *>(contextPtr);
		
		switch (code)
		{
			case T_OPENCOMPLETE:
			{
				TBind				bind;
				InetAddress		address;
				
				// Get the OT endpoint and put it in blocking mode
				EndpointRef endpoint = static_cast<EndpointRef>(cookie);
				networkEndpoint->endpoint = endpoint;
				networkEndpoint->created = true;
				OTSetBlocking(endpoint);
				
				// Bind the endpoint to our address
				address.fAddressType = AF_INET;
				address.fPort = networkEndpoint->address.port;
				address.fHost = networkEndpoint->address.host;
				for (long a = 0; a < 8; a++) address.fUnused[a] = 0;
				
				bind.addr.maxlen = sizeof(InetAddress);
				bind.addr.len = sizeof(InetAddress);
				bind.addr.buf = (UInt8 *) &address;
				bind.qlen = 0;
				
				OTBind(endpoint, &bind, 0);
				break;
			}
			
			case T_BINDCOMPLETE:
			
				networkEndpoint->bound = true;
				break;
			
			case T_DATA:
			
				// Receive packets here
				break;
			
			case T_GODATA:
				
				// Send packets here
				break;
			
			case T_UDERR:
				
				// Always have to receive these
				OTRcvUDErr(networkEndpoint->endpoint, 0);
				break;
		}
	}

#endif

// Send and Receive return true if they succeed,
// otherwise they return false
bool NetworkEndpoint::Send(NetworkPacket *packet)
{
	#if WINDOWS
	
		WSABUF			buf;
		SOCKADDR_IN		dest;
		DWORD				count;
		
		// Fill out the destination address
		dest.sin_family = AF_INET;
		dest.sin_port = htons(packet->address.port);
		dest.sin_addr.s_addr = htonl(packet->address.host);
		
		// Specify the data location and size
		buf.len = packet->size;
		buf.buf = packet->data;
		
		// Send the packet
		return (WSASendTo(socket, &buf, 1, &count, 0, (LPSOCKADDR) &dest,
			sizeof(SOCKADDR_IN), 0, 0) == 0);
		
	#elif MACOS
	
		TUnitData		data;
		InetAddress		dest;
		
		// Fill out the destination address
		dest.fAddressType = AF_INET;
		dest.fPort = packet->address.port;
		dest.fHost = packet->address.host;
		
		// Specify the address and data
		data.addr.maxlen = sizeof(InetAddress);
		data.addr.len = sizeof(InetAddress);
		data.addr.buf = (UInt8 *) &dest;
		data.opt.maxlen = 0;
		data.opt.len = 0;
		data.opt.buf = 0;
		data.udata.maxlen = maxPacketDataSize;
		data.udata.len = packet->size;
		data.udata.buf = (UInt8 *) packet->data;
		
		// Send the packet
		return (OTSndUData(endpoint, &data) == noErr);
		
	#endif
}

bool NetworkEndpoint::Receive(NetworkPacket *packet)
{
	for (;;)
	{
		#if WINDOWS
		
			WSABUF			buf;
			SOCKADDR_IN		from;
			DWORD				size;
			
			// Indicate that we'll receive from any address
			from.sin_family = AF_INET;
			from.sin_port = 0;
			from.sin_addr.s_addr = INADDR_ANY;
			INT fromLen = sizeof(SOCKADDR_IN);
			
			// Specify a receipt buffer
			buf.len = maxPacketDataSize;
			buf.buf = packet->data;
			
			// Receive a packet
			DWORD flags = 0;
			if (WSARecvFrom(socket, &buf, 1, &size, &flags,
				(LPSOCKADDR) &from, &fromLen, 0, 0) != 0)
			{
				// If this packet was oversized, ignore it and
				// try to receive another
				if (WSAGetLastError() == WSAEMSGSIZE) continue;
				return (false);
			}
			
			// Get the address that the packet came from
			packet->address.host = ntohl(from.sin_addr.s_addr);
			packet->address.port = ntohs(from.sin_port);
			packet->size = size;
			break;
		
		#elif MACOS
		
			TUnitData		data;
			InetAddress		from;
			OTFlags			flags;
			
			// Specify a recipt buffer
			data.addr.maxlen = sizeof(InetAddress);
			data.addr.len = sizeof(InetAddress);
			data.addr.buf = (UInt8 *) &from;
			data.opt.maxlen = 0;
			data.opt.len = 0;
			data.opt.buf = 0;
			data.udata.maxlen = maxPacketDataSize;
			data.udata.len = 0;
			data.udata.buf = (UInt8 *) packet->data;
			
			// Receive a packet
			if (OTRcvUData(endpoint, &data, &flags) != noErr) return (false);
			
			// If this packet was oversized, ignore it and
			// try to receive another
			if ((flags == T_MORE) || (data.addr.len == 0)) continue;
			
			// Get the address that the packet came from
			packet->address.host = from.fHost;
			packet->address.port = from.fPort;
			packet->size = data.udata.len;
			break;
			
		#endif
	}
	return (true);
}
