#if WINDOWS

	#include <wingdi.h>
	
	typedef HWND OpenGLWindow;
	typedef HGLRC OpenGLContext;

#elif MACOS

	#include <agl.h>
	
	typedef WindowPtr OpenGLWindow;
	typedef AGLContext OpenGLContext;

#endif

const long contextError = 1; // Generic error code

// The GraphicsContext object encapsulates information about
// an OpenGL context and its binding to a window.
class GraphicsContext
{
	private:
		
		OpenGLContext		openglContext;
		
		#if WINDOWS
		
			OpenGLWindow	openglWindow;
			HDC				deviceContext;
		
		#endif
	
	public:
		
		GraphicsContext(OpenGLWindow window);
		~GraphicsContext();
		
		void UpdateContext(void);
};

GraphicsContext::GraphicsContext(OpenGLWindow window)
{
	#if WINDOWS

		PIXELFORMATDESCRIPTOR	format;
		
		// Fill out information about our desired pixel format
		ZeroMemory(&format, sizeof(PIXELFORMATDESCRIPTOR));
		format.nSize = sizeof(PIXELFORMATDESCRIPTOR);
		format.nVersion = 1;
		format.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL
			| PFD_GENERIC_ACCELERATED | PFD_DOUBLEBUFFER;
		format.iPixelType = PFD_TYPE_RGBA;
		format.cColorBits = 24;
		format.cDepthBits = 32;
		format.cStencilBits = 0;
		format.iLayerType = PFD_MAIN_PLANE;
		
		deviceContext = GetDC(window);
		int pixelFormat = ChoosePixelFormat(deviceContext, &format);
		if (!pixelFormat)
		{
			// No matching pixel format available
			ReleaseDC(openglWindow, deviceContext);
			throw contextError;
		}
		if (!SetPixelFormat(deviceContext, pixelFormat, &format))
		{
			ReleaseDC(openglWindow, deviceContext);
			throw contextError;
		}
		
		// Create the OpenGL context and make it current
		openglContext = wglCreateContext(deviceContext);
		if (!wglMakeCurrent(deviceContext, openglContext))
		{
			wglDeleteContext(openglContext);
			ReleaseDC(openglWindow, deviceContext);
			throw contextError;
		}

	#elif MACOS

		GLint		attributes[16];
		
		// Fill out an array containing our desired
		// context attributes
		long x = 0;
		attributes[x++] = AGL_RGBA;
		attributes[x++] = AGL_ACCELERATED;
		attributes[x++] = AGL_DOUBLEBUFFER;
		attributes[x++] = AGL_PIXEL_SIZE;
		attributes[x++] = 32;
		attributes[x++] = AGL_DEPTH_SIZE;
		attributes[x++] = 24;
		attributes[x++] = AGL_NONE;
		
		AGLPixelFormat format = aglChoosePixelFormat(NULL, 0, attributes);
		if (!format) throw contextError;
		
		// Create the OpenGL context
		openglContext = aglCreateContext(format, NULL);
		if (!openglContext)
		{
			aglDestroyPixelFormat(format);
			throw contextError;
		}
		
		// Bind the OpenGL context to the window
		if (!aglSetDrawable(openglContext, (AGLDrawable) window))
		{
			aglDestroyContext(openglContext);
			aglDestroyPixelFormat(format);
			throw contextError;
		}
		
		// Make our OpenGL context current
		aglSetCurrentContext(openglContext);
		aglDestroyPixelFormat(format);
	
	#endif
}

GraphicsContext::~GraphicsContext()
{
	#if WINDOWS
	
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(openglContext);
		ReleaseDC(openglWindow, deviceContext);
	
	#elif MACOS
	
		aglSetDrawable(openglContext, NULL);
		aglDestroyContext(openglContext);
	
	#endif
}

// UpdateContext swaps displays the contents
// of the back buffer
void GraphicsContext::UpdateContext(void)
{
	#if WINDOWS
	
		SwapBuffers(deviceContext);
	
	#elif MACOS
	
		aglSwapBuffers(openglContext);
		
	#endif
}
