//
//   Copyright (c) 1999-2000 Intel Corporation.
//
// Permition is granted to use, copy, distribute and prepare derivative works 
// of this library for any purpose and without fee, provided, that the above 
// copyright notice and this statement appear in all copies.  
// Intel makes no representations about the suitability of this software for 
// any purpose, and specifically disclaims all warranties. 
// See LEGAL.TXT for all the legal information.
//

#include "PIIIBezier.h"
#include "beziertessaltion.h"
#include "scene.h"
#include <fvec.h>
#include "util.h"

CBezierPatch *CPIIIBezierSpline::CreatePatch(DWORD *patchIndices) {
	return new CPIIIBezierPatch(this,patchIndices);
}


CPIIIBezierSpline::CPIIIBezierSpline(CPIIIBezierTessalation *tesl) : CBezierSpline(tesl) 
{
	m_lights = (PIIILight *)alignedAlloc(sizeof(PIIILight)*8);	
}


CPIIIBezierSpline::~CPIIIBezierSpline() 
{
	alignedFree(m_lights);	
}


void CPIIIBezierSpline::PrepearSpline(CScene *scene) 
{
	DWORD i,t;
	CBezierSpline::PrepearSpline(scene);

	for (i=0,t=1;i<8;i++,t<<=1) {
		if (scene->GetLightingState() & t) {
			CScene::Light l = scene->GetLight(i);
			m_lights[i].color.x = F32vec4(l.color.x);
			m_lights[i].color.y = F32vec4(l.color.y);
			m_lights[i].color.z = F32vec4(l.color.z);
			m_lights[i].pos.x = F32vec4(l.pos.x);
			m_lights[i].pos.y = F32vec4(l.pos.y);
			m_lights[i].pos.z = F32vec4(l.pos.z);
			m_lights[i].type = l.type;
		}
	}
}

CPIIIBezierPatch::CPIIIBezierPatch(CBezierSpline *spline,DWORD *indices) : CBezierPatch(spline,indices) {}

bool	CPIIIBezierPatch::EvalSurface(LPDIRECT3DVERTEXBUFFER7 vb,CScene *scene,LPWORD &indices,DWORD &indicesCnt)
{
	
	static F32vec4 _255_ = F32vec4(255.0f);
	D3DTLVERTEX *vtx;
	DWORD i,j,k,t;
	DWORD teslLevel = m_spline->GetTeslLevel();
	Vector4	*projCtrlPts = m_spline->GetProjCtrlPts();
	Vector4	*worldCtrlPts = m_spline->GetViewCtrlPts();
	CBezierTessalation::TessLevel *tessLevel = m_spline->GetBezierTessalation()->GetTessalationLevel(teslLevel);
	F32vec4 *coeffs = (F32vec4 *)tessLevel->m_teslValues;
	DWORD coeffsSkip = tessLevel->m_teslValuesSkip;

	DWORD lightState = scene->GetLightingState();

	vb->Lock(DDLOCK_WAIT,(LPVOID *)&vtx,NULL);
	
	
	F32vec4 vertex;

	F32vec4 scale,translate;
	Vector3 scaling = scene->GetScreenScaling();
	Vector3 trans = scene->GetScreenTranslation();
	scale[0] = scaling.x;
	scale[1] = scaling.y;
	scale[2] = scaling.z;
	scale[3] = 1.0f;
	translate[0] = trans.x;
	translate[1] = trans.y;
	translate[2] = trans.z;
	translate[3] = 0;

	_mm_prefetch((char*)&vtx[1].sx,_MM_HINT_NTA);
	DWORD lastIter = teslLevel*teslLevel;
	DWORD iterCnt = 0;
	for (i=0;i<teslLevel;i++) {
		for (j=0;j<teslLevel;j++) {
			F32vec4 nk;
			F32vec4 nt;
			F32vec4 wpos;
			F32vec4 normx,normy,normz;
			F32vec4 posx,posy,posz;
			vertex = _mm_setzero_ps();
			nk = _mm_setzero_ps();
			nt = _mm_setzero_ps();
			wpos = _mm_setzero_ps();
			for (k=0;k<4;k++) {
				F32vec4 kCoeff = coeffs[k + i*8];
				for (t=0;t<4;t++) {
					F32vec4 tCoeff = coeffs[t + j*8];
					F32vec4 ktCoeff = kCoeff * tCoeff;
					F32vec4 *pt = ((F32vec4 *)&projCtrlPts[m_indices[k][t]]);
					vertex += *pt * ktCoeff;
					if (lightState) {
						F32vec4 ptw = *((F32vec4 *)&worldCtrlPts[m_indices[k][t]]);
						wpos += ptw * ktCoeff;
						nk += ptw * tCoeff * coeffs[k + 4 + i*8];
						nt += ptw * kCoeff*coeffs[t+4 + j*8];
					}
				}
			}
			__asm {
				movaps	xmm0,vertex
				mov		eax,vtx
				rcpss	xmm1,xmm0
				shufps	xmm1,xmm1,0
				movss	xmm2,xmm0
				mulps	xmm0,xmm1
				movss	xmm0,xmm2
				shufps	xmm0,xmm0,0x39
				mulps	xmm0,scale
				addps	xmm0,translate
				movups	[eax],xmm0
			}
			
			if (lightState) {
				F32vec4 n1,temp;
				DWORD mod = iterCnt & 3;

				n1 = _mm_shuffle_ps(nt,nt,_MM_SHUFFLE(1,3,2,0));
				temp = nk*n1;

				n1 = _mm_shuffle_ps(nk,nk,_MM_SHUFFLE(1,3,2,0));
				temp -= n1*nt;

				normx[mod] = temp[2];
				normy[mod] = temp[3];
				normz[mod] = temp[1];
				
				posx[mod] = wpos[1];
				posy[mod] = wpos[2];
				posz[mod] = wpos[3];
				
				iterCnt++;
				if ((iterCnt == lastIter) || !(iterCnt & 0x3)) {
					PIIIVector3 dir;
					PIIIVector3 color;
					F32vec4 size = normx*normx + normy*normy + normz*normz + F32vec4(1e-10);
					F32vec4 dir_size;

					size = rsqrt(size);
					color.x = color.y = color.z = F32vec4(0.0f);

					for (k = 1,t = 0;t< 8; t++,k <<= 1) {
						if (lightState & k) {
							const CPIIIBezierSpline::PIIILight& light = ((CPIIIBezierSpline *)m_spline)->m_lights[t];
							F32vec4 dot;
							if (light.type == CScene::Light::DIRECTIONAL_LIGHT) {
								dir = light.norm;
								dir_size = F32vec4(1.0f);
							} else {
								dir.x = posx-light.pos.x;
								dir.y = posy-light.pos.y;
								dir.z = posz-light.pos.z;
								dir_size = rsqrt(dir.x*dir.x + dir.y*dir.y + dir.z*dir.z);
							}
							dot = dir.x*normx + dir.y*normy + dir.z*normz;
							
							if (move_mask(dot) != 0xf) {
								dot = simd_max(dot,_mm_setzero_ps());
								dot *= size * dir_size;
								color.x += dot*light.color.x;
								color.y += dot*light.color.y;
								color.z += dot*light.color.z;
							}
						}
					}

					F32vec4 r = simd_min(color.x*_255_,_255_);
					F32vec4 g = simd_min(color.y*_255_,_255_);
					F32vec4 b = simd_min(color.z*_255_,_255_);

					DWORD Color[4];

					__m64 dr,dg,db,*outCol = (__m64 *)Color;

					// conversion of lower two
					dr = _mm_cvtt_ps2pi(r); 
					dg = _mm_cvtt_ps2pi(g); 
					db = _mm_cvtt_ps2pi(b); 
					// packing
					dr  = _m_psllqi(dr,16);	 
					dg =  _m_psllqi(dg,8);	 
					outCol[0]  = _m_por(dr,_m_por(dg,db));	
					// conversion of upper two
					dr = _mm_cvtt_ps2pi(_mm_shuffle_ps(r,r,0xe)); 
					dg = _mm_cvtt_ps2pi(_mm_shuffle_ps(g,g,0xe)); 
					db = _mm_cvtt_ps2pi(_mm_shuffle_ps(b,b,0xe)); 
					// packing
					dr  = _m_psllqi(dr,16);	 
					dg =  _m_psllqi(dg,8);	 
					outCol[1]  = _m_por(db,_m_por(dr,dg));	
					vtx -= mod;
					vtx[0].color = 0xff000000 + Color[0];
					vtx[1].color = 0xff000000 + Color[1];
					vtx[2].color = 0xff000000 + Color[2];
					vtx[3].color = 0xff000000 + Color[3];
					vtx += mod;


				}
			} else {
				vtx->color = 0xffffffff;
			}
			vtx->specular = 0x0;
			vtx->tu = 0;
			vtx->tv = 0;
			vtx++;
		}
	}

	_m_empty();
	indices = tessLevel->m_indices;
	indicesCnt = tessLevel->m_indicesCount;
	vb->Unlock();
	return true;
}


