// MirrorDemo.cpp
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
//  PURPOSE:
//    
// MirrorDemo.cpp: implementation the MirrorDemo class.
// This is a rendering to texture demo.
//
// ------------------------------------------------------------------------------------
//
// Author: Kim Pallister - Intel Developer Relations & Engineering Group
//
////////////////////////////////////////////////////////////////////////////////////////

#include "DRGShell.h"


// -----------------------------------------------------------------------
// CMirrorDemo::CMirrorDemo() 
// The MirrorDemo class constructor initializes some of it's member vars (which
// are instances of the DRGMatrix class
// -----------------------------------------------------------------------
CMirrorDemo::CMirrorDemo()
{

	// Set the mirror and central cube center location
	m_drgvMirrorCenter = DRGVector(0.0f, 0.0f, 3.0f);
	m_drgvCubeCenter = DRGVector(0.0f, 0.0f, 0.0f);
	
	// Set the cube's rotational
	// speeds in rotations/second
	//
	m_fRotPerSecX = 0.2f;
	m_fRotPerSecY = 0.3f;
	m_fRotPerSecZ = 0.14f;

	
	m_MxWorld.SetIdentity();
	

	//set the camera at 0,0,-5; the target at 0,0,0 and the up vector at 0,1,0
	m_MxView.SetView( DRGVector( 0, 0, -5 ), DRGVector( 0, 0, 0 ), DRGVector( 0, 1, 0 ) );
	

	// set the projection matrix to 1.57 field of view (radians), 1.33 aspect ratio 
	// (640x480), the near clip plane to 1 and the far to clip plane 100
	// Note: The mirror reflection view uses this projection matrix also
	//
	m_MxProjection.SetProjection( 1.57f, 1.333f, 1.0f, 50.0f );
	

	// Initialize the cube's local world
	// and rotation matrix
	m_MxCubeWorld.SetIdentity();
	m_MxCubeTransform.SetIdentity();
	m_MxCubeTransform.Translate(m_drgvCubeCenter.x, m_drgvCubeCenter.y, m_drgvCubeCenter.z);
	

	// Initialized the mirror's local world
	// and transformation matrices
	m_MxMirrorWorld.SetIdentity();

	// This is the matrix that describes the mirror's
	// rotation in its local space
	m_MxMirrorRotate.SetIdentity();
	
	// This is the matrix that describes the mirror's
	// translation in world space (so far we set it up
	// as a translation to the mirror's center location)
	m_MxMirrorTranslate.SetIdentity();
	m_MxMirrorTranslate.Translate(m_drgvMirrorCenter.x, m_drgvMirrorCenter.y, m_drgvMirrorCenter.z);
	

	m_pMirrorTexture = NULL;
	m_pMirrorSurfaceTexture = NULL;
	m_pFloorTexture = NULL;

	
	m_bDrawCube = true;				
	m_bDrawMirror = true;			
	m_bDrawFloor = true;			


	// We can draw little helper normals
	// and the view frustums (before/after shear)
	m_bDrawMirrorToCameraReflection = false;
	m_bDrawMirrorNormal = false;
	m_bDrawCamUpVector  = false;
	m_bDrawViewFrustum = false;
	m_bDrawCrosshairs = false;

	
	// Toggles auto-rotation of cube
	m_bAutoRotateCube = false;		
	
	// The arrow keys rotate the cube or mirror
	//
	m_bRotateMirror = false;		
	
	// Allows the scene to be drawn from the
	// mirror's point of view, to see what the 
	// dynamic texture looks like
	m_bViewFromMirror = false;		
	
	// The rendered texture is a 512x512 texture
	// but can be changed on the fly during the
	// application
	m_iTextureSize = 512;

	// m_iFrameNumber is a counter that counts
	// from 0 to m_iReflectionUpdateFrequency
	// Each time this counter flips back to zero,
	// the reflection texture is updated
	//
	m_iFrameNumber = 0;
	m_iReflectionUpdateFrequency = 1;

}

// -----------------------------------------------------------------------
// CMirrorDemo::~CMirrorDemo()
// destructor does nothing for now.
// -----------------------------------------------------------------------
CMirrorDemo::~CMirrorDemo()
{
	delete m_pMirrorTexture;
	delete m_pMirrorSurfaceTexture;
	delete m_pFloorTexture;
}

// -----------------------------------------------------------------------
// HRESULT CMirrorDemo::InitWorld( CDRGWrapper *pWrapper )
//
// -----------------------------------------------------------------------
HRESULT CMirrorDemo::InitWorld( CDRGWrapper *pWrapper )
{
	//
	// Make local copy of the pointer to the wrapper class
	//
	m_pWrapper = pWrapper;

	// Create a default material
	//
	ZeroMemory( &m_Material, sizeof(m_Material) );
  
    m_Material.ambient.r = 1.0f;
	m_Material.ambient.g = 1.0f;
	m_Material.ambient.b = 1.0f;
	m_Material.ambient.a = 1.0f;
	
	m_Material.diffuse.r = 1.0f;
	m_Material.diffuse.g = 1.0f;
	m_Material.diffuse.b = 1.0f;
	m_Material.diffuse.a = 1.0f;
	
	m_Material.specular.r = 1.0f;
	m_Material.specular.g = 1.0f;
	m_Material.specular.b = 1.0f;
	m_Material.specular.a = 1.0f;

	
	
	// Set up a default light
	//
    ZeroMemory( &m_Light, sizeof(m_Light) );

    m_Light.dltType       = D3DLIGHT_POINT;
    m_Light.dcvDiffuse.r  = 1.0f;
    m_Light.dcvDiffuse.g  = 1.0f;
    m_Light.dcvDiffuse.b  = 1.0f;
    m_Light.dcvSpecular.r  = 1.0f;
    m_Light.dcvSpecular.g  = 1.0f;
    m_Light.dcvSpecular.b  = 1.0f;
    m_Light.dcvAmbient.r  = 0.1f;
	m_Light.dcvAmbient.g  = 0.1f;
	m_Light.dcvAmbient.b  = 0.1f;
	
	m_Light.dvAttenuation0 = 1.0f; 
	m_Light.dvRange       = D3DLIGHT_RANGE_MAX;
    m_Light.dvPosition.x = 2.0f;
	m_Light.dvPosition.y = 2.0f;
	m_Light.dvPosition.z = -3.0f;

	// Set up the render-to-texture
	// viewport
	m_vpTexViewport.dwX = 0;
	m_vpTexViewport.dwY = 0;
	m_vpTexViewport.dwWidth = m_iTextureSize;
	m_vpTexViewport.dwHeight = m_iTextureSize;
	m_vpTexViewport.dvMinZ = 0.0f;
	m_vpTexViewport.dvMaxZ = 1.0f;

	// This function sets up stuff and restores textures after changing 
	// the device. We call it the first time in to the app.
	TargetChanged();


	// full scale
	m_fScale = 1.0f; 
	
	// m_fMirrorSize is half the mirror's width
	// and half the mirror's height
	m_fMirrorSize = 1.5f;

	
	SetupCubeVerts( m_fScale * 0.5f );
	SetupMirrorVerts( m_fScale * m_fMirrorSize );					
	SetupFloorVerts( m_fScale * 4.0f );


	// Create a blank texture to render the reflected scene to
	// This texture gets mapped to the mirror surface
	//
	m_pMirrorTexture = new CDRGTexture;

	if( m_pMirrorTexture->CreateTexture( m_pWrapper, m_iTextureSize, m_iTextureSize, 16,
								         (DWORD)0, DRGTEXTR_RENDERTARGET, NULL) == S_OK )
		m_pMirrorTexture->Restore(pWrapper->m_pD3DDevice7);


	//
	// Create a new texture, and init it, with an alpha channel. 
	// (try swapping the flag field in the line below for the remmed
	// one to compare the quality of a 4444 vs 8888 ARGB texture, if 
	// your device supports it (if the device doesn't, it will get 
	// converted to a supported format automatically))
	//
	m_pMirrorSurfaceTexture = new CDRGTexture;

	if( m_pMirrorSurfaceTexture->CreateTexture( CBMPLoader( CFileStream( "Images\\mirror.bmp" ) ),
						  CBMPLoader( CFileStream( "Images\\mirror_a.bmp" ) ),
						  (DWORD)0, (DWORD)0/*DRGTEXTR_32BITSPERPIXEL*//*DRGTEXTR_FORCEINAGP*/) == S_OK )

		m_pMirrorSurfaceTexture->Restore(pWrapper->m_pD3DDevice7);
	

	// Checkered floor texture
	m_pFloorTexture = new CDRGTexture;

	if( m_pFloorTexture->CreateTexture( CBMPLoader( CFileStream( "Images\\floor.bmp" ) ),
										(DWORD)0, (DWORD)0/*DRGTEXTR_FORCEINAGP*/ ) == S_OK )
		m_pFloorTexture->Restore(pWrapper->m_pD3DDevice7);

	
	m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_COLOROP,   D3DTOP_MODULATE );
	m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
		
	m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_MINFILTER, D3DTFN_LINEAR );
	m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_MAGFILTER, D3DTFN_LINEAR );

	

	
	return S_OK;
}



// -----------------------------------------------------------------------
// HRESULT CMirrorDemo::DestroyWorld( void )
//
// -----------------------------------------------------------------------
HRESULT CMirrorDemo::DestroyWorld( void )
{
	return S_OK;
}



// -----------------------------------------------------------------------
// HRESULT CMirrorDemo::UpdateWorld( DWORD dwTime, int InputFlag )
//
// -----------------------------------------------------------------------
HRESULT CMirrorDemo::UpdateWorld()
{
	
	// See if it is time to update the reflection texture
	// or not. We update the reflection every time m_iFrameNumber
	// reaches zero
	m_iFrameNumber = ( m_iFrameNumber + 1 ) % m_iReflectionUpdateFrequency;

	if ( m_iFrameNumber == 0 )
		m_bUpdateReflection = true;
	else
		m_bUpdateReflection = false;
	

	// Update the mirror local world matrix
	m_MxMirrorWorld = m_MxMirrorTranslate * m_MxMirrorRotate * m_MxWorld;

	
	// Skip all of the following calculations if we're not
	// updating the reflection texture
	if ( m_bUpdateReflection )
	{

		// Get the mirror's normal vector
		m_drgvMirrorNormal = (DRGVector(0, 0, -1) * m_MxMirrorWorld) - m_drgvMirrorCenter;


		// Get the camera's center location
		m_MxViewInv = m_MxView;
		m_MxViewInv.Invert();
		m_drgvCameraCenter = DRGVector(0, 0, 0) * m_MxViewInv;


		
		// Get the distance from the camera to the mirror's center
		m_fCameraToMirrorDist = (float)sqrt( (m_drgvMirrorCenter.x - m_drgvCameraCenter.x)*(m_drgvMirrorCenter.x - m_drgvCameraCenter.x  ) +
										     (m_drgvMirrorCenter.y - m_drgvCameraCenter.y)*(m_drgvMirrorCenter.y - m_drgvCameraCenter.y ) +
										     (m_drgvMirrorCenter.z - m_drgvCameraCenter.z)*(m_drgvMirrorCenter.z - m_drgvCameraCenter.z ) );
	
	
		// Get the camera's view vector
		m_drgvCameraView = Normalize( m_drgvMirrorCenter - m_drgvCameraCenter );
		
		// Reflect the camera's view using the mirror's normal
		m_drgvCameraViewReflected = ( ( 2 * m_drgvMirrorNormal ) * DotProduct( m_drgvMirrorNormal, m_drgvCameraView ) ) - m_drgvCameraView;
	
		// Get the camera's up vector from the view matrix
		// This vector is only used to draw the camera up vector 
		// in the scene as a reference
		m_drgvCameraUpVector = DRGVector( m_MxView.m_Mx(0,1), m_MxView.m_Mx(1,1), m_MxView.m_Mx(2,1 ) );;

		// Calculate the camera's virtual location behind the mirror to get the mirror's view
		m_drgvCameraCenterReflected = ( m_drgvCameraViewReflected * m_fCameraToMirrorDist ) + m_drgvMirrorCenter;

		// From the calculated reflected camera's location, we get the location of a point
		// that is behind the mirror, along the same vector as the mirror's normal. The initial
		// view matrix is set up from this point as a "look-from" point. The view matrix is then
		// sheared so that this point is moved to the reflected camera's location
		//
		m_drgvPointBehindMirror = ((DotProduct(-m_drgvMirrorNormal, (m_drgvCameraCenterReflected-m_drgvMirrorCenter)) * -m_drgvMirrorNormal)) + m_drgvMirrorCenter;
		

		// Distance from the point behind the mirror to the mirror's center
		m_fDistToFrontPlane = Magnitude(m_drgvPointBehindMirror - m_drgvMirrorCenter);

	
		// Get the field of view from the point behind the mirror and the
		// mirror's width
		m_fFov = 2.0f * (float)atan(m_fMirrorSize / m_fDistToFrontPlane);


		m_MxMirrorProjection.SetProjection( m_fFov, 1.0f, m_fDistToFrontPlane, 50.0f );
		
		

		// Get a matrix that is the inverse of the mirror's world matrix. This
		// matrix will undo all of the transformations applied to the mirror so far.
		// This matrix is used in order to shear the mirror about the mirror center
		m_MxMirrorWorldInv = m_MxMirrorWorld;
		m_MxMirrorWorldInv.Invert();
	
		// Get the mirror's up vector
		m_drgvMirrorUpVector = (DRGVector(0,1,0) * m_MxMirrorWorld) - m_drgvMirrorCenter;


		// Set the initial mirror view before the shear is applied
		m_MxMirrorView.SetView( m_drgvPointBehindMirror, m_drgvMirrorCenter, m_drgvMirrorUpVector );
		
	
		// The mirror view (_B_efore  _S_hear) inverted is used to transform coordinates
		// in the mirror view space to world space. This matrix is used for drawing the
		// view frustum before the shear is applied to the view matrix.
		m_MxMirrorViewInvBS = m_MxMirrorView;
		m_MxMirrorViewInvBS.Invert();
	
	
		// Transform the camera center location and the point behind the mirror
		// we calculated above to mirror view coordinates
		// (IMVC = In Mirror View Coordinates)
		m_drgvCameraCenterReflectedIMVC = m_drgvCameraCenterReflected * m_MxMirrorView;
		m_drgvPointBehindMirrorIMVC = m_drgvPointBehindMirror * m_MxMirrorView;
	

		// The differences in the X and Y values of the camera's reflected center (virtual)
		// and the point behind the mirror is equal to the shear we must perform on the
		// mirror view matrix. The X value is negated in order to properly simulate a mirror
		// because a real mirror flips x-values.
		m_fShearOnX = -(m_drgvCameraCenterReflectedIMVC.x - m_drgvPointBehindMirrorIMVC.x);
		m_fShearOnY = m_drgvCameraCenterReflectedIMVC.y - m_drgvPointBehindMirrorIMVC.y;

	
		// Shear -> shear on (x, y, z) using the shear matrix results in
		// (x + shearX * z, y + shearY * z, z). This is why the shear amounts are
		// divided by the z value of the point behind the mirror (m_fDistToFrontPlane).
		// This is so that the view frustrum is sheared correctly.
		//
		m_MxShear.SetIdentity();
		m_MxShear.m_Mx(2,0) = m_fShearOnX/(m_fDistToFrontPlane);
		m_MxShear.m_Mx(2,1) = m_fShearOnY/(m_fDistToFrontPlane);


		// Apply the transformations that were applied to the mirror
		// so that we shear around the mirror's center rather than
		// the world's center. Then multiply by the inverse of the 
		// mirror's world matrix to put it back to the right location.
		m_MxShear =  m_MxMirrorWorld * m_MxShear * m_MxMirrorWorldInv;
	
		// We must invert the shear matrix
		// before applying it to the view matrix.
		m_MxShear.Invert();
		
		m_MxMirrorView =  m_MxMirrorView * m_MxShear;

	}


	// Rotate the cube automatically each frame the amount specified
	// (speed is in rotations/second)
	if (m_bAutoRotateCube)
	{
		m_MxCubeTransform.RotateX((float)m_fRotPerSecX * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxCubeTransform.RotateY((float)m_fRotPerSecY * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxCubeTransform.RotateZ((float)m_fRotPerSecZ * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
	}

	// Update the cube and mirror local world matrices
	m_MxCubeWorld = m_MxCubeTransform * m_MxWorld;


	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CMirrorDemo::RenderWorld( void )
//
// -----------------------------------------------------------------------
HRESULT CMirrorDemo::RenderWorld( void )
{

	HRESULT hr;
	

	//m_pWrapper->m_pD3D7->EvictManagedTextures();


	// Save the viewport
	D3DVIEWPORT7 viewDataSave;
	m_pWrapper->m_pD3DDevice7->GetViewport( &viewDataSave );


	// We need to store the view frustum points
	// in a DRGVector array because we need to
	// multiply them by a DRGMatrix. We then convert
	// the DRGVectors to D3DLVERTEXs to draw the lines
	//
	DRGVector viewFrustumVerts[9];
	DRGVector viewFrustumVertsBS[9];
	D3DLVERTEX viewFrustum[9];
	D3DLVERTEX viewFrustumBS[9];

	
	// These are the indices for drawing the view frustums
	WORD viewFrustumIndices[] = { 0,1,1,3,3,2,2,0,1,5,3,7,2,6,0,4,4,5,5,7,7,
		                          6,6,4,0,8,1,8,2,8,3,8 };
	
	if ( m_bDrawViewFrustum )
	{
		
		// Get a matrix that is the inverse of the mirror view matrix so that we can 
		// get the values of the view frustum in world coordinates.
		m_MxMirrorViewInv = m_MxMirrorView;
		m_MxMirrorViewInv.Invert();

		// Half the front plane width is m_fMirrorSize
		float fHalfFrontPlaneWidth = m_fMirrorSize;
		float fHalfFrontPlaneHeight = m_fMirrorSize;
		
		// Get 1/2 the width and height of the back 
		// of the view frustum using a simple ratio
		float fHalfBackPlaneWidth = (m_fDistToFrontPlane+50)/(m_fDistToFrontPlane/fHalfFrontPlaneWidth);
		float fHalfBackPlaneHeight = (m_fDistToFrontPlane+50)/(m_fDistToFrontPlane/fHalfFrontPlaneHeight);


		// bottomLeft
		viewFrustumVerts[0] = DRGVector(-fHalfFrontPlaneWidth, -fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInv;
		// topLeft
		viewFrustumVerts[1] = DRGVector(-fHalfFrontPlaneWidth, fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInv;
		// bottomRight
		viewFrustumVerts[2] = DRGVector(fHalfFrontPlaneWidth, -fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInv;
		// topRight	
		viewFrustumVerts[3] = DRGVector(fHalfFrontPlaneWidth, fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInv;
	
		// back plane verts
		viewFrustumVerts[4] = DRGVector(-fHalfBackPlaneWidth, -fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInv;
		viewFrustumVerts[5] = DRGVector(-fHalfBackPlaneWidth, fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInv;
		viewFrustumVerts[6] = DRGVector(fHalfBackPlaneWidth, -fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInv;
		viewFrustumVerts[7] = DRGVector(fHalfBackPlaneWidth, fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInv;
	
		// mirror view origin
		viewFrustumVerts[8] = m_drgvCameraCenterReflected;

	
		// bottomLeft
		viewFrustumVertsBS[0] = DRGVector(-fHalfFrontPlaneWidth, -fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInvBS;
		// topLeft
		viewFrustumVertsBS[1] = DRGVector(-fHalfFrontPlaneWidth, fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInvBS;
		// bottomRight
		viewFrustumVertsBS[2] = DRGVector(fHalfFrontPlaneWidth, -fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInvBS;
		// topRight	
		viewFrustumVertsBS[3] = DRGVector(fHalfFrontPlaneWidth, fHalfFrontPlaneHeight, m_fDistToFrontPlane) * m_MxMirrorViewInvBS;
	
		// back plane verts
		viewFrustumVertsBS[4] = DRGVector(-fHalfBackPlaneWidth, -fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInvBS;
		viewFrustumVertsBS[5] = DRGVector(-fHalfBackPlaneWidth, fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInvBS;
		viewFrustumVertsBS[6] = DRGVector(fHalfBackPlaneWidth, -fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInvBS;
		viewFrustumVertsBS[7] = DRGVector(fHalfBackPlaneWidth, fHalfBackPlaneHeight, m_fDistToFrontPlane+50) * m_MxMirrorViewInvBS;
	
		// mirror view origin
		viewFrustumVertsBS[8] = m_drgvPointBehindMirror;

		// copy all of the DRGVector data to D3D vertices
		for(int k=0; k<9; k++)
		{
			viewFrustum[k].x = viewFrustumVerts[k].x;
			viewFrustum[k].y = viewFrustumVerts[k].y;
			viewFrustum[k].z = viewFrustumVerts[k].z;
			viewFrustum[k].color = D3DRGBA(0.5f, 0.5f, 0.5f, 1.0f);

			viewFrustumBS[k].x = viewFrustumVertsBS[k].x;
			viewFrustumBS[k].y = viewFrustumVertsBS[k].y;
			viewFrustumBS[k].z = viewFrustumVertsBS[k].z;
			viewFrustumBS[k].color = D3DRGBA(0.0f, 0.75f, 0.0f, 1.0f);
		}
	
	
	}

	// mirror center to virtual camera center
	D3DLVERTEX mirrorToCameraReflection[2];
	if ( m_bDrawMirrorToCameraReflection )
	{
		mirrorToCameraReflection[0].x = m_drgvMirrorCenter.x;
		mirrorToCameraReflection[0].y = m_drgvMirrorCenter.y;
		mirrorToCameraReflection[0].z = m_drgvMirrorCenter.z;
		mirrorToCameraReflection[1].x = m_drgvCameraCenterReflected.x;
		mirrorToCameraReflection[1].y = m_drgvCameraCenterReflected.y;
		mirrorToCameraReflection[1].z = m_drgvCameraCenterReflected.z;
		mirrorToCameraReflection[0].color = D3DRGBA(1.0f, 1.0f, 1.0f, 1.0f);
		mirrorToCameraReflection[1].color = D3DRGBA(1.0f, 1.0f, 1.0f, 1.0f);
	}


	// mirror normal
	D3DLVERTEX mirrorNormalLine[2];
	
	if ( m_bDrawMirrorNormal )
	{
		mirrorNormalLine[0].x = m_drgvMirrorCenter.x;
		mirrorNormalLine[0].y = m_drgvMirrorCenter.y;
		mirrorNormalLine[0].z = m_drgvMirrorCenter.z;
		mirrorNormalLine[1].x = m_drgvMirrorCenter.x + m_drgvMirrorNormal.x;
		mirrorNormalLine[1].y = m_drgvMirrorCenter.y + m_drgvMirrorNormal.y;
		mirrorNormalLine[1].z = m_drgvMirrorCenter.z + m_drgvMirrorNormal.z;
		mirrorNormalLine[0].color = D3DRGBA(0.0f, 0.0f, 1.0f, 1.0f);
		mirrorNormalLine[1].color = D3DRGBA(0.0f, 0.0f, 1.0f, 1.0f);
	}


	// cross hairs on the mirror to help see the mirror's orientation
	D3DLVERTEX crosshairsOnMirror[4];
	
	if ( m_bDrawCrosshairs )
	{
		crosshairsOnMirror[0].x = -m_fMirrorSize;
		crosshairsOnMirror[0].y = 0.0f;
		crosshairsOnMirror[0].z = 0.0f;
		crosshairsOnMirror[1].x = m_fMirrorSize;
		crosshairsOnMirror[1].y = 0.0f;
		crosshairsOnMirror[1].z = 0.0f;
		crosshairsOnMirror[2].x = 0.0f;
		crosshairsOnMirror[2].y = -m_fMirrorSize;
		crosshairsOnMirror[2].z = 0.0f;
		crosshairsOnMirror[3].x = 0.0f;
		crosshairsOnMirror[3].y = m_fMirrorSize;
		crosshairsOnMirror[3].z = 0.0f;

		for(int i=0; i<4; i++)
			crosshairsOnMirror[i].color = D3DRGBA(1.0f, 0.0f, 0.0f, 1.0f);
	}

	
	// camera up vector drawn at the mirror center
	D3DLVERTEX camUpVector[2];
	
	if ( m_bDrawCamUpVector )
	{
		camUpVector[0].x = m_drgvMirrorCenter.x;
		camUpVector[0].y = m_drgvMirrorCenter.y;
		camUpVector[0].z = m_drgvMirrorCenter.z;
		camUpVector[1].x = m_drgvCameraUpVector.x + m_drgvMirrorCenter.x;
		camUpVector[1].y = m_drgvCameraUpVector.y + m_drgvMirrorCenter.y;
		camUpVector[1].z = m_drgvCameraUpVector.z + m_drgvMirrorCenter.z;
		camUpVector[0].color = D3DRGBA(0.0f, 1.0f, 1.0f, 1.0f);
		camUpVector[1].color = D3DRGBA(0.0f, 1.0f, 1.0f, 1.0f);
	}

	
	
	// Call EndScene() because we began one in main.cpp
	// We must end the scene before changing the render target
	m_pWrapper->m_pD3DDevice7->EndScene();


	if ( m_bUpdateReflection )
	{
		// Set the render target to the mirror texture
		m_pWrapper->m_pD3DDevice7->SetRenderTarget( m_pMirrorTexture->ptexTexture, 0 );


		// Set the viewport to fit the rendered scene into the texture
		hr = m_pWrapper->m_pD3DDevice7->SetViewport( &m_vpTexViewport );
		
		if (hr != D3D_OK)
			OutputDebugString( "Setting the viewport to the texture size failed\n" );


		// Clear the viewport
		// Here we clear the mirror texture just a little lighter than black
		// so that we can see the mirror's boundaries in the scene
		hr = m_pWrapper->m_pD3DDevice7->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x00151515, 1.0f, 0L );

		m_pWrapper->m_pD3DDevice7->BeginScene();
	
	
		// From the mirror's view we want to draw the scene to the texture
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		if ( hr != D3D_OK )
			OutputDebugString( "SetTransform failed : World - MxWorld\n" );
	

		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxMirrorView.m_Mx );
		if ( hr != D3D_OK )
			OutputDebugString( "SetTransform failed : View - MxMirrorView\n" );


		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxMirrorProjection.m_Mx );
		if ( hr != D3D_OK )
			OutputDebugString( "SetTransform failed : Projection - MxMirrorProjection\n" );



		// Note: Even though we render the scene from the view of the mirror along the reflection
		// vector, the scene is still not like a true reflection. In order to make the mirror like
		// a real reflection we must flip the texture about the Y-Axis. This is done when the texture
		// coordinates are set up for the mirror surface.

		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pFloorTexture->dwStage, m_pFloorTexture->ptexTexture);	

	
	
		// Draw the floor and cube to the texture used for the mirror
		// The floor is drawn in world coordinates while the cube is drawn in cube
		// local coordinates
		//
		if (m_bDrawFloor)
		{
			hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
														   m_FloorVerts, NUMFLOORVERTS, 0);
		}
	

		// Now set the world to the cube's local coordinates
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxCubeWorld.m_Mx );
		if ( hr != D3D_OK )
			OutputDebugString( "SetTransform failed : World - MxCubeWorld\n" );


		if (m_bDrawCube) 
			DrawCube();
	
	
		m_pWrapper->m_pD3DDevice7->EndScene();

	
	
		// We now want to begin rendering to the back buffer
		m_pWrapper->m_pD3DDevice7->SetRenderTarget( m_pWrapper->m_pBackBuffer, 0 );

	}

	// Set the projection, view and world coordinates for rendering to the back buffer
	hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxProjection.m_Mx );
	if ( hr != D3D_OK )
		OutputDebugString( "SetTransform failed : Projection - MxProjection\n" );

	

	// Set the view matrix to the mirror's view if we're supposed to draw the
	// scene as seen from the mirror (using the 'V' key)
	if (!m_bViewFromMirror)
	{
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxProjection.m_Mx );
	}
	else
	{
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxMirrorView.m_Mx );
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxMirrorProjection.m_Mx );
	}

	if ( hr != D3D_OK )
		OutputDebugString( "SetTransform failed : View - Mx(Mirror)View\n" );



	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
	if ( hr != D3D_OK )
		OutputDebugString( "SetTransform failed : World - MxWorld\n" );

	
	// Set the viewport back to the original full screen
	m_pWrapper->m_pD3DDevice7->SetViewport( &viewDataSave );
	
	
	// Clear the viewport to black
	hr = m_pWrapper->m_pD3DDevice7->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0L );

	
	
	m_pWrapper->m_pD3DDevice7->BeginScene();


	
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, FALSE );
	

	// Draw the view frustum
	if ( m_bDrawViewFrustum )
		hr = m_pWrapper->m_pD3DDevice7->DrawIndexedPrimitive( D3DPT_LINELIST, D3DFVF_LVERTEX,
															  viewFrustum, 9, viewFrustumIndices,
															  32, 0);

	// Draw the view frustum before the shear
	if ( m_bDrawViewFrustum )
		hr = m_pWrapper->m_pD3DDevice7->DrawIndexedPrimitive( D3DPT_LINELIST, D3DFVF_LVERTEX,
															  viewFrustumBS, 9, viewFrustumIndices,
															  32, 0);

	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, TRUE );

	
	hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pFloorTexture->dwStage, m_pFloorTexture->ptexTexture);	


	if (m_bDrawFloor)
	{
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
													   m_FloorVerts, NUMFLOORVERTS, 0);
	}

	
	// Set the view back, but use the mirror's local coordinates drawing the mirror
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxMirrorWorld.m_Mx );
	if ( hr != D3D_OK )
		OutputDebugString( "SetTransform failed : World - MxMirrorWorld\n" );



	// We use our recently rendered texture on the mirror surface
	hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pMirrorTexture->dwStage, m_pMirrorTexture->ptexTexture);	
	

	// Don't use lighting on the mirror surface because the mirror's illumination
	// is all based on the illumination of the reflected objects
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, FALSE );
	
	if (m_bDrawMirror) 
	{
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_LVERTEX,
			 m_MirrorVerts, NUMMIRRORVERTS, 0);
	}

	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, TRUE );
	

	// Draw Crosshairs on the mirror
	// We have to turn off lighting here 
	// to disable the material properties
	// otherwise our lines won't be colored
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, FALSE );

		
	if (m_bDrawCrosshairs)
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_LINELIST, D3DFVF_LVERTEX,
													   crosshairsOnMirror, 4, 0);

	
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, TRUE );


	
	// We now want to draw the mirror surface with the cracked mirror texture
	hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pMirrorSurfaceTexture->dwStage, m_pMirrorSurfaceTexture->ptexTexture);	

	
	// The mirror uses D3DLVERTEX vertices requiring no lighting
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, FALSE );

	if (m_bDrawMirror) 
	{
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_LVERTEX,
													   m_MirrorVerts, NUMMIRRORVERTS, 0);
	}
	
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, TRUE );


	// Now we want to draw the cube in the scene. Therefore we must set the world matrix
	// to the cube's local world matrix
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxCubeWorld.m_Mx );
	if ( hr != D3D_OK )
		OutputDebugString( "SetTransform failed : World - MxCubeWorld\n" );


	if (m_bDrawCube) 
		DrawCube();


	// Set the world matrix back to the real world coordinates
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
	if ( hr != D3D_OK )
		OutputDebugString( "SetTransform failed : World - MxWorld\n" );


	
	
	// We need to turn off the lights again
	// because we're drawing more line segments
	//
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, FALSE );
	

	// Draw the mirror to reflected camera location line segment
	if ( m_bDrawMirrorToCameraReflection )
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_LINELIST, D3DFVF_LVERTEX,
													   mirrorToCameraReflection, 2, 0);

	// Draw the line segment for the mirror's normal
	if ( m_bDrawMirrorNormal )
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_LINELIST, D3DFVF_LVERTEX,
													   mirrorNormalLine, 2, 0);
	

	// Draw the camera up vector
	if ( m_bDrawCamUpVector )
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_LINESTRIP, D3DFVF_LVERTEX,
													   camUpVector, 2, 0);
		
	
	// enable lighting
	m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_LIGHTING, TRUE );
		
	return S_OK;
}


// -----------------------------------------------------------------------
// HRESULT CMirrorDemo::TargetChanged( void )
//
// -----------------------------------------------------------------------
HRESULT CMirrorDemo::TargetChanged( void )
{
	HRESULT hr;
	//
	// This indicates that the devices of the DRGWrapper class have changed.
	// If you have anything that depends on the current rendering device, you'll
	// have to re-initialize it here
	//

	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxProjection.m_Mx );

	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_SRCBLEND , D3DBLEND_SRCALPHA );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_DESTBLEND , D3DBLEND_INVSRCALPHA );
	
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_DITHERENABLE, TRUE );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_SPECULARENABLE, TRUE );


	// Restore textures
	if (m_pMirrorTexture)
		m_pMirrorTexture->Restore(m_pWrapper->m_pD3DDevice7);
	
	if (m_pMirrorSurfaceTexture)
		m_pMirrorSurfaceTexture->Restore(m_pWrapper->m_pD3DDevice7);

	if (m_pFloorTexture)
		m_pFloorTexture->Restore(m_pWrapper->m_pD3DDevice7);
	
	if (m_pMirrorTexture)
	{
		hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_COLOROP,   D3DTOP_MODULATE );
		hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
		hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_MINFILTER, D3DTFN_LINEAR );
		hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pMirrorTexture->dwStage, D3DTSS_MAGFILTER, D3DTFN_LINEAR );
	}

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&m_Material) ) )
		return hr;
	
	
	
    // Set the light
    if( FAILED(m_pWrapper->m_pD3DDevice7->SetLight( 0, &m_Light )))
		return hr;

	if( FAILED(m_pWrapper->m_pD3DDevice7->LightEnable( 0, TRUE )))
		return hr;

	if( FAILED(m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_AMBIENT, 0xff333333 ) ) )
		return hr;


	
	return S_OK;
}



// -----------------------------------------------------------------------
// void CMirrorDemo::UpdateMirrorTextureSize( int iTextureSize )
//
// -----------------------------------------------------------------------
void CMirrorDemo::UpdateMirrorTextureSize( int iTextureSize )
{
	delete m_pMirrorTexture;
	m_pMirrorTexture = NULL;
	
	m_iTextureSize = iTextureSize;

	m_pMirrorTexture = new CDRGTexture;

	if( m_pMirrorTexture->CreateTexture( m_pWrapper, m_iTextureSize, m_iTextureSize, 16,
								         (DWORD)0, DRGTEXTR_RENDERTARGET, NULL) == S_OK )
		m_pMirrorTexture->Restore(m_pWrapper->m_pD3DDevice7);

	// We also need to update the viewport size so that the scene fits entirely
	// within the texture
	m_vpTexViewport.dwWidth = m_iTextureSize;
	m_vpTexViewport.dwHeight = m_iTextureSize;


}




// -----------------------------------------------------------------------
// HRESULT CMirrorDemo::DrawCube()
//
//   Draws the cube with different colors for each face
// -----------------------------------------------------------------------
HRESULT CMirrorDemo::DrawCube()
{

	HRESULT hr;

	// Save the previously used textures and materials
	//
	LPDIRECTDRAWSURFACE7 prevTexture;
	m_pWrapper->m_pD3DDevice7->GetTexture( 0, &prevTexture );

	hr = m_pWrapper->m_pD3DDevice7->SetTexture( 0, NULL);	

	D3DMATERIAL7 prevMaterial;
	ZeroMemory( &prevMaterial, sizeof(prevMaterial) );
	m_pWrapper->m_pD3DDevice7->GetMaterial(&prevMaterial);


	// Create a material used for the 6 different sides of the cube
	//
	D3DMATERIAL7 tempmat;
	ZeroMemory( &tempmat, sizeof(tempmat) );
  
    tempmat.ambient.r = 1.0f;
	tempmat.ambient.g = 0.0f;
	tempmat.ambient.b = 0.0f;
	tempmat.ambient.a = 1.0f;
	
	tempmat.diffuse.r = 1.0f;
	tempmat.diffuse.g = 0.0f;
	tempmat.diffuse.b = 0.0f;
	tempmat.diffuse.a = 1.0f;
	
	tempmat.specular.r = 1.0f;
	tempmat.specular.g = 0.0f;
	tempmat.specular.b = 0.0f;
	tempmat.specular.a = 1.0f;
	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          m_CubeVerts, 6, 0);

    
    tempmat.ambient.r = 0.0f;
	tempmat.ambient.g = 1.0f;
	tempmat.ambient.b = 0.0f;
	tempmat.ambient.a = 1.0f;
	
	tempmat.diffuse.r = 0.0f;
	tempmat.diffuse.g = 1.0f;
	tempmat.diffuse.b = 0.0f;
	tempmat.diffuse.a = 1.0f;
	
	tempmat.specular.r = 0.0f;
	tempmat.specular.g = 1.0f;
	tempmat.specular.b = 0.0f;
	tempmat.specular.a = 1.0f;


	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[6], 6, 0);


    tempmat.ambient.r = 0.0f;
	tempmat.ambient.g = 0.0f;
	tempmat.ambient.b = 1.0f;
	tempmat.ambient.a = 1.0f;
	
	tempmat.diffuse.r = 0.0f;
	tempmat.diffuse.g = 0.0f;
	tempmat.diffuse.b = 1.0f;
	tempmat.diffuse.a = 1.0f;
	
	tempmat.specular.r = 0.0f;
	tempmat.specular.g = 0.0f;
	tempmat.specular.b = 1.0f;
	tempmat.specular.a = 1.0f;

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[12], 6, 0);
	
	tempmat.ambient.r = 1.0f;
	tempmat.ambient.g = 1.0f;
	tempmat.ambient.b = 0.0f;
	tempmat.ambient.a = 1.0f;
	
	tempmat.diffuse.r = 1.0f;
	tempmat.diffuse.g = 1.0f;
	tempmat.diffuse.b = 0.0f;
	tempmat.diffuse.a = 1.0f;
	
	tempmat.specular.r = 1.0f;
	tempmat.specular.g = 1.0f;
	tempmat.specular.b = 0.0f;
	tempmat.specular.a = 1.0f;

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[18], 6, 0);

	
	tempmat.ambient.r = 0.0f;
	tempmat.ambient.g = 1.0f;
	tempmat.ambient.b = 1.0f;
	tempmat.ambient.a = 1.0f;
	
	tempmat.diffuse.r = 0.0f;
	tempmat.diffuse.g = 1.0f;
	tempmat.diffuse.b = 1.0f;
	tempmat.diffuse.a = 1.0f;
	
	tempmat.specular.r = 0.0f;
	tempmat.specular.g = 1.0f;
	tempmat.specular.b = 1.0f;
	tempmat.specular.a = 1.0f;

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[24], 6, 0);

	
    tempmat.ambient.r = 1.0f;
	tempmat.ambient.g = 0.0f;
	tempmat.ambient.b = 1.0f;
	tempmat.ambient.a = 1.0f;
	
	tempmat.diffuse.r = 1.0f;
	tempmat.diffuse.g = 0.0f;
	tempmat.diffuse.b = 1.0f;
	tempmat.diffuse.a = 1.0f;
	
	tempmat.specular.r = 1.0f;
	tempmat.specular.g = 0.0f;
	tempmat.specular.b = 1.0f;
	tempmat.specular.a = 1.0f;

	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[30], 6, 0);


	hr = m_pWrapper->m_pD3DDevice7->SetTexture( 0, prevTexture);	
	hr = m_pWrapper->m_pD3DDevice7->SetMaterial( &prevMaterial );
	
	return S_OK;
}


// -----------------------------------------------------------------------
// void CMirrorDemo::SetupCubeVerts(float scale)
//
//   Sets up the vertices for the cube using texture coordinates
// -----------------------------------------------------------------------
void CMirrorDemo::SetupCubeVerts(float scale)
{

	int i;
	memset( m_CubeVerts, 0, sizeof( D3DVERTEX ) * NUMCUBEVERTS );

	
	// Front face
	m_CubeVerts[0].x = -scale;		m_CubeVerts[0].y = -scale;		m_CubeVerts[0].z = -scale;
	m_CubeVerts[0].tu = 0.0f;		m_CubeVerts[0].tv = 1.0f;
	m_CubeVerts[1].x = -scale;		m_CubeVerts[1].y = scale;		m_CubeVerts[1].z = -scale;
	m_CubeVerts[1].tu = 0.0f;		m_CubeVerts[1].tv = 0.0f;
	m_CubeVerts[2].x = scale;		m_CubeVerts[2].y =  -scale;		m_CubeVerts[2].z = -scale;
	m_CubeVerts[2].tu = 1.0f;		m_CubeVerts[2].tv = 1.0f;

	m_CubeVerts[3] = m_CubeVerts[1];
	m_CubeVerts[3].tu = 0.0f;		m_CubeVerts[3].tv = 0.0f;
	m_CubeVerts[4].x = scale;		m_CubeVerts[4].y = scale;		m_CubeVerts[4].z = -scale;
	m_CubeVerts[4].tu = 1.0f;		m_CubeVerts[4].tv = 0.0f;
	m_CubeVerts[5] = m_CubeVerts[2];
	m_CubeVerts[5].tu = 1.0f;		m_CubeVerts[5].tv = 1.0f;

	for(i=0; i<6; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = -1.0f;
	}
	
	
	// Right Face
	m_CubeVerts[6] = m_CubeVerts[2];
	m_CubeVerts[6].tu = 0.0f;		m_CubeVerts[6].tv = 1.0f;
	m_CubeVerts[7]= m_CubeVerts[4];
	m_CubeVerts[7].tu = 0.0f;		m_CubeVerts[7].tv = 0.0f;
	m_CubeVerts[8].x = scale;		m_CubeVerts[8].y =  -scale;		m_CubeVerts[8].z = scale;
	m_CubeVerts[8].tu = 1.0f;		m_CubeVerts[8].tv = 1.0f;
	m_CubeVerts[9] = m_CubeVerts[4];
	m_CubeVerts[9].tu = 0.0f;		m_CubeVerts[9].tv = 0.0f;
	m_CubeVerts[10].x = scale;		m_CubeVerts[10].y = scale;		m_CubeVerts[10].z = scale;
	m_CubeVerts[10].tu = 1.0f;		m_CubeVerts[10].tv = 0.0f;
	m_CubeVerts[11] = m_CubeVerts[8];
	m_CubeVerts[11].tu = 1.0f;		m_CubeVerts[11].tv = 1.0f;
	
	for(i=6; i<12; i++)
	{
		m_CubeVerts[i].nx = 1.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = 0.0f;
	}
	
	// Top face
	m_CubeVerts[12] = m_CubeVerts[1];
	m_CubeVerts[12].tu = 0.0f;	m_CubeVerts[12].tv = 0.0f;
	m_CubeVerts[13].x = -scale;	m_CubeVerts[13].y = scale;		m_CubeVerts[13].z = scale;
	m_CubeVerts[13].tu = 1.0f;	m_CubeVerts[13].tv = 0.0f;
	m_CubeVerts[14] = m_CubeVerts[4];
	m_CubeVerts[14].tu = 0.0f;	m_CubeVerts[14].tv = 1.0f;
	m_CubeVerts[15] = m_CubeVerts[4];
	m_CubeVerts[15].tu = 0.0f;	m_CubeVerts[15].tv = 1.0f;
	m_CubeVerts[16] = m_CubeVerts[13];
	m_CubeVerts[16].tu = 1.0f;	m_CubeVerts[16].tv = 0.0f;
	m_CubeVerts[17] = m_CubeVerts[10];
	m_CubeVerts[17].tu = 1.0f;	m_CubeVerts[17].tv = 1.0f;
	
	for(i=12; i<18; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = 1.0f;
		m_CubeVerts[i].nz = 0.0f;
	}

	// Left face
	m_CubeVerts[18] = m_CubeVerts[0];
	m_CubeVerts[18].tu = 1.0f;	m_CubeVerts[18].tv = 1.0f;
	m_CubeVerts[19] = m_CubeVerts[13];
	m_CubeVerts[19].tu = 0.0f;	m_CubeVerts[19].tv = 0.0f;
	m_CubeVerts[20] = m_CubeVerts[1];
	m_CubeVerts[20].tu = 1.0f;	m_CubeVerts[20].tv = 0.0f;
	m_CubeVerts[21] = m_CubeVerts[0];
	m_CubeVerts[21].tu = 1.0f;	m_CubeVerts[21].tv = 1.0f;
	m_CubeVerts[22].x = -scale;	m_CubeVerts[22].y = -scale;		m_CubeVerts[22].z = scale;
	m_CubeVerts[22].tu = 0.0f;	m_CubeVerts[22].tv = 1.0f;
	
	m_CubeVerts[23] = m_CubeVerts[13];
	m_CubeVerts[23].tu = 0.0f;	m_CubeVerts[23].tv = 0.0f;

	for(i=18; i<24; i++)
	{
		m_CubeVerts[i].nx = -1.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = 0.0f;
	}

	// Back face
	m_CubeVerts[24] = m_CubeVerts[8];
	m_CubeVerts[24].tu = 0.0f;	m_CubeVerts[24].tv = 1.0f;
	m_CubeVerts[25] = m_CubeVerts[13];
	m_CubeVerts[25].tu = 1.0f;	m_CubeVerts[25].tv = 0.0f;
	m_CubeVerts[26] = m_CubeVerts[22];
	m_CubeVerts[26].tu = 1.0f;	m_CubeVerts[26].tv = 1.0f;
	m_CubeVerts[27] = m_CubeVerts[8];
	m_CubeVerts[27].tu = 0.0f;	m_CubeVerts[27].tv = 1.0f;
	m_CubeVerts[28] = m_CubeVerts[10];
	m_CubeVerts[28].tu = 0.0f;	m_CubeVerts[28].tv = 0.0f;
	m_CubeVerts[29] = m_CubeVerts[13];
	m_CubeVerts[29].tu = 1.0f;	m_CubeVerts[29].tv = 0.0f;
	
	for(i=24; i<30; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = 1.0f;
	}

	
	// Bottom face
	m_CubeVerts[30] = m_CubeVerts[2];
	m_CubeVerts[30].tu = 0.0f;	m_CubeVerts[30].tv = 1.0f;
	m_CubeVerts[31] = m_CubeVerts[8];
	m_CubeVerts[31].tu = 0.0f;	m_CubeVerts[31].tv = 0.0f;
	m_CubeVerts[32] = m_CubeVerts[22];
	m_CubeVerts[32].tu = 1.0f;	m_CubeVerts[32].tv = 0.0f;
	m_CubeVerts[33] = m_CubeVerts[2];
	m_CubeVerts[33].tu = 0.0f;	m_CubeVerts[33].tv = 1.0f;
	m_CubeVerts[34] = m_CubeVerts[22];
	m_CubeVerts[34].tu = 1.0f;	m_CubeVerts[34].tv = 0.0f;
	m_CubeVerts[35] = m_CubeVerts[0];
	m_CubeVerts[35].tu = 1.0f;	m_CubeVerts[35].tv = 1.0f;
	
	for(i=30; i<36; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = -1.0f;
		m_CubeVerts[i].nz = 0.0f;
	}

	return;
}



// -----------------------------------------------------------------------
// void CMirrorDemo::SetupMirrorVerts(float scale)
//
//   Sets up the vertices for the mirror
// -----------------------------------------------------------------------
void CMirrorDemo::SetupMirrorVerts(float scale)
{

	memset( m_MirrorVerts, 0, sizeof( D3DLVERTEX ) * NUMMIRRORVERTS );

	m_MirrorVerts[0].x = -scale;	m_MirrorVerts[0].y = -scale;	m_MirrorVerts[0].z = 0.0f; 
	m_MirrorVerts[1].x = scale;		m_MirrorVerts[1].y = scale;		m_MirrorVerts[1].z = 0.0f; 
	m_MirrorVerts[2].x = scale;		m_MirrorVerts[2].y = -scale;	m_MirrorVerts[2].z = 0.0f;

	m_MirrorVerts[3].x = -scale;	m_MirrorVerts[3].y = -scale;	m_MirrorVerts[3].z = 0.0f;
	m_MirrorVerts[4].x = -scale;	m_MirrorVerts[4].y = scale;		m_MirrorVerts[4].z = 0.0f;
	m_MirrorVerts[5].x = scale;		m_MirrorVerts[5].y = scale;		m_MirrorVerts[5].z = 0.0f; 
	
	
	//
	// Here we set up the texture coordinates in a special way
	// In order to mimic a reflection correctly, we must flip the
	// texture rendered from the mirror's viewpoint over the Y-Axis.
	//
	m_MirrorVerts[0].tu = 1.0f; m_MirrorVerts[0].tv = 1.0f;
	m_MirrorVerts[1].tu = 0.0f; m_MirrorVerts[1].tv = 0.0f;
	m_MirrorVerts[2].tu = 0.0f; m_MirrorVerts[2].tv = 1.0f;
	
	m_MirrorVerts[3].tu = 1.0f; m_MirrorVerts[3].tv = 1.0f;
	m_MirrorVerts[4].tu = 1.0f; m_MirrorVerts[4].tv = 0.0f;
	m_MirrorVerts[5].tu = 0.0f; m_MirrorVerts[5].tv = 0.0f;


	// The mirror should be at full brightness
	for (int i=0;i<NUMMIRRORVERTS;i++)
	{
		m_MirrorVerts[i].color = D3DRGBA(1.0f, 1.0f, 1.0f, 1.0f);
	}

	return;
}



// -----------------------------------------------------------------------
// void CMirrorDemo::SetupFloorVerts(float scale)
//
//   Sets up the vertices for the floor
// -----------------------------------------------------------------------
void CMirrorDemo::SetupFloorVerts(float scale)
{
	memset( m_FloorVerts, 0, sizeof( D3DVERTEX ) * NUMFLOORVERTS );

	m_FloorVerts[0].x = -scale;		m_FloorVerts[0].y = -1.0f;	m_FloorVerts[0].z = -scale; 
	m_FloorVerts[1].x = scale;		m_FloorVerts[1].y = -1.0f;	m_FloorVerts[1].z = scale; 
	m_FloorVerts[2].x = scale;		m_FloorVerts[2].y = -1.0f;	m_FloorVerts[2].z = -scale;

	m_FloorVerts[3].x = -scale;		m_FloorVerts[3].y = -1.0f;	m_FloorVerts[3].z = -scale;
	m_FloorVerts[4].x = -scale;		m_FloorVerts[4].y = -1.0f;	m_FloorVerts[4].z = scale;
	m_FloorVerts[5].x = scale;		m_FloorVerts[5].y = -1.0f;	m_FloorVerts[5].z = scale; 
	
	
	m_FloorVerts[0].tu = 0.0f; m_FloorVerts[0].tv = 1.0f;
	m_FloorVerts[1].tu = 1.0f; m_FloorVerts[1].tv = 0.0f;
	m_FloorVerts[2].tu = 1.0f; m_FloorVerts[2].tv = 1.0f;
	
	m_FloorVerts[3].tu = 0.0f; m_FloorVerts[3].tv = 1.0f;
	m_FloorVerts[4].tu = 0.0f; m_FloorVerts[4].tv = 0.0f;
	m_FloorVerts[5].tu = 1.0f; m_FloorVerts[5].tv = 0.0f;

	// common elements
	for (int i=0;i<NUMFLOORVERTS;i++)
	{
		m_FloorVerts[i].nx = 0.0f;
		m_FloorVerts[i].ny = 1.0f;
		m_FloorVerts[i].nz = 0.0f;
	}

	return;

}






// -----------------------------------------------------------------------
// void CMirrorDemo::KeyUp(WPARAM wParam)
//
//   Keyboard keypress handler. 
// -----------------------------------------------------------------------
void CMirrorDemo::KeyDown(WPARAM wParam)
{
	//we need a variable to modify our rotation speed with frame rate	
	float scaler = 1.0f;

	//if exiting, the wrapper could be invalid, but we don't care, since we will exit anyway
	if (wParam == 'X') 
		return;

	if(m_pWrapper) //we could get into this loop after the wrapper has been invalidates
	{
		if (m_pWrapper->m_fFrameTime >0.0f)
			scaler = m_pWrapper->m_fFrameTime;  
	}



	switch (wParam) 
	{
	case VK_UP:
		if (!m_bRotateMirror)
		{
			if( !m_bAutoRotateCube)
				m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._11, m_MxCubeTransform.m_Mx._21, m_MxCubeTransform.m_Mx._31, -3.14159f*scaler );
		}
		else
			m_MxMirrorRotate.Rotate( DRGVector(1,0,0), -3.14159f*scaler );
		break;

	case VK_DOWN:
		if (!m_bRotateMirror)
		{
			if( !m_bAutoRotateCube)
				m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._11, m_MxCubeTransform.m_Mx._21, m_MxCubeTransform.m_Mx._31, 3.14159f*scaler );
		}
		else
		{
			m_MxMirrorRotate.Rotate( DRGVector(1,0,0), 3.14159f*scaler );
		}
		break;

	case VK_LEFT:
		if (!m_bRotateMirror)
		{
			if( !m_bAutoRotateCube)
				m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._12, m_MxCubeTransform.m_Mx._22, m_MxCubeTransform.m_Mx._32, -3.14159f*scaler );
		}
		else
			m_MxMirrorRotate.Rotate( m_MxMirrorRotate.m_Mx._12, m_MxMirrorRotate.m_Mx._22, m_MxMirrorRotate.m_Mx._32, -3.14159f*scaler );
		break;

	case VK_RIGHT:	
		if (!m_bRotateMirror)
		{
			if( !m_bAutoRotateCube)
				m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._12, m_MxCubeTransform.m_Mx._22, m_MxCubeTransform.m_Mx._32, 3.14159f*scaler );
		}
		else
			m_MxMirrorRotate.Rotate( m_MxMirrorRotate.m_Mx._12, m_MxMirrorRotate.m_Mx._22, m_MxMirrorRotate.m_Mx._32, 3.14159f*scaler );
		break;

	// Toggle cube auto-rotation
	case VK_SPACE:
		m_bAutoRotateCube = !m_bAutoRotateCube;
		break;

	// C, M, and O toggles drawing the objects
	case 'C':
		m_bDrawCube = !m_bDrawCube;		
		break;
	case 'M':
		m_bDrawMirror = !m_bDrawMirror;
		break;
	case 'O':
		m_bDrawFloor = !m_bDrawFloor;
		break;

	// Toggle between mirror and cube rotation with
	// the arrow keys
	case 'T':
		m_bRotateMirror = !m_bRotateMirror;
		break;
	
	// View the scene from the mirror's viewpoint
	// displaying what the dynamic texture looks like
	case 'V':
		m_bViewFromMirror = !m_bViewFromMirror;
		break;

	// Change the reflection update frequency
	// J - Decrement the frequency
	// K - Increment the frequency
	case 'J':
		m_iReflectionUpdateFrequency++;
		break;
	case 'K':
		if ( m_iReflectionUpdateFrequency > 1 )
			m_iReflectionUpdateFrequency--;
		break;

	
	// Change the size of the texture being used for the reflection	
	// If the texture is currently 512x512, change to 32x32. Otherwise
	// Just go to the next higher size
	case 'N':
		UpdateMirrorTextureSize( m_iTextureSize == 512 ? 32 : m_iTextureSize * 2);
		break;

	// Mirror center to virtual camera center
	case '1':
		m_bDrawMirrorToCameraReflection = !m_bDrawMirrorToCameraReflection;
		break;
	case '2':
		m_bDrawMirrorNormal = !m_bDrawMirrorNormal;
		break;
	case '3':
		m_bDrawCamUpVector  = !m_bDrawCamUpVector;
		break;
	// Crosshairs on mirror
	case '4':
		m_bDrawCrosshairs = !m_bDrawCrosshairs;
		break;
	// View frustums before and after the shear
	case '5':
		m_bDrawViewFrustum = !m_bDrawViewFrustum;
		break;

	}

	// Debug output about the reflection update frequency
	if ( wParam == 'J' || wParam == 'K' )
	{
		char string[50];
		sprintf( string, "Reflection Updated Every %d Frame(s)\n", m_iReflectionUpdateFrequency );
		OutputDebugString( string );
	}
	
	// Debug output about the mirror reflection texture size
	if ( wParam == 'N' )
	{
		char string[50];
		sprintf(string, "Texture Size = %d x %d\n", m_iTextureSize, m_iTextureSize );
		OutputDebugString( string );
	}

	if (m_pWrapper)
	{
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxCubeTransform.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );
	}

	
	return;
}

void CMirrorDemo::KeyUp(WPARAM wParam)
{		
	switch (wParam) 
	{
		//add code here to handle key releases, if you need them (as in remmed out example)
	case VK_UP:
		//OutputDebugString("The up arrow was released;");			
		break;
	}
	
	return;
}


void CMirrorDemo::MouseLeftUp(WPARAM wParam, LPARAM lParam)
{	
	return;
}

void CMirrorDemo::MouseLeftDown(WPARAM wParam, LPARAM lParam)
{		
	return;
}

void CMirrorDemo::MouseRightUp(WPARAM wParam, LPARAM lParam)
{	
	return;
}

void CMirrorDemo::MouseRightDown(WPARAM wParam, LPARAM lParam)
{		
	return;
}

void CMirrorDemo::MouseMove(WPARAM wParam, LPARAM lParam)
{
	if ( wParam & MK_LBUTTON )	// Left button down plus mouse move == rotate
	{
		//we need a variable to modify our rotation speed with frame rate	
		float scaler = 1.0f;
		if(m_pWrapper) //we could get into this loop after the wrapper has been invalidates
		{
			scaler = 2.0f / max(m_pWrapper->m_dwRenderWidth, m_pWrapper->m_dwRenderHeight);  
		}

		//we need to compare the last position to the current position
		int x = LOWORD(lParam);
		int y = HIWORD(lParam);
		
		if (x - m_iMousePrevX)
		{
			m_MxView.Rotate( m_MxView.m_Mx._12, m_MxView.m_Mx._22, m_MxView.m_Mx._32, 3.14159f*scaler*(float)(x - m_iMousePrevX) );
		}
			
		if (y - m_iMousePrevY)
		{
			m_MxView.Rotate( m_MxView.m_Mx._11, m_MxView.m_Mx._21, m_MxView.m_Mx._31, 3.14159f*scaler*(float)(y - m_iMousePrevY) );
		}

		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );

	
		m_iMousePrevX = x;
		m_iMousePrevY = y;

	}
	else if ( wParam & MK_RBUTTON )	// Right button down plus mouse move == Zoom camera in/out
	{
		//we need a variable to modify our rotation speed with frame rate	
		float scaler = 1.0f;
		if(m_pWrapper) //we could get into this loop after the wrapper has been invalidates
		{
			scaler = 5.0f / max(m_pWrapper->m_dwRenderWidth, m_pWrapper->m_dwRenderHeight);  
		}

		//we need to compare the last position to the current position
		int x = LOWORD(lParam);
		int y = HIWORD(lParam);
		
		if (y - m_iMousePrevY)
		{
			DRGMatrix Temp(true);

			Temp.Translate( 0,0,scaler * (y - m_iMousePrevY) );
			m_MxView.PostMultiply( Temp );
		}

		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );

	
		m_iMousePrevX = x;
		m_iMousePrevY = y;

	}
	else
	{
		//need to update these, or the next time someone clicks, the delta is massive
		m_iMousePrevX = LOWORD(lParam);
		m_iMousePrevY = HIWORD(lParam);	
	}
	return;
}

