// Sphere.cpp
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
//  PURPOSE:
//    
// Sphere.cpp: Sphere class implementation
// This class is used to create a sphere object
//
// ------------------------------------------------------------------------------------
//
// Author: Mike Rohman
// 
////////////////////////////////////////////////////////////////////////////////////////


#include "DRGShell.h"


// -----------------------------------------------------------------------
// Sphere::Sphere()
//
// Initialize a sphere to be drawn using DrawPrimitive,
// use unoptimized vertex buffers and rotate
// -----------------------------------------------------------------------
Sphere::Sphere()
{
	// The sphere verts and indices
	m_pIndexedVertices = NULL;
	m_pVertexIndices = NULL;

	// The sphere vertices transformed into
	// the reflection view space
	m_pRVSVertices = NULL;

}



// -----------------------------------------------------------------------
// Sphere::~Sphere()
//
//  This destructor releases the memory set aside for the vertex buffer
//  and for the array of vertices
// -----------------------------------------------------------------------
Sphere::~Sphere()
{
	delete[] m_pRVSVertices;
	delete[] m_pIndexedVertices;
	delete[] m_pVertexIndices;
}


// -----------------------------------------------------------------------
// HRESULT Sphere::DrawSphere()
//
//  This draws a sphere to the screen using DrawPrimitive. 
// -----------------------------------------------------------------------
HRESULT Sphere::DrawSphere()
{
	
	HRESULT hr;

	// Rotate the spheres the proper
	// amount independent of the framerate

	
	// The sphere maintains 2 sets of texture coordinates
	// One for the marble texture, another for the reflection texture
	if ( FAILED ( hr = m_pWrapper->m_pD3DDevice7->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX2,
										                                m_pIndexedVertices, 
																		m_iNumIndexedVertices,
																	    m_pVertexIndices, 
																	    m_iNumVertexIndices, 0) ) )
			return hr;
	

	
	return S_OK;
}



// -----------------------------------------------------------------------
// HRESULT Sphere::Init( float rad, int nDiv, CDRGWrapper pWrapper )
//
//  This function sets up all of the vertex structures used for the sphere
//  and also sets up the local matrix the sphere maintains. The sphere will
//  be located at (x, y, z) and have a radius of r and have nDiv divisions.  
//  The number of divisions is the number of longitude strips and also the 
//  number of latitude strips
// -----------------------------------------------------------------------
HRESULT Sphere::Init( float rad, int nDiv, CDRGWrapper *pWrapper )
{
	HRESULT hr;

	// Radius
	m_fRadius = rad;
	
	// Number of longitudinal and latitudinal sections
	// to use in rendering the sphere
	m_iNumDiv = nDiv;
	
	m_pWrapper = pWrapper;

	// Calculate the number of vertices and indices required for the sphere
	m_iNumIndexedVertices = ((m_iNumDiv) * 2) + ((m_iNumDiv + 1) * (m_iNumDiv - 1));
	m_iNumVertexIndices = 6 * (m_iNumDiv + (m_iNumDiv * (m_iNumDiv-2)));
	
	// Create vertices that maintain 2 sets of texture coordinates
	m_pIndexedVertices =	new DualTexVertex[m_iNumIndexedVertices];
	
	m_pVertexIndices =		new WORD[m_iNumVertexIndices];
	m_pRVSVertices =		new DRGVector[m_iNumIndexedVertices];
		
	if ( FAILED( hr = SetupIndexedVertices() ) )
		return hr;

	if ( FAILED( hr = SetupIndexedIndices() ) )
		return hr;

	return S_OK;

}



// -----------------------------------------------------------------------
// HRESULT Sphere::SetupIndexedVertices()
//
//  Given a pointer to the array of vertices that make up the sphere,
//  this function sets up the vertex locations and normals
// -----------------------------------------------------------------------
HRESULT Sphere::SetupIndexedVertices()
{

	// heh heh...
	float pi = 3.14159265358979323846f;

	
	// theta is the current latitudinal angle
	float theta = pi / 2.0f;
	
	// phi is the current longitudinal angle
	float phi = 0.0f;
	
	
	// delta values for the angles depending
	// on the number of divisions
	float dTheta = pi / m_iNumDiv;
	float dPhi = 2.0f * pi / m_iNumDiv;

	int curVert = 0;
	float x, y, z;

	// Initialize the vertices at the top of the sphere
	for(int i=0; i<m_iNumDiv; i++)
	{
		// We need multiple vertices at the top position in order
		// to correctly map a texture to the sphere
		m_pIndexedVertices[curVert].x = m_pIndexedVertices[curVert].z
									  = m_pIndexedVertices[curVert].nx 
									  = m_pIndexedVertices[curVert].nz = 0.0f;
		m_pIndexedVertices[curVert].y = m_fRadius;
		m_pIndexedVertices[curVert].ny = 1.0f;

		// Set the texture coordinates used for the base material
		m_pIndexedVertices[curVert].tu = (float)i / m_iNumDiv;
		m_pIndexedVertices[curVert++].tv = 0.0f;

	}

	
	// Intialize the vertices that make up the middle section
	for(i = 1; i<m_iNumDiv; i++)
	{
		theta += dTheta;	// Longitude angle
		phi = 0.0f;			// Latitude angle
		
		
		// Iterates m_iNumDiv+1 times because we need an extra vertex at
		// each latitude to map a texture completely around the sphere
		for(int j=0; j<m_iNumDiv+1; j++)
		{
			x = (float)cos(theta) * (float)cos(phi);
			y = (float)sin(theta);
			z = (float)cos(theta) * (float)sin(phi);
			
			m_pIndexedVertices[curVert].x = m_fRadius * x;
			m_pIndexedVertices[curVert].y = m_fRadius * y;
			m_pIndexedVertices[curVert].z = m_fRadius * z;
			m_pIndexedVertices[curVert].nx = x;
			m_pIndexedVertices[curVert].ny = y;
			m_pIndexedVertices[curVert].nz = z;
			
			m_pIndexedVertices[curVert].tu = (float)j / m_iNumDiv;
			m_pIndexedVertices[curVert++].tv = (float)i / m_iNumDiv;


			phi += dPhi;
		}
	
	}


	// Initialize the bottom section vertices
	for(i = 0; i<m_iNumDiv; i++)
	{
		m_pIndexedVertices[curVert].x = m_pIndexedVertices[curVert].z
									  = m_pIndexedVertices[curVert].nx
									  = m_pIndexedVertices[curVert].nz = 0.0f;
		m_pIndexedVertices[curVert].y = -m_fRadius;
		m_pIndexedVertices[curVert].ny = -1.0f;
		
		m_pIndexedVertices[curVert].tu = (float)i / m_iNumDiv;
		m_pIndexedVertices[curVert++].tv = 1.0f;

	}
		

	return S_OK;  
	
}	



// -----------------------------------------------------------------------
// HRESULT Sphere::SetupIndexedVertices()
//
//  Given a pointer to the array of vertex indices, this function sets up
//  the correct indices in that array
// -----------------------------------------------------------------------
HRESULT Sphere::SetupIndexedIndices()
{
	
	int iCurLat, iCurLong;
				
	int curVert = 0;

	// Top section
	for(iCurLat=0; iCurLat<m_iNumDiv; iCurLat++)
	{

		m_pVertexIndices[curVert++] = iCurLat;
		
		// Skip past the vertices at the top of the sphere (+m_iNumDiv)
		// and also skip to the next vertice (+1)
		m_pVertexIndices[curVert++] = m_iNumDiv + (iCurLat + 1);

		// Skip past the vertices at the top of the sphere (+m_iNumDiv)
		m_pVertexIndices[curVert++] = m_iNumDiv + iCurLat;	

		
	}


	// Middle section
	for(iCurLat=1; iCurLat<m_iNumDiv-1; iCurLat++)
	{
		for(iCurLong=0; iCurLong<m_iNumDiv; iCurLong++)
		{
			
			// There are m_iNumDiv vertices at each latitude so we need to 
			// multiply by that amount to skip to a specific latitude
			// *----* <- Each quadrilateral space is composed of 2 triangles
		    // | \ 2|    Triangle 1 is initialized after Triangle 2
			// |1  \|    
			// *----*    

			// first triangle of the quad area
			m_pVertexIndices[curVert++] = (iCurLat * m_iNumDiv) + (iCurLat-1) + iCurLong;
			m_pVertexIndices[curVert++] = ((iCurLat + 1) * m_iNumDiv) + iCurLat + iCurLong + 1;
			m_pVertexIndices[curVert++] = ((iCurLat + 1) * m_iNumDiv) + iCurLat + iCurLong;

			// second triangle of the quad area
			m_pVertexIndices[curVert++] = (iCurLat * m_iNumDiv) + (iCurLat-1) + iCurLong;
			m_pVertexIndices[curVert++] = (iCurLat * m_iNumDiv) + (iCurLat-1) + iCurLong + 1;
			m_pVertexIndices[curVert++] = ((iCurLat + 1) * m_iNumDiv) + iCurLat + iCurLong + 1;
		}
	}


	// Bottom section
	for(iCurLat = 0; iCurLat<m_iNumDiv; iCurLat++)
	{
		// Go to the second to last latitude (last is the bottom-most point)
		m_pVertexIndices[curVert++] = ((m_iNumDiv * m_iNumDiv) - 2) + iCurLat;
		m_pVertexIndices[curVert++] = ((m_iNumDiv * m_iNumDiv) - 2) + (iCurLat + 1);	
		m_pVertexIndices[curVert++] = ((m_iNumDiv * (m_iNumDiv + 1)) - 1) + iCurLat;
	}
	
	return S_OK;
}


// -----------------------------------------------------------------------
// void Sphere::UpdateTextureCoords( DRGMatrix mirrorViewMatrix, bool straightProjection )
//
//  This function updates the texture coordinates for the sphere
// -----------------------------------------------------------------------
void Sphere::UpdateTextureCoords( DRGMatrix mirrorViewMatrix, DRGMatrix mirrorTransform, 
								  DRGVector cameraCenter)
{
	float fMaxX, fMaxY, fMinX, fMinY;			// Max/Min texture coordinate values
												// used for scaling tex-coords to [0,1]
	
	float u, v, phi;							// used in case 0
	
	DRGVector refVec;						
	DRGVector referenceVec(0,0,1);				
	DRGVector mirrorToCamera;
	DRGVector mirrorToCameraXZ;
	DRGVector mirrorToCameraYZ;
	DRGVector vertexVectorXZ;
	DRGVector vertexVectorYZ;
	DRGVector cameraToMirror;
	DRGVector mirrorNormal;
	DRGVector reflectionVector;
	
	bool bAngleXZZero = false;
	bool bAngleYZZero = false;

	int i;

	// Cycle through all of the vertices
	for(i=0; i<m_iNumIndexedVertices; i++)
	{
			
			
		// Copy the D3DVERTEX vertices to DRGVector vertices in order to be able
		// multiply by matrices. m_pRVSVertices are transformed to view space.
		// Normalize the vertex coords in order to use them as unit vectors.
		m_pRVSVertices[i] = Normalize(DRGVector( m_pIndexedVertices[i].x,
												 m_pIndexedVertices[i].y,
												 m_pIndexedVertices[i].z ) * mirrorTransform * mirrorViewMatrix);
			
		// Make all of the vertex Z values positive otherwise the
		// texture would be wrapped around the entire sphere. We
		// only want the texture to be wrapped around the visible 
		// portion of the sphere
		//
		if (m_pRVSVertices[i].z < 0.0f)
			m_pRVSVertices[i].z = -m_pRVSVertices[i].z;

		
		
			// Get the camera to mirror vector
			cameraToMirror = Normalize(m_pRVSVertices[i]-(cameraCenter * mirrorViewMatrix));
			// The mirror normal for getting the reflection vector is m_pRVSVertices[i]
			reflectionVector = Normalize(( ( 2 * m_pRVSVertices[i])  * DotProduct( m_pRVSVertices[i], cameraToMirror ) ) - cameraToMirror);
			
			// This is the method Dean had me try out
			phi = (float)acos(reflectionVector.y);
			v = phi / 3.1415926f;

			// If v is zero, or v is 1, the u texture coord is 0
			if ( ( ( v < 1e-6 ) && ( v > -1e-6 ) ) || ( ( v < 1.0+1e-6 ) && ( v > 1.0-1e-6 ) ) )
				u = 0.0f;
			else
			{
				// store this result to check to see if it is invalid
				// as input to arc cosine
				float fResult = reflectionVector.x / (float)sin(phi);
				
				// Sometimes the result is invalid as input to the
				// arc cos function (over 1.0, under -1.0).
				// When this occurs, the result is very close to being
				// valid. This occurance is probably due to rounding
				// error. When this does occur, I simply made the result
				// valid by changing it's value.
				//
				if (fResult > 1.0f-1e-6)
				{
					//char string[20];
					//sprintf(string, "over: %f\n", fResult);
					//OutputDebugString(string);
					fResult = 1.0f - (float)1e-6;
				}

				if (fResult < -(1.0f-1e-6) )
				{
					//char string[20];
					//sprintf(string, "under: %f\n", fResult);
					//OutputDebugString(string);
					fResult = -(1.0f - (float)1e-6);
				}

				u = (float)acos(fResult)/3.1415926f;
			}

			// texture coordinates are stored back into the vertex array
			// scaled to [0,1] and then placed back into the D3DVERTEX array
			m_pRVSVertices[i].x = u;
			m_pRVSVertices[i].y = -v;  // flip the sign of the v tex coord
									   // to get a correct mapping
		

	}
	
	
	// set fMaxX and fMaxY to the first values in the array
	fMaxX = fMinX = m_pRVSVertices[0].x;
	fMaxY = fMinY = m_pRVSVertices[0].y;

	// Get the largest and smallest values that exist
	// and multiply each vertex by the viewMatrix to put
	// the vertices in view coordinates
	for(i=1; i<m_iNumIndexedVertices; i++)
	{
		if (m_pRVSVertices[i].x > fMaxX)
			fMaxX = m_pRVSVertices[i].x;
		if (m_pRVSVertices[i].y > fMaxY)
			fMaxY = m_pRVSVertices[i].y;
		if (m_pRVSVertices[i].x < fMinX)
			fMinX = m_pRVSVertices[i].x;
		if (m_pRVSVertices[i].y < fMinY)
			fMinY = m_pRVSVertices[i].y;
	
	}
	

	// Use the maximum value and minimum value to scale the values to
	// range between 0.0 and 1.0
	for(i=0; i<m_iNumIndexedVertices; i++)
	{
		// Update the texture coords used for the reflection
		m_pIndexedVertices[i].tu2 = (m_pRVSVertices[i].x - fMinX)/(fMaxX-fMinX);
		m_pIndexedVertices[i].tv2 = (m_pRVSVertices[i].y - fMinY)/(fMaxY-fMinY);
	}


}


