// SphericalMirror.cpp
// ------------------------------------------------------------------------------------
// Copyright  1999 Intel Corporation
// All Rights Reserved
// 
// Permission is granted to use, copy, distribute and prepare derivative works of this 
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  This software is provided "AS IS." 
//
// Intel specifically disclaims all warranties, express or implied, and all liability,
// including consequential and other indirect damages, for the use of this software, 
// including liability for infringement of any proprietary rights, and including the 
// warranties of merchantability and fitness for a particular purpose.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
// ------------------------------------------------------------------------------------
//
//  PURPOSE:
//    
// SphericalMirror.h: implementation the SphericalMirror class.
// This class is the "blank" class created by the DRG 3D Appwizard, where you will put your stuff
//
// ------------------------------------------------------------------------------------
//
// Author: Kim Pallister - Intel Developer Relations & Engineering Group
//
////////////////////////////////////////////////////////////////////////////////////////

#include "DRGShell.h"


// -----------------------------------------------------------------------
// CSphericalMirror::CSphericalMirror() 
// The SphericalMirror class constructor initializes some of it's member vars (which
// are instances of the DRGMatrix class
// -----------------------------------------------------------------------
CSphericalMirror::CSphericalMirror()
{
	m_MxWorld.SetIdentity();
	//set the camera at 0,0,-5; the target at 0,0,0 and the up vector at 0,1,0
	m_MxView.SetView( DRGVector( 0, 0, -6 ), DRGVector( 0, 0, 0 ), DRGVector( 0, 1, 0 ) );
	//set the projection matrix to 1.57 field of view (radians), 1.33 aspect ratio 
	// (640x480), //the near clip plane to 1 and the far to clip plane 100
	m_MxProjection.SetProjection( 1.57f, 1.0f, 1.0f, 100.0f );

	
	m_drgvMirrorCenter = DRGVector( 0.0f, 0.0f, 0.0f );
	
	// Central cube center
	m_drgvCubeCenter = DRGVector( 0.0f, 0.0f, -3.0f );

	
	// Set the cubes' rotational
	// speeds in rotations/second
	//
	m_fRotPerSecX = 0.2f;
	m_fRotPerSecY = 0.3f;
	m_fRotPerSecZ = 0.14f;

	
	// The spheres' rotational
	// speeds
	m_fSphereRotPerSecX = 0.0f;
	m_fSphereRotPerSecY = 0.1f;
	m_fSphereRotPerSecZ = 0.0f;

	m_fMirrorRadius = 1.5f;
	
	// The field of view for the mirror reflection
	m_fFov = 3.1415926f/1.5f;	
	
	// Mirror local matrix
	m_MxMirrorWorld.SetIdentity();
	m_MxMirrorTransform.SetIdentity();
	m_MxMirrorTransform.Translate( m_drgvMirrorCenter.x, m_drgvMirrorCenter.y, m_drgvMirrorCenter.z );

	// Cube local matrices
	m_MxCubeWorld.SetIdentity();
	m_MxCubeTransform.SetIdentity();

	m_MxCubeTransform.Translate( m_drgvCubeCenter.x, m_drgvCubeCenter.y, m_drgvCubeCenter.z );
	m_MxLeftCubeWorld.SetIdentity();
	m_MxLeftCubeTransform.SetIdentity();
	m_MxLeftCubeTransform.Translate( m_drgvCubeCenter.x - 2.0f, m_drgvCubeCenter.y, m_drgvCubeCenter.z );
	
	m_MxRightCubeWorld.SetIdentity();
	m_MxRightCubeTransform.SetIdentity();
	m_MxRightCubeTransform.Translate( m_drgvCubeCenter.x + 2.0f, m_drgvCubeCenter.y, m_drgvCubeCenter.z );

	m_pMirrorTexture = NULL;
	m_pFloorTexture = NULL;
	m_pCubeTexture = NULL;
	m_pBaseTexture = NULL;

	m_iTextureSize = 512;

	m_bViewFromMirror = false;	// controls whether to draw scene from the mirror's
								// point of view or not

	m_bAutoRotateCube = false;
	m_bKeysRotateCube = true;
	
	m_bRotateSphere = false;

	m_bDrawFloor = true;
	m_bDrawCubes = true;
	m_bDrawMirror = true;
	m_bDrawWalls = true;
	
	// Begin the application showing a mirror sphere
	// 0 - Mirror
	// 1 - Normal sphere
	// 2 - Shiny sphere
	m_iTypeOfSphere = 2;


	// Update reflection every frame
	m_iReflectionUpdateFrequency = 1;
	m_iFrameNumber = 0;
}

// -----------------------------------------------------------------------
// CSphericalMirror::~CSphericalMirror()
// destructor does nothing for now.
// -----------------------------------------------------------------------
CSphericalMirror::~CSphericalMirror()
{
	delete m_pFloorTexture;
	delete m_pMirrorTexture;
	delete m_pCubeTexture;
	delete m_pBaseTexture;

}

// -----------------------------------------------------------------------
// HRESULT CSphericalMirror::InitWorld( CDRGWrapper *pWrapper )
//
// -----------------------------------------------------------------------
HRESULT CSphericalMirror::InitWorld( CDRGWrapper *pWrapper )
{
	//
	// Make local copy of the pointer to the wrapper class
	//
	m_pWrapper = pWrapper;

	// Initialize a material used by all objects
	//
	ZeroMemory( &m_Material, sizeof(m_Material) );
    m_Material.dcvAmbient.r = 1.0f;
	m_Material.dcvAmbient.g = 1.0f;
	m_Material.dcvAmbient.b = 1.0f;
	m_Material.dcvAmbient.a = 1.0f;
	
	m_Material.dcvDiffuse.r = 1.0f;
	m_Material.dcvDiffuse.g = 1.0f;
	m_Material.dcvDiffuse.b = 1.0f;
	m_Material.dcvDiffuse.a = 1.0f;
	
	m_Material.dcvSpecular.r = 1.0f;
	m_Material.dcvSpecular.g = 1.0f;
	m_Material.dcvSpecular.b = 1.0f;
	m_Material.dcvSpecular.a = 1.0f;
	m_Material.dvPower = (float)50.0;
	
	// Initialize a material used by all objects
	//
	ZeroMemory( &m_MaterialLowAlpha, sizeof(m_MaterialLowAlpha) );
    
	// A float the reflects the shininess of the base sphere
	m_fBaseShininess = 0.1f;

	m_MaterialLowAlpha.dcvAmbient.r = 1.0f;
	m_MaterialLowAlpha.dcvAmbient.g = 1.0f;
	m_MaterialLowAlpha.dcvAmbient.b = 1.0f;
	m_MaterialLowAlpha.dcvAmbient.a = m_fBaseShininess;
	
	m_MaterialLowAlpha.dcvDiffuse.r = 1.0f;
	m_MaterialLowAlpha.dcvDiffuse.g = 1.0f;
	m_MaterialLowAlpha.dcvDiffuse.b = 1.0f;
	m_MaterialLowAlpha.dcvDiffuse.a = m_fBaseShininess;
	
	m_MaterialLowAlpha.dcvSpecular.r = 1.0f;
	m_MaterialLowAlpha.dcvSpecular.g = 1.0f;
	m_MaterialLowAlpha.dcvSpecular.b = 1.0f;
	m_MaterialLowAlpha.dcvSpecular.a = m_fBaseShininess;
	m_MaterialLowAlpha.dvPower = (float)50.0;
	
	// Set up a default light
    ZeroMemory( &m_Light, sizeof(m_Light) );
    m_Light.dltType       = D3DLIGHT_POINT;
    m_Light.dcvDiffuse.r  = 1.0f;
    m_Light.dcvDiffuse.g  = 1.0f;
    m_Light.dcvDiffuse.b  = 1.0f;
    m_Light.dcvSpecular.r = 1.0f;
    m_Light.dcvSpecular.g = 1.0f;
    m_Light.dcvSpecular.b = 1.0f;
 
	m_Light.dcvAmbient.r  = 0.4f;
	m_Light.dcvAmbient.g  = 0.4f;
	m_Light.dcvAmbient.b  = 0.4f;
    
	m_Light.dvAttenuation0 = 1.0f; 
    m_Light.dvRange = D3DLIGHT_RANGE_MAX;
    m_Light.dvPosition.x = 2.0f;
	m_Light.dvPosition.y = 2.0f;
	m_Light.dvPosition.z = -3.0f;


	m_vpTexViewport.dwX = 0;
	m_vpTexViewport.dwY = 0;
	m_vpTexViewport.dwWidth = m_iTextureSize;
	m_vpTexViewport.dwHeight = m_iTextureSize;
	m_vpTexViewport.dvMinZ = 0.0f;
	m_vpTexViewport.dvMaxZ = 1.0f;

	
	// This function sets up stuff and restores textures after changing 
	// the device. We call it the first time in to the app.
	TargetChanged();


	float scale = 1.0f; 
	
	SetupFloorVerts(scale * 4.0f); 
	SetupCubeVerts(scale / 2.5f);
	SetupWallVerts(scale * 4.0f);
	

	m_sphSphere.Init( m_fMirrorRadius, 24, m_pWrapper);
	
	
	// Create a blank texture to render the reflected scene to
	// This texture gets mapped to the mirror surface
	//
	m_pMirrorTexture = new CDRGTexture;

	if( m_pMirrorTexture->CreateTexture( m_pWrapper, m_iTextureSize, m_iTextureSize, 16,
										 (DWORD)0, DRGTEXTR_RENDERTARGET, NULL ) == S_OK )
		m_pMirrorTexture->Restore(pWrapper->m_pD3DDevice7);


	m_pFloorTexture = new CDRGTexture;
	
	if( m_pFloorTexture->CreateTexture( CBMPLoader( CFileStream( "Images\\floor.bmp" ) ), (DWORD)0,
									    (DWORD)0/*DRGTEXTR_FORCEINAGP*/ ) == S_OK )
		m_pFloorTexture->Restore(pWrapper->m_pD3DDevice7);
	

	m_pCubeTexture = new CDRGTexture;
	
	if( m_pCubeTexture->CreateTexture( CBMPLoader( CFileStream( "Images\\test.bmp" ) ),
										(DWORD)0, (DWORD)0/*DRGTEXTR_FORCEINAGP*/ ) == S_OK )
		m_pCubeTexture->Restore(pWrapper->m_pD3DDevice7);


	m_pBaseTexture = new CDRGTexture;
	
	if( m_pBaseTexture->CreateTexture( CBMPLoader( CFileStream( "Images\\base.bmp" ) ),
								   (DWORD)0,(DWORD)0/*DRGTEXTR_FORCEINAGP*/ ) == S_OK )
		m_pBaseTexture->Restore(pWrapper->m_pD3DDevice7);



	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CSphericalMirror::DestroyWorld( void )
//
// -----------------------------------------------------------------------
HRESULT CSphericalMirror::DestroyWorld( void )
{
	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CSphericalMirror::UpdateWorld( DWORD dwTime, int InputFlag )
//
// -----------------------------------------------------------------------
HRESULT CSphericalMirror::UpdateWorld()
{
	
	// Auto rotate the cubes
	if (m_bAutoRotateCube)
	{
		m_MxCubeTransform.RotateX((float)m_fRotPerSecX * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxCubeTransform.RotateY((float)m_fRotPerSecY * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxCubeTransform.RotateZ((float)m_fRotPerSecZ * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxLeftCubeTransform.RotateX((float)m_fRotPerSecX * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxLeftCubeTransform.RotateY((float)m_fRotPerSecY * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxLeftCubeTransform.RotateZ((float)m_fRotPerSecZ * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxRightCubeTransform.RotateX((float)m_fRotPerSecX * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxRightCubeTransform.RotateY((float)m_fRotPerSecY * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxRightCubeTransform.RotateZ((float)m_fRotPerSecZ * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
	}

	
	// Rotate both the base sphere and the mirror reflection sphere
	// note: if only the base sphere is rotated, bands will appear in
	// the texture blend. When we rotate both, we get a sort of
	// jiggle in the reflection
	if (m_bRotateSphere)
	{
		m_MxMirrorTransform.RotateX((float)m_fSphereRotPerSecX * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxMirrorTransform.RotateY((float)m_fSphereRotPerSecY * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
		m_MxMirrorTransform.RotateZ((float)m_fSphereRotPerSecZ * 2.0f * 3.141592653f * m_pWrapper->m_fFrameTime);
	}

	// Update all of the local matrices
	m_MxCubeWorld =	m_MxCubeTransform * m_MxWorld;
	m_MxLeftCubeWorld =	m_MxLeftCubeTransform * m_MxWorld;
	m_MxRightCubeWorld = m_MxRightCubeTransform * m_MxWorld;
	m_MxMirrorWorld = m_MxMirrorTransform * m_MxWorld;


	// The following gets various vectors that we need in order to update
	// the sphere's texture coordinates and update the mirror view matrix

	// Get the camera's center location
	m_MxViewInverse = m_MxView;
	m_MxViewInverse.Invert();
	m_drgvCameraCenter = DRGVector( 0, 0, 0 ) * m_MxViewInverse;
	m_drgvCameraUpVector = Normalize( ( DRGVector( 0, 1, 0 ) * m_MxViewInverse ) - m_drgvCameraCenter );

	
	// Set up a new view from the camera's center location to the mirror's center location.
	// This is because the mirror can be positioned in places other than the origin.
	// This matrix allows us to get an up vector required for setting up the mirror reflection
	// view
	//
	m_MxCameraToMirrorViewInv.SetView(m_drgvCameraCenter, m_drgvMirrorCenter, m_drgvCameraUpVector );
	m_MxCameraToMirrorViewInv.Invert();
	m_drgvCameraToMirrorUpVector = Normalize( ( DRGVector( 0, 1, 0) * m_MxCameraToMirrorViewInv ) - m_drgvCameraCenter );
	

	// Set the mirror's view using the up vector from above
	m_MxMirrorView.SetView( m_drgvMirrorCenter, m_drgvCameraCenter, m_drgvCameraToMirrorUpVector );

		
	// Update the sphere's texture coordinates once per frame
	m_sphSphere.UpdateTextureCoords(m_MxMirrorView, m_MxMirrorTransform, 
		                            m_drgvCameraCenter);

												
	/*
	//test for available vidmem
	DDSCAPS2	ddscaps;
    ZeroMemory( &ddscaps, sizeof(DDSCAPS2) );
    ddscaps.dwCaps         = DDSCAPS_VIDEOMEMORY | DDSCAPS_LOCALVIDMEM;
    //use to get available vidmem
	DWORD dwFree, dwTotal;
	m_pWrapper->m_pDD7->GetAvailableVidMem(&ddscaps, &dwTotal,&dwFree);
	char str[255];
	sprintf(str,"total: %d, free: %d \n", dwTotal,dwFree);
	OutputDebugString(str);
	*/


	// See if it is time to update the reflection or not
	// Whenever m_iFrameNumber reaches 0, the reflection is updated
	m_iFrameNumber = ( m_iFrameNumber + 1 ) % m_iReflectionUpdateFrequency;
	
	if ( m_iFrameNumber == 0 )
		m_bUpdateReflection = true;
	else
		m_bUpdateReflection = false;


	
	return S_OK;
}

// -----------------------------------------------------------------------
// HRESULT CSphericalMirror::RenderWorld( void )
//
// -----------------------------------------------------------------------
HRESULT CSphericalMirror::RenderWorld( void )
{
	HRESULT hr;

	//	m_pWrapper->m_pD3D7->EvictManagedTextures();

	// Set the front clip plane a little within the sphere to help reduce some of the
	// strange clipping that occurs. The strange clipping occurs because the field of view
	// is large. Having a large field of view causes the front clipping plane to be large in area.
	//
	m_MxMirrorProjection.SetProjection( m_fFov, 1.0f, m_fMirrorRadius*0.3f, 100.0f );

	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW );

	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_SRCBLEND , D3DBLEND_SRCALPHA );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_DESTBLEND , D3DBLEND_INVSRCALPHA );
	
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_DITHERENABLE, TRUE );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_SPECULARENABLE, FALSE );

	hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pFloorTexture->dwStage, D3DTSS_MINFILTER, D3DTFN_LINEAR );
	hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pFloorTexture->dwStage, D3DTSS_MAGFILTER, D3DTFN_LINEAR );
		
    hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pFloorTexture->dwStage, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pFloorTexture->dwStage, D3DTSS_COLORARG2, D3DTA_CURRENT);
    hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( m_pFloorTexture->dwStage, D3DTSS_COLOROP,   D3DTOP_MODULATE );	

	

	m_pWrapper->m_pD3DDevice7->EndScene();


	// We only enter this code if it is time to update
	// the reflection texture
	if ( m_bUpdateReflection )
	{
		
		// Save the viewport
		D3DVIEWPORT7 viewDataSave;
		m_pWrapper->m_pD3DDevice7->GetViewport( &viewDataSave );


		// Set render target to the mirror texture
		m_pWrapper->m_pD3DDevice7->SetRenderTarget( m_pMirrorTexture->ptexTexture, 0 );

	
		// Set the viewport to nicely fit the scene in the texture
		hr = m_pWrapper->m_pD3DDevice7->SetViewport( &m_vpTexViewport );

		if ( hr != D3D_OK )
			OutputDebugString( "Setting the viewport to the texture size failed" );


		// Clear the viewport to a value just off black so that the mirror can be seen
		m_pWrapper->m_pD3DDevice7->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xff0f0f0f, 1.0f, 0L );
	
	
		m_pWrapper->m_pD3DDevice7->BeginScene();


		// Render the scene from the mirror's view to the mirror texture
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxMirrorView.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxMirrorProjection.m_Mx );


		// Use the checkered floor texture on both the floors and the walls
		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pFloorTexture->dwStage, m_pFloorTexture->ptexTexture);	

		if (m_bDrawFloor) 
		{
			hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
														   m_FloorVerts, NUMFLOORVERTS, 0);
		}


		if (m_bDrawWalls)
		{
			hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
														   m_WallVerts, NUMWALLVERTS, 0);
		}


	
		if (m_bDrawCubes)
		{
	
			// The center cube is drawn first by calling DrawCube()
			// DrawCube draws the cube with different colored sides
			// and no texture
			m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxCubeWorld.m_Mx );

			DrawCube();
		
			// Draw the Left cube using the center cube's vertices, but a different world matrix
			// This cube uses the checkered texture that was activated when drawing the floor and walls
			m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxLeftCubeWorld.m_Mx );

			m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
				                                      m_CubeVerts, 36, 0);

	
			// Change the texture for the last cube drawn
			hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pCubeTexture->dwStage, m_pCubeTexture->ptexTexture);	

	
			// Draw the Right cube using the center cube's vertices again
			m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxRightCubeWorld.m_Mx );

			m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
													  m_CubeVerts, 36, 0);
		}
	
	

		m_pWrapper->m_pD3DDevice7->EndScene();

	
		// Done rendering the reflection texture, now off to render the scene
		m_pWrapper->m_pD3DDevice7->SetRenderTarget( m_pWrapper->m_pBackBuffer, 0 );

		m_pWrapper->m_pD3DDevice7->SetViewport( &viewDataSave );
	}

	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
	
	// Use the mirror view matrix if the 'V' key is pressed so that the appearance
	// of the texture applied to the mirror can be seen
	if (!m_bViewFromMirror)
	{
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxProjection.m_Mx );
	}
	else
	{
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxMirrorView.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxMirrorProjection.m_Mx );
	}

	

	
	m_pWrapper->m_pD3DDevice7->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xff000000, 1.0f, 0L );


	m_pWrapper->m_pD3DDevice7->BeginScene();

	hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pFloorTexture->dwStage, m_pFloorTexture->ptexTexture );

	
	
	if ( m_bDrawFloor )
	{
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
			m_FloorVerts, NUMFLOORVERTS, 0);
	}

	
	if ( m_bDrawWalls )
	{
		hr = m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
			m_WallVerts, NUMWALLVERTS, 0);
	}



	// This switch() is for toggling between a spherical mirror, a marble sphere, and
	// a shiny marble sphere
	//
	switch ( m_iTypeOfSphere )
	{

	case 0: //chrome sphere

		// Use the mirror reflection texture coordinates
		m_pWrapper->m_pD3DDevice7->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 1);

		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxMirrorWorld.m_Mx );
		
		// Use the dynamically created texture to make a mirror
		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pMirrorTexture->dwStage, m_pMirrorTexture->ptexTexture );
		if ( m_bDrawMirror )
		{

			m_sphSphere.DrawSphere();
		
		}
		
		// Set the texture coord index back
		m_pWrapper->m_pD3DDevice7->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);

		break;

	case 1: //no reflection

		
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxMirrorWorld.m_Mx );

		
		// Use the base texture
		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pBaseTexture->dwStage, m_pBaseTexture->ptexTexture );
		hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE );


		if ( m_bDrawMirror )
		{
			m_sphSphere.DrawSphere();
		}

		break;

	case 2: //reflection on marble base

		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxMirrorWorld.m_Mx );

		// Use the base texture and mirror texture to make a shiny sphere
		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pBaseTexture->dwStage, m_pBaseTexture->ptexTexture );
		hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE );


		if ( m_bDrawMirror )
		{
			m_sphSphere.DrawSphere();
		}


		// Use the reflection texture coordinates
		m_pWrapper->m_pD3DDevice7->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 1);

		D3DMATERIAL7 oldMaterial;

		ZeroMemory( &oldMaterial, sizeof(oldMaterial) );

		if( FAILED( hr = m_pWrapper->m_pD3DDevice7->GetMaterial(&oldMaterial) ) )
			return hr;
	

		if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&m_MaterialLowAlpha) ) )
			return hr;

		// Use the reflection texture
		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pMirrorTexture->dwStage, m_pMirrorTexture->ptexTexture );
		hr = m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxMirrorWorld.m_Mx );

		if ( m_bDrawMirror )
		{
			//setup the texture cascade to get it's alpha output from the material, and color from the texture.
			hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
			hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
    
			hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_SELECTARG1);
			hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);

			hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA);
			hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_ONE );
			hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE );
			
			m_sphSphere.DrawSphere();

			//reset our alpha output from the texture cascade so that our text will still look OK.
			hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE );
			hr = m_pWrapper->m_pD3DDevice7->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);

		}

		if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&oldMaterial) ) )
			return hr;

		// set the texture coord index back
		m_pWrapper->m_pD3DDevice7->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);

		break;
	}

	
	
	if (m_bDrawCubes)
	{
	
		// The center cube is drawn first by calling DrawCube()
		// DrawCube draws the cube with different colored sides
		// and no texture
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxCubeWorld.m_Mx );

		DrawCube();
		
		// Draw the Left cube using the center cube's vertices, but a different world matrix

		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pFloorTexture->dwStage, m_pFloorTexture->ptexTexture);	
		
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxLeftCubeWorld.m_Mx );

		m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
			                                      m_CubeVerts, 36, 0);

	
		// Change the texture for the last cube drawn
		hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pCubeTexture->dwStage, m_pCubeTexture->ptexTexture);	

	
		// Draw the Right cube using the center cube's vertices again
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxRightCubeWorld.m_Mx );

		m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
			                                      m_CubeVerts, 36, 0);
	}


	return S_OK;

}


// -----------------------------------------------------------------------
// HRESULT CSphericalMirror::TargetChanged( void )
//
// -----------------------------------------------------------------------
HRESULT CSphericalMirror::TargetChanged( void )
{
	HRESULT hr;
	//
	// This indicates that the devices of the DRGWrapper class have changed.
	// If you have anything that depends on the current rendering device, you'll
	// have to re-initialize it here
	//

	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );
	m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &m_MxProjection.m_Mx );

	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_SRCBLEND , D3DBLEND_SRCALPHA );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_DESTBLEND , D3DBLEND_INVSRCALPHA );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState(D3DRENDERSTATE_LIGHTING, TRUE);

	hr = m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_DITHERENABLE, TRUE );
	hr = m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_SPECULARENABLE, TRUE );


	// Restore textures
	if (m_pMirrorTexture)
		m_pMirrorTexture->Restore(m_pWrapper->m_pD3DDevice7);
	
	if (m_pFloorTexture)
		m_pFloorTexture->Restore(m_pWrapper->m_pD3DDevice7);
		
	if (m_pCubeTexture)
		m_pCubeTexture->Restore(m_pWrapper->m_pD3DDevice7);

	if (m_pBaseTexture)
		m_pBaseTexture->Restore(m_pWrapper->m_pD3DDevice7);

	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&m_Material) ) )
		return hr;
	

    // Set the light
    if( FAILED(m_pWrapper->m_pD3DDevice7->SetLight( 0, &m_Light )))
		return hr;

	if( FAILED(m_pWrapper->m_pD3DDevice7->LightEnable( 0, TRUE )))
		return hr;

	if( FAILED(m_pWrapper->m_pD3DDevice7->SetRenderState( D3DRENDERSTATE_AMBIENT, 0xff333333 ) ) )
		return hr;


	return S_OK;

}



// -----------------------------------------------------------------------
// void CSphericalMirror::UpdateMirrorTextureSize( int iTextureSize )
//
// -----------------------------------------------------------------------
void CSphericalMirror::UpdateMirrorTextureSize( int iTextureSize )
{
	delete m_pMirrorTexture;
	m_pMirrorTexture = NULL;
	
	m_iTextureSize = iTextureSize;

	m_pMirrorTexture = new CDRGTexture;

	if( m_pMirrorTexture->CreateTexture( m_pWrapper, iTextureSize, iTextureSize, 16,
								         (DWORD)0, DRGTEXTR_RENDERTARGET, NULL) == S_OK )
		m_pMirrorTexture->Restore(m_pWrapper->m_pD3DDevice7);

	// We also need to update the viewport size so that the scene
	// fills the entire texture
	m_vpTexViewport.dwWidth = iTextureSize;
	m_vpTexViewport.dwHeight = iTextureSize;


}



// -----------------------------------------------------------------------
// HRESULT CSphericalMirror::DrawCube()
//
//   Draws the cube using many different materials
// -----------------------------------------------------------------------
HRESULT CSphericalMirror::DrawCube()
{

	HRESULT hr;

	// Save the previously used textures and materials
	//
	LPDIRECTDRAWSURFACE7 prevTexture;
	m_pWrapper->m_pD3DDevice7->GetTexture( 0, &prevTexture );

	hr = m_pWrapper->m_pD3DDevice7->SetTexture( m_pCubeTexture->dwStage, m_pCubeTexture->ptexTexture);	
	hr = m_pWrapper->m_pD3DDevice7->SetTexture( 0, NULL);	
	
	D3DMATERIAL7 prevMaterial;
	ZeroMemory( &prevMaterial, sizeof(prevMaterial) );
	m_pWrapper->m_pD3DDevice7->GetMaterial(&prevMaterial);


	// Create material that is changed 
	// for each of the 6 sides of the cube
	//
	D3DMATERIAL7	tempmat;
	ZeroMemory( &tempmat, sizeof(tempmat) );
  
    tempmat.dcvAmbient.r = 1.0f;
	tempmat.dcvAmbient.g = 0.0f;
	tempmat.dcvAmbient.b = 0.0f;
	tempmat.dcvAmbient.a = 1.0f;
	
	tempmat.dcvDiffuse.r = 1.0f;
	tempmat.dcvDiffuse.g = 0.0f;
	tempmat.dcvDiffuse.b = 0.0f;
	tempmat.dcvDiffuse.a = 1.0f;
	
	tempmat.dcvSpecular.r = 1.0f;
	tempmat.dcvSpecular.g = 0.0f;
	tempmat.dcvSpecular.b = 0.0f;
	tempmat.dcvSpecular.a = 1.0f;
	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          m_CubeVerts, 6, 0);

    
    tempmat.dcvAmbient.r = 0.0f;
	tempmat.dcvAmbient.g = 1.0f;
	tempmat.dcvAmbient.b = 0.0f;
	tempmat.dcvAmbient.a = 1.0f;
	
	tempmat.dcvDiffuse.r = 0.0f;
	tempmat.dcvDiffuse.g = 1.0f;
	tempmat.dcvDiffuse.b = 0.0f;
	tempmat.dcvDiffuse.a = 1.0f;
	
	tempmat.dcvSpecular.r = 0.0f;
	tempmat.dcvSpecular.g = 1.0f;
	tempmat.dcvSpecular.b = 0.0f;
	tempmat.dcvSpecular.a = 1.0f;


	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[6], 6, 0);


    tempmat.dcvAmbient.r = 0.0f;
	tempmat.dcvAmbient.g = 0.0f;
	tempmat.dcvAmbient.b = 1.0f;
	tempmat.dcvAmbient.a = 1.0f;
	
	tempmat.dcvDiffuse.r = 0.0f;
	tempmat.dcvDiffuse.g = 0.0f;
	tempmat.dcvDiffuse.b = 1.0f;
	tempmat.dcvDiffuse.a = 1.0f;
	
	tempmat.dcvSpecular.r = 0.0f;
	tempmat.dcvSpecular.g = 0.0f;
	tempmat.dcvSpecular.b = 1.0f;
	tempmat.dcvSpecular.a = 1.0f;

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[12], 6, 0);
	
	tempmat.dcvAmbient.r = 1.0f;
	tempmat.dcvAmbient.g = 1.0f;
	tempmat.dcvAmbient.b = 0.0f;
	tempmat.dcvAmbient.a = 1.0f;
	
	tempmat.dcvDiffuse.r = 1.0f;
	tempmat.dcvDiffuse.g = 1.0f;
	tempmat.dcvDiffuse.b = 0.0f;
	tempmat.dcvDiffuse.a = 1.0f;
	
	tempmat.dcvSpecular.r = 1.0f;
	tempmat.dcvSpecular.g = 1.0f;
	tempmat.dcvSpecular.b = 0.0f;
	tempmat.dcvSpecular.a = 1.0f;

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[18], 6, 0);

	
	tempmat.dcvAmbient.r = 0.0f;
	tempmat.dcvAmbient.g = 1.0f;
	tempmat.dcvAmbient.b = 1.0f;
	tempmat.dcvAmbient.a = 1.0f;
	
	tempmat.dcvDiffuse.r = 0.0f;
	tempmat.dcvDiffuse.g = 1.0f;
	tempmat.dcvDiffuse.b = 1.0f;
	tempmat.dcvDiffuse.a = 1.0f;
	
	tempmat.dcvSpecular.r = 0.0f;
	tempmat.dcvSpecular.g = 1.0f;
	tempmat.dcvSpecular.b = 1.0f;
	tempmat.dcvSpecular.a = 1.0f;

	
	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[24], 6, 0);

	
    tempmat.dcvAmbient.r = 1.0f;
	tempmat.dcvAmbient.g = 0.0f;
	tempmat.dcvAmbient.b = 1.0f;
	tempmat.dcvAmbient.a = 1.0f;
	
	tempmat.dcvDiffuse.r = 1.0f;
	tempmat.dcvDiffuse.g = 0.0f;
	tempmat.dcvDiffuse.b = 1.0f;
	tempmat.dcvDiffuse.a = 1.0f;
	
	tempmat.dcvSpecular.r = 1.0f;
	tempmat.dcvSpecular.g = 0.0f;
	tempmat.dcvSpecular.b = 1.0f;
	tempmat.dcvSpecular.a = 1.0f;

	if( FAILED( hr = m_pWrapper->m_pD3DDevice7->SetMaterial(&tempmat) ) )
		return hr;

	m_pWrapper->m_pD3DDevice7->DrawPrimitive( D3DPT_TRIANGLELIST, D3DFVF_VERTEX,
	                                          &m_CubeVerts[30], 6, 0);


	hr = m_pWrapper->m_pD3DDevice7->SetTexture( 0, prevTexture);	
	hr = m_pWrapper->m_pD3DDevice7->SetMaterial( &prevMaterial );
	
	return S_OK;
}



// -----------------------------------------------------------------------
// void CSphericalMirror::SetupCubeVerts(float scale)
//
//   Sets up the vertices for the cube using texture coordinates
// -----------------------------------------------------------------------
void CSphericalMirror::SetupCubeVerts(float scale)
{

	int i;
	memset( m_CubeVerts, 0, sizeof( D3DVERTEX ) * NUMCUBEVERTS );

	
	// Front face
	m_CubeVerts[0].x = -scale;		m_CubeVerts[0].y = -scale;		m_CubeVerts[0].z = -scale;
	m_CubeVerts[0].tu = 0.0f;		m_CubeVerts[0].tv = 1.0f;
	m_CubeVerts[1].x = -scale;		m_CubeVerts[1].y = scale;		m_CubeVerts[1].z = -scale;
	m_CubeVerts[1].tu = 0.0f;		m_CubeVerts[1].tv = 0.0f;
	m_CubeVerts[2].x = scale;		m_CubeVerts[2].y =  -scale;		m_CubeVerts[2].z = -scale;
	m_CubeVerts[2].tu = 1.0f;		m_CubeVerts[2].tv = 1.0f;

	m_CubeVerts[3] = m_CubeVerts[1];
	m_CubeVerts[3].tu = 0.0f;		m_CubeVerts[3].tv = 0.0f;
	m_CubeVerts[4].x = scale;		m_CubeVerts[4].y = scale;		m_CubeVerts[4].z = -scale;
	m_CubeVerts[4].tu = 1.0f;		m_CubeVerts[4].tv = 0.0f;
	m_CubeVerts[5] = m_CubeVerts[2];
	m_CubeVerts[5].tu = 1.0f;		m_CubeVerts[5].tv = 1.0f;

	for(i=0; i<6; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = -1.0f;
	}
	
	
	// Right Face
	m_CubeVerts[6] = m_CubeVerts[2];
	m_CubeVerts[6].tu = 0.0f;		m_CubeVerts[6].tv = 1.0f;
	m_CubeVerts[7]= m_CubeVerts[4];
	m_CubeVerts[7].tu = 0.0f;		m_CubeVerts[7].tv = 0.0f;
	m_CubeVerts[8].x = scale;		m_CubeVerts[8].y =  -scale;		m_CubeVerts[8].z = scale;
	m_CubeVerts[8].tu = 1.0f;		m_CubeVerts[8].tv = 1.0f;
	m_CubeVerts[9] = m_CubeVerts[4];
	m_CubeVerts[9].tu = 0.0f;		m_CubeVerts[9].tv = 0.0f;
	m_CubeVerts[10].x = scale;		m_CubeVerts[10].y = scale;		m_CubeVerts[10].z = scale;
	m_CubeVerts[10].tu = 1.0f;		m_CubeVerts[10].tv = 0.0f;
	m_CubeVerts[11] = m_CubeVerts[8];
	m_CubeVerts[11].tu = 1.0f;		m_CubeVerts[11].tv = 1.0f;
	
	for(i=6; i<12; i++)
	{
		m_CubeVerts[i].nx = 1.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = 0.0f;
	}
	
	// Top face
	m_CubeVerts[12] = m_CubeVerts[1];
	m_CubeVerts[12].tu = 0.0f;	m_CubeVerts[12].tv = 0.0f;
	m_CubeVerts[13].x = -scale;	m_CubeVerts[13].y = scale;		m_CubeVerts[13].z = scale;
	m_CubeVerts[13].tu = 1.0f;	m_CubeVerts[13].tv = 0.0f;
	m_CubeVerts[14] = m_CubeVerts[4];
	m_CubeVerts[14].tu = 0.0f;	m_CubeVerts[14].tv = 1.0f;
	m_CubeVerts[15] = m_CubeVerts[4];
	m_CubeVerts[15].tu = 0.0f;	m_CubeVerts[15].tv = 1.0f;
	m_CubeVerts[16] = m_CubeVerts[13];
	m_CubeVerts[16].tu = 1.0f;	m_CubeVerts[16].tv = 0.0f;
	m_CubeVerts[17] = m_CubeVerts[10];
	m_CubeVerts[17].tu = 1.0f;	m_CubeVerts[17].tv = 1.0f;
	
	for(i=12; i<18; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = 1.0f;
		m_CubeVerts[i].nz = 0.0f;
	}

	// Left face
	m_CubeVerts[18] = m_CubeVerts[0];
	m_CubeVerts[18].tu = 1.0f;	m_CubeVerts[18].tv = 1.0f;
	m_CubeVerts[19] = m_CubeVerts[13];
	m_CubeVerts[19].tu = 0.0f;	m_CubeVerts[19].tv = 0.0f;
	m_CubeVerts[20] = m_CubeVerts[1];
	m_CubeVerts[20].tu = 1.0f;	m_CubeVerts[20].tv = 0.0f;
	m_CubeVerts[21] = m_CubeVerts[0];
	m_CubeVerts[21].tu = 1.0f;	m_CubeVerts[21].tv = 1.0f;
	m_CubeVerts[22].x = -scale;	m_CubeVerts[22].y = -scale;		m_CubeVerts[22].z = scale;
	m_CubeVerts[22].tu = 0.0f;	m_CubeVerts[22].tv = 1.0f;
	
	m_CubeVerts[23] = m_CubeVerts[13];
	m_CubeVerts[23].tu = 0.0f;	m_CubeVerts[23].tv = 0.0f;

	for(i=18; i<24; i++)
	{
		m_CubeVerts[i].nx = -1.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = 0.0f;
	}

	// Back face
	m_CubeVerts[24] = m_CubeVerts[8];
	m_CubeVerts[24].tu = 0.0f;	m_CubeVerts[24].tv = 1.0f;
	m_CubeVerts[25] = m_CubeVerts[13];
	m_CubeVerts[25].tu = 1.0f;	m_CubeVerts[25].tv = 0.0f;
	m_CubeVerts[26] = m_CubeVerts[22];
	m_CubeVerts[26].tu = 1.0f;	m_CubeVerts[26].tv = 1.0f;
	m_CubeVerts[27] = m_CubeVerts[8];
	m_CubeVerts[27].tu = 0.0f;	m_CubeVerts[27].tv = 1.0f;
	m_CubeVerts[28] = m_CubeVerts[10];
	m_CubeVerts[28].tu = 0.0f;	m_CubeVerts[28].tv = 0.0f;
	m_CubeVerts[29] = m_CubeVerts[13];
	m_CubeVerts[29].tu = 1.0f;	m_CubeVerts[29].tv = 0.0f;
	
	for(i=24; i<30; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = 0.0f;
		m_CubeVerts[i].nz = 1.0f;
	}

	
	// Bottom face
	m_CubeVerts[30] = m_CubeVerts[2];
	m_CubeVerts[30].tu = 0.0f;	m_CubeVerts[30].tv = 1.0f;
	m_CubeVerts[31] = m_CubeVerts[8];
	m_CubeVerts[31].tu = 0.0f;	m_CubeVerts[31].tv = 0.0f;
	m_CubeVerts[32] = m_CubeVerts[22];
	m_CubeVerts[32].tu = 1.0f;	m_CubeVerts[32].tv = 0.0f;
	m_CubeVerts[33] = m_CubeVerts[2];
	m_CubeVerts[33].tu = 0.0f;	m_CubeVerts[33].tv = 1.0f;
	m_CubeVerts[34] = m_CubeVerts[22];
	m_CubeVerts[34].tu = 1.0f;	m_CubeVerts[34].tv = 0.0f;
	m_CubeVerts[35] = m_CubeVerts[0];
	m_CubeVerts[35].tu = 1.0f;	m_CubeVerts[35].tv = 1.0f;
	
	for(i=30; i<36; i++)
	{
		m_CubeVerts[i].nx = 0.0f;
		m_CubeVerts[i].ny = -1.0f;
		m_CubeVerts[i].nz = 0.0f;
	}

	return;
}



// -----------------------------------------------------------------------
// void CSphericalMirror::SetupFloorVerts(float scale)
//
//   Sets up the vertices for the floor
// -----------------------------------------------------------------------
void CSphericalMirror::SetupFloorVerts(float scale)
{
	memset( m_FloorVerts, 0, sizeof( D3DVERTEX ) * NUMFLOORVERTS );

	m_FloorVerts[0].x = -scale;		m_FloorVerts[0].y = -1.75f;	m_FloorVerts[0].z = -scale; 
	m_FloorVerts[1].x = scale;		m_FloorVerts[1].y = -1.75f;	m_FloorVerts[1].z = scale; 
	m_FloorVerts[2].x = scale;		m_FloorVerts[2].y = -1.75f;	m_FloorVerts[2].z = -scale;

	m_FloorVerts[3].x = -scale;		m_FloorVerts[3].y = -1.75f;	m_FloorVerts[3].z = -scale;
	m_FloorVerts[4].x = -scale;		m_FloorVerts[4].y = -1.75f;	m_FloorVerts[4].z = scale;
	m_FloorVerts[5].x = scale;		m_FloorVerts[5].y = -1.75f;	m_FloorVerts[5].z = scale; 
	
	
	m_FloorVerts[0].tu = 0.0f; m_FloorVerts[0].tv = 1.0f;
	m_FloorVerts[1].tu = 1.0f; m_FloorVerts[1].tv = 0.0f;
	m_FloorVerts[2].tu = 1.0f; m_FloorVerts[2].tv = 1.0f;
	
	m_FloorVerts[3].tu = 0.0f; m_FloorVerts[3].tv = 1.0f;
	m_FloorVerts[4].tu = 0.0f; m_FloorVerts[4].tv = 0.0f;
	m_FloorVerts[5].tu = 1.0f; m_FloorVerts[5].tv = 0.0f;

	// common elements
	for (int i=0;i<NUMFLOORVERTS;i++)
	{
		m_FloorVerts[i].nx = 0.0f;
		m_FloorVerts[i].ny = 1.0f;
		m_FloorVerts[i].nz = 0.0f;
	}

	return;

}


// -----------------------------------------------------------------------
// void CSphericalMirror::SetupWallVerts(float scale)
//
//   Sets up the vertices for the floor
// -----------------------------------------------------------------------
void CSphericalMirror::SetupWallVerts(float scale)
{
	memset( m_WallVerts, 0, sizeof( D3DVERTEX ) * NUMWALLVERTS );

	
	// back wall
	m_WallVerts[0].x = -scale;	m_WallVerts[0].y = -1.75f;				m_WallVerts[0].z = -scale;
	m_WallVerts[1].x = scale;	m_WallVerts[1].y = -1.75f;				m_WallVerts[1].z = -scale;
	m_WallVerts[2].x = scale;	m_WallVerts[2].y = -1.75f+(2*scale);	m_WallVerts[2].z = -scale;
	m_WallVerts[3].x = -scale;	m_WallVerts[3].y = -1.75f+(2*scale);	m_WallVerts[3].z = -scale;

	m_WallVerts[0].tu = 1.0f;	m_WallVerts[0].tv = 1.0f;
	m_WallVerts[1].tu = 0.0f;	m_WallVerts[1].tv = 1.0f;
	m_WallVerts[2].tu = 0.0f;	m_WallVerts[2].tv = 0.0f;
	m_WallVerts[3].tu = 1.0f;	m_WallVerts[3].tv = 0.0f;
	
	m_WallVerts[4] = m_WallVerts[0];
	m_WallVerts[5] = m_WallVerts[2];

	for(int i=0; i<6; i++)
	{
		m_WallVerts[i].nx = 0.0f;
		m_WallVerts[i].ny = 0.0f;
		m_WallVerts[i].nz = 1.0f;
	}

	// right wall
	m_WallVerts[6].x = scale;	m_WallVerts[6].y = -1.75f;				m_WallVerts[6].z = -scale;
	m_WallVerts[7].x = scale;	m_WallVerts[7].y = -1.75f;				m_WallVerts[7].z = scale;
	m_WallVerts[8].x = scale;	m_WallVerts[8].y = -1.75f+(2*scale);	m_WallVerts[8].z = scale;
	m_WallVerts[9].x = scale;	m_WallVerts[9].y = -1.75f+(2*scale);	m_WallVerts[9].z = -scale;

	m_WallVerts[6].tu = 1.0f;	m_WallVerts[6].tv = 1.0f;
	m_WallVerts[7].tu = 0.0f;	m_WallVerts[7].tv = 1.0f;
	m_WallVerts[8].tu = 0.0f;	m_WallVerts[8].tv = 0.0f;
	m_WallVerts[9].tu = 1.0f;	m_WallVerts[9].tv = 0.0f;
	
	m_WallVerts[10] = m_WallVerts[6];
	m_WallVerts[11] = m_WallVerts[8];

	for(i=6; i<12; i++)
	{
		m_WallVerts[i].nx = -1.0f;
		m_WallVerts[i].ny = 0.0f;
		m_WallVerts[i].nz = 0.0f;
	}

	// left wall
	m_WallVerts[12].x = -scale;	m_WallVerts[12].y = -1.75f;				m_WallVerts[12].z = scale;
	m_WallVerts[13].x = -scale;	m_WallVerts[13].y = -1.75f;				m_WallVerts[13].z = -scale;
	m_WallVerts[14].x = -scale;	m_WallVerts[14].y = -1.75f+(2*scale);	m_WallVerts[14].z = -scale;
	m_WallVerts[15].x = -scale;	m_WallVerts[15].y = -1.75f+(2*scale);	m_WallVerts[15].z = scale;

	m_WallVerts[12].tu = 0.0f;	m_WallVerts[12].tv = 1.0f;
	m_WallVerts[13].tu = 1.0f;	m_WallVerts[13].tv = 1.0f;
	m_WallVerts[14].tu = 1.0f;	m_WallVerts[14].tv = 0.0f;
	m_WallVerts[15].tu = 0.0f;	m_WallVerts[15].tv = 0.0f;
	
	m_WallVerts[16] = m_WallVerts[12];
	m_WallVerts[17] = m_WallVerts[14];

	for(i=12; i<18; i++)
	{
		m_WallVerts[i].nx = 1.0f;
		m_WallVerts[i].ny = 0.0f;
		m_WallVerts[i].nz = 0.0f;
	}

	// back wall
	m_WallVerts[18].x = scale;	m_WallVerts[18].y = -1.75f;				m_WallVerts[18].z = scale;
	m_WallVerts[19].x = -scale;	m_WallVerts[19].y = -1.75f;				m_WallVerts[19].z = scale;
	m_WallVerts[20].x = -scale;	m_WallVerts[20].y = -1.75f+(2*scale);	m_WallVerts[20].z = scale;
	m_WallVerts[21].x = scale;	m_WallVerts[21].y = -1.75f+(2*scale);	m_WallVerts[21].z = scale;

	m_WallVerts[18].tu = 0.0f;	m_WallVerts[18].tv = 1.0f;
	m_WallVerts[19].tu = 1.0f;	m_WallVerts[19].tv = 1.0f;
	m_WallVerts[20].tu = 1.0f;	m_WallVerts[20].tv = 0.0f;
	m_WallVerts[21].tu = 0.0f;	m_WallVerts[21].tv = 0.0f;
	
	m_WallVerts[22] = m_WallVerts[18];
	m_WallVerts[23] = m_WallVerts[20];

	for(i=18; i<24; i++)
	{
		m_WallVerts[i].nx = 0.0f;
		m_WallVerts[i].ny = 0.0f;
		m_WallVerts[i].nz = -1.0f;
	}

}









// -----------------------------------------------------------------------
// void CSphericalMirror::KeyUp(WPARAM wParam)
//
//   Keyboard keypress handler. 
// -----------------------------------------------------------------------
void CSphericalMirror::KeyDown(WPARAM wParam)
{
	//we need a variable to modify our rotation speed with frame rate	
	float scaler = 1.0f;

	//if exiting, the wrapper could be invalid, but we don't care, since we will exit anyway
	if (wParam == 'X') 
		return;

	if(m_pWrapper) //we could get into this loop after the wrapper has been invalidates
	{
		if (m_pWrapper->m_fFrameTime >0.0f)
			scaler = m_pWrapper->m_fFrameTime;  
	}
		
	// Variables used in 'A' and 'Z' cases
	char string[50];

	switch (wParam) 
	{
	case VK_UP:
		
		// If the arrow keys should be used for rotating the cubes
		// and they're not currently set to auto rotate, then
		// rotate the cubes. Otherwise rotate the view with the
		// arrow keys
		//
		if (m_bKeysRotateCube && !m_bAutoRotateCube)
		{
			m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._11, m_MxCubeTransform.m_Mx._21, m_MxCubeTransform.m_Mx._31, -3.14159f*scaler );
			m_MxLeftCubeTransform.Rotate( m_MxLeftCubeTransform.m_Mx._11, m_MxLeftCubeTransform.m_Mx._21, m_MxLeftCubeTransform.m_Mx._31, -3.14159f*scaler );
			m_MxRightCubeTransform.Rotate( m_MxRightCubeTransform.m_Mx._11, m_MxRightCubeTransform.m_Mx._21, m_MxRightCubeTransform.m_Mx._31, -3.14159f*scaler );
		}
		else
			m_MxView.Rotate( m_MxView.m_Mx._11, m_MxView.m_Mx._21, m_MxView.m_Mx._31, -3.14159f*scaler );
		break;
	
	case VK_DOWN:
		
		if (m_bKeysRotateCube && !m_bAutoRotateCube)
		{
			m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._11, m_MxCubeTransform.m_Mx._21, m_MxCubeTransform.m_Mx._31, 3.14159f*scaler );
			m_MxRightCubeTransform.Rotate( m_MxRightCubeTransform.m_Mx._11, m_MxRightCubeTransform.m_Mx._21, m_MxRightCubeTransform.m_Mx._31, 3.14159f*scaler );
			m_MxLeftCubeTransform.Rotate( m_MxLeftCubeTransform.m_Mx._11, m_MxLeftCubeTransform.m_Mx._21, m_MxLeftCubeTransform.m_Mx._31, 3.14159f*scaler );
		}
		else
			m_MxView.Rotate( m_MxView.m_Mx._11, m_MxView.m_Mx._21, m_MxView.m_Mx._31, 3.14159f*scaler );
		break;
	
	case VK_LEFT:
		
		if (m_bKeysRotateCube && !m_bAutoRotateCube)
		{
			m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._12, m_MxCubeTransform.m_Mx._22, m_MxCubeTransform.m_Mx._32, -3.14159f*scaler );
			m_MxRightCubeTransform.Rotate( m_MxRightCubeTransform.m_Mx._12, m_MxRightCubeTransform.m_Mx._22, m_MxRightCubeTransform.m_Mx._32, -3.14159f*scaler );
			m_MxLeftCubeTransform.Rotate( m_MxLeftCubeTransform.m_Mx._12, m_MxLeftCubeTransform.m_Mx._22, m_MxLeftCubeTransform.m_Mx._32, -3.14159f*scaler );
		}
		else
			m_MxView.Rotate( m_MxView.m_Mx._12, m_MxView.m_Mx._22, m_MxView.m_Mx._32, -3.14159f*scaler );
		break;
	
	case VK_RIGHT:	
		
		if (m_bKeysRotateCube && !m_bAutoRotateCube)
		{
			m_MxCubeTransform.Rotate( m_MxCubeTransform.m_Mx._12, m_MxCubeTransform.m_Mx._22, m_MxCubeTransform.m_Mx._32, 3.14159f*scaler );
			m_MxRightCubeTransform.Rotate( m_MxRightCubeTransform.m_Mx._12, m_MxRightCubeTransform.m_Mx._22, m_MxRightCubeTransform.m_Mx._32, 3.14159f*scaler );
			m_MxLeftCubeTransform.Rotate( m_MxLeftCubeTransform.m_Mx._12, m_MxLeftCubeTransform.m_Mx._22, m_MxLeftCubeTransform.m_Mx._32, 3.14159f*scaler );
		}
		else
			m_MxView.Rotate( m_MxView.m_Mx._12, m_MxView.m_Mx._22, m_MxView.m_Mx._32, 3.14159f*scaler );
		break;

	case VK_SPACE:
		m_bAutoRotateCube = !m_bAutoRotateCube;
		break;

	case 'D':
		m_bDrawFloor = !m_bDrawFloor;		
		break;

	case 'C':
		m_bDrawCubes = !m_bDrawCubes;
		break;

	case 'M':
		m_bDrawMirror = !m_bDrawMirror;
		break;
	
	case 'V':
		m_bViewFromMirror = !m_bViewFromMirror;
		break;

	case 'W':
		m_bDrawWalls = !m_bDrawWalls;
		break;

	case 'B':
		m_iTypeOfSphere = ( m_iTypeOfSphere + 1 ) % 3;
		break;

	case 'R':
		m_bRotateSphere = !m_bRotateSphere;
		break;

	case 'J':
		m_iReflectionUpdateFrequency++;
		break;

	case 'K':
		if ( m_iReflectionUpdateFrequency > 1 )
			m_iReflectionUpdateFrequency--;
		break;

	case 'N':
		
		// If the texture is currently 512x512, change to 32x32. Otherwise
		// Just go to the next higher size
		UpdateMirrorTextureSize( m_iTextureSize == 512 ? 32 : m_iTextureSize * 2);
		break;

	// change the texture mapping method
		
	case 'A':
		m_fFov += 3.14f/24;
		if (m_fFov > 3.14f)
			m_fFov = 3.14f;
		sprintf( string, "FOV = %03f\n", m_fFov );
		OutputDebugString(string);
		
		break;

	case 'Z':
		if (m_fFov < 3.14f/4)
			m_fFov = 3.14f/4;
		m_fFov -= 3.14f/24;
		sprintf( string, "FOV = %03f\n", m_fFov );
		OutputDebugString(string);
		break;

	case 'T':
		m_bKeysRotateCube = !m_bKeysRotateCube;
		break;

	case 'O':
		m_fBaseShininess += .02f;

		// Update the base material shininess
		m_MaterialLowAlpha.dcvAmbient.a = m_fBaseShininess;
		m_MaterialLowAlpha.dcvDiffuse.a = m_fBaseShininess;
		m_MaterialLowAlpha.dcvSpecular.a = m_fBaseShininess;

		break;
	
	case 'P':
		m_fBaseShininess -= .02f;

		// Update the base material shininess
		m_MaterialLowAlpha.dcvAmbient.a = m_fBaseShininess;
		m_MaterialLowAlpha.dcvDiffuse.a = m_fBaseShininess;
		m_MaterialLowAlpha.dcvSpecular.a = m_fBaseShininess;

		break;

	}

	
	if ( wParam == 'J' || wParam == 'K' )
	{
		sprintf( string, "Reflection Updated Every %d Frame(s)\n", m_iReflectionUpdateFrequency );
		OutputDebugString( string );
	}
	

	if ( wParam == 'N' )
	{
		char string[50];
		sprintf(string, "Texture Size = %d x %d\n", m_iTextureSize, m_iTextureSize );
		OutputDebugString( string );
	}

	
	if (m_pWrapper)
	{
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );
	}

	
	return;
}

void CSphericalMirror::KeyUp(WPARAM wParam)
{		
	switch (wParam) 
	{
		//add code here to handle key releases, if you need them (as in remmed out example)
	case VK_UP:
		//OutputDebugString("The up arrow was released;");			
		break;
	}
	
	return;
}


void CSphericalMirror::MouseLeftUp(WPARAM wParam, LPARAM lParam)
{	
	return;
}

void CSphericalMirror::MouseLeftDown(WPARAM wParam, LPARAM lParam)
{		
	return;
}

void CSphericalMirror::MouseRightUp(WPARAM wParam, LPARAM lParam)
{	
	return;
}

void CSphericalMirror::MouseRightDown(WPARAM wParam, LPARAM lParam)
{		
	return;
}

void CSphericalMirror::MouseMove(WPARAM wParam, LPARAM lParam)
{
	if ( wParam & MK_LBUTTON )	// Left button down plus mouse move == rotate
	{
		//we need a variable to modify our rotation speed with frame rate	
		float scaler = 1.0f;
		if(m_pWrapper) //we could get into this loop after the wrapper has been invalidates
		{
			scaler = 2.0f / max(m_pWrapper->m_dwRenderWidth, m_pWrapper->m_dwRenderHeight);  
		}

		//we need to compare the last position to the current position
		int x = LOWORD(lParam);
		int y = HIWORD(lParam);
		
		if (x - m_iMousePrevX)
		{
			m_MxView.Rotate( m_MxView.m_Mx._12, m_MxView.m_Mx._22, m_MxView.m_Mx._32, 3.14159f*scaler*(float)(x - m_iMousePrevX) );
		}
			
		if (y - m_iMousePrevY)
		{
			m_MxView.Rotate( m_MxView.m_Mx._11, m_MxView.m_Mx._21, m_MxView.m_Mx._31, 3.14159f*scaler*(float)(y - m_iMousePrevY) );
		}

		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );

	
		m_iMousePrevX = x;
		m_iMousePrevY = y;

	}
	else if ( wParam & MK_RBUTTON )	// Right button down plus mouse move == Zoom camera in/out
	{
		//we need a variable to modify our rotation speed with frame rate	
		float scaler = 1.0f;
		if(m_pWrapper) //we could get into this loop after the wrapper has been invalidates
		{
			scaler = 5.0f / max(m_pWrapper->m_dwRenderWidth, m_pWrapper->m_dwRenderHeight);  
		}

		//we need to compare the last position to the current position
		int x = LOWORD(lParam);
		int y = HIWORD(lParam);
		
		if (y - m_iMousePrevY)
		{
			DRGMatrix Temp(true);

			Temp.Translate( 0,0,scaler * (y - m_iMousePrevY) );
			m_MxView.PostMultiply( Temp );
		}

		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_WORLD, &m_MxWorld.m_Mx );
		m_pWrapper->m_pD3DDevice7->SetTransform( D3DTRANSFORMSTATE_VIEW, &m_MxView.m_Mx );

	
		m_iMousePrevX = x;
		m_iMousePrevY = y;

	}
	else
	{
		//need to update these, or the next time someone clicks, the delta is massive
		m_iMousePrevX = LOWORD(lParam);
		m_iMousePrevY = HIWORD(lParam);	
	}
	return;
}

