//-----------------------------------------------------------------------------
// Matrix.h
//
// Started 06/21/2000
//
// Shrike Framework -  Rob Wyatt 
//-----------------------------------------------------------------------------
#if !defined(MATRIX_H_INCLUDED)
#define MATRIX_H_INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Vector.h"


//----------------------------------------------------------------------------
// Matrices are ordered row major and are post multiplied with vectors.
// V' = V * A * B
// Means transform V by matrix A and then by matrix B.
class Matrix4 : public D3DMATRIX
{
public:
    //
    // Constructors and Destructor
    //

    //----------------------------------------------------------------------------
    Matrix4
    (
    )
    //--------------------------------------
    {
    };


    //----------------------------------------------------------------------------
    // Sets the matrix to a perpsective projection matrix built
    // from the field-of-view (fov, in y), aspect ratio, near plane ,
    // and far plane .
    Matrix4
    (
        float fov,          // field of view in Y (full angle)
        float aspect,       // aspect ratio of the render target (Y/X)
        float near_p,       // near clip plane,  camera Z at this distance will be projected to 0.0
        float far_p         // far clip plane,   camera Z at the distance will be projected to 1.0
    )
    //--------------------------------------
    {
        SetPerspective(fov,aspect,near_p,far_p);
    }


    //----------------------------------------------------------------------------
    // Given an eye point, a lookat point, and an up vector, this
    // function builds a 4x4 view matrix.
    Matrix4
    (
        Vector3& from,
        Vector3& at,
        Vector3& world_up
    );

    //------------------------------------------------------------------------------
    // Construct a rotation matrix given by a vector and a rotation around that
    // vector. The vector passed in must be a unit vector specifying a direction.
    Matrix4::Matrix4
    (
        Vector3&    dir,       // must be a unit vector
        float       rad        // angle around the above vector in radians
    );


    //----------------------------------------------------------------------------
    // Construct  matrix from 16 values, uses by matrix multiply to construct the
    // resulting matrix.
    inline Matrix4
    (
        float   xx, float   xy, float   xz, float   xw,
        float   yx, float   yy, float   yz, float   yw,
        float   zx, float   zy, float   zz, float   zw,
        float   wx, float   wy, float   wz, float   ww
    )
    //--------------------------------------
    {
        _11 = xx;   _12 = xy;   _13 = xz; _14 = xw;
        _21 = yx;   _22 = yy;   _23 = yz; _24 = yw;
        _31 = zx;   _32 = zy;   _33 = zz; _34 = zw;
        _41 = wx;   _42 = wy;   _43 = wz; _44 = ww;
    }


    //
    // The basic vector*matrix multiplication.
    //

    //
    // Multilying a 3D vector by a 4D matrix requires special care. This inline
    // operator is optimized for speed in the common case and the W component
    // of the matrix is ignored while the W component of the vector is assumed to
    // be unity. This results in the following operation:
    // X' = (M11*X) + (M21*Y) + (M31*Z) + (M41*1)
    // Y' = (M12*X) + (M22*Y) + (M32*Z) + (M42*1)
    // Z' = (M13*X) + (M23*Y) + (M33*Z) + (M43*1)
    // We also assume W is unity on exit which means the matrix 4th colum must
    // be 0,0,0,1.

    //----------------------------------------------------------------------------
    friend inline Vector3 operator *
    (
        const Vector3& v,
        const Matrix4& m
    )
    //--------------------------------------
    {
        // Ensure our matrix is going to affect W
        Assert(m._14 == 0.0f);
        Assert(m._24 == 0.0f);
        Assert(m._34 == 0.0f);
        Assert(m._44 == 1.0f);

        // Construct the rotated point directly at the return address
        that_ptr->x = (m._11*v.x) + (m._21*v.y) + (m._31*v.z) + m._41;
        that_ptr->y = (m._12*v.x) + (m._22*v.y) + (m._32*v.z) + m._42;
        that_ptr->z = (m._13*v.x) + (m._23*v.y) + (m._33*v.z) + m._43;

        return *that_ptr;
    }


    //----------------------------------------------------------------------------
    friend inline Vector3& operator *=
    (
        Vector3&        v,
        const Matrix4&  m
    )
    //--------------------------------------
    {
        // Ensure our matrix is going to affect W
        Assert(m._14 == 0.0f);
        Assert(m._24 == 0.0f);
        Assert(m._34 == 0.0f);
        Assert(m._44 == 1.0f);

        Vector3 vec = v;

        v.x = (m._11*vec.x) + (m._21*vec.y) + (m._31*vec.z) + m._41;
        v.y = (m._12*vec.x) + (m._22*vec.y) + (m._32*vec.z) + m._42;
        v.z = (m._13*vec.x) + (m._23*vec.y) + (m._33*vec.z) + m._43;

        return v;
    }

    //----------------------------------------------------------------------------
    // Vector4's can be multiple by nay matrix, therefore to rotate a vector3
    // by an arbitary matrix first convert it to a vector4.
    friend inline Vector4 operator *
    (
        const Vector4& v,
        const Matrix4& m
    )
    //--------------------------------------
    {
        // Construct the rotated point directly at the return address
        that_ptr->x = (m._11*v.x) + (m._21*v.y) + (m._31*v.z) + (m._41*v.w);
        that_ptr->y = (m._12*v.x) + (m._22*v.y) + (m._32*v.z) + (m._42*v.w);
        that_ptr->z = (m._13*v.x) + (m._23*v.y) + (m._33*v.z) + (m._43*v.w);
        that_ptr->w = (m._14*v.x) + (m._24*v.y) + (m._34*v.z) + (m._44*v.w);

        return *that_ptr;
    }


    //----------------------------------------------------------------------------
    friend inline Vector4& operator *=
    (
        Vector4&        v,
        const Matrix4&  m
    )
    //--------------------------------------
    {
        Vector4 vec = v;

        v.x = (m._11*v.x) + (m._21*v.y) + (m._31*v.z) + (m._41*v.w);
        v.y = (m._12*v.x) + (m._22*v.y) + (m._32*v.z) + (m._42*v.w);
        v.z = (m._13*v.x) + (m._23*v.y) + (m._33*v.z) + (m._43*v.w);
        v.w = (m._14*v.x) + (m._24*v.y) + (m._34*v.z) + (m._44*v.w);

        return v;
    }


    //
    // Matrix * Matrix
    //

    // A = X*Y
    // Means a is the composite of transform X then transform Y. As usual with
    // matrices X*Y != Y*X.

    //----------------------------------------------------------------------------
    inline Matrix4 operator *
    (
        const Matrix4& y
    ) const
    //--------------------------------------
    {
        Matrix4 res;

        res._11 = _11 * y._11 + _12 * y._21 + _13 * y._31 + _14 * y._41;
        res._12 = _11 * y._12 + _12 * y._22 + _13 * y._32 + _14 * y._42;
        res._13 = _11 * y._13 + _12 * y._23 + _13 * y._33 + _14 * y._43;
        res._14 = _11 * y._14 + _12 * y._24 + _13 * y._34 + _14 * y._44;

        res._21 = _21 * y._11 + _22 * y._21 + _23 * y._31 + _24 * y._41;
        res._22 = _21 * y._12 + _22 * y._22 + _23 * y._32 + _24 * y._42;
        res._23 = _21 * y._13 + _22 * y._23 + _23 * y._33 + _24 * y._43;
        res._24 = _21 * y._14 + _22 * y._24 + _23 * y._34 + _24 * y._44;

        res._31 = _31 * y._11 + _32 * y._21 + _33 * y._31 + _34 * y._41;
        res._32 = _31 * y._12 + _32 * y._22 + _33 * y._32 + _34 * y._42;
        res._33 = _31 * y._13 + _32 * y._23 + _33 * y._33 + _34 * y._43;
        res._34 = _31 * y._14 + _32 * y._24 + _33 * y._34 + _34 * y._44;

        res._41 = _41 * y._11 + _42 * y._21 + _43 * y._31 + _44 * y._41;
        res._42 = _41 * y._12 + _42 * y._22 + _43 * y._32 + _44 * y._42;
        res._43 = _41 * y._13 + _42 * y._23 + _43 * y._33 + _44 * y._43;
        res._44 = _41 * y._14 + _42 * y._24 + _43 * y._34 + _44 * y._44;

        return res;
    }


    //----------------------------------------------------------------------------
    // With this operator we need to be careful not to corrupt the original
    // matrix before we have finished using it!
    inline Matrix4& operator *=
    (
        const Matrix4& m
    )
    //--------------------------------------
    {
        *this = *this * m;

        return *this;
    }

    //----------------------------------------------------------------------------
    // Scale whole matrix  (multiply by constant)
    inline Matrix4 operator *
    (
        const float& y
    ) const
    //--------------------------------------
    {
        return Matrix4( _11 * y, _12 * y, _13 * y, _14 * y,
                        _21 * y, _22 * y, _23 * y, _24 * y,
                        _31 * y, _32 * y, _33 * y, _34 * y,
                        _41 * y, _42 * y, _43 * y, _44 * y );
    }


    //----------------------------------------------------------------------------
    // Scale whole matrix in place (multiply by constant)
    inline Matrix4& operator *=
    (
        const float& y
    )
    //--------------------------------------
    {
        _11 = _11 * y;
        _12 = _11 * y;
        _13 = _11 * y;
        _14 = _11 * y;

        _21 = _21 * y;
        _22 = _21 * y;
        _23 = _21 * y;
        _24 = _21 * y;

        _31 = _31 * y;
        _32 = _31 * y;
        _33 = _31 * y;
        _34 = _31 * y;

        _41 = _41 * y;
        _42 = _41 * y;
        _43 = _41 * y;
        _44 = _41 * y;

        return *this;
    }

    //----------------------------------------------------------------------------
    // Scale who matrix (divide by constant)
    inline Matrix4 operator /
    (
        const float& y
    ) const
    //--------------------------------------
    {
        return Matrix4( _11 / y, _12 / y, _13 / y, _14 / y,
                        _21 / y, _22 / y, _23 / y, _24 / y,
                        _31 / y, _32 / y, _33 / y, _34 / y,
                        _41 / y, _42 / y, _43 / y, _44 / y );
    }


    //----------------------------------------------------------------------------
    // Scale whole matrix in place (divide by constant)
    inline Matrix4& operator /=
    (
        const float& y
    )
    //--------------------------------------
    {
        _11 = _11 / y;
        _12 = _11 / y;
        _13 = _11 / y;
        _14 = _11 / y;

        _21 = _21 / y;
        _22 = _21 / y;
        _23 = _21 / y;
        _24 = _21 / y;

        _31 = _31 / y;
        _32 = _31 / y;
        _33 = _31 / y;
        _34 = _31 / y;

        _41 = _41 / y;
        _42 = _41 / y;
        _43 = _41 / y;
        _44 = _41 / y;

        return *this;
    }


    //----------------------------------------------------------------------------
    // Return the inverse of the current matrix, if the current matrix is singular
    // then the identity matrix is returned.
    Matrix4 operator~
    (
    ) const;


    //
    // Helper functions
    //

    //----------------------------------------------------------------------------
    // Set the matrix to be the 4x4 identity
    inline void SetIdentity
    (
    )
    //--------------------------------------
    {
        *this = Identity;
    }


    //----------------------------------------------------------------------------
    // Set the matrix to be a scaled identity matrix
    inline void SetScale
    (
        float   fScale
    )
    //--------------------------------------
    {
        _12 = _13 = _14 = _21 = _23 = _24 = 0.0f;
        _31 = _32 = _34 = _41 = _42 = _43 = 0.0f;
        _11 = _22 = _33 = fScale;
        _44 = 1.0f;
    }


    //----------------------------------------------------------------------------
    // Set an identity matrix with a translation
    inline void SetTranslation
    (
        Vector3&    vLoc
    )
    //--------------------------------------
    {
        *this = Identity;
        _41 = vLoc.x;
        _42 = vLoc.y;
        _43 = vLoc.z;
    }


    //----------------------------------------------------------------------------
    // Return the transpose of the matrix
    inline Matrix4 Transpose
    (
    ) const
    //--------------------------------------
    {
        return Matrix4
        (
            _11, _21, _31, _41,
            _12, _22, _32, _42,
            _13, _23, _33, _43,
            _14, _24, _34, _44
        );
    }


    //----------------------------------------------------------------------------
    // Set a perspective frustum directly, used in exactly the same way as the
    // standard glFrustum() call in OpenGL. 
    // Note that the projection matrix is normalized for element [3][4] to be 1.0.
    // This is performed so that W-based range fog will work correctly.
    void SetFrustum
    (
        float   left,
        float   right,
        float   bottom,
        float   top,
        float   near,
        float   far
    );


    //----------------------------------------------------------------------------
    // Sets the matrix to a perpsective projection matrix built
    // from the field-of-view (fov, in y), aspect ratio, near plane ,
    // and far plane .
    void SetPerspective
    (
        float fov,          // field of view in Y (full angle)
        float aspect,       // aspect ratio of the render target
        float near_plane,   // Position of near clip plane, output Z at this distance will be 0.0
        float far_plane     // Position of far clip plane, output Z at the distance will be 1.0
    );

    //
    // Debug asist functions
    //
    const static Matrix4 Identity;
};


// Include the correct library
#if defined(_DEBUG)
#pragma comment(lib, "DirectXD.lib")
#else
#pragma comment(lib, "DirectXR.lib")
#endif

#endif //SDK_HEADER_MATRIX
