//-----------------------------------------------------------------------------
// PixelFormat.h
//
// Started 06/25/2000
//
// Shrike Framework -  Rob Wyatt
//-----------------------------------------------------------------------------

#if !defined(PIXELFORMAT_H_INCLUDED)
#define PIXELFORMAT_H_INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "Common.h"
#include "DDraw.h"
#include "D3d.h"
#include "Colors.h"

//-----------------------------------------------------------------------------
// Predefined texture formats, note these do not include things like Z buffer
// formats
enum TextureFormat
{
    TF_NONE     = -1,       //Used as error codes and also when restoring a surface

    TF_RGBA8888 = 0,
    TF_RGB888,
    TF_RGB565,
    TF_RGB555,
    TF_RGBA5551,
    TF_RGBA4444
};



//----------------------------------------------------------------------------
class PixelFormat
{
public:

    //----------------------------------------------------------------------------
    PixelFormat
    (
        TextureFormat   fmt
    );

    //----------------------------------------------------------------------------
    PixelFormat
    (
        DDPIXELFORMAT&  fmt
    );

    //----------------------------------------------------------------------------
    // If we have some flags set it is safe to assume it is a valid pixel format
    bool Valid
    (
    ) const
    //--------------------------------------
    {
        return (m_DDFmt.dwFlags != 0) && (m_DDFmt.dwSize == sizeof(DDPIXELFORMAT));
    }

    //----------------------------------------------------------------------------
    // Given an RGBColor this function wil return a uint32 holding the same colour
    // quantized to the current pixel format. The result is always 32 bits so the
    // caller is repsonible for truncating it.
    uint32 PixelColorFromRGB
    (
        const RGBColor& col
    ) const;

    //
    // Inline member functions
    //

    //----------------------------------------------------------------------------
    // Get the number of bits consumed by each stored pixel, this may not be how
    // big the colours of the pixel are. For example most video cards store RGB888
    // in a 32 bit pixel and waste 8 bits. In this case this member would return 32
    // and not 24.
    uint32 GetBitsPerPixel
    (
    ) const
    //--------------------------------------
    {
        return m_DDFmt.dwRGBBitCount;
    }


    //
    // Return the bitmasks for each color component
    //

    //----------------------------------------------------------------------------
    // Return the bit mask of the red bits
    uint32 GetRedMask
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_DDFmt.dwRBitMask;
    }

    //----------------------------------------------------------------------------
    // Return the bit mask of the green bits
    uint32 GetGreenMask
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_DDFmt.dwGBitMask;
    }

    //----------------------------------------------------------------------------
    // Return the bit mask of the blue bits
    uint32 GetBlueMask
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_DDFmt.dwBBitMask;
    }

    //----------------------------------------------------------------------------
    // Return the bit mask of the alpha bits or 0 if no alpha
    uint32 GetAlphaMask
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_DDFmt.dwRGBAlphaBitMask;
    }


    //
    // Return the bitpositions for each color component
    //

    //----------------------------------------------------------------------------
    // Return the number of bits in the red mask
    uint32 NumberRedBits
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_NumRedBits;
    }

    //----------------------------------------------------------------------------
    // Return the start bit position of the red mask
    uint32 RedBitsStartPos
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_RedShift;
    }

    //----------------------------------------------------------------------------
    // Return the number of bits in the green mask
    uint32 NumberGreenBits
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_NumGreenBits;
    }

    //----------------------------------------------------------------------------
    // Return the start bit position of the green mask
    uint32 GreenBitsStartPos
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_GreenShift;
    }

    //----------------------------------------------------------------------------
    // Return the number of bits in the blue mask
    uint32 NumberBlueBits
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_NumBlueBits;
    }

    //----------------------------------------------------------------------------
    // Return the start bit position of the blue mask
    uint32 BlueBitsStartPos
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_BlueShift;
    }

    //----------------------------------------------------------------------------
    // Return the number of bits in the alpha mask
    uint32 NumberAlphaBits
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_NumAlphaBits;
    }

    //----------------------------------------------------------------------------
    // Return the start bit position of the alpha mask
    uint32 AlphaBitsStartPos
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwFlags & DDPF_RGB);
        return m_AlphaShift;
    }

    //----------------------------------------------------------------------------
    operator const DDPIXELFORMAT&
    (
    ) const
    //--------------------------------------
    {
        Assert(m_DDFmt.dwSize == sizeof(DDPIXELFORMAT));
        return m_DDFmt;
    }

private:

    //----------------------------------------------------------------------------
    bool InitFromD3DPixelFormat
    (
        DDPIXELFORMAT& fmt
    );

    DDPIXELFORMAT   m_DDFmt;

    // Number of bits per component
    uint32       m_NumRedBits;
    uint32       m_NumGreenBits;
    uint32       m_NumBlueBits;
    uint32       m_NumAlphaBits;

    // Number of bits to shift each component down to get the LSB in bit 0
    uint32       m_RedShift;
    uint32       m_GreenShift;
    uint32       m_BlueShift;
    uint32       m_AlphaShift;
};

#endif
