//-----------------------------------------------------------------------------
// Tessellate2D.h
//
// Started 06/24/2000
//
// Shrike Framework -  Rob Wyatt
//-----------------------------------------------------------------------------

#if !defined(TESSELLATE2D_H_INCLUDED)
#define TESSELLATE2D_H_INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Common.h"
#include "Vector.h"

//-----------------------------------------------------------------------------
const uint32 TESS_POSITION   = 0x00000001;
const uint32 TESS_NORMAL     = 0x00000002;

//-----------------------------------------------------------------------------
// Base class for all geometric objects that can be tessellated by 2D parameric
// inputs, such as quadrics or patches.
class Tessellate2D
{
public:
    //-----------------------------------------------------------------------------
    // Constructor
    Tessellate2D
    (
        uint32  tess_u,
        uint32  tess_v
    );

    //-----------------------------------------------------------------------------
    // Virtual destructor
    virtual ~Tessellate2D
    (
    );

    //-----------------------------------------------------------------------------
    // The default implementation of this function will return a valid list for
    // meshes that are tessellated as a grid along lines of U then V. If a derived
    // tessellator cannot generate points in this order then it will have to also
    // implement this function to return the index list.
    virtual uint16* GenerateTriangleListIndices
    (
        uint32& count           // holds the list element count on exit
    );

    //-----------------------------------------------------------------------------
    // This is the only function that a derived class must implement, this function
    // should return a position on the mesh at the given u,v. The default
    // implementation of functions that return normals, binormals or tangents use
    // this function with tiny u and v delta's and then calculate the normal via
    // a cross product.
    // The U,V co-oordinates passed to this function are normalzied in the range
    // 0.0 to 1.0, these will have to be scaled to the correct parametric range
    // by any derived classes.
    virtual Vector3 GeneratePosition
    (
        float   u,
        float   v
    ) = 0;

    //-----------------------------------------------------------------------------
    // The default implementation of this function will take tiny U,V deltas and
    // calculate two other positons on the surface, these will then be crossed to
    // give a normal. If for a given surface type it is possible to calculate a
    // normal faster or more accurate than this then implement a version of this
    // function.
    virtual Vector3 GenerateNormal
    (
        float   u,
        float   v
    );

    //-----------------------------------------------------------------------------
    // Tessellate the whole mesh at the specified tessellation resolution, each
    // new vertex is placed 'vert_stride' bytes apart and the flags element control
    // what elements get inserted into a specific vertex. See the Vertex Components
    // above.
    // A pointer to aa freshly allocated block of memory is returned and the caller
    // is responsible for deleting this, the number of vertices created is placed
    // in vert_count and the size of the block is vert_count*vert_stride. If a 
    // particular mesh wants to perform a specific form of tessellation then it
    // can implement its own version of this function. The default implementation
    // implements a grid tessellation.
    virtual void* Tessellate
    (
        uint32  vert_stride,
        uint32  flags,
        uint32& vert_count
    );

    //-----------------------------------------------------------------------------
    // Helper function to generate a position at one of the specified tessellation
    // points.  
    Vector3 GeneratePosition
    (
        uint32 u,
        uint32 v
    )
    //--------------------------------------
    {
        Assert(u<m_TessU);
        Assert(v<m_TessV);

        return GeneratePosition(m_StartU+(m_StepU*u),m_StartV+(m_StepV*v));
    }

    //-----------------------------------------------------------------------------
    // Helper function to generate a normal at one of the specified tessellation
    // points.  
    Vector3 GenerateNormal
    (
        uint32 u,
        uint32 v
    )
    //--------------------------------------
    {
        Assert(u<m_TessU);
        Assert(v<m_TessV);

        return GenerateNormal(m_StartU+(m_StepU*u),m_StartV+(m_StepV*v));
    }

private:
    float   m_StartU;
    float   m_StartV;
    float   m_StepU;
    float   m_StepV;
    uint32  m_TessU;
    uint32  m_TessV;
};

// Include the correct library
#if defined(_DEBUG)
#pragma comment(lib, "GeomD.lib")
#else
#pragma comment(lib, "GeomR.lib")
#endif

#endif
