//-----------------------------------------------------------------------------
// Vector.h
//
// Started 06/21/2000
//
// Shrike Framework -  Rob Wyatt 
//-----------------------------------------------------------------------------
#if !defined(VECTOR_H_INCLUDED)
#define VECTOR_H_INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif

#include "Common.h"
#include "D3d.h"
#include "Debug.h"
#include "ShrikeMath.h"

//----------------------------------------------------------------------------
class Vector3 : public D3DVECTOR
{
public:
    //----------------------------------------------------------------------------
    Vector3()
    {
    }

    //----------------------------------------------------------------------------
    inline Vector3
    (
        float x_pos,
        float y_pos,
        float z_pos
    )
    //--------------------------------------
    {
        x = x_pos;
        y = y_pos;
        z = z_pos;
    }

    //----------------------------------------------------------------------------
    inline Vector3& operator +=
    (
        const Vector3& v
    )
    //--------------------------------------
    {
        x += v.x;
        y += v.y;
        z += v.z;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector3 operator +
    (
        const Vector3& v
    ) const
    //--------------------------------------
    {
        return Vector3(x+v.x, y+v.y, z+v.z);
    }

    //----------------------------------------------------------------------------
    inline Vector3& operator -=
    (
        const Vector3& v
    )
    //--------------------------------------
    {
        x -= v.x;
        y -= v.y;
        z -= v.z;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector3 operator -
    (
        const Vector3& v
    ) const
    //--------------------------------------
    {
        return Vector3(x - v.x, y - v.y, z - v.z);
    }

    //----------------------------------------------------------------------------
    inline Vector3 operator -
    (
    ) const
    //--------------------------------------
    {
        return Vector3(-x, -y, -z);
    }

    //----------------------------------------------------------------------------
    inline Vector3& operator *=
    (
        float f
    )
    //--------------------------------------
    {
        x *= f;
        y *= f;
        z *= f;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector3 operator *
    (
        float f
    ) const
    //--------------------------------------
    {
        return Vector3(x*f, y*f, z*f);
    }

    //----------------------------------------------------------------------------
    inline Vector3& operator /=
    (
        float f
    )
    //--------------------------------------
    {
        // Cannot divide by zero
        Assert(f != 0.0f);

        // It is quicker to take 1/f the multiply but the multiply will cause a
        // stall because the divide will not have finished.
        f = 1.0f/f;
        x *= f;
        y *= f;
        z *= f;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector3 operator /
    (
        float f
    ) const
    //--------------------------------------
    {
        Assert(f != 0.0f);

        f = 1.0f/f;
        return Vector3(x*f, y*f, z*f);
    }

    //----------------------------------------------------------------------------
    // The dot-product is performed by the * operator
    inline float operator *
    (
        const Vector3& v
    ) const
    //--------------------------------------
    {
        return x*v.x + y*v.y + z*v.z;
    }

    //----------------------------------------------------------------------------
    // The cross-product is performed by the ^ operator
    inline Vector3 operator ^
    (
        const Vector3& v
    ) const
    //--------------------------------------
    {
        return Vector3(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
    }

    //----------------------------------------------------------------------------
    // The reflection operation is performed by the | operator
    inline Vector3 operator |
    (
        const Vector3& normal
    ) const
    //--------------------------------------
    {
        // reflection = 2(vec dot normal)normal-Vector
        return Vector3( (normal - (*this)) * 2.0f * ((*this)*normal) );
    }

    //----------------------------------------------------------------------------
    inline bool operator ==
    (
        const Vector3& v
    ) const
    //--------------------------------------
    {
        return (x==v.x) && (y==v.y) && (z==v.z);
    }

    //
    // Vector3 Helper functions
    //

    //----------------------------------------------------------------------------
    inline float Magnitude
    (
    )
    //--------------------------------------
    {
        return sqrtf(x * x + y * y + z * z);
    }


    //----------------------------------------------------------------------------
    inline Vector3& Normalize
    (
    )
    //--------------------------------------
    {
        float f = 1.0f / Magnitude();

        x *= f;
        y *= f;
        z *= f;

        return *this;
    }

    //----------------------------------------------------------------------------
    inline bool IsNormalized
    (
    )
    //--------------------------------------
    {
        return fabsf(Magnitude() - 1.0f) < NORMAL_EPSILON;
    }
};



//----------------------------------------------------------------------------
// At the moment D3D does not have a vector type that we can derive a
// 4 element homogeneous vector from, although a D3DTLVERTEX is a homogeneous
// position by definition.
class Vector4
{
public:
    //----------------------------------------------------------------------------
    inline Vector4()
    {
    }

    //----------------------------------------------------------------------------
    inline Vector4
    (
        float x_pos,
        float y_pos,
        float z_pos,
        float w_pos
    ) : x(x_pos), y(y_pos), z(z_pos), w(w_pos)
    //--------------------------------------
    {
    }

    //----------------------------------------------------------------------------
    inline Vector4& operator +=
    (
        const Vector4& v
    )
    //--------------------------------------
    {
        x += v.x;
        y += v.y;
        z += v.z;
        w += v.w;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector4 operator +
    (
        const Vector4& v
    ) const
    //--------------------------------------
    {
        return Vector4(x+v.x, y+v.y, z+v.z, w+v.w);
    }

    //----------------------------------------------------------------------------
    inline Vector4& operator -=
    (
        const Vector4& v
    )
    //--------------------------------------
    {
        x -= v.x;
        y -= v.y;
        z -= v.z;
        w -= v.w;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector4 operator -
    (
        const Vector4& v
    ) const
    //--------------------------------------
    {
        return Vector4(x - v.x, y - v.y, z - v.z, z - v.w);
    }

    //----------------------------------------------------------------------------
    inline Vector4 operator -
    (
    ) const
    //--------------------------------------
    {
        return Vector4(-x, -y, -z, -w);
    }

    //----------------------------------------------------------------------------
    inline Vector4& operator *=
    (
        float f
    )
    //--------------------------------------
    {
        x *= f;
        y *= f;
        z *= f;
        w *= f;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector4 operator *
    (
        float f
    ) const
    //--------------------------------------
    {
        return Vector4(x*f, y*f, z*f, w*f);
    }

    //----------------------------------------------------------------------------
    inline Vector4& operator /=
    (
        float f
    )
    //--------------------------------------
    {
        // Cannot divide by zero
        Assert(f != 0.0f);

        // It is quicker to take 1/f the multiply but the multiply will cause a
        // stall because the divide will not have finished.
        f = 1.0f/f;
        x *= f;
        y *= f;
        z *= f;
        w *= f;
        return *this;
    }

    //----------------------------------------------------------------------------
    inline Vector4 operator /
    (
        float f
    ) const
    //--------------------------------------
    {
        Assert(f != 0.0f);

        f = 1.0f/f;
        return Vector4(x*f, y*f, z*f, w*f);
    }

    //----------------------------------------------------------------------------
    // The dot-product is performed by the * operator
    inline float operator *
    (
        const Vector4& v
    ) const
    //--------------------------------------
    {
        return x*v.x + y*v.y + z*v.z + w*v.w;
    }

    //----------------------------------------------------------------------------
    // Vector equal
    inline bool operator ==
    (
        const Vector4& v
    ) const
    //--------------------------------------
    {
        return (x==v.x) && (y==v.y) && (z==v.z) && (w==v.w);
    }

    //
    // Vector4 Helper functions
    //

    //----------------------------------------------------------------------------
    inline float Magnitude
    (
    )
    //--------------------------------------
    {
        return sqrtf(x * x + y * y + z * z + w * w);
    }



    //----------------------------------------------------------------------------
    inline Vector4& Normalize
    (
    )
    //--------------------------------------
    {
        float f = 1.0f / Magnitude();

        x *= f;
        y *= f;
        z *= f;
        w *= f;

        return *this;
    }

    //----------------------------------------------------------------------------
    // Returns a boolean indicating if the length of the vector is 1.0f, or
    // within epsilon amount.
    inline bool IsNormalized
    (
    )
    //--------------------------------------
    {
        return fabsf(Magnitude() - 1.0f) < NORMAL_EPSILON;
    }

    float   x;
    float   y;
    float   z;
    float   w;
};



#endif


