//-----------------------------------------------------------------------------
// WinShell.h
//
// Started 06/09/2000
//
// Shrike Framework -  Rob Wyatt 
//-----------------------------------------------------------------------------
#if !defined(WINSHELL_H_INCLUDED)
#define WINSHELL_H_INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Common.h"
#include "Debug.h"
#include "Timer.h"
#include "3DDevice.h"

#define WINSHELL_CLASS "SHRIKE_WINDOW_CLASS"

//------------------------------------------------------------------------------
// System abstraction shell
class Shell
{
public:
    //------------------------------------------------------------------------------
    Shell
    (
    );

    //------------------------------------------------------------------------------
    virtual ~Shell
    (
    );


    //
    // No virtual Member functions
    //

    //------------------------------------------------------------------------------
    // Return true if the app is to start otherwise false, calling Run after this
    // function returns 'false' will cause it to exit immediately.
    bool Initialize
    (
        const char* app_title,
        const char* menu_title,
        const char* reg_key_name,
        uint32  xsize,                  // size of the back buffer
        uint32  ysize,
        uint32  render_bpp,
        uint32  depth_bpp,
        bool    stencil
    );

    //------------------------------------------------------------------------------
    // Start the application running, the return value is the exit code of the
    // framework if Exit() is called or will return 0 if the application is closed
    // naturally (Alt/F4, close icon etc). If the application exits for an unknown
    // reason 0xffffffff is returned.
    uint32 Run
    (
    );


protected:
    //------------------------------------------------------------------------------
    // Enum containing mouse actions, protected for this class only
    enum MouseAction
    {
        MA_MOVING,      // Moving over window with no buttons pressed
        MA_CLICK,       // Initial click
        MA_DRAG,        // Moving with button pressed
        MA_RELEASE      // Button released
    };

    //------------------------------------------------------------------------------
    // Exit the framework with the given exit code. Simialr to exit() of stdlib.
    void Exit
    (
        uint32  exit_code
    );

    //------------------------------------------------------------------------------
    // Returns the D3D interface pointer associatd with the render device class
    // that this window uses. Useful in functions that do not get passed a device
    // pointer.
    IDirect3DDevice7* GetD3DDevice
    (
    )
    //--------------------------------------
    {
        Assert(m_3DDevice);
        return m_3DDevice->GetDevice();
    }

    //------------------------------------------------------------------------------
    // Returns the class of the render device to enable its members to be called
    RenderDevice* GetDeviceClass
    (
    )
    //--------------------------------------
    {
        Assert(m_3DDevice);
        return m_3DDevice;
    }

    //------------------------------------------------------------------------------
    // Set the wait for vsync status, set to true and the window update will be
    // ayncronized with the monitor refresh otherwise window updates occur as fast
    // as the video can produce the frames.
    void SetWaitVSync
    (
        bool wait
    )
    //--------------------------------------
    {
        m_WaitVSync = wait;
    }

    //------------------------------------------------------------------------------
    // Disables/enables the automatic window title update with frame rate.
    // In some video card configurations updating the title bar causes serious frame
    // rate problems, if any frame rate glitches occur with the update of the title
    // (2 second intervals) then call this function to disble the frame rate counter.
    void EnableFrameCounter
    (
        bool enable
    )
    //--------------------------------------
    {
        m_ShowTitleUpdates = enable;
    }

    //------------------------------------------------------------------------------
    // Shows or removes a selection tick from the specified menu item, set enable to
    // 'true' to tick the item or 'false' to untick it.
    void SelectMenu
    (
        uint32  id,      // ID of the menu item
        bool    enable   // 'true' to tick, 'false' to untick
    );


    //------------------------------------------------------------------------------
    // Read data from the registry under value 'key', if the key (or parent key)
    // cannot be found or any other error occurs such as the key not containing
    // enough data false is returned. True is returned for success. If false is
    // returned then size bytes from def_ptr are copied to the dest pointer if
    // def_ptr is set.
    bool ReadRegData
    (
        const char* key,
        void*       ptr,
        uint32      size,
        void*       def_ptr
    );

    //------------------------------------------------------------------------------
    // Write size bytes to the registry under key 'key'.
    // Returns true for success otherwise false
    bool WriteRegData
    (
        const char* key,
        void*       ptr,
        uint32      size
    );

    //------------------------------------------------------------------------------
    // This will delete the current render device in the usual way and then recreate
    // it. The DeviceCheck function will be called as usuall. This enables the logic
    // in the device check function to change to accept or reject other devices
    // such as force hardware or force software.
    // 'True' is returned if the restart is successful otherwise false is returned,
    // in this case the app is suspended and a message displayed in the window.
    // The new device will be in the old window at the current size.
    // NOTE: The app is not fully shutdown, DeviceClose() is called, followed by
    // DeviceCheck() for each device and then DeviceSetup().
    bool RestartDevice
    (
    );

    //
    // Virtual overrides
    //

    //------------------------------------------------------------------------------
    // Called while the render device is being initialized, this function is called
    // for every device enumerated. The app can override this function and check the
    // device caps to see if this device is capable of doing what the app requires,
    // 'true' should be returned to consider the device or 'false' to ignore it.
    // The devices passed are HW T&l, HW, SW, Reference, if all devices are rejected
    // then an error is displayed and the application exits. This function is called
    // whenever a render device is created including mode changes so it is advisable
    // not to have user input within this function. Also D3D is not fully setup and
    // should not be used.
    // The default behaviour if this member is not implemented is to accept every
    // device.
    virtual bool DeviceCheck
    (
        D3DDEVICEDESC7* device
    );

    //------------------------------------------------------------------------------
    // Application setup, similar to a class constructor except that the registry and
    // window are open and functioning so windows code can be called. There is no
    // 2D or 3D device created so D3d code cannot be executed.
    virtual bool Setup
    (
        bool    first_run
    );

    //------------------------------------------------------------------------------
    // This is called after all devices have been destroyed and resources freed, this
    // is the last entry point before the Run() member returns to the caller. The
    // registry key passed into this function is the same as the key to the Setup()
    // function and it is intended to be used to record persistent data so it can
    // be reused next time the application is run.
    virtual bool Closedown
    (
    );

    //------------------------------------------------------------------------------
    // This member is called after the 3D device and rendering surfaces have been
    // created. In this function you should setup rendering states and create any
    // additional surfaces such as textures and vertex buffers. If the screen mode
    // is changes this member will be called again to initialize the new device,
    // this member is also called when the window is resized and the render surface
    // is resized. This is so you can reset the projection matrix and any viewport
    // parameters.
    virtual bool DeviceSetup
    (
        IDirect3DDevice7*   device,
        bool                resize
    );

    //------------------------------------------------------------------------------
    // This member is called before the 3d device and rendering surface is destroyed,
    // this could be at the end of the application or for example when the screen mode
    // changes. In this function destroy any addition surfaces, textures and vertex
    // buffers that were created in DeviceSetup()
    virtual bool DeviceClose
    (
        IDirect3DDevice7*   device
    );

    //------------------------------------------------------------------------------
    // Called to restore the contents of textures and VBs after they have been lost.
    // If you cannot restore then return 'false' and the application will exit.
    // Restore is called before the Step() function of the frame that needs restoring,
    // the render buffers (primary, render and depth) are automatically restored for
    // you.
    virtual bool Restore
    (
        IDirect3DDevice7*   device
    );

    //------------------------------------------------------------------------------
    // Called once per frame with the timing info from the last frame, this function
    // should be used to update animation and other time based actions.
    virtual bool Step
    (
        IDirect3DDevice7*   device,
        float               frame_time,         // Time stamp of this frame
        float               frame_step          // Time elapsed since the last frame
    );

    //------------------------------------------------------------------------------
    // Called to render the scene, the current render target is not cleared but this
    // function is not responsible for copying the back buffer to the window, that 
    // will be done by the shell if this function returns true.
    virtual bool Render
    (
        IDirect3DDevice7*   device
    );

    //------------------------------------------------------------------------------
    // Called when a menu item is selected, enables derived classes to get input
    // from the user. If the command is known return true otherwise return false.
    virtual bool MenuSelect
    (
        uint32  menu_item_id
    );

    //------------------------------------------------------------------------------
    // Called for various mouse actions (defined in enum above).
    // Co-ordnates are window position apart from when MA_DRAG is the action then
    // x and y are deltas from the last call.
    virtual bool MouseClick
    (
        MouseAction action,
        int32       x,          // a delta if 'action' is MA_DRAG otherwise a position
        int32       y           // a delta if 'action' is MA_DRAG otherwise a position
    );

    //------------------------------------------------------------------------------
    // Standard Windows message pump, if this is overridden then the base class
    // *MUST* be called for the default behaviour.
    virtual LRESULT MessagePump
    (
        HWND    hwnd,
        UINT    msg,
        WPARAM  wp,
        LPARAM  lp
    );


private:

    //------------------------------------------------------------------------------
    // Callback from RenderDevice Initialize() for render device selection, this
    // functions passes the call on to a virtual member and therefore to the app.
    static bool DeviceCallback
    (
        void* context,
        D3DDEVICEDESC7* device
    );

    //------------------------------------------------------------------------------
    // Initial callback function for all windows of the the WINSHELL class type
    static LRESULT CALLBACK SceneProc
    (
        HWND    hwnd,
        UINT    msg,
        WPARAM  wp,
        LPARAM  lp
    );


    //------------------------------------------------------------------------------
    // Perform basic validation of the screen mode, such as compare the requested
    // bit depth to the windows desktop bit depth. This function displays messages
    // and errors in windows message boxes.
    bool ValidateScreenMode
    (
        uint32  screen_width,
        uint32  screen_height,
        uint32  screen_bpp,
        uint32  width,
        uint32  height,
        uint32  bpp,
        bool    mode_change
    );

    //------------------------------------------------------------------------------
    // Given the GUID of the device that has been created this function will create
    // a name suitable for displaying in the window title and place it in the
    // m_DeviceTitle array
    void NameDevice
    (
        const GUID&   dev
    );


    HINSTANCE       m_AppInstance;
    HWND            m_MainWindow;
    HKEY            m_AppRegKey;
    bool            m_HasMenu;
    bool            m_Paused;
    bool            m_Quit;
    bool            m_FirstFrame;
    bool            m_Capture;
    bool            m_TitleUpdate;
    bool            m_ShowTitleUpdates;
    bool            m_BadSetup;
    bool            m_WaitVSync;
    Timer           m_FrameTimer;
    float           m_LastFrameTime;
    float           m_LastTitleUpdate;
    char            m_AppTitle[64];
    char            m_DeviceTitle[32];
    char            m_SuspendMsg[128];
    RenderDevice*   m_3DDevice;

    // Setup parameters
    uint32          m_ClientWidth;
    uint32          m_ClientHeight;
    uint32          m_RenderBpp;
    uint32          m_ZBits;
    bool            m_CreateWithStencil;
};


// Include the correct library
#if defined(_DEBUG)
#pragma comment(lib, "WinShellD.lib")
#else
#pragma comment(lib, "WinShellR.lib")
#endif

#endif
