//-----------------------------------------------------------------------------
// Main.cpp
//
// Started 06/09/2000
//
// Shrike Framework -  Rob Wyatt 
//-----------------------------------------------------------------------------
#include "TestApp.h"

//------------------------------------------------------------------------------
// Standard macro to reroute new allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//------------------------------------------------------------------------------
bool TestApp::DeviceCheck
(
    D3DDEVICEDESC7* device
)
//--------------------------------------
{
    // We allow any device (this is the default action of the base class but we
    // implement it here in the test app for completeness and debugging).
    return true;
}


//------------------------------------------------------------------------------
bool TestApp::Setup
(
    bool    first_run
)
//--------------------------------------
{
    m_MouseX = 0.0f;
    m_MouseY = 0.0f;

    // Read the world matrix from the registry, if it fails or is not present
    // then copy the Idenity instead.
    ReadRegData("WorldRot",&m_WorldMat,sizeof(Matrix4),(void*)&Matrix4::Identity);

    // Create one of the shrike test objects
    m_UnitCube = TestObject::MakeVertexList(TestObject::TEST_OBJECT_INDEXED_UNIQUE_CUBE,m_VertCount);
    m_Index = TestObject::MakeTriangleListIndices(TestObject::TEST_OBJECT_INDEXED_UNIQUE_CUBE,m_IndexCount);

    return true;
}


//------------------------------------------------------------------------------
bool TestApp::Closedown
(
)
//--------------------------------------
{
    // Delete the test object
    delete m_UnitCube;
    delete m_Index;

    // Write the world matrix to the registry under our apps key
    WriteRegData("WorldRot",&m_WorldMat,sizeof(Matrix4));

    return true;
}


//------------------------------------------------------------------------------
bool TestApp::DeviceSetup
(
    IDirect3DDevice7*   device,
    bool                resize
)
//--------------------------------------
{
    // Set Projection Matrix
    D3DVIEWPORT7 vp;
    device->GetViewport(&vp);
    Matrix4 projection = Matrix4(PI_DIV_4, ((float)vp.dwHeight) / ((float)vp.dwWidth), 1.0f, 20.0f);
    device->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &projection);

    // If we are just resizing the all we need to do is reset the projection matrix
    // account for the new viewport.
    if (resize)
        return true;

    // Set the view transform to be looking straight at the origin
    Vector3 eye = Vector3(0.0f, 0.0f, -8.5f);
    Vector3 lookat = Vector3(0.0f, 0.0f,  0.0f);
    Vector3 up = Vector3(0.0f, 1.0f,  0.0f);
    Matrix4 view_mat(eye, lookat, up );
    device->SetTransform(D3DTRANSFORMSTATE_VIEW, &view_mat);

    // Set up a point light (D3D light 0) and set the ambient light level
    PointLight light = PointLight(Vector3(0.0f,0.0f,-10.0f), RGBColor(1.0f,1.0f,1.0f) );
    device->SetLight( 0, &light );
    
    device->SetRenderState(D3DRENDERSTATE_AMBIENT, RGBColor(0.15f,0.15f,0.15f) );
    device->LightEnable( 0, true );

    // By default both diffuse and ambient colors come from the diffuse element of
    // a vertex, therefore no material is required.

    // Load an RGB888 texture and set it as texture 0, also set texture 0's filter state
    m_BoxTexture = RenderDevice::CreateTexture(ID_BITMAP_TEST, RenderDevice::MatchFormat(TF_RGB888) );

    device->SetTexture(0,m_BoxTexture->GetTexture());
    device->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTFN_LINEAR );
    device->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTFN_LINEAR );

    return true;
}


//------------------------------------------------------------------------------
bool TestApp::DeviceClose
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    delete m_BoxTexture;
    return true;
}


//------------------------------------------------------------------------------
bool TestApp::Restore
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    m_BoxTexture->LoadFile(ID_BITMAP_TEST);
    return true;
}


//------------------------------------------------------------------------------
bool TestApp::Step
(
    IDirect3DDevice7*   device,
    float               frame_time,
    float               frame_step
)
//--------------------------------------
{
    // Make a world space aligned rotation from the mouse deltas that are calculated
    // above. Keep a copy of the world matrix because we need it on the next frame.
    Matrix4 xrot = EulerAngle(EULER_X,-m_MouseY * frame_step*PI_MUL_2);
    Matrix4 yrot = EulerAngle(EULER_Y,-m_MouseX * frame_step*PI_MUL_2);
    m_WorldMat = m_WorldMat*yrot*xrot;
    device->SetTransform(D3DTRANSFORMSTATE_WORLD, &m_WorldMat);

    return true;
}


//------------------------------------------------------------------------------
bool TestApp::Render
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    device->Clear(0,NULL,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0x00000000,1.0f,0);
   
    if (device->BeginScene() != S_OK)
        return false;

    // Draw the cube with a single call, the vertices are in system memory
    device->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, DiffuseTextureVertex1::FORMAT,
        m_UnitCube, m_VertCount,
        m_Index, m_IndexCount,
        0);

    device->EndScene();

    return true;
}


//------------------------------------------------------------------------------
bool TestApp::MenuSelect
(
    uint32 menu_id
)
//--------------------------------------
{
    switch (menu_id)
    {
    case IDM_EXIT:              // File->Exit has been selected
        Exit( EXIT_SUCCESS );
        break;

    default:
        return false;
    }

    return true;
}


//------------------------------------------------------------------------------
bool TestApp::MouseClick
(
    MouseAction action,
    int32       x,
    int32       y
)
//--------------------------------------
{
    if (action == MA_CLICK)
    {
        // When we first click set the rotation to be still
        m_MouseX = 0.0f;
        m_MouseY = 0.0f;
    }
    else if (action == MA_DRAG)
    {
        // When captured the mouse position is a delta, capture is effectively
        // set while the mouse is being dragged. Divide the deltas by some
        // randomly choosen factor for the mouse sensitivity
        m_MouseX = ((float)x)/50.0f;    
        m_MouseY = ((float)y)/50.0f;
    }

    // When the mouse is released the object will continue rotating at the
    // current speed.

    return true;
}


//------------------------------------------------------------------------------
int APIENTRY WinMain
(
    HINSTANCE instance,
    HINSTANCE prev_instance,
    LPSTR     cmd_line,
    int       cmd_show
)
//--------------------------------------
{
    TestApp   test_app;

    if (!test_app.Initialize(
            "Shrike 7 Test",                    // Application Title
            MAKEINTRESOURCE(IDR_SHRIKE_MENU),   // Menu or NULL for no menu
            "TestApp",                          // Registry key or NULL for no registry
            640,                                // Window width
            480,                                // Window height
            0xffffffff,                         // Render buffer bit depth, 0xffffffff for don't care
            0xffffffff,                         // Z Buffer bit depth, 0 for no Z, 0xffffffff for don't care
            false                               // Stencil buffer
            ))
    {
        return EXIT_FAILURE;
    }

	return test_app.Run();
}
