//-----------------------------------------------------------------------------
// 3DDevice.cpp
//
// Started 06/20/2000
//
// Shrike Framework -  Rob Wyatt
//-----------------------------------------------------------------------------

#include "3DDevice.h"
#include "Debug.h"
#include "HlpLib.h"
#include "Textures.h"
#include "PixelFormat.h"

//------------------------------------------------------------------------------
// Standard macro to reroute new allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


#define MAX_FORMATS 64

//-----------------------------------------------------------------------------
// Struct used internally to this file for pixel format callbacks
struct PixelFormatCallbackData
{
    RenderDevice*   m_Device;
    DDPIXELFORMAT*  m_FormatArray;
    uint32*         m_FormatCount;
};

//-----------------------------------------------------------------------------
// Struct used internally to this file for device callbacks
struct DeviceCallbackData
{
    RenderDevice*           m_Device;
    RenderDeviceCallback    m_UserFunction;
    void*                   m_FunctionContext;
};


//-----------------------------------------------------------------------------
IDirectDraw7*           RenderDevice::m_DD = NULL;
IDirect3D7*             RenderDevice::m_D3D = NULL;
 bool                    RenderDevice::m_UsingHardware = true;
GUID                    RenderDevice::m_DeviceType = {0};
uint32                  RenderDevice::m_NumTextureFormats;
DDPIXELFORMAT*          RenderDevice::m_TexturePixelFormats = NULL;
IDirectDrawSurface7*    RenderDevice::m_DDPrimarySurface = NULL;
IDirectDrawSurface7*    RenderDevice::m_DDRenderSurface = NULL;
IDirectDrawSurface7*    RenderDevice::m_DepthBuffer = NULL;
IDirect3DDevice7*       RenderDevice::m_D3DDevice = NULL;

//-----------------------------------------------------------------------------
RenderDevice::RenderDevice
(
)
//--------------------------------------
{
    m_DD   = NULL;
    m_D3D  = NULL;
    m_DDPrimarySurface = NULL;
    m_DDRenderSurface = NULL;
    m_DepthBuffer = NULL;
    m_D3DDevice = NULL;
    m_TexturePixelFormats = NULL;
}


//-----------------------------------------------------------------------------
RenderDevice::~RenderDevice
(
)
//--------------------------------------
{
    delete m_TexturePixelFormats;

    // Before we delete the device evit any D3D managed textures that could
    // be remaining. By the this destructor is reached all textures should
    // be long gone.
    if (m_D3D)
    {
        m_D3D->EvictManagedTextures(); 
    }

    if (m_D3DDevice != NULL)
    {
        m_D3DDevice->Release();
        m_D3DDevice = NULL;
    }

    if (m_DepthBuffer != NULL)
    {
        // remove the Z Buffer attachment
        if (m_DDRenderSurface)
            m_DDRenderSurface->DeleteAttachedSurface(0,m_DepthBuffer);

        // Now release the Z buffer
        m_DepthBuffer->Release();
        m_DepthBuffer = NULL;
    }

    // Free the render surface
    if (m_DDRenderSurface)
    {
        m_DDRenderSurface->Release();
        m_DDRenderSurface = NULL;
    }

    // Free the primary surface
    if (m_DDPrimarySurface)
    {
        m_DDPrimarySurface->Release();
        m_DDPrimarySurface = NULL;
    }

    // Free the 3D device
    if (m_D3D)
    {
        m_D3D->Release();
        m_D3D = NULL;
    }

    // Free the 2D device
    if (m_DD)
    {
        m_DD->Release();
        m_DD = NULL;
    }
}


//-----------------------------------------------------------------------------
bool RenderDevice::Initialize
(
    HWND                    hwnd,
    uint32                  width,
    uint32                  height,
    uint32                  colorbpp,
    uint32                  zbpp,
    bool                    stencil,
    RenderDeviceCallback    func,
    void*                   func_context
)
//--------------------------------------
{
    // Copy the input parameters to class members
    m_RenderWindow  = hwnd;
    m_RenderWidth   = width;
    m_RenderHeight  = height;
    m_RenderBpp     = colorbpp;
    m_ZBits         = zbpp;
    m_StencilBuffer = stencil;

    // We currently have no texture formats;
    m_NumTextureFormats = 0;
    Assert(m_TexturePixelFormats == NULL);
    

    // Create the DD component and check its flags for being useful..
    if (!Create2D())
    {
        return false;
    }

    // Create a suitable device along with render surfaces
    if (!Create3D(func,func_context))
    {
        return false;
    }

    //
    // Create a clipper object and attach it to the window
    //

    IDirectDrawClipper* clipper;
    if (m_DD->CreateClipper( 0, &clipper, NULL) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Initialize() Create clipper\n");
        return false;
    }

    // Associate the clipper with the window
    clipper->SetHWnd( 0, m_RenderWindow );
    m_DDPrimarySurface->SetClipper( clipper );

    //delete the clipper object because no need it no more
    clipper->Release();

    // Create the viewport at the size passed in, this can be less than
    // the the size of the surface but cannot be more
    D3DVIEWPORT7 vp = { 0, 0, width, height, 0.0f, 1.0f };

    if (m_D3DDevice->SetViewport( &vp ) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Initialize() SetViewport()\n");
        return false;
    }

    // Initially both the diffuse and ambient colors come from the vertices.
    m_D3DDevice->SetRenderState(D3DRENDERSTATE_DIFFUSEMATERIALSOURCE,D3DMCS_COLOR1);
    m_D3DDevice->SetRenderState(D3DRENDERSTATE_AMBIENTMATERIALSOURCE,D3DMCS_COLOR1);

    return true;
}

//-----------------------------------------------------------------------------
bool RenderDevice::Create2D
(
)
//--------------------------------------
{
    DDSURFACEDESC2          desc;

    // We must use CreateEx for DD7 otherwise D3D7 will fail to create
    // We create the NULL device for the device associated with GDI.
    if (DirectDrawCreateEx( NULL, (VOID**)&m_DD, IID_IDirectDraw7, NULL ) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Create2D() DirectDrawCreate()\n");
        return false;
    }

    // Get the caps
    DDCAPS  caps;
    caps.dwSize = sizeof(caps);
    if (m_DD->GetCaps(&caps,NULL) != S_OK)
    {
        return false;
    }

    if ((caps.dwCaps & DDCAPS_3D) == 0)
    {
        DBGSTR("ERROR: RenderDevice::Create2D() No 3D acceleration\n");
        return false;
    }

    if ((caps.dwCaps2 & DDCAPS2_CANRENDERWINDOWED) == 0)
    {
        DBGSTR("ERROR: RenderDevice::Create2D() Device cannot render to a window\n");
        return false;
    }

    // Set the co-operative level to normal, reuqired for running in a window
    if (m_DD->SetCooperativeLevel(m_RenderWindow, DDSCL_NORMAL) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Create2D() Set Co-op level\n");
        return false;
    }

    //
    // Create the primary surface
    //

    memset(&desc, 0, sizeof(desc));
    desc.dwSize = sizeof(desc);
    desc.dwFlags = DDSD_CAPS;
    desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

    if (m_DD->CreateSurface(&desc, &m_DDPrimarySurface, NULL) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Create2D() Create Primary surface\n");
        return false;
    }

    return true;
}


//-----------------------------------------------------------------------------
HRESULT CALLBACK RenderDevice::DeviceCallback
(
    LPSTR               device_des, 
    LPSTR               device_name, 
    LPD3DDEVICEDESC7    device_desc,
    LPVOID              context
)
//--------------------------------------
{
    DeviceCallbackData* dev_data = (DeviceCallbackData*)context;

    // We first vet this device against what was asked for by the creator, then
    // check some standard features.
    if (dev_data->m_Device->m_RenderBpp != 0xffffffff)
    {
        // We have specified a particualr bit depth, if a device cannot do it
        // then we reject it.
        if (dev_data->m_Device->m_RenderBpp == 16)
        {
            if ((device_desc->dwDeviceRenderBitDepth & DDBD_16) == 0)
            {
                // This device is no good to us, go to the next
                return D3DENUMRET_OK;
            }
        }
        else
        {
            Assert(dev_data->m_Device->m_RenderBpp == 32); //we only support 16 and 32 bit rendering

            if ((device_desc->dwDeviceRenderBitDepth & DDBD_32) == 0)
            {
                // This device is no good to us, go to the next
                return D3DENUMRET_OK;
            }
        }
    }

    // 0xffffffff for don't care, 0 for no Z buffer
    if ((dev_data->m_Device->m_ZBits != 0xffffffff) && (dev_data->m_Device->m_ZBits != 0))
    {
        // We have specified a particualr Z bit depth, if a device cannot do it
        // then we reject it.
        if (dev_data->m_Device->m_ZBits == 16)
        {
            if ((device_desc->dwDeviceZBufferBitDepth & DDBD_16) == 0)
            {
                // This device is no good to us, go to the next
                return D3DENUMRET_OK;
            }
        }
        else if (dev_data->m_Device->m_ZBits == 24)
        {
            if ((device_desc->dwDeviceZBufferBitDepth & DDBD_24) == 0)
            {
                // This device is no good to us, go to the next
                return D3DENUMRET_OK;
            }
        }
        else
        {
            if ((device_desc->dwDeviceZBufferBitDepth & DDBD_32) == 0)
            {
                // This device is no good to us, go to the next
                return D3DENUMRET_OK;
            }
        }
    }

    if (dev_data->m_Device->m_StencilBuffer)
    {
        // We have specified that we want a stencil buffer,check that this device
        // has a stencil buffer. There is no cap bit for a stencil buffer so we
        // just look to see if there are any stencil operators
        if (device_desc->dwStencilCaps == 0)
        {
            // This device is no good to us, go to the next
            return D3DENUMRET_OK;
        }
    }

    // We then ask the creator via a callback if this device is to be used,
    // then we check its device type and add it to the list.
    if (dev_data->m_UserFunction)
    {
        if (!dev_data->m_UserFunction(dev_data->m_FunctionContext,device_desc))
        {
            // The app supplied callback function has rejected this device
            return D3DENUMRET_OK;
        }
    }

    // Which device type is this??
    if (device_desc->deviceGUID == IID_IDirect3DTnLHalDevice)
    {
        if ((device_desc->dwDevCaps & D3DDEVCAPS_HWRASTERIZATION) &&
            (device_desc->dwDevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT))
        {
            dev_data->m_Device->m_FoundTnL = true;
        }
    }
    else if (device_desc->deviceGUID == IID_IDirect3DHALDevice)
    {
        if (device_desc->dwDevCaps & D3DDEVCAPS_HWRASTERIZATION)
        {
            dev_data->m_Device->m_FoundHW = true;
        }
    }
    else if (device_desc->deviceGUID == IID_IDirect3DRGBDevice)
    {
        dev_data->m_Device->m_FoundSW = true;
    }
    else if (device_desc->deviceGUID == IID_IDirect3DRefDevice)
    {
        dev_data->m_Device->m_FoundRef = true;
    }


    return D3DENUMRET_OK;       // give us the next device
}


//-----------------------------------------------------------------------------
HRESULT CALLBACK RenderDevice::ZBufferCallback
(
    LPDDPIXELFORMAT pixel_format,  
    LPVOID          context    
)
//--------------------------------------
{
    PixelFormatCallbackData* format_data = (PixelFormatCallbackData*)context;
    RenderDevice*            device = format_data->m_Device;

    if (device->m_StencilBuffer)
    {
        // We have requested a stencil buffer so only allow Z buffer formats that
        // have stencil bits.
        if ((pixel_format->dwStencilBitDepth) == 0)
        {
            // No stencil buffer
            return D3DENUMRET_OK;
        }
    }

    if (device->m_ZBits != 0xffffffff)
    {
        // We have requested a particular Z bit depth so we must match it but 
        // pixel_format->dwZBufferBitDepth inlcudes any stencil bits we may
        // not care about as it is OK to create a stencil buffer even if
        // one if one is not required. Therefore we should subtract the
        // stencil bit count
        if ((pixel_format->dwZBufferBitDepth-pixel_format->dwStencilBitDepth) != device->m_ZBits)
        {
            // Z bit depth does not match requested
            return D3DENUMRET_OK;
        }
    }

    // We can use this format so copy it to the destination block
    memcpy(format_data->m_FormatArray+*(format_data->m_FormatCount),pixel_format,sizeof(DDPIXELFORMAT));
    (*format_data->m_FormatCount)++;

    return D3DENUMRET_OK;
}


//-----------------------------------------------------------------------------
bool RenderDevice::Create3D
(
    RenderDeviceCallback    func,
    void*                   func_context
)
//--------------------------------------
{
    // Create D3D
    if (m_DD->QueryInterface(IID_IDirect3D7, (void **)&m_D3D) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Create3D() QI D3D 7\n");
        return false;
    }

    m_FoundHW  = false;
    m_FoundSW  = false;
    m_FoundTnL = false;
    m_FoundRef = false;

    DeviceCallbackData  dev_data = {this,func,func_context};

    // Enumerate the devices so we can see what is here
    if (m_D3D->EnumDevices(DeviceCallback,(void*)&dev_data) != S_OK)
    {
        DBGSTR("ERROR: RenderDevice::Create3D() EnumDevices()\n");
        return false;
    }

    if ((!m_FoundHW) && (!m_FoundSW) && (!m_FoundTnL) && (!m_FoundRef))
    {
        // No devices passed the enumeration... We have to quit
        // we have no choice...
        DBGSTR("ERROR: RenderDevice::Create3D() No suitable devices found\n");
        return false;
    }

    // Create an order to try the devices in, we go from best to worst
    // in the normal case.

    GUID    device_type[4];
    uint32  device_count = 0;

    if (m_FoundTnL)
    {
        device_type[device_count] = IID_IDirect3DTnLHalDevice;
        device_count++;
    }

    if (m_FoundHW)
    {
        device_type[device_count] = IID_IDirect3DHALDevice;
        device_count++;
    }

    if (m_FoundSW)
    {
        device_type[device_count] = IID_IDirect3DRGBDevice;
        device_count++;
    }

    if (m_FoundRef)
    {
        device_type[device_count] = IID_IDirect3DRefDevice;
        device_count++;
    }

    uint32  next_device = 0;
    bool    created = false;
    while (next_device<device_count)
    {
        // For the current device set the using hardware flag so we
        // know if we need to create surfaces and vertex buffers in video
        // or system memory
        if ((device_type[next_device] == IID_IDirect3DRGBDevice) || 
            (device_type[next_device] == IID_IDirect3DRefDevice))
        {
            m_UsingHardware = false;
        }
        else if ((device_type[next_device] == IID_IDirect3DHALDevice) || 
                 (device_type[next_device] == IID_IDirect3DTnLHalDevice))
        {
            m_UsingHardware = true;
        }
        else
        {
            Assert(0);
        }

        // Create the back buffer as a render target, in the correct memory type
        // for hardware or software
        if (!CreateRenderBuffer(m_RenderWidth,m_RenderHeight,m_RenderBpp,&m_DDRenderSurface))
        {
            // We failed to create the render back buffer
            return false;
        }

        DDPIXELFORMAT           z_formats[MAX_FORMATS];
        uint32                  z_count = 0;
        PixelFormatCallbackData callback_data = {this,z_formats,&z_count};

        // For the current device enumerate all the Z buffer formats, if we asked for a Z buffer
        if (m_ZBits)
        {
            m_D3D->EnumZBufferFormats(device_type[next_device],ZBufferCallback,(void*)&callback_data);

            // All of the formats in the z_formats list are potentially good formats for
            // the current requested settings, assuming there are some formats in the list!
            if (z_count == 0)
            {
                // There are no useful Z buffer formats, cannot create a Z buffer so there
                // is no point in continuing as we have to create something.
                if (m_DDRenderSurface)
                {
                    m_DDRenderSurface->Release();                 
                    m_DDRenderSurface = NULL;
                }
                return false;
            }
        }
        else
        {
            // We did not ask for a Z buffer so we cannot have a stencil buffer
            Assert(!m_StencilBuffer);
        }

        uint32 next_z = 0;

        while ((next_z<z_count) || (z_count == 0))
        {
            // We requested a Z buffer so lets create a Z buffer in the format indicated
            // by 'next_z' in the format list.
            if (m_ZBits)
            {
                DDSURFACEDESC2 desc;
                memset(&desc, 0, sizeof(desc));
                desc.dwSize = sizeof(desc);
                desc.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_PIXELFORMAT;
                desc.dwWidth = m_RenderWidth;
                desc.dwHeight = m_RenderHeight;
                desc.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | (UsingHardware()?DDSCAPS_VIDEOMEMORY:DDSCAPS_SYSTEMMEMORY);
                desc.ddpfPixelFormat = z_formats[next_z];
                m_ZCreateFormat = z_formats[next_z];        // Keep a copy of the current Z format (for resize)

                if (m_DD->CreateSurface(&desc, &m_DepthBuffer, NULL) != S_OK)
                {
                    // Failed to the create the surface...
                    // Try the next Z buffer format
                    next_z++;
                    continue;
                }

                // Add the Z buffer to the render buffer
                if (m_DDRenderSurface->AddAttachedSurface(m_DepthBuffer) != S_OK)
                {
                    // Failed to attach the Z buffer to the render target, delete the Z buffer
                    // and try the next Z buffer format
                    m_DepthBuffer->Release();
                    m_DepthBuffer = NULL;
                    next_z++;
                    continue;
                }
            }
            else
            {
                m_DepthBuffer = NULL;
            }

            // Create the Direct3D device. This can fail because of a bad Z buffer format
            // attached to the back buffer, to complicate matters legal Z buffer formats
            // change with the scren mode. Alot of video cards only allow Z buffer formats
            // that match the screen bit depth but this is changing.
            if (m_D3D->CreateDevice(device_type[next_device], m_DDRenderSurface, &m_D3DDevice) == S_OK)
            {
                m_DeviceType = device_type[next_device];
                created = true;
                break;      // All is good.
            }

            if (m_DepthBuffer)
            {
                m_DepthBuffer->Release();                 
                m_DepthBuffer = NULL;
            }

            if (z_count == 0)
            {
                // We have failed to create and we did not specify a Z buffer, this is
                // an unfixable error so delete the render target surface and return
                return false;
            }

            // Try the next Z buffer format
            next_z++;
        }

        // No need to try the next device we have created one
        if (created)
        {
            break;
        }

        next_device++;
    }

    // We have tried every device and Z buffer format and could not create a render device
    // so we have to quit.
    if (!created)
    {
        if (m_DDRenderSurface)
        {
            m_DDRenderSurface->Release();
            m_DDRenderSurface = NULL;
        }
        return false;
    }

    EnumerateTextureFormats();

    return true;

}

//-----------------------------------------------------------------------------
bool RenderDevice::CreateRenderBuffer
(
    uint32                  width,
    uint32                  height,
    uint32                  bpp,
    IDirectDrawSurface7**   render_surface
)
//--------------------------------------
{
    DDSURFACEDESC2          desc_render;
    uint32                  render_bpp;
        
    //
    // Now create the initial back buffer as a plain surface
    //
    memset(&desc_render, 0, sizeof(desc_render));
    desc_render.dwSize         = sizeof(desc_render);
    desc_render.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
    desc_render.dwWidth        = width;
    desc_render.dwHeight       = height;
    desc_render.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE | 
        (RenderDevice::UsingHardware()?DDSCAPS_VIDEOMEMORY:DDSCAPS_SYSTEMMEMORY);

    if (m_DD->CreateSurface(&desc_render, render_surface, NULL) != S_OK)
    {
        *render_surface = NULL;
        return false;
    }

    if ( (*render_surface)->GetSurfaceDesc(&desc_render) != S_OK)
    {
        // Can't get the surface descriptor
        (*render_surface)->Release();
        *render_surface = NULL;
        return false;
    }

    render_bpp = desc_render.ddpfPixelFormat.dwRGBBitCount;
    if (render_bpp < 16)
    {
        // Less than 16 bpp is not an allowed format
        (*render_surface)->Release();
        *render_surface = NULL;
        return false;
    }

    if (bpp != 0xffffffff)
    {
        if (render_bpp != bpp)
        {
            // If we specified a bit depth then we must match it
            (*render_surface)->Release();
            *render_surface = NULL;
            return false;
        }
    }

    return true;
}


//-----------------------------------------------------------------------------
bool RenderDevice::CopyTargetToTexture
(
    uint32      render_x,
    uint32      render_y,
    uint32      render_width,
    uint32      render_height,
    uint32      dest_x,
    uint32      dest_y,
    Texture2D*  dest
)
//--------------------------------------
{
    RECT    rc;

    // Build the src rect, no clipping is performed
    rc.left = render_x;
    rc.right = render_x+render_width;
    rc.top = render_y;
    rc.bottom = render_y+render_height;

    // We have to use BLT for this and not D3D->Load because D3D->load only
    // operates on textures.
    if (dest->GetTexture()->BltFast(dest_x,dest_y,m_DDRenderSurface,&rc,DDBLTFAST_WAIT) != S_OK)
        return false;

    return true;
}


//-----------------------------------------------------------------------------
void RenderDevice::DisplayBackBuffer
(
    bool wait_vsync
)
//--------------------------------------
{
    if (wait_vsync)
    {
        // Wait for the start of the vblank
        m_DD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
    }

    RECT    rc;
    POINT   pt = {0,0};
    ClientToScreen(m_RenderWindow,&pt);
    rc.left = pt.x;
    rc.top = pt.y;
    rc.right = rc.left+m_RenderWidth;
    rc.bottom = rc.top+m_RenderHeight;
    m_DDPrimarySurface->Blt(&rc,m_DDRenderSurface, NULL, DDBLT_WAIT, NULL);
}


//-----------------------------------------------------------------------------
bool RenderDevice::Resize
(
    int32   width,
    int32   height
)
//--------------------------------------
{
    IDirectDrawSurface7*    render = NULL;
    IDirectDrawSurface7*    depth = NULL;

    // Make a new render buffer at the new size
    if (!CreateRenderBuffer(width,height,m_RenderBpp,&render))
    {
        // We failed to make our new colour buffer, we can do no more
        return false;
    }

    // Only create a Z buffer if we are supposed to have one
    if (m_ZBits)
    {
        DDSURFACEDESC2 desc;
        memset(&desc, 0, sizeof(desc));
        desc.dwSize = sizeof(desc);
        desc.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_PIXELFORMAT;
        desc.dwWidth = width;
        desc.dwHeight = height;
        desc.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | (UsingHardware()?DDSCAPS_VIDEOMEMORY:DDSCAPS_SYSTEMMEMORY);
        desc.ddpfPixelFormat = m_ZCreateFormat;

        if (m_DD->CreateSurface(&desc, &depth, NULL) != S_OK)
        {
            // We failed to create the new Z buffer, again there is nothing we can do
            // especially because the new buffer is in the same format as the one we
            // originally created.
            render->Release();
            return false;
        }

        // Add the Z buffer to the render buffer
        if (render->AddAttachedSurface(depth) != S_OK)
        {
            depth->Release();
            render->Release();
            return false;
        }
    }

    // We now have two attached buffers that need to be set as the render target
    if (m_D3DDevice->SetRenderTarget(render,0) != S_OK)
    {
        // Failed to attach the buffers to the render device therefore we failed
        if (render)
            render->Release();

        if (depth)
            depth->Release();

        return false;
    }

    D3DVIEWPORT7 vp = { 0, 0, width, height, 0.0f, 1.0f };

    if (m_D3DDevice->SetViewport( &vp ) != S_OK)
    {
        return false;
    }

    if (m_D3DDevice->Clear(0,NULL,D3DCLEAR_TARGET,0x00000000,0.0f,0)!=S_OK)
    {
        return false;
    }

    // If clear then it is pretty safe to assume that the device is working
    // OK at the new size so lets adjust the class members.

    // Now we can delete the old Z and Depth buffers and reset the class member variables
    if (m_DepthBuffer)
    {
        // remove the Z Buffer attachment
        Assert(m_DDRenderSurface);
        m_DDRenderSurface->DeleteAttachedSurface(0,m_DepthBuffer);

        // Now release the Z buffer
        m_DepthBuffer->Release();
    }

    // Free the render surface
    if (m_DDRenderSurface)
    {
        m_DDRenderSurface->Release();
    }

    m_DDRenderSurface = render;
    m_DepthBuffer = depth;

    m_RenderWidth = width;
    m_RenderHeight = height;

    return false;
}


//-----------------------------------------------------------------------------
bool RenderDevice::Restore
(
)
//--------------------------------------
{
    if ((m_DDPrimarySurface->IsLost() != DDERR_SURFACELOST) && 
        (m_DDRenderSurface->IsLost() != DDERR_SURFACELOST))
    {
        // All is good no need to restore
        return false;
    }

    // Restore the main rendering surfaces
    m_DDPrimarySurface->Restore();
    m_DDRenderSurface->Restore();
    if (m_DepthBuffer)
        m_DepthBuffer->Restore();

    m_D3DDevice->Clear(0,NULL,D3DCLEAR_TARGET,0x00000000,0.0f,0);

    return true;
}


//-----------------------------------------------------------------------------
HRESULT CALLBACK RenderDevice::PixelFormatCountCallback
(
    LPDDPIXELFORMAT pixel_format,  
    LPVOID          context    
)
//--------------------------------------
{
    (*(uint32*)context)++;
    return D3DENUMRET_OK;
}


//-----------------------------------------------------------------------------
HRESULT CALLBACK RenderDevice::PixelFormatCopyCallback
(
    LPDDPIXELFORMAT pixel_format,  
    LPVOID          context    
)
//--------------------------------------
{
    PixelFormatCallbackData* format_data = (PixelFormatCallbackData*)context;

    // We can use this format so copy it to the destination block
    memcpy(format_data->m_FormatArray+*(format_data->m_FormatCount),pixel_format,sizeof(DDPIXELFORMAT));
    (*format_data->m_FormatCount)++;

    return D3DENUMRET_OK;
}


//-----------------------------------------------------------------------------
// Make a list of all the supported texture formats
bool RenderDevice::EnumerateTextureFormats
(
)
//--------------------------------------
{
    // Do the first callback to count the texture formats
    if (m_D3DDevice->EnumTextureFormats(PixelFormatCountCallback,&m_NumTextureFormats)!=S_OK)
    {
        return false;
    }
    Assert(m_NumTextureFormats != 0);

    m_TexturePixelFormats = new DDPIXELFORMAT[m_NumTextureFormats];

    uint32 count = 0;
    PixelFormatCallbackData    callback_data = {this,m_TexturePixelFormats,&count};

    // Do the second callback to copy the formats into our allocated block,
    // this should not fail because its the same as the first one.
    if (m_D3DDevice->EnumTextureFormats(PixelFormatCopyCallback,&callback_data)!=S_OK)
    {
        delete m_TexturePixelFormats;
        m_TexturePixelFormats = NULL;
        return false;
    }

    // Make sure we have the same numbr of textures as last time, otherwise we
    //have probably written over the end of the memory block
    Assert(count == m_NumTextureFormats);

    return true;
}


//-----------------------------------------------------------------------------
// Static function to create a 2D texture directly from a file, you can pass
// any of the known extensions to this function and a texture will be created.
Texture2D* RenderDevice::CreateTexture
(
    const char*     bitmap,
    TextureFormat   format
)
//--------------------------------------
{
    Assert(bitmap);     // DO not pas a null filename

    Texture2D*  texture = new Texture2D;
    if (texture == NULL)
        return NULL;

    if (!texture->LoadFile(bitmap,format))
    {
        // We have failed to load the texture but the Texture2D class is
        // still valid, it just contains a NULL pointer so we shall return
        // it. Any atempt to use it will reuslt in a NULL (all white) texture
        // being selected. First we will give the user a nice error.
        char buf[1024];
        wsprintf(buf,"Texture file '%s' could not be loaded",bitmap);
        MessageBox(GetActiveWindow(),buf,"Create Texture",MB_OK|MB_ICONINFORMATION);
    }

    return texture;
}


//-----------------------------------------------------------------------------
// Static function to create a 2D texture directly from a file, you can pass
// any of the known extensions to this function and a texture will be created.
Texture2D* RenderDevice::CreateTexture
(
    uint32          bitmap,
    TextureFormat   format
)
//--------------------------------------
{
    Texture2D*  texture = new Texture2D;
    if (texture == NULL)
        return NULL;

    if (!texture->LoadFile(bitmap,format))
    {
        // We have failed to load the texture but the Texture2D class is
        // still valid, it just contains a NULL pointer so we shall return
        // it. Any atempt to use it will reuslt in a NULL (all white) texture
        // being selected. First we will give the user a nice error.
        char buf[1024];
        wsprintf(buf,"Texture resource '%d' could not be loaded",bitmap);
        MessageBox(GetActiveWindow(),buf,"Create Texture",MB_OK|MB_ICONINFORMATION);
    }

    return texture;
}


//-----------------------------------------------------------------------------
// Static function to match a given texture format to what a device can actually
// do. Most devices do most formats so the work of this funciton is quite easy.
// If a texture cannot be matched then TF_NONE is returned
TextureFormat RenderDevice::MatchFormat
(
    TextureFormat   fmt
)
//--------------------------------------
{
    PixelFormat     src(fmt);
    TextureFormat   list[8];

    if (src.Valid())
    {
        // fmt exists as a D3D pixel format so its OK to use
        return fmt;
    }

    // Based on the source format make a list in order of preference of
    // other formats that could be used in place.
    switch (fmt)
    {
    case TF_RGBA8888:
        list[0] = TF_RGBA4444;
        list[1] = TF_NONE;
        break;

    case TF_RGB888:
        list[0] = TF_RGBA8888;
        list[1] = TF_RGB565;
        list[2] = TF_RGB555;
        list[3] = TF_RGBA5551;
        list[4] = TF_RGBA4444;
        list[5] = TF_NONE;
        break;

    case TF_RGB565:
        list[0] = TF_RGB555;
        list[1] = TF_RGBA5551;
        list[2] = TF_RGBA4444;
        list[3] = TF_RGB888;
        list[4] = TF_RGBA8888;
        list[5] = TF_NONE;
        break;

    case TF_RGB555:
        list[0] = TF_RGBA5551;
        list[1] = TF_RGB565;
        list[2] = TF_RGBA4444;
        list[3] = TF_RGB888;
        list[4] = TF_RGBA8888;
        list[5] = TF_NONE;
        break;

    case TF_RGBA5551:
        list[0] = TF_RGBA4444;
        list[1] = TF_RGBA8888;
        break;

    case TF_RGBA4444:
        list[0] = TF_RGBA8888;
        list[1] = TF_NONE;
        break;

    default:
        DBGSTR("RenderDevice::MatchFormat() unknown format or TF_NONE.\n");
        return TF_NONE;
    }

    // Now go through the replacement list and see if any of those listed are
    // valid texture formats
    uint32 count = 0;
    while (list[count] != TF_NONE)
    {
        PixelFormat* replacement = new PixelFormat(list[count]);
        if (replacement->Valid())
        {
            // If the new pixel format is valid then the format
            // must be one of those supported by this version of D3D
            // so it can be used.
            delete replacement;
            return list[count];
        }
        delete replacement;
    }

    // If the loop exits then there is no suitable replacement for the specifed
    // texture type
    return TF_NONE;
}


//-----------------------------------------------------------------------------
// Ideally this should use some sort of cache to save having to search everytime.
bool RenderDevice::FindCompatibleD3DTextureFormat
(
    TextureFormat   fmt,
    DDPIXELFORMAT&  ddfmt
)
//--------------------------------------
{
    uint32 required_flags;
    uint32 red_bits;
    uint32 green_bits;
    uint32 blue_bits;
    uint32 alpha_bits;

    switch (fmt)
    {
    case TF_NONE:
        DBGSTR("ERROR: RenderDevice::FindCompatibleD3DTextureFormat() Cannot make D3D format for TF_NONE\n");
        return false;

    case TF_RGBA8888:
        required_flags = DDPF_RGB|DDPF_ALPHAPIXELS;
        red_bits   = 8;
        green_bits = 8;
        blue_bits  = 8;
        alpha_bits = 8;
        break;

    case TF_RGB888:
        required_flags = DDPF_RGB;
        red_bits   = 8;
        green_bits = 8;
        blue_bits  = 8;
        alpha_bits = 0;
        break;

    case TF_RGB565:
        required_flags = DDPF_RGB;
        red_bits   = 5;
        green_bits = 6;
        blue_bits  = 5;
        alpha_bits = 0;
        break;

    case TF_RGB555:
        required_flags = DDPF_RGB;
        red_bits   = 5;
        green_bits = 5;
        blue_bits  = 5;
        alpha_bits = 0;
        break;

    case TF_RGBA5551:
        required_flags = DDPF_RGB|DDPF_ALPHAPIXELS;
        red_bits   = 5;
        green_bits = 5;
        blue_bits  = 5;
        alpha_bits = 1;
        break;

    case TF_RGBA4444:
        required_flags = DDPF_RGB|DDPF_ALPHAPIXELS;
        red_bits   = 4;
        green_bits = 4;
        blue_bits  = 4;
        alpha_bits = 4;
        break;

    default:
        Assert(0);
        return false;
    };

    // Go through the enumerated texture formats within the deivce and
    // match the texture format to it.
    for (uint32 count = 0; count<m_NumTextureFormats;count++)
    {
        if ( (m_TexturePixelFormats[count].dwFlags & (DDPF_RGB|DDPF_ALPHAPIXELS)) != required_flags)
        {
            // Does not have the required format of color bits so ignore it.
            continue;
        }

        uint32 r_bits = BitCount(m_TexturePixelFormats[count].dwRBitMask);
        uint32 g_bits = BitCount(m_TexturePixelFormats[count].dwGBitMask);
        uint32 b_bits = BitCount(m_TexturePixelFormats[count].dwBBitMask);
        uint32 a_bits = BitCount(m_TexturePixelFormats[count].dwRGBAlphaBitMask);

        //check the bits per component of this pixel format
        if ((r_bits == red_bits) && (b_bits == blue_bits) && (g_bits == green_bits) && (a_bits == alpha_bits))
        {
            ddfmt = m_TexturePixelFormats[count];
            return true;
        }
    }

    // We got to the end of the array so there must not be a matching format
    return false;
}
