//-----------------------------------------------------------------------------
// Matrix.cpp
//
// Started 06/21/2000
//
// Shrike Framework -  Rob Wyatt 
//-----------------------------------------------------------------------------
#include "Matrix.h"
#include "ShrikeMath.h"

//------------------------------------------------------------------------------
// Standard macro to reroute new allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//----------------------------------------------------------------------------
Matrix4::Matrix4
(
    Vector3& from,
    Vector3& look_at,
    Vector3& world_up
)
//--------------------------------------
{
    // Get the z basis vector, which points straight ahead. This is the
    // difference from the eyepoint to the lookat point.
    Vector3 view = look_at - from;

    float length = view.Magnitude();
    Assert(length >= FLOAT_SMALL);

    // Normalize the z basis vector
    view /= length;

    // Get the dot product, and calculate the projection of the z basis
    // vector onto the up vector. The projection is the y basis vector.
    float dp = world_up*view;

    Vector3 up_vec = world_up - (view * dp);

    // If this vector has near-zero length because the input specified a
    // bogus up vector, let's try a default up vector
    if (  (length = up_vec.Magnitude()) <= 1e-6f )
    {
        up_vec = Vector3(0.0f, 1.0f, 0.0f) - (view * view.y);

        // If we still have near-zero length, resort to a different axis.
        if (  (length = up_vec.Magnitude()) <= 1e-6f )
        {
            up_vec = Vector3(0.0f, 0.0f, 1.0f) - (view * view.z);

            Assert(up_vec.Magnitude() >= FLOAT_SMALL);
        }
    }

    // Normalize the y basis vector
    up_vec /= length;

    // The x basis vector is found simply with the cross product of the y
    // and z basis vectors
    Vector3 right = (up_vec^view);

    // Start building the matrix. The first three rows contains the basis
    // vectors used to rotate the view to point at the lookat point
    SetIdentity();

    _11 = right.x;    _12 = up_vec.x;    _13 = view.x;
    _21 = right.y;    _22 = up_vec.y;    _23 = view.y;
    _31 = right.z;    _32 = up_vec.z;    _33 = view.z;

    // Do the translation values (rotations are still about the eyepoint)
    _41 = - from*right;
    _42 = - from*up_vec;
    _43 = - from*view;
}



//------------------------------------------------------------------------------
Matrix4::Matrix4
(
    Vector3&    dir,       // must be a unit vector
    float       fRad        // angle in radians
)
//--------------------------------------
{
    // The rotation vector must be direction/unit vector
    Assert( dir.IsNormalized() );

    SinCosPair   sc;
    SinCos(fRad,sc);

    _11 = ( dir.x * dir.x ) * ( 1.0f - sc.c ) + sc.c;
    _12 = ( dir.x * dir.y ) * ( 1.0f - sc.c ) - (dir.z * sc.s);
    _13 = ( dir.x * dir.z ) * ( 1.0f - sc.c ) + (dir.y * sc.s);

    _21 = ( dir.y * dir.x ) * ( 1.0f - sc.c ) + (dir.z * sc.s);
    _22 = ( dir.y * dir.y ) * ( 1.0f - sc.c ) + sc.c ;
    _23 = ( dir.y * dir.z ) * ( 1.0f - sc.c ) - (dir.x * sc.s);

    _31 = ( dir.z * dir.x ) * ( 1.0f - sc.c ) - (dir.y * sc.s);
    _32 = ( dir.z * dir.y ) * ( 1.0f - sc.c ) + (dir.x * sc.s);
    _33 = ( dir.z * dir.z ) * ( 1.0f - sc.c ) + sc.c;

    _14 = _24 = _34 = 0.0f;
    _41 = _42 = _43 = 0.0f;
    _44 = 1.0f;
}



//----------------------------------------------------------------------------
// Code similar to Mark Kilgards GLUT for GL
void Matrix4::SetPerspective
(
    float fov,      
    float aspect,   
    float near_plane, 
    float far_plane
)
//--------------------------------------
{
    Assert( fabs(far_plane-near_plane) >= FLOAT_EPSILON );
    Assert( fabs(sinf(fov/2.0f)) >= FLOAT_EPSILON );

    SinCosPair   sc;
    SinCos(fov/2,sc); 
    float top = (near_plane / (sc.c/sc.s));  // (sc.c/sc.s) = cotangent

    SetFrustum(-(top/aspect),top/aspect, -top, top, near_plane, far_plane);
}



//----------------------------------------------------------------------------
// Set frustum GL style, code obtained from MESA GL
void Matrix4::SetFrustum
(
    float   left,
    float   right,
    float   bottom,
    float   top,
    float   near_p,
    float   far_p
)
//--------------------------------------
{
    *this = Identity;
    _11 = (2.0f*near_p) / (right-left);
    _22 = (2.0f*near_p) / (top-bottom);
    _33 = (far_p+near_p) / (far_p-near_p);
    _44 = 0.0f;
    _34 = 1.0f;
    _31 = (right+left) / (right-left);
    _32 = (top+bottom) / (top-bottom);
    _43 = (-2.0f*near_p*far_p) / (far_p-near_p);
}


//------------------------------------------------------------------------------
//
// Find the inverse of the 4 by 4 matrix b using gausian elimination, probably
// not the fastest way of doing it but it does invert the full 4x4.
//
// Code contributed by Jacques Leroy <jle@star.be>
// Code lifted from Brian Paul's Mesa freeware OpenGL implementation.
//
Matrix4 Matrix4::operator~
(
) const
//--------------------------------------
{
#define MAT(m,r,c) ( ((float*)(&m))[(r)*4+(c)] )

    float   val,val2;
    int     i,j,k,ind;

    Matrix4 res;
    res.SetIdentity();

    Matrix4 tmp = *this;

    for (i = 0; i != 4; i++)
    {
        // find pivot
        val = MAT(tmp,i,i);
        ind = i;
        for (j = i+1; j!=4; j++)
        {
            if (fabs(MAT(tmp,j,i)) > fabs(val))
            {
                ind = j;
                val = MAT(tmp,j,i);
            }
        }

        // swap columns
        if (ind != i)
        {
            for (j = 0; j != 4; j++)
            {
                val2 = MAT(res,i,j);
                MAT(res,i,j) = MAT(res,ind,j);
                MAT(res,ind,j) = val2;
                val2 = MAT(tmp,i,j);
                MAT(tmp,i,j) = MAT(tmp,ind,j);
                MAT(tmp,ind,j) = val2;
            }
        }

        if (val == 0.0F)
        {
            // The matrix is singular (has no inverse).  This isn't really
            // an error since singular matrices can be used for projecting
            // shadows, etc.  We let the inverse be the identity matrix.
            res.SetIdentity();
            return res;
        }

        for (j = 0; j != 4; j++)
        {
            MAT(tmp,i,j) /= val;
            MAT(res,i,j) /= val;
        }

        // eliminate column
        for (j = 0; j != 4; j++)
        {
            if (j == i)
                continue;
            val = MAT(tmp,j,i);
            for (k = 0; k != 4; k++)
            {
                MAT(tmp,j,k) -= MAT(tmp,i,k) * val;
                MAT(res,j,k) -= MAT(res,i,k) * val;
            }
        }
    }
    return res;

#undef MAT
}


//----------------------------------------------------------------------------
// Static Indentity Maxtrix
const Matrix4 Matrix4::Identity = Matrix4
(
    1.0f,0.0f,0.0f,0.0f,
    0.0f,1.0f,0.0f,0.0f,
    0.0f,0.0f,1.0f,0.0f,
    0.0f,0.0f,0.0f,1.0f
);


