//-----------------------------------------------------------------------------
// Texture.cpp
//
// Started 06/22/2000
//
// Shrike Framework -  Rob Wyatt
//-----------------------------------------------------------------------------

#include "Textures.h"
#include "3DDevice.h"
#include "PixelFormat.h"
#include "HlpLib.h"
#include "Debug.h"

//------------------------------------------------------------------------------
// Standard macro to reroute new allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//-----------------------------------------------------------------------------
Texture2D::Texture2D
(
)
//--------------------------------------
{
    m_Format = TF_NONE;
    m_TextureSurface = NULL;
    m_Width  = 0;
    m_Height = 0;
}


//-----------------------------------------------------------------------------
Texture2D::~Texture2D
(
)
//--------------------------------------
{
    if (m_TextureSurface)
    {
        m_TextureSurface->Release();
        m_TextureSurface = NULL;
    }
}


//-----------------------------------------------------------------------------
bool Texture2D::Initialize
(
    uint32          width,
    uint32          height,
    TextureFormat   format
)
//--------------------------------------
{
    Assert (m_TextureSurface == NULL);

    // This is the biggest texture that we support.
    Assert (width<=4096);
    Assert (width<=4096);

    DDSURFACEDESC2  ddsd;
    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize   = sizeof(ddsd);
    ddsd.dwFlags  = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_PIXELFORMAT;
    ddsd.dwWidth  = width;
    ddsd.dwHeight = height;

    ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE | 
        (RenderDevice::UsingHardware()?DDSCAPS_VIDEOMEMORY:DDSCAPS_SYSTEMMEMORY);

    // Get the device caps
    D3DDEVICEDESC7 dev_desc;
    if( RenderDevice::GetDevice()->GetCaps(&dev_desc) != S_OK )
        return false;

    // Adjust width and height to be powers of 2, if the device requires it
    if( dev_desc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2 )
    {
        for( ddsd.dwWidth=1;  width>ddsd.dwWidth;   ddsd.dwWidth<<=1 );
        for( ddsd.dwHeight=1; height>ddsd.dwHeight; ddsd.dwHeight<<=1 );
    }

    // Limit max texture sizes, if the device can't handle large textures (all devices can handle 256x256)
    ddsd.dwWidth  = Min( ddsd.dwWidth,  ( dev_desc.dwMaxTextureWidth  ? dev_desc.dwMaxTextureWidth  : 256 ) );
    ddsd.dwHeight = Min( ddsd.dwHeight, ( dev_desc.dwMaxTextureHeight ? dev_desc.dwMaxTextureHeight : 256 ) );

    // Make the texture square, if the device requires it
    if( dev_desc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY )
    {
        if( ddsd.dwWidth > ddsd.dwHeight ) ddsd.dwHeight = ddsd.dwWidth;
        else                               ddsd.dwWidth  = ddsd.dwHeight;
    }

    PixelFormat pixfmt(format);
    if (!pixfmt.Valid())
    {
        // No pixel matching pixel format was found, cannot create texture
        DBGSTR("ERROR: Texture2D::Initialize() Invalid format\n");
        return false;
    }
    else
    {
        // We found a good format, use it to create the texture
        ddsd.ddpfPixelFormat = pixfmt;
    }

    if (RenderDevice::GetDD()->CreateSurface(&ddsd, &m_TextureSurface, NULL) != S_OK)
    {
        DBGSTR("ERROR: Texture2D::Initialize() CreateSurface()\n");
        m_TextureSurface = NULL;
        return false;
    }

    m_Width = width;
    m_Height = height;
    m_Format = format;
    return true;
}



//----------------------------------------------------------------------------
// Resource file load
bool Texture2D::LoadFile
(
    uint32          bitmap,
    TextureFormat   format
)
//--------------------------------------
{
    HRSRC res = FindResource(NULL,MAKEINTRESOURCE(bitmap),RT_BITMAP);
    if (res == NULL)
    {
        DBGSTR("ERROR: Texture2D::LoadFile() FindResource() failed\n");
        return false;
    }

    HGLOBAL file = LoadResource(NULL,res);
    if (file == NULL)
    {
        DBGSTR("ERROR: Texture2D::LoadFile() LoadResource() failed\n");
        return false;
    }

    BITMAPINFO*	bmp = (BITMAPINFO*)LockResource(file);

    return LoadBMP(bmp,format);
}


//----------------------------------------------------------------------------
// Disk file load
bool Texture2D::LoadFile
(
    const char*     bitmap,
    TextureFormat   format
)
//--------------------------------------
{
    Assert(bitmap);

    HANDLE      hfile;
    uint32      read;
    uint32      size;

    hfile = CreateFile(bitmap,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
        FILE_ATTRIBUTE_READONLY|FILE_FLAG_SEQUENTIAL_SCAN,
        NULL
    );

    if (hfile == INVALID_HANDLE_VALUE)
        return false;

    size=GetFileSize(hfile,NULL);

    uint8* bmpfile= new uint8[size];

    if (!ReadFile(hfile,bmpfile,size,(DWORD*)&read,NULL))
    {
        delete bmpfile;
        CloseHandle(hfile);
        return false;
    }

    CloseHandle(hfile);

    if (size != read)
    {
        // we did not read the whole file
        delete bmpfile;
        return false;
    }

    BITMAPINFO* bmp = (BITMAPINFO*)(bmpfile+sizeof(BITMAPFILEHEADER));

    bool res = LoadBMP( bmp,format);

    delete bmpfile;
    return res;
}


//----------------------------------------------------------------------------
bool Texture2D::LoadBMP
(
    BITMAPINFO*     bmp,
    TextureFormat   format
)
//--------------------------------------
{
    // The surface must exist and the bmp passed in must be valid, at this
    // point we do not care if the data is from a file or a resource
    Assert(bmp != NULL);

    if (m_TextureSurface)
    {
        // We already have a texture surface so just restore it and reload the data.
        // Also verify that the surface is the same size as the BMP;
        Assert((int)m_Width == bmp->bmiHeader.biWidth);
        Assert((int)m_Height == Abs(bmp->bmiHeader.biHeight));

        // Restore the surface
        if (m_TextureSurface->Restore() != S_OK)
        {
            DBGSTR("ERROR: Texture2D::LoadBMP() surface restore() failed\n");
            // we failed to restore this surface and therefore it is pointless trying
            // to load the data into it. Lets just delete the surface and assume we
            // never loaded it in the first place. The default texture is a white texture.
            m_TextureSurface->Release();
            m_TextureSurface = NULL;
            m_Width  = 0;
            m_Height = 0;
            m_Format = TF_NONE;

            // return false for failed to load but the texture surface is safe to use.
            return false;
        }
    }
    else
    {
        // Initialize the texture surface
        if (!Initialize(bmp->bmiHeader.biWidth,Abs(bmp->bmiHeader.biHeight),format))
        {
            DBGSTR("ERROR: Texture2D::LoadBMP() initialize surface failed\n");
            return false;
        }
    }

    RGBQUAD*			pPalette;
    BYTE*               pBits;      // pointer to the start of the picture bits

    PixelFormat         surface_desc(GetTextureFormat());
    if (!surface_desc.Valid())
    {
        // Not a vlid pixel format
        DBGSTR("ERROR: Texture2D::LoadBMP() Invalid surface pixel format\n");
        return false;
    }

    // Bit positions of the individual colors in the source bitmap
    uint32              bmp_red_mask;
    uint32              bmp_red_start;
    uint32              bmp_red_bits;

    uint32              bmp_green_mask;
    uint32              bmp_green_start;
    uint32              bmp_green_bits;

    uint32              bmp_blue_mask;
    uint32              bmp_blue_start;
    uint32              bmp_blue_bits;

    bool                palette;

    uint32              value_mask;         //LSB mask for data from a dword
    uint32              value_shift;        //number of bits to shift for next data
    uint32              value_size;         //size of data element
    uint32              value_elements;     //number of elements in a dword
    uint32              height;             //height of bitmap (absolute)
    int32               line_stride;        //can be negative for bottom up bitmaps
    uint32              copy_width;
    uint32              count;
    uint32              items;
    uint32              source;
    uint32              data;
    uint32              red;
    uint32              green;
    uint32              blue;
    uint32              final_color;

	pBits       = ( ((BYTE*)bmp->bmiColors)+(bmp->bmiHeader.biClrUsed*sizeof(RGBQUAD)) );
	pPalette    = (RGBQUAD*)((BYTE*)bmp+(DWORD)(bmp->bmiHeader.biSize)); 

    // validate the bmp
    Assert(bmp->bmiHeader.biCompression == BI_RGB);
    Assert (bmp->bmiHeader.biPlanes  == 1);

    height = (bmp->bmiHeader.biHeight<0)?(uint32)(-bmp->bmiHeader.biHeight):(uint32)bmp->bmiHeader.biHeight;

    // Get the source bit depth
    switch (bmp->bmiHeader.biBitCount)
    {
    case 8:
        palette         = true;
        value_mask      = 0xff;         // Data is a byte
        value_shift     = 8;            // we shift down by 8 bits to get the next data element
        value_size      = 4;            // The value holding the data is a DWORD
        value_elements  = 4;            // There are 4 data elements in a DWORD
        line_stride     = (bmp->bmiHeader.biWidth + 3) & ~0x00000003;

        // for palettized data the color format is that after the palette lookup
        bmp_blue_mask   = (0xff)<<0;
        bmp_green_mask  = (0xff)<<8;
        bmp_red_mask    = (0xff)<<16;
        bmp_blue_start  = 0;
        bmp_green_start = 8;
        bmp_red_start   = 16;
        bmp_blue_bits   = 8;
        bmp_green_bits  = 8;
        bmp_red_bits    = 8;
        break;

    case 16:
        palette         = false;
        value_mask      = 0x0000ffff;
        value_shift     = 16;   
        value_size      = 4;   
        value_elements  = 2;
        line_stride     = ((bmp->bmiHeader.biWidth*2) + 3) & ~0x00000003;       // A BMP line is always a multiple of a DWORD long

        bmp_blue_mask   = (0x1f)<<0;
        bmp_green_mask  = (0x1f)<<5;
        bmp_red_mask    = (0x1f)<<10;
        bmp_blue_start  = 0;
        bmp_green_start = 5;
        bmp_red_start   = 10;
        bmp_blue_bits   = 5;
        bmp_green_bits  = 5;
        bmp_red_bits    = 5;
        break;

    case 24:
        palette         = false;
        value_mask      = 0x00ffffff;
        value_shift     = 0;   
        value_size      = 3;   
        value_elements  = 1;
        line_stride     = ((bmp->bmiHeader.biWidth*3) + 3) & ~0x00000003;

        bmp_blue_mask   = (0xff)<<0;
        bmp_green_mask  = (0xff)<<8;
        bmp_red_mask    = (0xff)<<16;
        bmp_blue_start  = 0;
        bmp_green_start = 8;
        bmp_red_start   = 16;
        bmp_blue_bits   = 8;
        bmp_green_bits  = 8;
        bmp_red_bits    = 8;
        break;

    case 32:
        palette         = false;
        value_mask      = 0xffffffff;
        value_shift     = 0;   
        value_size      = 4;   
        value_elements  = 1;
        line_stride     = bmp->bmiHeader.biWidth*4;

        bmp_blue_mask   = (0xff)<<0;
        bmp_green_mask  = (0xff)<<8;
        bmp_red_mask    = (0xff)<<16;
        bmp_blue_start  = 0;
        bmp_green_start = 8;
        bmp_red_start   = 16;
        bmp_blue_bits   = 8;
        bmp_green_bits  = 8;
        bmp_red_bits    = 8;
        break;

    default:
        DBGSTR("ERROR: Texture2D::LoadBMP() Invalid BMP format\n");
        return false;
    }

    //Bottom up BMP
    if (bmp->bmiHeader.biHeight>0)
    {
        pBits += ((height-1)*line_stride);
        line_stride = -line_stride;
    }

    // We set all bits in dst alpha, BMPs have no alpha channel so we assume
    // opaque colors.
    uint32   dst_alpha = surface_desc.GetAlphaMask();

    int32 dest_blue_shift = surface_desc.NumberBlueBits() - bmp_blue_bits;
    if (dest_blue_shift<0)
    {
        bmp_blue_start+= -dest_blue_shift;
        dest_blue_shift = 0;
    }
    dest_blue_shift += surface_desc.BlueBitsStartPos();

    int32 dest_green_shift = surface_desc.NumberGreenBits() - bmp_green_bits;
    if (dest_green_shift<0)
    {
        bmp_green_start+= -dest_green_shift;
        dest_green_shift = 0;
    }
    dest_green_shift += surface_desc.GreenBitsStartPos();

    int32 dest_red_shift = surface_desc.NumberRedBits() - bmp_red_bits;
    if (dest_red_shift<0)
    {
        bmp_red_start+= -dest_red_shift;
        dest_red_shift = 0;
    }
    dest_red_shift += surface_desc.RedBitsStartPos();

    // When loading we do not stretch so only load the region where the surface
    // and BMP overlap, ie the smallest of the two.
    DWORD       read_width  = Min(m_Width,(uint32)bmp->bmiHeader.biHeight);
    DWORD       read_height = Min(m_Height,(uint32)height);

    DDSURFACEDESC2 ddsd;
    ddsd.dwSize = sizeof(DDSURFACEDESC2);

    // Now fill in our simple texture map
    m_TextureSurface->Lock(NULL,&ddsd,DDLOCK_WAIT|DDLOCK_SURFACEMEMORYPTR|DDLOCK_WRITEONLY,NULL);

    uint8*      dest;
    uint8*      dest_line = (uint8*)ddsd.lpSurface;
    uint32      dest_stride = (uint32)ddsd.lPitch;

    // read the source bitmap
    for (uint32 y=0;y<read_height;y++)
    {
        // point to the current DWORD of data
        count = 0;
        dest = dest_line;
        copy_width = read_width;
        while (copy_width>0)
        {
            items = value_elements;
            source = *((uint32*)(pBits+count));
            while ((copy_width>0) && (items>0))
            {
                data = (source & value_mask);
                source>>=value_shift;
                items--;

                if (palette)
                {
                    // this is a paletted bmp so look up the index in the palette
                    // to give a 32 bit color
                    data = *(uint32*)(pPalette+data);
                }
                
                red   = (data & bmp_red_mask)>>bmp_red_start;
                green = (data & bmp_green_mask)>>bmp_green_start;
                blue  = (data & bmp_blue_mask)>>bmp_blue_start;

                final_color = (red<<dest_red_shift) | (green<<dest_green_shift) | (blue<<dest_blue_shift) | dst_alpha;

                if (surface_desc.GetBitsPerPixel() == 16)
                {
                    *((uint16*)dest) = (uint16)final_color;
                    dest+=2;
                }
                else
                {
                    *((uint32*)dest) = final_color;
                    dest+=4;
                }

                copy_width--;
            }
            count+=value_size;
        }

        // mode to the next scan line
        pBits += line_stride;      
        dest_line += dest_stride;
    }

    m_TextureSurface->Unlock(NULL);

    return true;
}
