//-----------------------------------------------------------------------------
// Tessellate2D.cpp
//
// Started 06/24/2000
//
// Shrike Framework -  Rob Wyatt
//-----------------------------------------------------------------------------

#include "Tessellate2D.h"
#include "Debug.h"

//------------------------------------------------------------------------------
// Standard macro to reroute new allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//-----------------------------------------------------------------------------
Tessellate2D::Tessellate2D
(
    uint32  tess_u,
    uint32  tess_v
)
//--------------------------------------
{
    // More than 1 vertex in each direction has to be generated before
    // any triangle data can be created.
    Assert(tess_u>1);
    Assert(tess_v>1);

    m_TessU = tess_u;
    m_TessV = tess_v;
    m_StartU = 0.0f;
    m_StartV = 0.0f;
    m_StepU = 1.0f/(tess_u-1);
    m_StepV = 1.0f/(tess_v-1);
}


//-----------------------------------------------------------------------------
Tessellate2D::~Tessellate2D
(
)
//--------------------------------------
{
}


//-----------------------------------------------------------------------------
// This default triangle list generator makes a list of triangles that run in
// the U (horizontol) direction followed by V. At each position a quad is made
// from the neighboring points and that quad is tessellated as a pair of triangles
// in a clockwise order.
uint16* Tessellate2D::GenerateTriangleListIndices
(
    uint32& count           // holds the list element count on exit
)
//--------------------------------------
{
    DWORD   quad = 0;
    DWORD   x;
    DWORD   y;

    // Allocate the memory for the given number if triangle indices
    count = ((m_TessU-1) * (m_TessV-1) * 2)*3;
    uint16* indices = new uint16[count];

    if (indices == NULL)
        return NULL;        // This should never happen

    for (y = 0; y<m_TessV-1;y++)
    {
        for (x = 0; x<m_TessU-1;x++)
        {
            // desribe the currnet quad in a clockwise order as a pair
            // of triangles

            // tri 1
            indices[(quad*6)+0] = (uint16)(((y+0)*m_TessU) + x+0);
            indices[(quad*6)+1] = (uint16)(((y+0)*m_TessU) + x+1);
            indices[(quad*6)+2] = (uint16)(((y+1)*m_TessU) + x+0);

            // tri 2
            indices[(quad*6)+3] = (uint16)(indices[(quad*6)+1]);
            indices[(quad*6)+4] = (uint16)(((y+1)*m_TessU) + x+1);
            indices[(quad*6)+5] = (uint16)(indices[(quad*6)+2]);

            quad++;
        }
    }

    return indices;
}


//-----------------------------------------------------------------------------
Vector3 Tessellate2D::GenerateNormal
(
    float   u,
    float   v
)
//--------------------------------------
{
    Vector3     res;

    Assert(u>(0.0f-FLOAT_EPSILON));
    Assert(v>(0.0f-FLOAT_EPSILON));
    Assert(u<(1.0f+FLOAT_EPSILON));
    Assert(v<(1.0f+FLOAT_EPSILON));

	const float epsilon = 0.005f;        // the du,dv deltas
    Vector3 pos = GeneratePosition(u,v); // The central position

	// if our delta vector will take us off the edge of the patch then reverse the direction and flip
    // the resulting normal by changing the order of the cross product
	if (u<epsilon)
	{
		if (v<epsilon)
		{
            res = ((GeneratePosition(u+epsilon,v)-pos)^(GeneratePosition(u,v+epsilon)-pos)).Normalize();
		}
		else
		{
			res = ((GeneratePosition(u,v-epsilon)-pos)^(GeneratePosition(u+epsilon,v)-pos)).Normalize();
		}
	}
	else
	{
		if (v<epsilon)
		{
			res = ((GeneratePosition(u,v+epsilon)-pos)^(GeneratePosition(u-epsilon,v)-pos)).Normalize();
		}
		else
		{
			res = ((GeneratePosition(u-epsilon,v)-pos)^(GeneratePosition(u,v-epsilon)-pos)).Normalize();
		}
	}

    return res;
}


//-----------------------------------------------------------------------------
void* Tessellate2D::Tessellate
(
    uint32  vert_stride,
    uint32  flags,
    uint32& vert_count
)
//--------------------------------------
{
    vert_count = m_TessU*m_TessV;

    uint8* block = new uint8[vert_count*vert_stride];
    uint8* next = block;
    uint8* element;
    uint32 v_count;
    uint32 u_count;

    for (v_count = 0;v_count<m_TessV;v_count++)
    {
        for (u_count = 0;u_count<m_TessU;u_count++)
        {
            element = next;

            //
            // Position
            //
            if (flags & TESS_POSITION)
            {
                *((Vector3*)element) = GeneratePosition(u_count,v_count);
                element+=sizeof(Vector3);
            }

            //
            // Normal
            //
            if (flags & TESS_NORMAL)
            {
                *((Vector3*)element) = GenerateNormal(u_count,v_count);
                element+=sizeof(Vector3);
            }

            next+=vert_stride;
        }
    }

    return block;
}
