//-----------------------------------------------------------------------------
// WinShell.cpp
//
// Started 06/09/2000
//
// Shrike Framework -  Rob Wyatt
//-----------------------------------------------------------------------------
#include "WinShell.h"
#include "stdio.h"

//------------------------------------------------------------------------------
// Standard macro to reroute new allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//-----------------------------------------------------------------------------
// Conatant registry keys
#define REG_TOP_LEVEL   HKEY_LOCAL_MACHINE
#define REG_PARENT      "Software\\Shrike7\\"



//------------------------------------------------------------------------------
Shell::Shell
(
)
//--------------------------------------
{
    int flag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
    // Keep freed memory blocks in theheaps linked list and mark them as freed
    flag |= _CRTDBG_DELAY_FREE_MEM_DF;
    // Check for leaks at the end of the program
    flag |= _CRTDBG_LEAK_CHECK_DF;
    // prevent _CrtCheckMemory from being called at every allocation request
    flag &= ~_CRTDBG_CHECK_ALWAYS_DF;
    _CrtSetDbgFlag( flag );

    // Get the instance handle of the app and keep it
    m_AppInstance = GetModuleHandle(NULL);
    m_MainWindow = NULL;
    m_Paused = false;
    m_Quit = false;
    m_HasMenu = false;
    m_FirstFrame = true;
    m_Capture = false;
    m_AppRegKey = NULL;
    m_LastFrameTime = 0.0f;
    m_LastTitleUpdate = 0.0f;
    m_3DDevice = NULL;
    m_TitleUpdate = true;
    m_ShowTitleUpdates = true;
    m_BadSetup = false;
    m_WaitVSync = true;
    m_SuspendMsg[0] = 0;
    m_DeviceTitle[0] = 0;
}


//------------------------------------------------------------------------------
Shell::~Shell
(
)
//--------------------------------------
{
    // The 3d device should have been deleted long before we get to here.
    if (m_3DDevice)
        delete m_3DDevice;

    // The Key should also have been closed
    if (m_AppRegKey)
        RegCloseKey(m_AppRegKey);

    UnregisterClass(WINSHELL_CLASS,m_AppInstance);
}


//------------------------------------------------------------------------------
bool Shell::Initialize
(
    const char* app_title,
    const char* menu_title,
    const char* reg_key_name,
    uint32  xsize,
    uint32  ysize,
    uint32  render_bpp,
    uint32  z_bpp,
    bool    stencil
)
//--------------------------------------
{
    //
    // Before we create a window lets check the initial screen mode, and if it is
    // not what we want we can exit now.
    //
    HDC     hdc = GetDC(NULL);
    int32   bpp = GetDeviceCaps(hdc,BITSPIXEL);
    int32   width = GetDeviceCaps(hdc,HORZRES);
    int32   height = GetDeviceCaps(hdc,VERTRES);
    ReleaseDC(NULL,hdc);

    Assert(m_3DDevice == NULL);

    // Check basic proprties of the desktop mode
    if (!ValidateScreenMode(width,height,bpp,xsize,ysize,render_bpp,false))
    {
        return false;
    }

    m_RenderBpp = render_bpp;

    //
    // Create the Window class and main window
    //
    WNDCLASS Class;

    Class.style = CS_HREDRAW | CS_VREDRAW;
    Class.lpfnWndProc = SceneProc;
    Class.cbClsExtra = 0;
    Class.cbWndExtra = 0;
    Class.hInstance = m_AppInstance;
    Class.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    Class.hCursor = LoadCursor(NULL, IDC_ARROW);
    Class.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    Class.lpszMenuName = menu_title;
    Class.lpszClassName = WINSHELL_CLASS;

    if (RegisterClass(&Class) == 0)
    {
        DBGSTR("ERROR: Shell::Initialize() Register Class\n");
        return false;
    }

    m_HasMenu = menu_title != NULL;

    RECT rc = {0,0,xsize,ysize};
    AdjustWindowRect(&rc,WS_OVERLAPPEDWINDOW,m_HasMenu);

    m_MainWindow = CreateWindowEx(0, WINSHELL_CLASS, "Shrike", WS_OVERLAPPEDWINDOW,
                          0,0, rc.right-rc.left, rc.bottom-rc.top,
                          NULL, NULL, m_AppInstance, NULL);

    if (m_MainWindow == NULL)
    {
        DBGSTR("ERROR: Shell::Initialize() CreateWindowEx\n");
        return false;
    }

    SetWindowLong(m_MainWindow,GWL_USERDATA,(DWORD)this);
    SetActiveWindow(m_MainWindow);

    Assert(strlen(app_title)<64);
    SetWindowText(m_MainWindow,app_title);
    strcpy(m_AppTitle,app_title);

    // Open the registry
    bool created = false;

    if (reg_key_name)
    {
        char regbuf[MAX_PATH];
        wsprintf(regbuf,"%s%s",REG_PARENT,reg_key_name);
        if (RegOpenKey(REG_TOP_LEVEL,regbuf,&m_AppRegKey)!=ERROR_SUCCESS)
        {
            created = true;
            if (RegCreateKey(REG_TOP_LEVEL,regbuf,&m_AppRegKey)!=ERROR_SUCCESS)
            {
                // We have no registry because we failed to create it!
                m_AppRegKey = NULL;
            }
        }
    }
    else
    {
        // We have no registry
        m_AppRegKey = NULL;
        created = true;
    }
   
    //
    // Call the initial setup function, to setup class variables.
    // DeviceSetup() will be called after the main render surface has
    // been created.
    //
    Setup(created);

    RenderDevice* dev = new RenderDevice();

    //
    // We never care about the Z format only that it has a stencil buffer
    //
    if (!dev->Initialize(m_MainWindow,xsize,ysize,render_bpp,z_bpp,stencil,DeviceCallback,(void*)this))
    {
        // We failed to create the 3D render device
        // The now empty device class will get destroyed when
        // this Shell class is destroyed
        Closedown();
        RegCloseKey(m_AppRegKey);
        m_AppRegKey = NULL;
        delete dev;        
        return false;
    }

    m_3DDevice = dev;

    // Set up the actual device we have created
    if (!DeviceSetup(dev->GetDevice(),false))
    {
        // the actual device init started
        Closedown();
        RegCloseKey(m_AppRegKey);
        m_AppRegKey = NULL;
        delete dev;
        m_3DDevice = NULL;
        return false;
    }

    ShowWindow(m_MainWindow,SW_SHOWNORMAL);
    NameDevice(m_3DDevice->DeviceType());

    m_ClientWidth = xsize;
    m_ClientHeight = ysize;
    m_ZBits = z_bpp;
    m_CreateWithStencil = stencil;
    return true;
}


//------------------------------------------------------------------------------
bool Shell::ValidateScreenMode
(
    uint32  screen_width,
    uint32  screen_height,
    uint32  screen_bpp,
    uint32  width,
    uint32  height,
    uint32  bpp,
    bool    mode_change
)
//--------------------------------------
{
    if (screen_bpp<=8)
    {
        // The bit depth is 8 bits or less in which case we always fail.
        MessageBox(NULL,"Desktop must be in a high or true color mode, change mode and try again.","Screen Error", MB_OK|MB_ICONERROR);
        strcpy(m_SuspendMsg,"Application cannot run with an 8 bit desktop");
        return false;
    }

    if (bpp!=0xffffffff)
    {
        // We have requested a specific bit depth for the back buffer, therefore
        // the front buffer has to be in a compatible format.
        if (screen_bpp != bpp)
        {
            char buf[1024];
            wsprintf(buf, "A %d bit mode was requested but your destop is in a %d bit mode, change to a %d bit mode and try again", bpp, screen_bpp,bpp);
            strcpy(m_SuspendMsg,"Incompatible bit depth");
            MessageBox(NULL,buf,"Screen Error",MB_OK|MB_ICONERROR);
            return false;
        }
    }

    return true;
}


//------------------------------------------------------------------------------
bool Shell::DeviceCallback
(
    void*           context,
    D3DDEVICEDESC7* device
)
//--------------------------------------
{
    Shell*  shell_base = (Shell*)context;

    return shell_base->DeviceCheck(device);
}


//------------------------------------------------------------------------------
LRESULT CALLBACK Shell::SceneProc
(
    HWND    hwnd,
    UINT    msg,
    WPARAM  wp,
    LPARAM  lp
)
//--------------------------------------
{
    Shell*  shell = (Shell*)GetWindowLong(hwnd,GWL_USERDATA);
    if (shell == NULL)
    {
        return DefWindowProc(hwnd, msg, wp, lp);
    }

    return shell->MessagePump(hwnd,msg,wp,lp);
}


//------------------------------------------------------------------------------
uint32 Shell::Run
(
)
//--------------------------------------
{
    MSG     msg;
    uint32  exit_code = 0xffffffff;
    uint32  frames = 0;

    // Ensure that teh class is ready to be run
    Assert(m_3DDevice);

    while (!m_Quit)
    {
        // Suck up messages for this frame in the usuall Windows way.
        while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!GetMessage(&msg, NULL, 0, 0))
            {
                // A WM_CLOSE message has been found, its time to exit.
                m_Quit = true;
                exit_code = (uint32)msg.wParam;
            }

            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }

        //
        // If we are in a paused state then we do not process the main loop,
        // we just block on messages and then continue with the message pump.
        //
        if (m_Paused || m_BadSetup)
        {
            WaitMessage();
            continue;
        }

        //
        // Do main loop things
        //

        // Before we run the first frame we need to reset the timer so we
        // do not time the startup time
        if (m_FirstFrame)
        {
            m_FirstFrame = false;
            m_FrameTimer.Reset();
            m_LastFrameTime = 0.0f;
        }

        float   frame_elapsed = m_FrameTimer.GetElapsed();
        float   frame_step = frame_elapsed - m_LastFrameTime;
        m_LastFrameTime = frame_elapsed;

        frames++;

        // Check if we need to restore a surface...
        if (m_3DDevice->Restore())
        {
            // Something was lost so call the app's restore function
            if (!Restore(m_3DDevice->GetDevice()))
            {
                m_BadSetup = true;
                strcpy(m_SuspendMsg,"Exiting");
                Exit(0xffffffff);
                continue;
            }
        }

        if (Step(m_3DDevice->GetDevice(),frame_elapsed,frame_step))
        {
            // Step returned true indicating that we should go on and
            // render the scene.
            if ( Render(m_3DDevice->GetDevice()) )
            {
                // Render returned true so display the back buffer
                m_3DDevice->DisplayBackBuffer(m_WaitVSync);
            }
        }

        //
        // Is it time to update the title bar with the screen size and frame rate;
        //
        if (m_ShowTitleUpdates)
        {
            if ( (frame_elapsed > m_LastTitleUpdate + 2.0f) || m_TitleUpdate)
            {            
                // We have auto updates on so display the frame counter
                char buff[1024];
                static float    last_time = 0.0f;
                float fps = (frame_elapsed-last_time)<0.0001?0.0f:((float)frames / (frame_elapsed-last_time));
                sprintf(buff,"%s - [%d,%d] - [%s] - %.2f fps",m_AppTitle,m_ClientWidth,m_ClientHeight,m_DeviceTitle,fps);
                SetWindowText(m_MainWindow,buff);
                m_LastTitleUpdate = frame_elapsed;
                frames = 0;
                last_time = frame_elapsed;
                m_TitleUpdate = false;
            }
        }
        else
        {
            if (m_TitleUpdate)
            {
                char buff[1024];
                sprintf(buff,"%s - [%d,%d] - [%s]",m_AppTitle,m_ClientWidth,m_ClientHeight,m_DeviceTitle);
                SetWindowText(m_MainWindow,buff);
                m_TitleUpdate = false;
            }
        }
    }

    //
    // Call the device close down function
    //
    if (m_3DDevice)
        DeviceClose(m_3DDevice->GetDevice());

    //
    // Destroy the render device
    //
    delete m_3DDevice;
    m_3DDevice = NULL;

    //
    // The main loop is about to exit the 3d device has been destroyed.
    //
    Closedown();
    RegCloseKey(m_AppRegKey);
    m_AppRegKey = NULL;

    return exit_code;
}


//------------------------------------------------------------------------------
bool Shell::RestartDevice
(
)
//--------------------------------------
{
    SetWindowText(m_MainWindow,m_AppTitle);
    m_Paused = true;
    m_FrameTimer.Pause(true);

    // close and destroy the device (if we have one)
    if (m_3DDevice)
    {
        DeviceClose(m_3DDevice->GetDevice());
        delete m_3DDevice;
    }

    m_3DDevice = NULL;
    InvalidateRect(m_MainWindow,NULL,true);

    // create and initialize a new device in the new mode
    RenderDevice* dev = new RenderDevice();

    // Initialize the device again with the current window size and the same
    // Z and stencil parameters that we originally started with.
    if (!dev->Initialize(m_MainWindow,m_ClientWidth,m_ClientHeight,m_RenderBpp,m_ZBits,m_CreateWithStencil,DeviceCallback,(void*)this))
    {
        strcpy(m_SuspendMsg,"Bad Device Setup");
        InvalidateRect(m_MainWindow,NULL,true);
        m_BadSetup = true;
        delete dev;
        m_3DDevice = NULL;
        return false;
    }
    
    // Set up the new device that we have created
    if (!DeviceSetup(dev->GetDevice(),false))
    {
        strcpy(m_SuspendMsg,"Bad Device Setup");
        InvalidateRect(m_MainWindow,NULL,true);
        m_BadSetup = true;
        delete dev;
        m_3DDevice = NULL;
        return false;
    }
    
    m_3DDevice = dev;

    if (m_3DDevice != NULL)
    {
        NameDevice(m_3DDevice->DeviceType());
        m_3DDevice->GetDevice()->Clear(0,NULL,D3DCLEAR_TARGET,0x00000000,0.0f,0);
        m_3DDevice->DisplayBackBuffer(false);
    }

    // We have a good setup
    m_BadSetup = false;
    m_Paused = false;
    m_FrameTimer.Pause(false);
    m_TitleUpdate = true;

    return true;
}


//------------------------------------------------------------------------------
// Given the device type guid make a short name that can be displayed in the
// title of the window
void Shell::NameDevice
(
    const GUID&   dev
)
//--------------------------------------
{
    if (dev == IID_IDirect3DTnLHalDevice)
    {
        strcpy(m_DeviceTitle,"Hardware T&L");
    }
    else if (dev == IID_IDirect3DHALDevice)
    {
        strcpy(m_DeviceTitle,"Hardware");
    }
    else if (dev == IID_IDirect3DRGBDevice)
    {
        strcpy(m_DeviceTitle,"Software");
    }
    else if (dev == IID_IDirect3DRefDevice)
    {
        strcpy(m_DeviceTitle,"Reference");
    }
    else
    {
        strcpy(m_DeviceTitle,"D3D");
    }
}


//------------------------------------------------------------------------------
// Exit the framework
void Shell::Exit
(
    uint32  exit_code
)
//--------------------------------------
{
    PostQuitMessage(exit_code);
}


//------------------------------------------------------------------------------
// Tick or untick the specified menu item
void Shell::SelectMenu
(
    uint32 id,      // ID of the menu item
    bool enable     // 'true' to tick, 'false' to untick
)
//--------------------------------------
{
    MENUITEMINFO info;
    info.cbSize = sizeof(info);
    info.fMask = MIIM_STATE;
    info.fState = enable?MFS_CHECKED:MFS_UNCHECKED;
    SetMenuItemInfo(GetMenu(m_MainWindow),id,false,&info);
}


//------------------------------------------------------------------------------
// Read data from the registry
bool Shell::ReadRegData
(
    const char* key,
    void*       ptr,
    uint32      size,
    void*       def_ptr
)
//--------------------------------------
{
    Assert(ptr);
    Assert(size);

    uint32 read_size = size;
    uint32 type;

    if (m_AppRegKey == NULL)
        goto RegError;

    // Read the data from the registry
    if (RegQueryValueEx(m_AppRegKey,key,0,(DWORD*)&type,(uint8*)ptr,(DWORD*)&size) != ERROR_SUCCESS)
    {
        goto RegError;
    }

    // This function reads binary data
    Assert(type == REG_BINARY);

    if (read_size != size)
    {
        goto RegError;
    }

    return true;

RegError:
    // If we have a default ptr copy it
    if (def_ptr)
        memcpy(ptr,def_ptr,size);
    return false;
}

//------------------------------------------------------------------------------
// Write data to the registry
bool Shell::WriteRegData
(
    const char* key,
    void*       ptr,
    uint32      size
)
//--------------------------------------
{
    Assert(ptr);
    Assert(size);

    if (m_AppRegKey == NULL)
        return false;

    // Write to the registry
    if (RegSetValueEx(m_AppRegKey,key,0,REG_BINARY,(uint8*)ptr,size) != ERROR_SUCCESS)
    {
        return false;
    }

    return true;
}


//
// Default implementations of virtual functions
//

//------------------------------------------------------------------------------
LRESULT Shell::MessagePump
(
    HWND    hwnd,
    UINT    msg,
    WPARAM  wp,
    LPARAM  lp
)
//--------------------------------------
{
    uint16 active;
    uint16 minimized;
    int32 newxpos;
    int32 newypos;
    int32 width;
    int32 height;
    static int32 oldxpos;
    static int32 oldypos;

    
    switch (msg)
    {
    case WM_ACTIVATE:
        active = LOWORD(wp);         // activation flag 
        minimized = HIWORD(wp);      // minimized flag 

        if ((active != WA_INACTIVE) && (m_Paused == true))
        {
            if (!m_BadSetup)
            {
                // If we have a bad config do not unpause the app
                m_Paused = false;
                m_FrameTimer.Pause(false);
            }
        }
        else if ((active == WA_INACTIVE) && (m_Paused == false))
        {
            m_Paused = true;
            m_FrameTimer.Pause(true);
        }
        return 0;


    case WM_SIZE:
        width = LOWORD(lp);  // new width of client area 
        height = HIWORD(lp); // new height of client area 

        // We do not have the device yet so we are not going to process this message properly
        // as we are in a bad configuration. We just keep a copy of the current window size
        if ((m_3DDevice != NULL) && (!m_BadSetup))
        {
            if (!m_3DDevice->Resize(width,height))
            {
                // We failed to resize so the old sized surfaces are still set, just reset
                // the window to this old size
                width = m_3DDevice->GetRenderWidth();
                height = m_3DDevice->GetRenderHeight();

                RECT rc = {0,0,width,height};
                AdjustWindowRect(&rc,WS_OVERLAPPEDWINDOW,m_HasMenu);
                SetWindowPos(m_MainWindow,NULL,0,0,rc.right-rc.left,rc.bottom-rc.top,SWP_NOMOVE|SWP_NOZORDER);
            }

            // Now call the device init function with 'true' to indicate that the render
            // surface has changed size
            if (!DeviceSetup(m_3DDevice->GetDevice(),true))
            {
                Exit(0xffffffff);
                strcpy(m_SuspendMsg,"Exiting");
                m_BadSetup = true;
            }

            m_TitleUpdate = true;
        }
        m_ClientWidth = width;        
        m_ClientHeight = height;
        return 0;

    case WM_LBUTTONDOWN:
        if (!m_BadSetup)
        {
            oldxpos = ((short)LOWORD(lp));       // horizontal position of cursor at click
            oldypos = ((short)HIWORD(lp));       // vertical position of cursor at click
            MouseClick(MA_CLICK,oldxpos,oldypos);
            SetCapture(m_MainWindow);
            m_Capture = true;
        }
        return 0;

    case WM_MOUSEMOVE:
        if (!m_BadSetup)
        {
            newxpos = ((short)LOWORD(lp));       // current horizontal position of cursor 
            newypos = ((short)HIWORD(lp));       // current vertical position of cursor 
            if (m_Capture)
            {
                MouseClick(MA_DRAG,newxpos-oldxpos,newypos-oldypos);
            }
            else
            {
                MouseClick(MA_MOVING,newxpos,newypos);
            }
            oldxpos = newxpos;
            oldypos = newypos;
        }
        return 0;

    case WM_LBUTTONUP:
        ReleaseCapture();
        m_Capture = false;
        newxpos = ((short)LOWORD(lp));       // horizontal position of cursor at click
        newypos = ((short)HIWORD(lp));       // vertical position of cursor at click
        MouseClick(MA_RELEASE,newxpos,newypos);
        return 0;

    case WM_COMMAND:
        if ((m_HasMenu) && (!m_BadSetup))
        {
            // If we have a menu then we can go ahead and process the message,
            // if we don't have a menu then this message is probably from either
            // a child window or an accelerator
            if ( MenuSelect((uint32)LOWORD(wp)) )
                return 0;
        }
        break;

    case WM_CLOSE:
        PostQuitMessage( EXIT_SUCCESS );
        return 0;

    case WM_ERASEBKGND:
        // If we have not drawn the first frame yet then allow window clears,
        // also allow them if we have no 3D device or have a bad setup
        if ((!m_FirstFrame) && (m_3DDevice != NULL) && (!m_BadSetup))
            return 1;
        break;

    case WM_ENTERMENULOOP:
        if (m_BadSetup)
            return 0;
        m_Paused = true;
        m_FrameTimer.Pause(true);
        return 0;
       
    case WM_EXITMENULOOP:
        if (m_BadSetup)
            return 0;
        m_Paused = false;
        m_FrameTimer.Pause(false);
        return 0;

    case WM_ENTERSIZEMOVE:
        if (m_BadSetup)
            return 0;
        m_Paused = true;
        m_FrameTimer.Pause(true);
        return 0;

    case WM_EXITSIZEMOVE:
        if (m_BadSetup)
            return 0;
        m_Paused = false;
        m_FrameTimer.Pause(false);
        return 0;

    case WM_PAINT:
        // If we have not drawn the first frame then do not use the device pointer as it
        // may be NULL.
        if ((!m_FirstFrame) && (m_3DDevice != NULL) && (!m_BadSetup))
        {
            m_3DDevice->DisplayBackBuffer(false);
        }
        else
        {
            // We have not drawn the first frame yet so do nothing at the moment
            if (m_FirstFrame)
                break;

            // We have to perform a real paint in the window, show a message to indicate
            // why the app is not functioning properly.
            SetWindowText(m_MainWindow,m_AppTitle);

            PAINTSTRUCT paint;
            HDC hdc = BeginPaint(hwnd,&paint);
            SetBkMode(hdc, TRANSPARENT); 
            SetTextColor(hdc,0xffffffff);
            SetTextAlign(hdc,TA_CENTER);
            TextOut(hdc,m_ClientWidth/2,20,m_AppTitle,strlen(m_AppTitle));
            TextOut(hdc,m_ClientWidth/2,44,"Suspended",9);
            TextOut(hdc,m_ClientWidth/2,80,m_SuspendMsg,strlen(m_SuspendMsg));
            EndPaint(hwnd,&paint);
            return 0;
        }
        break;

    case WM_DISPLAYCHANGE:
        // Pause
        SetWindowText(m_MainWindow,m_AppTitle);
        m_Paused = true;
        m_FrameTimer.Pause(true);

        // close and destroy the device (if we have one)
        if (m_3DDevice)
        {
            DeviceClose(m_3DDevice->GetDevice());
            delete m_3DDevice;
        }

        m_3DDevice = NULL;
        InvalidateRect(m_MainWindow,NULL,true);

        // Check the new screen mode for basic compatibility
        if (!ValidateScreenMode(LOWORD(lp),HIWORD(lp),wp,m_ClientWidth,m_ClientHeight,m_RenderBpp,true))
        {
            InvalidateRect(m_MainWindow,NULL,true);
            m_BadSetup = true;
            break;
        }

        // create and initialize a new device in the new mode
        m_3DDevice = new RenderDevice();

        // Initialize the device again with the current window size and the same
        // Z and stencil parameters that we originally started with.
        if (!m_3DDevice->Initialize(m_MainWindow,m_ClientWidth,m_ClientHeight,m_RenderBpp,m_ZBits,m_CreateWithStencil,DeviceCallback,(void*)this))
        {
            strcpy(m_SuspendMsg,"Bad Device Setup");
            InvalidateRect(m_MainWindow,NULL,true);
            m_BadSetup = true;
            delete m_3DDevice;
            m_3DDevice = NULL;
            break;
        }
        
        // Set up the new device that we have created
        if (!DeviceSetup(m_3DDevice->GetDevice(),false))
        {
            strcpy(m_SuspendMsg,"Bad Device Setup");
            InvalidateRect(m_MainWindow,NULL,true);
            m_BadSetup = true;
            delete m_3DDevice;
            m_3DDevice = NULL;
            break;
        }
        
        if (m_3DDevice != NULL)
        {
            NameDevice(m_3DDevice->DeviceType());
            m_3DDevice->GetDevice()->Clear(0,NULL,D3DCLEAR_TARGET,0x00000000,0.0f,0);
            m_3DDevice->DisplayBackBuffer(false);
        }

        // We have a good setup
        m_BadSetup = false;

        break;
    }

    return DefWindowProc(hwnd, msg, wp, lp);
}

//------------------------------------------------------------------------------
bool Shell::DeviceCheck
(
    D3DDEVICEDESC7* device
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::Setup
(
    bool    just_created
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::Closedown
(
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::DeviceSetup
(
    IDirect3DDevice7*   device,
    bool                resize
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::DeviceClose
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::Restore
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::Step
(
    IDirect3DDevice7*   device,
    float               frame_time, 
    float               frame_step
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::Render
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    return true;
}


//------------------------------------------------------------------------------
bool Shell::MenuSelect
(
    uint32  menu_item_id
)
//--------------------------------------
{
    return false;   // By default we do not process anything
}


//------------------------------------------------------------------------------
bool Shell::MouseClick
(
    MouseAction action,
    int32       xpos,
    int32       ypos
)
//--------------------------------------
{
    return true;    // all is OK
}
