#if WINDOWS

	#include <dsound.h>

#elif MACOS

	#include <Sound.h>

#endif

const long soundError = 1;	// Generic error code

class Sound
{
	private:
		
		#if WINDOWS
		
			// Secondary buffer object
			LPDIRECTSOUNDBUFFER		playBuffer;
		
		#elif MACOS
		
			SndChannelPtr				channel;
			ExtSoundHeader				header;
			SndCallBackUPP				callBack;
			
			// Callback function used for looping
			static pascal void CallBack(SndChannelPtr chan, SndCommand *cmd);
		
		#endif
	
	public:
		
		Sound(const void *sound, unsigned long length);
		~Sound();
		
		void Play(bool loop);
};

#if WINDOWS

	class SoundMgr
	{
		friend class Sound;
		
		private:
			
			LPDIRECTSOUND				directSound;
			LPDIRECTSOUNDBUFFER		primaryBuffer;
		
		public:
			
			SoundMgr(HWND window);
			~SoundMgr();
	};
	
	SoundMgr *TheSoundMgr = 0;
	
	SoundMgr::SoundMgr(HWND window)
	{
		WAVEFORMATEX	format;
		DSBUFFERDESC	bufferDesc;
		
		if (FAILED(DirectSoundCreate(NULL, &directSound, NULL)))
			throw soundError;
		directSound->SetCooperativeLevel(window, DSSCL_EXCLUSIVE);
		
		// Fill out the format that we are using
		format.wFormatTag = WAVE_FORMAT_PCM;
		format.nChannels = 2;
		format.nSamplesPerSec = 22050;
		format.nAvgBytesPerSec = 88200;
		format.nBlockAlign = 4;
		format.wBitsPerSample = 16;
		format.cbSize = 0;
		
		// Create the primary play buffer
		bufferDesc.dwSize = sizeof(DSBUFFERDESC);
		bufferDesc.dwFlags = DSBCAPS_PRIMARYBUFFER;
		bufferDesc.dwBufferBytes = 0;
		bufferDesc.dwReserved = 0;
		bufferDesc.lpwfxFormat = NULL;
		if (FAILED(directSound->CreateSoundBuffer(&bufferDesc,
			&primaryBuffer, NULL)))
		{
			directSound->Release();
			throw soundError;
		}
		
		// Set the primary buffer format and begin playing
		// Until we play secondary buffers, this will
		// just play silence
		primaryBuffer->SetFormat(&format);
		primaryBuffer->Play(0, 0, DSBPLAY_LOOPING);
	}
	
	SoundMgr::~SoundMgr()
	{
		primaryBuffer->Stop();
		primaryBuffer->Release();
		directSound->Release();
	}

#endif

Sound::Sound(const void *sound, unsigned long length)
{
	#if WINDOWS
	
		WAVEFORMATEX	format;
		DSBUFFERDESC	bufferDesc;
		LPVOID			audioPtr1;
		LPVOID			audioPtr2;
		DWORD				audioBytes1;
		DWORD				audioBytes2;
		
		format.wFormatTag = WAVE_FORMAT_PCM;
		format.nChannels = 2;
		format.nSamplesPerSec = 22050;
		format.nAvgBytesPerSec = 88200;
		format.nBlockAlign = 4;
		format.wBitsPerSample = 16;
		format.cbSize = 0;
		
		// Create a secondary play buffer
		bufferDesc.dwSize = sizeof(DSBUFFERDESC);
		bufferDesc.dwFlags = DSBCAPS_CTRLDEFAULT;
		bufferDesc.dwBufferBytes = length;
		bufferDesc.dwReserved = 0;
		bufferDesc.lpwfxFormat = &format;
		
		// Assume TheSoundMgr points to a valid SoundMgr instance
		if (FAILED(TheSoundMgr->directSound->CreateSoundBuffer(&bufferDesc,
			&playBuffer, NULL))) throw soundError;
		
		// Copy the audio data into the new buffer
		playBuffer->Lock(0, length, &audioPtr1, &audioBytes1, &audioPtr2,
			&audioBytes2, DSBLOCK_ENTIREBUFFER);
		CopyMemory(audioPtr1, sound, length);
		playBuffer->Unlock(audioPtr1, length, audioPtr2, 0);
	
	#elif MACOS
	
		// Create a sound channel
		channel = NULL;
		callBack = NewSndCallBackUPP(&CallBack);
		OSStatus err = SndNewChannel(&channel, sampledSynth,
			initStereo, callBack);
		if (err != noErr) throw soundError;
		
		// Fill out the sound header
		BlockZero(&header, sizeof(ExtSoundHeader));
		header.samplePtr = (Ptr) sound;
		header.numChannels = 2;
		header.sampleRate = 22050 << 16;
		header.encode = extSH;
		header.numFrames = length >> 2;
		header.sampleSize = 16;
	
	#endif
}

Sound::~Sound()
{
	#if WINDOWS
	
		playBuffer->Stop();
		playBuffer->Release();
	
	#elif MACOS
	
		SndDisposeChannel(channel, true);
		DisposeSndCallBackUPP(callBack);
	
	#endif
}

void Sound::Play(bool loop)
{
	#if WINDOWS
	
		playBuffer->Play(0, 0, (loop) ? DSBPLAY_LOOPING : 0);
	
	#elif MACOS
	
		SndCommand	command;
		
		// This command plays the sound
		command.cmd = bufferCmd;
		command.param2 = (long) &header;
		SndDoCommand(channel, &command, true);
		
		if (loop)
		{
			// This command causes a callback to occur
			// when the bufferCmd has completed
			command.cmd = callBackCmd;
			command.param2 = (long) this;
			SndDoCommand(channel, &command, true);
		}
	
	#endif
}

#if MACOS

	pascal void Sound::CallBack(SndChannelPtr chan, SndCommand *cmd)
	{
		SndCommand	command;
		
		Sound *sound = (Sound *) cmd->param2;
		
		// Loop the sound by playing it again and
		// then calling back to this function again
		command.cmd = bufferCmd;
		command.param2 = (long) &sound->header;
		SndDoCommand(chan, &command, true);
		
		command.cmd = callBackCmd;
		command.param2 = (long) sound;
		SndDoCommand(chan, &command, true);
	}

#endif
