//-----------------------------------------------------------------------------
// TGReflect.cpp
//
// Started 06/23/2000
//
//  Rob Wyatt 2000 for Game Developer Magazine/Gamasutra
//-----------------------------------------------------------------------------
#include "TGReflect.h"
#include "Lights.h"
#include "Material.h"
#include "SuperQuadric.h"
#include "BiQuadraticPatch.h"
#include "Torus.h"
#include "MeshHelp.h"
#include "Euler.h"


//------------------------------------------------------------------------------
// Macro to reroute 'new[]' allocations to the debug allocator
#if defined(_DEBUG)
#define new DEBUG_NEW
#define DEBUG_NEW new(_NORMAL_BLOCK,THIS_FILE,__LINE__)
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



//------------------------------------------------------------------------------
bool Reflect::Setup
(
    bool first_run
)
//--------------------------------------
{
    m_MouseX    = 0.0f;
    m_MouseY    = 0.0f;
    m_WorldMat  = Matrix4::Identity;

    m_MeshVB         = NULL;
    m_TriListIndices = NULL;
    m_MeshType       = 0;

    // Wait for a vsync before displaying render buffer
    SetWaitVSync(true);

    // On some cards updating the title of the window seriously stalls the hardware
    // rendering. If you are encountering periodic stalls at 2 second intervals
    // (the update period of the title) then uncomment the line below to disable
    // the frame counter.

    //EnableFrameCounter(false);

    return true;
}


//------------------------------------------------------------------------------
bool Reflect::Closedown
(
)
//--------------------------------------
{
    delete m_TriListIndices;

    return true;
}


//------------------------------------------------------------------------------
bool Reflect::DeviceSetup
(
    IDirect3DDevice7*   device,
    bool                resize
)
//--------------------------------------
{
    // Set Projection Matrix
    D3DVIEWPORT7 vp;
    device->GetViewport(&vp);
    Matrix4 projection = Matrix4(PI_DIV_4, ((float)vp.dwHeight) / ((float)vp.dwWidth), 1.0f, 20.0f);
    device->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &projection);

    // If we are just resizing the all we need to do is reset the projection matrix
    // account for the new viewport.
    if (resize)
        return true;

    // Set the view transform to be looking straight at the origin
    Vector3 eye = Vector3(0.0f, 0.0f, -8.5f);
    Vector3 lookat = Vector3(0.0f, 0.0f,  0.0f);
    Vector3 up = Vector3(0.0f, 1.0f,  0.0f);
    Matrix4 view_mat(eye, lookat, up );
    device->SetTransform(D3DTRANSFORMSTATE_VIEW, &view_mat);

    // Set up a point light and the ambient light.
    // This adds just a little bit of detail but lighting can be disabled and the reflections
    // still work.
    PointLight light = PointLight( Vector3(0.0f,0.0f,-10.0f), RGBColor(1.0f,1.0f,1.0f) );
    device->SetLight( 0, &light );
    
    device->SetRenderState(D3DRENDERSTATE_AMBIENT, RGBColor(0.2f,0.2f,0.2f) );
    device->LightEnable( 0, true );

    Material mat = Material(RGBColor(1.0f,1.0f,1.0f));
    device->SetMaterial(&mat);

    // Make the pipeline take diffuse and ambient colors from the material not the vertices
    // because the vertices used to draw the primitives in this program do not contain any
    // color components.
    device->SetRenderState(D3DRENDERSTATE_DIFFUSEMATERIALSOURCE,D3DMCS_MATERIAL);
    device->SetRenderState(D3DRENDERSTATE_AMBIENTMATERIALSOURCE,D3DMCS_MATERIAL);

    // Load a texture and set it as texture 0, also set texture 0's render state
    m_ReflectionTexture = RenderDevice::CreateTexture(ID_BITMAP_SPHERE, RenderDevice::MatchFormat(TF_RGB888) );

    device->SetTexture(0,m_ReflectionTexture->GetTexture());

    // Set the filter mode for texture 0
    device->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTFN_LINEAR );
    device->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTFN_LINEAR );

    // Create the mesh in a vertex buffer
    CreateReflectionMesh(m_MeshType);
    SetTextureMatrix(device);

    return true;
}


//------------------------------------------------------------------------------
bool Reflect::DeviceClose
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    if (m_MeshVB)
    {
        m_MeshVB->Release();
        m_MeshVB = NULL;
    }

    delete m_ReflectionTexture;
    return true;
}


//------------------------------------------------------------------------------
bool Reflect::Restore
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    // Reload our texture
    m_ReflectionTexture->LoadFile(ID_BITMAP_SPHERE);

    // You cannot restore vertex buffers but you can lose them!!! So lets just
    // re-run the mesh function which will clean up the old buffer and create a
    // new one.
    CreateReflectionMesh(m_MeshType);
    return true;
}


//------------------------------------------------------------------------------
void Reflect::CreateReflectionMesh
(
    uint32  model
)
//--------------------------------------
{
    const uint32 tess_u = 30;
    const uint32 tess_v = 30;

    // Create a standard tessellator for the requested mesh type
    Tessellate2D* mesh;

    if (model == 0)
    {
        mesh = new Torus(1.0f,2.0f,tess_u,tess_v);
    }
    else if (model == 1)
    {
        mesh = new SuperQuadric(2.0f,2.0f,2.0f,1.8f,1.65f,tess_u,tess_v);
    }
    else if (model == 2)
    {
        mesh = new QuadraticPatch( Vector3(-2.5f,2.5f,1.5f),Vector3(0.0f,2.5f,-0.4f),Vector3(3.0f,1.5f,0.0f),
                                   Vector3(-1.0f,0.0f,0.0f),Vector3(0.0f,0.0f,-6.0f),Vector3(2.5f,0.0f,0.0f),
                                   Vector3(-1.5f,-4.0f,0.0f),Vector3(0.0f,-2.5f,3.0f),Vector3(2.5f,-2.5f,0.0f),
                                   tess_u,tess_v);
    }   

    ReflectVertex* local_mesh = (ReflectVertex*)mesh->Tessellate(sizeof(ReflectVertex),TESS_POSITION|TESS_NORMAL,m_VertCount);

    // If there is already an existing vertex buffer delete it
    if (m_MeshVB)
        m_MeshVB->Release();

    // Create a vertex buffer from the local vertex data (local_mesh) that has just been generated.
    m_MeshVB = MeshHelp::GenerateVertexBuffer(local_mesh,m_VertCount,sizeof(ReflectVertex),ReflectVertex::FORMAT);

    Assert(m_MeshVB);

    // delete the old index list and create a new one for the new mesh
    delete m_TriListIndices;
    m_TriListIndices = mesh->GenerateTriangleListIndices(m_IndexCount);

    delete mesh;
    delete local_mesh;
}


//------------------------------------------------------------------------------
void Reflect::SetTextureMatrix
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    Matrix4 tex = Matrix4
    (
        0.40f,  0.0f,  0.0f, 0.0f,
        0.0f,  -0.40f, 0.0f, 0.0f,
        0.0f,   0.0f,  0.0f, 0.0f,
        0.50,  -0.50,  0.0f, 1.0f
    );

    device->SetTransform(D3DTRANSFORMSTATE_TEXTURE0, &tex);

    // Tell D3D to use transformed reflection vectors as texture co-ordinate 0
    // and then transform this coordinate by the specified texture matrix, also
    // tell D3D that only the first two coordinates of the output are valid.
    device->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2 );    
    device->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0|D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR);    

    // Set default menu state
    SelectMenu(IDM_MAPPING_REFLECTIONVECTOR,true);
    SelectMenu(IDM_MAPPING_NORMALVECTOR,false);
}


//------------------------------------------------------------------------------
bool Reflect::Step
(
    IDirect3DDevice7*   device,
    float               frame_time,
    float               frame_step
)
//--------------------------------------
{
    if (frame_step > 0.016f)
        frame_step = 0.016f;

    // Make a world space aligned rotation from the mouse deltas that are calculated
    // above. Keep a copy of the world matrix because we need it on the next frame.
    Matrix4 xrot = EulerAngle(EULER_X,-m_MouseY * frame_step*PI_DIV_2);
    Matrix4 yrot = EulerAngle(EULER_Y,-m_MouseX * frame_step*PI_DIV_2);

    m_WorldMat = m_WorldMat*yrot*xrot;
    device->SetTransform(D3DTRANSFORMSTATE_WORLD, &m_WorldMat);

    return true;
}


//------------------------------------------------------------------------------
bool Reflect::Render
(
    IDirect3DDevice7*   device
)
//--------------------------------------
{
    device->Clear(0,NULL,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0x00000000,1.0f,0);
   
    if (device->BeginScene() != S_OK)
        return false;

    device->DrawIndexedPrimitiveVB(D3DPT_TRIANGLELIST, m_MeshVB,
        0, m_VertCount,
        m_TriListIndices, m_IndexCount,
        0);

    device->EndScene();

    return true;
}


//------------------------------------------------------------------------------
bool Reflect::MenuSelect
(
    uint32 menu_id
)
//--------------------------------------
{
    switch (menu_id)
    {
    // File->Exit has been selected
    case IDM_EXIT:
        Exit( EXIT_SUCCESS );
        break;

    // Object->Torus has been selected, reset the world transform and generate the
    // the new mesh.
    case IDM_OBJECT_TORUS:
        m_MouseX = 0.0f;
        m_MouseY = 0.0f;
        m_WorldMat = Matrix4::Identity;
        m_MeshType = 0;
        CreateReflectionMesh(0);
        break;

    // Object->'Super Quadric' has been selected, reset the world transform and generate the
    // the new mesh.
    case IDM_OBJECT_SUPERQUADRIC:
        m_MouseX = 0.0f;
        m_MouseY = 0.0f;
        m_WorldMat = Matrix4::Identity;
        m_MeshType = 1;
        CreateReflectionMesh(1);
        break;

    // Object->'Quadratic Patch' has been selected, reset the world transform and generate the
    // the new mesh.
    case IDM_OBJECT_BIQUADRATICPATCH:
        m_MouseX = 0.0f;
        m_MouseY = 0.0f;
        m_WorldMat = Matrix4::Identity;
        m_MeshType = 2;
        CreateReflectionMesh(2);
        break;

    // Mapping->Reflection vector has been selected, just change the type of vector passed to the
    // texture transform and update the menus
    case IDM_MAPPING_REFLECTIONVECTOR:
        GetD3DDevice()->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0|D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR);    
        SelectMenu(IDM_MAPPING_REFLECTIONVECTOR,true);
        SelectMenu(IDM_MAPPING_NORMALVECTOR,false);
        break;

    // Mapping->Normal vector has been selected, just change the type of vector passed to the
    // texture transform and update the menus
    case IDM_MAPPING_NORMALVECTOR:
        GetD3DDevice()->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0|D3DTSS_TCI_CAMERASPACENORMAL);    
        SelectMenu(IDM_MAPPING_REFLECTIONVECTOR,false);
        SelectMenu(IDM_MAPPING_NORMALVECTOR,true);
        break;

    default:
        return false;
    }

    return true;
}


//------------------------------------------------------------------------------
bool Reflect::MouseClick
(
    MouseAction action,
    int32       x,
    int32       y
)
//--------------------------------------
{
    if (action == MA_CLICK)
    {
        // When we first click set the rotation to be still
        m_MouseX = 0.0f;
        m_MouseY = 0.0f;
    }
    else if (action == MA_DRAG)
    {
        // When captured the mouse position is a delta, capture is effectively
        // set while the mouse is being dragged. Divide the deltas by some
        // randomly choosen factor for the mouse sensitivity
        m_MouseX = ((float)x)/10.0f;    
        m_MouseY = ((float)y)/10.0f;
    }

    return true;
}


//------------------------------------------------------------------------------
int APIENTRY WinMain
(
    HINSTANCE instance,
    HINSTANCE prev_instance,
    LPSTR     cmd_line,
    int       cmd_show
)
//--------------------------------------
{
    // Create an instance of our class, derived from the framework
    Reflect   test_app;

    if (!test_app.Initialize(
                "Spherical Env Mapping",            // Title
                MAKEINTRESOURCE(IDR_REFLECT_MENU),  // Menu resource ID (or NULL)
                NULL,                               // No registry
                640,                                // Window width (client area)
                480,                                // Window height (client area)
                0xffffffff,                         // Render bit depth (0xffffffff for don't care)
                0xffffffff,                         // We want a Z buffer but don't care about format
                false))                             // We do not want a stencil buffer
    {
        return EXIT_FAILURE;
    }

	return test_app.Run();
}
