
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CLinkedList
	\ingroup g_data_structures
	\author  Sean O'Neil
	\warning Conversion work in progress
	\date    2000

	This class implements a doubly-linked list that allows the objects within it
	to insert and remove themselves without access to the actual list (to insert,
	a pointer to an item already in the list is needed). To accomplish this,
	static head and tail nodes must be kept instead of pointers, and no count
	variable can used (because the objects have no way to update them). The head
	and tail nodes point to each other when the list is empty. See CListNode above
	for more information.

	Important Note:
	The static head and tail nodes mark the end of the list, not NULL. Use
	CListNode::IsInList() to see when you've reached the end, like this:
	for(CObject *p = list.GetHead(); p->IsInList(); p = p->GetNext()) { ... }
*/


#ifndef TEDDY_DATA_STRUCTURES_C_LINKED_LIST_H
#define TEDDY_DATA_STRUCTURES_C_LINKED_LIST_H


#include "DataStructures/CListNode.h"


namespace DataStructures {


template <class A> class CLinkedList {
protected:
	CListNode<A> m_nHead;	// Points to head item
	CListNode<A> m_nTail;	// Points to tail item

public:
	// Constructors/destructors
	CLinkedList()		{ CListNode<A>::InitList(m_nHead, m_nTail); }
	~CLinkedList()		{ RemoveAll(); }

	// Members to get the head and tail items of the list
	A *GetHead(){ return m_nHead.GetNext(); }
	A *GetTail(){ return m_nTail.GetPrevious(); }

	// Members to add items to the list
	void AddHead         ( A *p ){ ASSERT(p); p->InsertAfter((A *)&m_nHead); }
	void AddTail         ( A *p ){ ASSERT(p); p->InsertBefore((A *)&m_nTail); }
	void InsertBefore    ( A *p, A *pNode ){ ASSERT(p); p->InsertBefore(pNode); }
	void InsertAfter     ( A *p, A *pNode ){ ASSERT(p); p->InsertAfter(pNode); }
	void InsertAscending ( A *p ){ ASSERT(p); AddTail(p); p->SortLeftAscending(); }
	void InsertDescending( A *p ){ ASSERT(p); AddTail(p); p->SortLeftDescending(); }

	// Members to remove items from the list
	A    *Remove    ( A *p){ ASSERT(p); p->Remove(); return p; }
	A    *RemoveHead()     { return Remove(GetHead()); }
	A    *RemoveTail()     { return Remove(GetTail()); }
	void  RemoveAll ()     { while(GetHead()->IsInList()) delete GetHead(); }

	// Members to get the size of the list
	bool IsEmpty(){ return !GetHead()->IsInList(); }

	int GetCount(){
		register A *pNode = GetHead();
		int         i;

		for( i=0; pNode->IsInList(); i++ ){
			pNode = pNode->GetNext();
		}
		return i;
	}

	// Members to find items in the list by ID or name
	A *FindID( unsigned long dwID, A *pNode=NULL ){
		if( !pNode ){
			pNode = GetHead();
		}
		for( ; pNode->IsInList(); pNode=pNode->GetNext() ){
			if( dwID == pNode->GetID() ){
				return pNode;
			}
		}
		return NULL;
	}

	A *FindName( const char *pszName, A *pNode=NULL ){
		if( !pNode ){
			pNode = GetHead();
		}
		for( ; pNode->IsInList(); pNode=pNode->GetNext() ){
			if(stricmp(pszName, pNode->GetName()) == 0)
				return pNode;
		}
		return NULL;
	}
};


};  //  namespace DataStructures


#endif  //  TEDDY_DATA_STRUCTURES_C_LINKED_LIST_H

