
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CListNode
	\ingroup g_data_structures
	\author  Sean O'Neil
	\warning Conversion work in progress
	\date    2000

	Deriving from this class turns your class into a node to use in a doubly-
	linked list. It was designed to work with CLinkedList (defined below) in such
	a way that the object can insert and remove itself from the list. This can be
	very useful when one part of your code maintains a list and other parts of the
	code can add or delete to it. Add a reference counter and you can implement
	a caching mechanism that maintains itself. It is templatized to keep you from
	having to cast pointers manually, making the code easier to read and maintain.

	Example:
	#define CObjectList CLinkedList<CObject>
	class CObject : public CListNode<CObject> { ... };
*/


#ifndef TEDDY_DATA_STRUCTURES_C_LIST_NODE_H
#define TEDDY_DATA_STRUCTURES_C_LIST_NODE_H


#include <string.h>


#define ASSERT(x) // FIX


namespace DataStructures {


template <class A> class CListNode {
protected:
	A *m_pNext;			//!< Points to next item in the list
	A *m_pPrevious;		//!< Points to previous item in the list

public:
	// Constructors/destructors
	CListNode(){ m_pNext = m_pPrevious = NULL; }
	~CListNode(){ Remove(); }

	static void InitList( CListNode<A> &nHead, CListNode<A> &nTail ){
		nHead.m_pNext     = (A *)&nTail;
		nTail.m_pPrevious = (A *)&nHead;
	}

	// Members to get the next and previous pointers (must use Insert/Remove members to set them)
	A *GetNext()				{ return m_pNext; }
	A *GetPrevious()			{ return m_pPrevious; }

	// Functions for the list search, override these functions to use
	unsigned long  GetID      (){ return 0; }
	const    char *GetName    (){ return ""; }
	float          GetPriority(){ return 0; }

	// Members to insert into, remove from, and check status in a list
	bool IsInList()				{ return (m_pNext && m_pPrevious); }
	void InsertBefore( A *pNode ){
		// First make sure this is not already in the list and pNode is not NULL
		ASSERT(!IsInList() && pNode);
		m_pNext     = pNode;
		m_pPrevious = pNode->m_pPrevious;
		pNode->m_pPrevious->m_pNext = (A *)this;
		pNode->m_pPrevious          = (A *)this;
	}
	void InsertAfter( A *pNode ){
		// First make sure this is not already in the list and pNode is not NULL
		ASSERT(!IsInList() && pNode);
		m_pPrevious = pNode;
		m_pNext     = pNode->m_pNext;
		pNode->m_pNext->m_pPrevious = (A *)this;
		pNode->m_pNext              = (A *)this;
	}
	void Remove(){
		if( m_pPrevious ){
			m_pPrevious->m_pNext = m_pNext;
		}
		if( m_pNext ){
			m_pNext->m_pPrevious = m_pPrevious;
		}
		m_pNext = m_pPrevious = NULL;
	}

	// Members to help keep a list sorted
	void SortLeftAscending(){
		ASSERT( IsInList() );
		A *p;
		for( p=GetPrevious(); p->IsInList(); p=p->GetPrevious() ){
			if( p->GetPriority() <= ((A *)this)->GetPriority() ){
				break;
			}
		}
		if( p != GetPrevious() ){
			Remove();
			InsertAfter(p);
		}
	}

	void SortLeftDescending(){
		ASSERT( IsInList() );
		A *p;
		for( p=GetPrevious(); p->IsInList(); p=p->GetPrevious() ){
			if( p->GetPriority() >= ((A *)this)->GetPriority() ){
				break;
			}
		}
		if( p != GetPrevious() ){
			Remove();
			InsertAfter(p);
		}
	}
	void SortRightAscending(){
		A *p;
		for( p=GetNext(); p->IsInList(); p=p->GetNext() ){
			if( p->GetPriority() >= ((A *)this)->GetPriority() ){
				break;
			}
		}
		if( p != GetNext() ){
			Remove();
			InsertBefore(p);
		}
	}
	void SortRightDescending(){
		A *p;
		for( p=GetNext(); p->IsInList(); p=p->GetNext() ){
			if( p->GetPriority() <= ((A *)this)->GetPriority() ){
				break;
			}
		}
		if( p != GetNext() ){
			Remove();
			InsertBefore(p);
		}
	}
};


};  //  namespace DataStructures


#endif  //  TEDDY_DATA_STRUCTURES_C_LIST_NODE_H

