
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\class   Renderable
	\ingroup g_graphics
	\author  Timo Suoranta
	\brief   Base class for all renderable objects
	\date    2000

	Renderable is interface and base class for all graphics
	objects. Renderable maintains drawing instructions, which
	are dedicated to specific RenderContext. Object that wants
	to be renderable simultaneously on several RenderContext
	must maintain separate Renderables for each of them. For
	example, Mesh maintains set of MeshRenders, so that the
	very same Mesh can be rendered on different Projection
	Areas (which implement RenderContext interface).

	The base class provides optional service to store rendering
	code to displaylist. To use it, the subclass must implement
	update() method. In this case, drawing code should be placed
	into update(), beginUpdate() and endUpdate() should be called
	from there, and draw() method should not be overridden.

	If the subclass does not want to use the displaylist service,
	it does not need to implement update() method, and should
	place drawing code to draw() method.

	Optional drawDebug() method is part of the interface, but
	does not have clear use schema yet.
*/


#ifndef TEDDY_GRAPHICS_RENDERABLE_H
#define TEDDY_GRAPHICS_RENDERABLE_H


#include "Graphics/OpenGL.h"
#include "Graphics/RenderContext.h"
#include <map>
using namespace std;


namespace Graphics {


class Renderer;
typedef map<RenderContext*,Renderer*,less<RenderContext* > > RenderContext_to_Renderer;
			   

class Renderable {
public:
	Renderable();
	virtual ~Renderable();

	virtual void  draw       ( RenderContext *rc );
	virtual void  beginUpdate( RenderContext *rc );
	virtual void  update     ( RenderContext *rc ) = 0;
	virtual void  endUpdate  ( RenderContext *rc );
	Renderer     *getRenderer( RenderContext *rc );

protected:
	RenderContext_to_Renderer renderers;
};


};  //  namespace Graphics


#endif  //  TEDDY_GRAPHICS_RENDERABLE_H

