
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Graphics/OpenGL.h"
#include "Graphics/View.h"
#include "Graphics/Font.h"
#include "PhysicalComponents/Layer.h"
#include "PhysicalComponents/Style.h"
#include "Maths/Vector.h"
#include "SysSupport/EndianIn.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <list>
#include <algorithm>
using namespace SysSupport;
using namespace PhysicalComponents;
using namespace std;


int SCREEN_X = 640;
int SCREEN_Y = 480;


namespace Graphics {


View *View::active = NULL;


//!  View constructor
View::View( char *title, bool fullscreen ){
	active = this;

    //  Initialize SDL for video output
    if( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE) < 0 ){
        cerr << "Unable to initialize SDL:" << SDL_GetError() << endl;;
		cerr.flush();
        exit( 1 );
    }else{
	    atexit( SDL_Quit );
	}

    //  Create OpenGL screen
	width  = SCREEN_X;
	height = SCREEN_Y;
	if( fullscreen ){
		if( SDL_SetVideoMode(width, height, 0, SDL_OPENGL|SDL_FULLSCREEN) == NULL ) {
			cerr  << "Unable to create OpenGL screen: " << SDL_GetError() << endl;;
			cerr.flush();
			exit( 2 );
		}
	}else{
		if( SDL_SetVideoMode(width, height, 0, SDL_OPENGL) == NULL ) {
			cerr  << "Unable to create OpenGL screen: " << SDL_GetError() << endl;;
			cerr.flush();
			exit( 3 );
		}
	}

    SDL_WM_SetCaption( title, NULL );

	ratio  = (GLfloat)width/(GLfloat)height;
	frames = 0;
	glGetIntegerv( GL_VIEWPORT, viewport );
	reshape( width, height );

	if( Font::default_font == NULL ){
		Font::default_font = new Font( "fonts.raw" );
	}
	if( Style::default_style == NULL ){
		Style::default_style = new Style();
	}

	frames                = 0;
	gl_polygon_mode       = GL_FILL;
	gl_shade_model        = GL_FLAT;
	gl_blend_func_sfactor = GL_SRC_ALPHA;
	gl_blend_func_dfactor = GL_ONE_MINUS_SRC_ALPHA;
	gl_fog_mode           = GL_LINEAR;
	gl_fog_color[0]       = 0.0;
	gl_fog_color[1]       = 0.0;
	gl_fog_color[2]       = 0.0;
	gl_fog_start          = 0.0;
	gl_fog_end            = 1.0;
	gl_clear_depth        = 1.0;

}


//!  Insert Layer to View
void View::insert( Layer *layer ){
	layers.push_back( layer );
}


/*static*/ void View::leave(){
}


//!  Static view::display() calls active->display();
void View::displayOne(){
	active->display();
}


//!  Return viewport ratio
float View::getRatio(){
	return ratio;
}


//!  Return viewport
GLint *View::getViewport(){
	return viewport;
}


//!  Return viewport height  FIX (-viewport[0])
GLint View::getHeight(){
	return viewport[3];
}


//!  OpenGL window reshape callback function
void View::reshape( int w, int h ){
	glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
	glGetIntegerv( GL_VIEWPORT, viewport );
	ratio  = (GLfloat)w/(GLfloat)h;
	width  = w;
	height = h;
	list<Layer*>::iterator l_it = layers.begin();
	while( l_it != layers.end() ){
		(*l_it)->update(this);
		l_it++;
	}
	displayOne();
}


void View::display(){
	//  Clear buffers
#ifdef USE_WINDOW_SYSTEM
	glClearColor( C_GRAY_50,1.0 );
	//glClearDepth( 1.0 );
	//glDepthRange( 0.0, 1.0 );
	glClear( GL_COLOR_BUFFER_BIT /*| GL_DEPTH_BUFFER_BIT*/ );
#else
	glClearColor( C_BLACK,1.0 );
	glClearDepth( 1.0 );
	glDepthRange( 0.0, 1.0 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
#endif

	//  Draw layers
	list<Layer*>::iterator l_it = layers.begin();
	while( l_it != layers.end() ){
		(*l_it)->draw();
		l_it++;
	}

	//  Flush OpenGL and swap buffers
	glFlush();
	frames++;
	SDL_GL_SwapBuffers();

	//  Check for errors
	check();
}


//!  Static view::reshape
void View::reshapeOne( int w, int h ){
	active->reshape( w, h );
}


//!  View interface - Prepare Area for rendering 2D graphics
void View::begin2d(){
	glViewport( viewport[0], viewport[1], viewport[2], viewport[3] );
	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glColor4f( 1.0, 1.0, 1.0, 1.0 );
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_CULL_FACE );
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0.0f, viewport[2]-viewport[0], viewport[3]-viewport[1], 0.0f, 0.0f, 1.0f );

	glMatrixMode( GL_MODELVIEW );

	glLoadIdentity();
	glTranslatef( 0.375, 0.375, 0.0 );

}


//!  View interface - End rendering 2D graphics
void View::end2d(){
	glPopAttrib();
}


//!  View interface - Prepare Area for rendering 3D graphics
void View::begin3d(){
	glPushAttrib( GL_ALL_ATTRIB_BITS );
}


//!  View interface - End rendering 3D graphics
void View::end3d(){
	glPopAttrib();
}


//!  Set gl polygon mode
/*static*/ void View::setPolygonMode( const GLenum polygon_mode ){
	if( gl_polygon_mode != polygon_mode ){
		gl_polygon_mode = polygon_mode;
		glPolygonMode( GL_FRONT_AND_BACK, polygon_mode );
	}
}


//!  Set gl shade model
/*static*/ void View::setShademodel( const GLenum shade_model ){
	if( gl_shade_model != shade_model ){
		gl_shade_model = shade_model;
		glShadeModel( shade_model );
	}
}


//!  Set blending function
/*static*/ void View::setBlendFunc( const GLenum sfactor, const GLenum dfactor ){
	if( gl_blend_func_sfactor != sfactor ||
	    gl_blend_func_dfactor != dfactor    ){
		gl_blend_func_sfactor = sfactor;
		gl_blend_func_dfactor = dfactor;
		glBlendFunc( sfactor, dfactor );
	}
}


//!  Set fog mode
/*static*/ void View::setFogMode( const GLenum mode ){
	glFogi( GL_FOG_MODE, mode );
}


//!  Set fog color
/*static*/ void View::setFogColor( GLfloat *color ){
	glFogfv( GL_FOG_COLOR, color );
}


//!  Set fog start
/*static*/ void View::setFogStart( GLfloat start ){
	glFogf( GL_FOG_START, start );
}


//!  Set fog end
/*static*/ void View::setFogEnd( GLfloat end ){
	glFogf( GL_FOG_END, end );
}


//!  Check if there are any OpenGL errors.
/*static*/ void View::check(){	   
	int e;

	e = glGetError();
	switch( e ){
	case GL_NO_ERROR:
		break;
	default:
//		err_v->setLine( 11, (char*)gluErrorString((GLenum)e) );
		cout << "OpenGL error: " << gluErrorString((GLenum)e) << endl;
		cout.flush();
		break;
	}
}


};  //  namespace Graphics


