
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Imports/LWMesh.h"
#include "Imports/LWSurface.h"
#include "Models/PolygonMesh.h"
#include "Models/SmoothFace.h"
#include "Models/SmoothVertex.h"
#include "SysSupport/Exception.h"
#include <iostream>
#include <algorithm>
using namespace std;


#define USE_LIGHTWAVE_SCALE   1
#define SCALE                 1.0f


namespace Imports {


//!  Constructor which loads LightWave object file
LWMesh::LWMesh( const string &fname ):Mesh(fname){
	ID4 form;
	U4  file_len;

	num_vertices    = 0;
	num_faces       = 0;
	num_surfaces    = 0;
	num_envelopes   = 0;
	num_clips       = 0;
	num_tags        = 0;
	current_surface = 0;
	max_len         = 0;

	try{
		f = new LWFile( fname.c_str() );

		f->pushDomain( 8 );           // Master domain only allows reading of first 8 bytes of FORM and length
		form     = f->read_ID4();     // FORM
		file_len = f->read_U4();      // file length-8

		f->pushDomain( file_len );    // File data domain allows reading of rest of the file
		file_type = f->read_ID4();    // LWOB

		did( file_type );
		cout << " file " << fname << " length " << file_len << endl;

		while( f->bytesRead()-8 < file_len ){
			processChunk();
		}
	}catch( Exception &e ){
		cout << e.tellMsg() << endl;
	}
	cout.flush();
	f->close();

	//  Post processing - smooth vertices
	list<Mesh*>::iterator m_it;
	m_it = submeshes.begin();
	while( m_it != submeshes.end() ){
		((PolygonMesh*)(*m_it))->PolygonMesh::smoothVertices();
		m_it++;
	}

	//  Post processing - set clip radius
	this->setClipRadius( max_len );
	//  FIX update();
}


//!  Destructor
/*virtual*/ LWMesh::~LWMesh(){
	//  FIX
}


//!  Process a LWOB & LWO2 file chunks
void LWMesh::processChunk(){
	U4 chunk_id     = f->read_ID4();
	U4 chunk_length = f->read_U4();
	f->pushDomain( chunk_length );

//	did( chunk_id );
//	cout << " chunk of length " << chunk_length << endl;
	switch( file_type ){
		case ID_LWOB:
		case ID_LWLO:
			switch( chunk_id ){
				case ID_LAYR: layer_U2_U2_S0(); break;  //  Actually LWLO only
				case ID_PNTS: pointList();      break;
				case ID_SRFS: surfaceList();    break;
				case ID_POLS: faceList();       break;
				case ID_CRVS: curveList();      break;
				case ID_PCHS: patchList();      break;
				case ID_SURF: surface_sc();     break;
				default: break;
			}
			break;
		case ID_LWO2:
			switch( chunk_id ){
				case ID_LAYR: layer_U2_U2_VEC12_S0_U2();   break;
				case ID_PNTS: pointList();                 break;
				case ID_VMAP: vertexMapping_ID4_U2_S0_d(); break;
				case ID_POLS: polygonList();               break;
				case ID_TAGS: tags_d();                    break;
				case ID_PTAG: polygonTags_ID4_d();         break;
				case ID_ENVL: envelope_U4_sc();            break;
				case ID_CLIP: clip_U4_sc();                break;
				case ID_SURF: surf_S0_S0_sc();             break;
				case ID_BBOX: boundingBox_VEC12_VEC12();   break;
				case ID_DESC: descriptionLine_S0();        break;
				case ID_TEXT: comments_S0();               break;
				case ID_ICON: thumbnail_U2_U2_d();         break;
				default: break;
			}
			break;
		default:
			break;
	}

	f->popDomain( true );
}


/*!
	LWOB and LOW2  PNTS { point-location[VEC12] * } 
	
	This chunk contains a list of the X, Y, and Z coordinates of all the points 
	in an object. Since each coordinate has three components, and each component 
	is stored as a four byte floating point number, the number of points in an 
	object can be determined by dividing the size in bytes of the PNTS chunk by 
	12. 
	
	By convention, the +X direction is to the right or east, the +Y direction is 
	upward, and the +Z direction is forward or north. For models of real-world 
	objects, the unit size is usually considered to be one meter. The 
	coordinates are specified relative to an object's pivot point. See the 
	LightWave Modeler manual for more information about LightWave 3D's geometric 
	conventions. 
	
	Points in the PNTS chunk are numbered in the order they occur, starting with 
	zero. This index is then used by polygons to define their vertices. The PNTS 
	chunk must be before the POLS, CRVS, and PCHS chunks in the file. 
*/
void LWMesh::pointList(){
	SmoothVertex *vertex;
	U4            current_v_index = 0;
	U4            i;

	num_vertices = f->domainLeft()/12;

	for( i=0; i<num_vertices; i++ ){
		if( f->domainLeft()<3*4 ){
			cout << "Not enough floats found in chunk" << endl;
			break;
		}

		vertex = new SmoothVertex( f->read_VEC12() );
#ifdef USE_LIGHTWAVE_SCALE
		*vertex *= SCALE; 
#endif
		vertex->setNormal( Vector(0,0,0) );
		if( vertex->Len() > max_len ){
			max_len = vertex->Len();
		}
		this->vertices.insert( pair<U4,SmoothVertex*>(current_v_index,vertex) );
		current_v_index++;
	}
	//cout << num_vertices << " points read " << this->vertices.size() << " in map" << endl;
}


/*
	LWOB SRFS { surf-name[S0] * } 
	
	This chunk contains a list of the names of all the surfaces in an object. 
	Each surface name appears as a null-terminated character string. If the 
	length of the string (including the null) is odd, an extra null byte is 
	added. Surface names should be read from the file until as many bytes as the 
	chunk size specifies have been read. 
	
	In LightWave 3D terminology, a "surface" is defined as a named set of 
	shading attributes. Each polygon contains a reference to the surface used to 
	color the polygon. The names as listed in the SRFS chunk are numbered in the 
	order they occur, starting from 1, and this index is used by polygons to 
	define their surface. The SRFS chunk must be before the POLS, CRVS, and PCHS 
	chunks in the file.
*/
void LWMesh::surfaceList(){
	string       surface_name;
	LWSurface   *surface;
	PolygonMesh *mesh;
	char         buffer[16];

	while( f->domainLeft() > 0 ){
		num_surfaces++;
		surface_name = f->read_S0();
		surface = new LWSurface  ( f, file_type, surface_name );
		mesh    = new PolygonMesh( surface_name );
		sprintf( buffer, "%d", (int)(num_surfaces) );
		surfaces.insert( pair<string,LWSurface  *>(surface_name,surface) );
		meshes  .insert( pair<string,PolygonMesh*>(string(buffer),mesh ) );
		mesh->setMaterial( surface );
		submeshes.push_back( mesh );
	}
	//cout << num_surfaces << " surfaces in this object" << endl;
}


/*!
	LWOB POLS { ( numvert[U2], vert[U2] # numvert, surf[I2] )* } 
	
	This chunk contains a list of all the polygons in an object. Each entry 
	consists of a short integer specifying the number of vertices in the polygon 
	followed by that many short integers specifying the vertices themselves (as 
	indices into the points list) followed by a short integer specifying which 
	surface is used by the polygon (as an index into the surfaces list). The 
	number of vertices in a polygon currently may vary from one to 200. The 
	vertex list for each polygon should begin at a convex vertex and proceed 
	clockwise as seen from the visible side of the polygon (LightWave 3D 
	polygons are single-sided, except for those whose surfaces have the 
	double-sided flag set). Polygons should be read from the file until as many 
	bytes as the chunk size specifies have been read. 
	
	Since the points in the PNTS chunk are referenced using two-byte integers, 
	the effective maximum number of points in a LightWave object file is 65,536. 
	This is an inherient limitation of this current format. 
	
	A negative surface number for a polygon indicates that the polygon has 
	detail polygons (which are drawn on top of the main polygon and may be 
	coplanar with it). In this case, the next number in the file is a short 
	integer specifying how many detail polygons belong to the current polygon. 
	This is followed by a list of those detail polygons, where each entry is of 
	the same format as described above for regular polygons (except that the 
	detail polygons cannot have details of their own). The list of regular 
	polygons then resumes. To determine which surface is used by a polygon with 
	a negative surface number, the absolute value of that number should be used. 
	Note, however, that detail polygons are mostly obsolete, so even though they 
	may be recognized by LightWave and old files contain them, they should be 
	ignored.
*/
void LWMesh::faceList(){
	SmoothFace   *face;
	SmoothVertex *vertex;
	PolygonMesh  *mesh;
	U4            num_vertices;
	U4            surface_index;
	U4            vertex_index;
	U4            i;
	bool          good;
	U4            bad_count   = 0;
	U4            point_count = 0;
	U4            line_count  = 0;

	while( f->domainLeft() > 0 ){
		num_vertices = f->read_U2();

		if( num_vertices<3 ){  //  FIX
			bad_count++;
			switch( num_vertices ){
			case 0: cout << "Bad polygon - 0 vertices " << endl; break;
			case 1: point_count++; break;
			case 2: line_count++; break;
			}			
			for( i=0; i<num_vertices; i++ ){
				vertex_index = f->read_U2();
			}
			f->read_U2();
			continue;
		}

		face = new SmoothFace();
		good = true;
		for( i=0; i<num_vertices; i++ ){
			if( f->domainLeft()<2 ){
				cout << "Not enough shorts found in chunk" << endl;
				good = false;
				break;
			}
			vertex_index = f->read_U2();

			if( vertex_index > this->num_vertices || vertex_index<0 ){
				cout << "Not understood as vertex index" << endl;
				good = false;
				break;
			}

			//  Find vertex for this index and insert to face
			U4_to_SmoothVertex::iterator v_it = vertices.find( vertex_index );
			if( v_it!=vertices.end() ){
				vertex = (*v_it).second;
				if( vertex != NULL ){
					face->append( vertex );
				}else{
					cout << "Vertex " << vertex_index << " pointer is bad" << endl;
					good = false;
				}
			}else{
				cout << "Vertex " << vertex_index << " not found" << endl;
				good = false;
			}

		}
		if( good ){
			surface_index = f->read_U2();
			if( surface_index > num_surfaces ){

				bad_count++;
				delete face;
				continue;
				//  FIX error
			}

			string_to_PolygonMesh::iterator m_it;
			string                          surface_name;
			char                            buffer[16];

			sprintf( buffer, "%d", (int)(surface_index) );
			surface_name = string( buffer );

			m_it = meshes.find( surface_name );
			if( m_it!=meshes.end() ){
				mesh = (*m_it).second;
				if( mesh==NULL ){
					cout << "Mesh-surface " << surface_name << " found as NULL" << endl;
					bad_count++;
					delete face;
					continue;
				}

				face->makeNormal();
				mesh->insert( face );
				faces.insert( pair<U4,SmoothFace*>(num_faces,face) );
				num_faces++;
			}else{
				cout << "Mesh-surface " << surface_name << " not found" << endl;
				bad_count++;
				delete face;
				continue;
			}
		}else{
			bad_count++;
			delete face;
		}
	}	
/*	cout << num_faces << " good polygons read " << bad_count << " bad polygons skipped";
	if( point_count>0 ){
		cout << " (" << point_count << " points, not yet supported)";
	}
	if( line_count>0 ){
		cout << " (" << line_count << " lines, not yet supported)";
	}
	cout << endl;
	cout.flush();*/
}


//!  LWOB Read curve list
void LWMesh::curveList(){
	//  FIX not implemented for LWOB
}


//!  LWOB Read patch list
void LWMesh::patchList(){
	//  FIX not implemented for LWOB
}


/*!
	LWOB SURF { name[S0], attributes[SUB-CHUNK] * } 
	
	Each SURF chunk describes the surface attributes of a particular surface. 
	These chunks begin with the name of the surface being described. Following 
	the name is a series of sub-chunks, which are like normal IFF chunks except 
	that their sizes are specified by short integers instead of longs. It is 
	likely that the variety of sub-chunks will grow as new surface attributes 
	are added to the program, but any unknown sub-chunks may be skipped over by 
	using the size. Sub-chunks should be read from the file until as many bytes 
	as the chunk size specifies have been read. 
*/
void LWMesh::surface_sc(){
	string_to_LWSurface::iterator  s_it;
	string                         surface_name;
	LWSurface                     *surface; 

	surface_name   = f->read_S0();                               

	s_it = surfaces.find( surface_name );
	if( s_it != surfaces.end() ){
		surface = (*s_it).second;
		if( surface == NULL ){
			cout << "NULL surface " << surface_name << " in surfacelist" << endl;
			return;
		}
	}else{
		surface = new LWSurface( f, file_type, surface_name );
	}
	
	while( f->domainLeft() > 0 ){
		surface->processSubChunk();
	}
}


/*! LWLO LAYR Chunk

	An LAYR chunk must precede each set of PNTS, POLS and CRVS data
	chunks and indicates to which layer those data belong.  An LAYR
	chunk consists of two unsigned short values and a string.  The first
	is the layer number which should be from 1 to 10 to operate
	correctly in Modeler.  This restriction may be lifted in future
	versions of the format.  The second value is a bitfield where only
	the lowest order bit is defined and all others should be zero.  This
	bit is one if this is an active layer and zero if it is a background
	layer.  The string which follows is the name of the layer and should
	be null-terminated and padded to even length.
*/
void LWMesh::layer_U2_U2_S0(){
	U2     layer_number = f->read_U2();
	U2     flags        = f->read_U2();
	string name         = f->read_S0();
/*	cout << "Start Layer " << name << endl;
	cout.flush();*/
}


/*!
	LWO2 Start Layer 

	LAYR { number[U2], flags[U2], pivot[VEC12], name[S0], parent[U2] } 

	This chunk signals the start of a new layer. All the data chunks which
	follow will be included in this layer until another layer chunk is
	encountered. If data is encountered before a layer chunk, it goes into
	an arbitrary layer. The parent index indicates the default parent for
	this layer and can be -1 or missing to indicate no parent. 
*/
void LWMesh::layer_U2_U2_VEC12_S0_U2(){
	U2     layer_number = f->read_U2();
	U2     layer_flags  = f->read_U2();
	Vector layer_pivot  = f->read_VEC12();
	string layer_name   = f->read_S0();

	if( f->domainLeft() >= 2 ){
		U2 layer_parent = f->read_U2();
	}
	num_layers++;
//	cout << "Layer " << num_layers << ":" << layer_name << " start" << endl;
}


/*void pointList(){  Same for LWO2 as LWOB
} */


/*!
	LWO2 Vertex Mapping 

	VMAP { type[ID4], dimension[U2], name[S0], ( vert[VX], value[F4] # dimension )* } 

	This chunk contains a set of floating point vectors associated with a set of vertices.
	Each one has a type which is a four-character ID code, a dimension and a name string.
	After that follows a list of vertex / vector pairs, with the vertex given by a vertex
	index in VX format and an array of "dimension" floating-point values. There can be any
	number of these chunks, although they should all have different types or names. 

	Some possible type codes are TXUV for UV mapping coordinates, MNVW for MetaNURBS
	vertex weights, MORF for vector offsets defining an alternate object shape, SPOT
	for alternate vertex positions, RGBA for coloring, etc. 
*/
void LWMesh::vertexMapping_ID4_U2_S0_d(){
	ID4    vmap_type = f->read_ID4();
	U2     vmap_dim  = f->read_U2();
	string vmap_name = f->read_S0();

	while( f->domainLeft() > 0 ){
		VX vertex_id = f->read_VX();
		for( int i=0; i<vmap_dim; i++ ){
			F4 value = f->read_F4();
		}		
	}

/*
	switch( vmap_type ){
		case ID_TXUV: break;
		case ID_MNVW: break;
		case ID_MORF: break;
		case ID_SPOT: break;
		case ID_RGBA: break;
		default:
			break;
	}
*/
}


/*!
	LWO2 Polygon List 

	POLS { type[ID4], ( numvert+flags[U2], vert[VX] # numvert )* } 

	This chunk contains a list of polygons for the current layer. The 'type' code indicates
	the polygon type and can be FACE, CURV, PACH, or possibly other values. Each entry starts
	with a short integer specifying the number of vertices in the polygon. The 6 high-order
	bits of the number of vertices are flags bits with different meaning for each polygon
	type, so there are a maximum of 1023 vertices per polygon. The vertex list is the
	specified number of two or four byte VX vertex indices (relative to the most recent
	points list). The vertex list for each polygon should begin at a convex vertex and proceed
	clockwise as seen from the visible side of the polygon (LightWave 3D polygons are
	single-sided, except for those whose surfaces have the double-sided flag set). Polygons
	should be read from the file until as many bytes as the chunk size specifies have been read. 

	128 64 32 16 8 4 2 1  123456789abcdef

*/
void LWMesh::polygonList(){
	SmoothFace   *face;
	SmoothVertex *vertex;
	ID4           polygons_type = f->read_ID4();
	U4            vertex_index;
	int           i;
	bool          good;
	int           bad_count   = 0;
	int           point_count = 0;
	int           line_count  = 0;

	while( f->domainLeft() > 0 ){
		U2 data    = f->read_U2();
		U2 numvert = data & 0x03ff;
		U2 flags   = (data & ~numvert) >> 10;

		if( numvert<3 ){  //  FIX
			bad_count++;
			switch( num_vertices ){
			case 0: cout << "Bad polygon - 0 vertices " << endl; break;
			case 1: point_count++; break;
			case 2: line_count++; break;
			}			
			for( i=0; i<numvert; i++ ){
				vertex_index = f->read_VX();
			}
			continue;
		}

		face = new SmoothFace();
		good = true;

		for( i=0; i<numvert; i++ ){

			if( f->domainLeft()<2 ){
				cout << "Not enough bytes left in chunk" << endl;
				good = false;
				break;
			}

			vertex_index = f->read_VX();
			if( vertex_index > this->num_vertices || vertex_index<0 ){
				cout << vertex_index <<" not understood as vertex index" << endl;
				good = false;
				break;
			}

			U4_to_SmoothVertex::iterator v_it = vertices.find( vertex_index);
			if( v_it!=vertices.end() ){
				vertex = (*v_it).second;
				if( vertex != NULL ){
					face->append( vertex );
				}else{
					cout << "Vertex " << vertex_index << " pointer is bad" << endl;
					good = false;
				}
			}else{
				cout << "Vertex " << vertex_index << " not found" << endl;
				good = false;
			}
		}
		if( good ){
			face->makeNormal();
			faces.insert( pair<U4,SmoothFace*>(num_faces,face) );
			num_faces++;
		}else{
			bad_count++;
			delete face;
		}

/*
		switch( polygons_type ){
			case ID_FACE: break;
			case ID_CURV: break;
			case ID_PACH: break;
			default:
				break;
		}
*/
	}
/*	cout << num_faces << " good polygons read " << bad_count << " bad polygons skipped";
	if( point_count>0 ){
		cout << " (" << point_count << " points, not yet supported)";
	}
	if( line_count>0 ){
		cout << " (" << line_count << " lines, not yet supported)";
	}
	cout << endl;
	cout.flush();*/
}


/*!
	LWO2 Tag Strings 

	TAGS { tag-string[S0] * }
	 
	This chunk lists the tags strings that can be associated with polygons by the PTAG chunk.
	Strings should be read until as many bytes as the chunk size specifies have been read, and
	each string is assigned an index starting from zero. 

*/
void LWMesh::tags_d(){

	while( f->domainLeft() > 0 ){	
		string tag = f->read_S0();
//		cout << "Tag " << num_tags << " " << tag << endl;
		tags.insert( pair<U4,string>(num_tags,tag) );
		num_tags++;
	}
}


/*!
	LWO2 Polygon Tag Mapping 

	PTAG { type[ID4], ( poly[VX], tag[U2] )* } 

	This chunk contains the all the tags of a given type for some subset of the polygons
	defined in the preceding POLS chunk. The type of the tag association is given by the
	first element of the chunk and is an normal 4-character ID code. The rest of the chunk
	is a list of polygon/tag associations. The polygon is identified by its index into the
	previous POLS chunk, and the tag is given by its index into the previous TAGS chunk.
	Any number of polygons may be mapped with this type of tag, and mappings should be read
	from the file until as many bytes as the chunk size specifies have been read. 

	Polygon tags types and their values are extensible, but there are some pre-defined types.
	The SURF type tags each polygon with the name of its surface. In LightWave 3D terminology,
	a "surface" is defined as a named set of shading attributes and may be described in the
	object file in SURF chunks. Another pre-defined type is PART which describes what aspect
	of the mesh each polygon belongs to, and SMGP which names the smoothing group for each
	polygon. Not all polygons have a value for every tag type, and the behavior for polygon
	which lack a certain tag depends on the type. 
*/
void LWMesh::polygonTags_ID4_d(){
	string_to_LWSurface  ::iterator s_it;
	string_to_PolygonMesh::iterator m_it;
	U4_to_SmoothFace     ::iterator f_it;
	U4_to_string         ::iterator t_it;

	SmoothFace    *face;
	LWSurface     *surface;
	PolygonMesh   *mesh;
	VX             polygon_index;
	U4             tag_index;
	ID4            tag_type;
	string         tag_value;

	tag_type = f->read_ID4();

	while( f->domainLeft() > 0 ){

		polygon_index = f->read_VX();
		tag_index     = f->read_U2();
		t_it          = tags.find( tag_index );
		if( t_it!= tags.end() ){
			tag_value  = (*t_it).second;

			switch( tag_type ){

				case ID_SURF:
					f_it = faces   .find( polygon_index );
					s_it = surfaces.find( tag_value     );
					m_it = meshes  .find( tag_value     );

					//  We absolutely need the face to be added
					if( f_it != faces.end() ){
						face = (*f_it).second;
						if( face == NULL ){
							break;
						}
					}else{
						break;
					}

					//  If first polygon to be added to the surface,
					//  then the surface has not been created yet.
					if( s_it == surfaces.end() ){
						surface = new LWSurface( f, file_type, tag_value );
						surfaces.insert( pair<string,LWSurface*>(tag_value,surface) );
					}else{
						surface = (*s_it).second;
					}

					if( m_it == meshes.end() ){
						mesh = new PolygonMesh( tag_value );
						meshes.insert( pair<string,PolygonMesh*>(tag_value,mesh) );
						mesh->setMaterial( surface );
						this->insert( mesh );
					}else{
						mesh = (*m_it).second;
					}

					mesh->insert( face );
					break;

				default:
					break;
			}
		}
	}
}


/*!
	LWO2 Envelope Definition 

	ENVL { index[U4], attributes[SUB-CHUNK] * } 

	Each ENVL chunk defines the time-varying curve for a single parameter channel.
	The index is used to identify this envelope uniquely, and can have any non-zero
	value less than 0x1000000. Following the index is a series of sub-chunks, which
	are like normal IFF chunks except that their sizes are specified by short
	integers instead of longs. The number of sub-chunks may vary depending on the
	complexity of the envelope, so sub-chunks should be read from the file until as
	many bytes as the chunk size specifies have been read. New types of sub-chunks
	may be introduced as the program capabilities increase, but any unknown sub-chunks
	may be skipped over by using their size. 

*/
void LWMesh::envelope_U4_sc(){
	U4          envelope_index = f->read_U4();
	LWEnvelope *envelope       = new LWEnvelope( f );

	envelopes.insert( pair<U4,LWEnvelope*>(num_envelopes,envelope) );
	num_envelopes++;
	while( f->domainLeft() > 0 ){
		envelope->processEnvelope();
	}
}


/*!
	LWO2 Image or Sequence Definition 

	CLIP { index[U4], attributes[SUB-CHUNK] * } 

	Each CLIP chunk defines a image which can be used for applying as a texture
	map in surfaces. The term 'clip' is used to describe these since they may be
	time-varying sequences or animations rather than just stills. The index
	identifies this clip uniquely and may be any non-zero value less than 0x1000000.
	The attributes which define the source imagery and modifiers follow as a variable
	list of sub-chunks
*/
void LWMesh::clip_U4_sc(){
	U4      clip_index = f->read_U4();
	LWClip *clip       = new LWClip( f );

	if( clip_index >= 0x1000000 ){
		cout << "Clip index not below 0x1000000" << endl;
	}

	clips.insert( pair<U4,LWClip*>(clip_index,clip) );
	num_clips++;
	while( f->domainLeft() > 0 ){
		clip->processClip();
	}
}


/*!
	LWO2 Surface Definition 

	SURF { name[S0], source[S0], attributes[SUB-CHUNK] * } 

	Each SURF chunk describes the shading attributes of a particular surface.
	These chunks begin with the name of the surface being described plus the
	name of the source surface. If the source name is non-null, then this surface
	is derived from the source surface. The base attributes of the source surface
	can be overridden by this surface, and texture blocks can be added to the
	source surface. The material attributes follow as a variable list of sub-chunks. 

*/
void LWMesh::surf_S0_S0_sc(){
	string_to_LWSurface::iterator  s_it;
	string                         surface_name;
	string                         surface_source;
	LWSurface                     *surface; 

	surface_name   = f->read_S0();                               
	surface_source = f->read_S0();

	s_it = surfaces.find( surface_name );
	if( s_it != surfaces.end() ){
		surface = (*s_it).second;
		if( surface == NULL ){
			cout << "NULL surface " << surface_name << " in surfacelist" << endl;
			return;
		}
	}else{
		surface = new LWSurface( f, file_type, surface_name );
	}
	
	if(	surface_source != "" ){
		cout << "NO MATERIAL INHERITANCE YET" << endl;
		//  FIX copy material properties
	}
	while( f->domainLeft() > 0 ){
		surface->processSubChunk();
	}
}


/*!
	LWO2 Bounding Box 

	BBOX { min[VEC12], max[VEC12] } 

	This is an optional chunk which can be included to store the bounding box
	for the vertex data in a layer. The min and max vectors are the lower and
	upper corners of the bounding box. 
*/
void LWMesh::boundingBox_VEC12_VEC12(){
	bbox_min = f->read_VEC12();
	bbox_max = f->read_VEC12();
}


/*!
	LWO2 Description Line 

	DESC { description-line[S0] } 

	This is an optional chunk which can be used to hold an object description.
	This should be a simple line of upper and lowercase characters, punctuation
	and spaces which describes the contents of the object file. There should be
	no control characters in this text string and it should generally be kept
	short. 
*/
void LWMesh::descriptionLine_S0(){
	description_line = f->read_S0();
}


/*!
	LWO2 Commentary Text 

	TEXT { comment[S0] } 

	This is another optional chunk which can be used to hold comments about the
	object. The text is just like the DESC chunk, but it can be about any subject,
	it may contain newline characters and it does not need to be particularly short. 
*/
void LWMesh::comments_S0(){
	commentary_text = f->read_S0();
}


/*!
	LWO2 Thumbnail Icon Image 

	ICON { encoding[U2], width[U2], data[U1] * } 

	This optional chunk contains an iconic or thumbnail image for the object
	which can be used when viewing the file in a browser. The encoding is a
	code for the data format which can only be zero for now meaning uncompressed,
	unsigned bytes as RGB triples. The width specifies the number of pixels
	contained in each row of the image. The data consists of rows of pixels
	(RGB triples for now), and the height of the image is determined by the
	length of the data.
*/
void LWMesh::thumbnail_U2_U2_d(){
	U2 encoding = f->read_U2();
	U2 width    = f->read_U2();

	if( encoding != 0 ){
		cout << "Unknown thumbnail encoding" << endl;
		return;
	}

	while( f->domainLeft() > 0 ){
		for( int x=0; x<width; x++ ){
			U1 red   = f->read_U1();
			U1 green = f->read_U1();
			U1 blue  = f->read_U1();
			//  FIX store data to texture
		}
	}
}


};  //  namespace Imports


