
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\class   LWMesh
	\ingroup g_imports
	\author  Timo Suoranta
	\brief   LightWave object file loader
	\warning Many features are not yet implemented
	\todo    Detail polygons
	\todo    Points
	\todo    Lines
	\todo    Polygon tessalation
	\todo    Layers
	\todo    Texture and Image maps
	\bug     Destructors missing
	\date    1999, 2000

	This class implements loader for LightWave object files.
	It is a subclass of Mesh, which will contain no Elements,
	only PolygonMeshes as submeshes.

	Parts LightWave object file format specifications are placed
	into comments into source-code. Briefly, the file has several
	chunks, which further may contain subchunks.

	Currently many features are not yet implemented. Some of those
	output some information to cout but do not use nor store that
	information in any way. Others simply do nothing.

	Some important information, which is properly processed, is also
	outputted to standard output. This is a debug feature and will
	be removed once classes are finished.

	One of the unfinished feature, layers, makes it almost useless
	to use layered files at the moment. I also seem to have trouble
	with polygon sidedness.
*/


#ifndef TEDDY_IMPORTS_LW_MESH_H
#define TEDDY_IMPORTS_LW_MESH_H


#include "Models/Mesh.h"
#include "Models/PolygonMesh.h"
#include "Models/SmoothVertex.h"
#include "Models/SmoothFace.h"
#include "Models/Mesh.h"
#include "Imports/LWSurface.h"
#include "Imports/LWFile.h"
#include "Imports/LWEnvelope.h"
#include "Imports/LWClip.h"
#include <map>
#include <string>
using namespace Models;
using namespace std;


namespace Imports {


typedef map<string, PolygonMesh *,   less<string> > string_to_PolygonMesh;
typedef map<string, LWSurface *,     less<string> > string_to_LWSurface;
typedef map<U4,     SmoothVertex *,  less<U4>     > U4_to_SmoothVertex;
typedef map<U4,     SmoothFace *,    less<U4>     > U4_to_SmoothFace;
typedef map<U4,     LWEnvelope *,    less<U4>     > U4_to_LWEnvelope;
typedef map<U4,     LWClip *,        less<U4>     > U4_to_LWClip;
typedef map<U4,     string,          less<U4>     > U4_to_string;


class LWMesh : public Mesh {
public:
	LWMesh( const string &fname );
	virtual ~LWMesh();

protected:
	void processChunk();

	//  LWOB & LWLO
	void pointList                 ();  //  LWOB::PNTS & LWO2::PNTS
	void polygonList               ();  //  LWO2::POLS
	void faceList                  ();  //  LWOB::POLS
	void surfaceList               ();  //  SRFS
	void curveList                 ();  //  CRVS
	void patchList                 ();  //  PCHS
	void surface_sc                ();  //  SURF
	void layer_U2_U2_S0            ();  //  LWLO::LAYR

	//  LWO2
	void layer_U2_U2_VEC12_S0_U2   ();  //  LWO2::LAYR
	void vertexMapping_ID4_U2_S0_d ();  //  VMAP
	void tags_d                    ();  //  TAGS
	void polygonTags_ID4_d         ();  //  PTAG
	void envelope_U4_sc            ();  //  ENVL
	void clip_U4_sc                ();  //  CLIP
	void surf_S0_S0_sc             ();  //  SURF
	void boundingBox_VEC12_VEC12   ();  //  BBOX
	void descriptionLine_S0        ();  //  DESC
	void comments_S0               ();  //  TEXT
	void thumbnail_U2_U2_d         ();  //  ICON

	LWFile *f;
	ID4     file_type;
	F4      max_len;

	string_to_PolygonMesh meshes;
	string_to_LWSurface   surfaces;
	U4_to_string          tags;
	U4_to_SmoothVertex    vertices;
	U4_to_SmoothFace      faces;
	U4_to_LWEnvelope      envelopes;
	U4_to_LWClip          clips;
	U4                    num_layers;
	U4                    num_vertices;
	U4                    num_faces;
	U4                    num_surfaces;
	U4                    num_envelopes;
	U4                    num_clips;
	U4                    num_tags;
	U4                    current_surface;
	Vector                bbox_min;
	Vector                bbox_max;
	string                description_line;
	string                commentary_text;	
};


};  //  namespace Imports


#endif  //  TEDDY_IMPORTS_LW_MESH_H


