
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Graphics/GlFeature.h"
#include "Graphics/View.h"
#include "Materials/Material.h"
#include "PhysicalComponents/Projection.h"
#include <iostream>
using namespace std;


namespace Materials {


/*static*/ unsigned short Material::master_flags =
	USE_GL_POLYGON_MODE |
	USE_GL_SHADE_MODEL 	|
	USE_GL_FRONT_FACE  	|
	USE_GL_BLEND        |
	USE_GL_ALPHA_TEST  	|
	USE_GL_CULL_FACE   	|
	USE_GL_DEPTH_TEST   |
	USE_GL_STENCIL_TEST	|
	USE_GL_LIGHTING    	|
	USE_GL_FOG         	|
	USE_GL_NORMALIZE   	|
	USE_GL_TEXTURE_2D;  


//!  Constructor with name
Material::Material( const string &name ){
	this->name         = name;
	this->shininess    = 0.0;
	this->ambient      = Color::GRAY_50;
	this->diffuse      = Color::GRAY_50;
	this->specular     = Color::BLACK;
	this->emission     = Color::BLACK;
	this->polygon_mode = GL_LINE;
	this->shade_model  = GL_FLAT;
	this->front_face   = GL_CW;
	this->flags =
		USE_GL_CULL_FACE |
		USE_GL_DEPTH_TEST;
}


//!  Constructor with name, lighting and shading modes
Material::Material( const string &name, const bool lighting, const int shade_model ){
	this->name         = name;
	this->shininess    = 0.0;
	this->ambient      = Color::GRAY_50;
	this->diffuse      = Color::GRAY_50;
	this->specular     = Color::BLACK;
	this->emission     = Color::BLACK;
	this->polygon_mode = GL_LINE;
	this->shade_model  = shade_model;
	this->front_face   = GL_CW;
	this->flags =
		USE_GL_CULL_FACE  |
		USE_GL_DEPTH_TEST |
		(lighting ? USE_GL_LIGHTING : 0);
}


//!  Constructor with name and diffuse component
Material::Material( const string &name, const Color &diffuse ){
	this->name         = name;
	this->shininess    = 1.0;
	this->ambient      = Color::GRAY_50;
	this->diffuse      = diffuse;
	this->specular     = Color::BLACK;
	this->emission     = Color::BLACK;
	this->polygon_mode = GL_LINE;
	this->shade_model  = GL_SMOOTH;
	this->front_face   = GL_CW;
	this->flags =
		USE_GL_CULL_FACE  |
		USE_GL_DEPTH_TEST |
		USE_GL_LIGHTING;
}



//!  Set flags
void Material::setFlags( unsigned short flags ){
	this->flags = flags;
}


//!  Enable Flags
void Material::enableFlags( unsigned short flags ){
	this->flags |= flags;
}


//!  Disable flags
void Material::disableFlags( unsigned short flags ){
	this->flags &= ~flags;
}


//!  Set material ambient component
void Material::setAmbient( const Color &a ){
	this->ambient = a;
}


//!  Set material diffuse component
void Material::setDiffuse( const Color &d ){
	this->diffuse = d;
}


//!  Set material specular component
void Material::setSpecular( const Color &s ){
	this->specular = s;
}


/*!
	Set material emission component
	(OpenGL emission does not emit light to other surfaces)
*/
void Material::setEmission( const Color &e ){
	this->emission = e;
}


/*!
	Set material shininess
	(OpenGL shininess from 1.0 to 128.0)
*/
void Material::setShininess( const float &s ){
	this->shininess = s;
}


//!  Debugging information
void Material::debug(){
	cout << "Material:  " << name.c_str() << endl;
	cout << "Ambient:   "; ambient .debug(); cout << endl;
	cout << "Diffuse:   "; diffuse .debug(); cout << endl;
	cout << "Specular:  "; specular.debug(); cout << endl;
	cout << "Emission:  "; emission.debug(); cout << endl;
	cout << "Shininess: " << shininess << endl;
	
	switch( polygon_mode ){
	case GL_LINE: cout << "Polygon mode GL_LINE" << endl; break;
	case GL_FILL: cout << "Polygon mode GL_FILL" << endl; break;
	default:      cout << "Unknown polygon mode " << polygon_mode << endl; break;
	}
	
	switch( shade_model ){
	case GL_FLAT:   cout << "Shade model GL_FLAT"   << endl; break;
	case GL_SMOOTH: cout << "Shade model GL_SMOOTH" << endl; break;
	default:        cout << "Unknown shade model "  << shade_model << endl; break;
	}
	
	switch( front_face ){
	case GL_CW:  cout << "Front face GL_CW"    << endl; break;
	case GL_CCW: cout << "Front face GL_CCW"   << endl; break;
	default:     cout << "Unknown front face " << front_face << endl; break;
	}
	
	if( (flags & USE_GL_BLEND )==USE_GL_BLEND ){
		cout << "Uses Blend" << endl;
	}
	if( (flags & USE_GL_CULL_FACE )==USE_GL_CULL_FACE ){
		cout << "Uses Cull Face" << endl;
	}
	if( (flags & USE_GL_DEPTH_TEST )==USE_GL_DEPTH_TEST ){
		cout << "Uses Depth Test" << endl;
	}
	if( (flags & USE_GL_FOG )==USE_GL_FOG ){
		cout << "Uses Fog" << endl;
	}
	if( (flags & USE_GL_LIGHTING )==USE_GL_LIGHTING ){
		cout << "Uses Lighting" << endl;
	}
	if( (flags & USE_GL_NORMALIZE )==USE_GL_NORMALIZE ){
		cout << "Uses Normalize" << endl;
	}
	if( (flags & USE_GL_TEXTURE_2D )==USE_GL_TEXTURE_2D ){
		cout << "Uses Texture 2D" << endl;
	}
	cout.flush();
}


//!  Apply material to OpenGL
void Material::glApply( RenderContext *rc ){
	//return;  //  FIX (Testing material change effect on performance - this call costs!)
	View::check();
	Projection *p = dynamic_cast<Projection*>( rc );
	if( p == NULL ){
		cerr << "Material::glApply(): RenderContext not Projection" << endl;
		cerr.flush();
		return;
	}
	View *view = p->getView();
	if( view == NULL ){
		cerr << "Material::glApply(): Projection has no View" << endl;
		cerr.flush();
		return;
	}

	/*
	switch( front_face ){
	case GL_CW:  cout << "Applying GL_CW Front face"    << endl; break;
	case GL_CCW: cout << "Applying GL_CCW Front face"   << endl; break;
	default:     cout << "Applying Unknown " << front_face << " front face " << endl; break;
	}
	switch( polygon_mode ){
	case GL_FILL: cout << "Applying GL_FILL polygon mode" << endl; break;
	case GL_LINE: cout << "Applying GL_LINE polygon mode" << endl; break;
	default:      cout << "Applying unknown " << polygon_mode << " polygon mode" << endl; break;
	}
	cout.flush();
	*/

/*	view->setPolygonMode( polygon_mode );
	view->setShademodel( shade_model );
	glFrontFace( front_face );
	GlFeature::blend     .apply( (flags & USE_GL_BLEND      )==USE_GL_BLEND );
	GlFeature::cull_face .apply( (flags & USE_GL_CULL_FACE  )==USE_GL_CULL_FACE );
	GlFeature::depth_test.apply( (flags & USE_GL_DEPTH_TEST )==USE_GL_DEPTH_TEST );
	GlFeature::fog       .apply( (flags & USE_GL_FOG        )==USE_GL_FOG );
	GlFeature::lighting  .apply( (flags & USE_GL_LIGHTING   )==USE_GL_LIGHTING );
	GlFeature::normalize .apply( (flags & USE_GL_NORMALIZE  )==USE_GL_NORMALIZE );
	GlFeature::texture_2d.apply( (flags & USE_GL_TEXTURE_2D )==USE_GL_TEXTURE_2D );
*/	if( (GlFeature::lighting.isMasterEnabled()     ) &&
	    ((flags & USE_GL_LIGHTING)==USE_GL_LIGHTING)    ){
		glMaterialfv( GL_FRONT, GL_AMBIENT,   ambient .rgba );
		glMaterialfv( GL_FRONT, GL_DIFFUSE,   diffuse .rgba );
		glMaterialfv( GL_FRONT, GL_SPECULAR,  specular.rgba );
		glMaterialfv( GL_FRONT, GL_EMISSION,  emission.rgba );
		glMaterialf ( GL_FRONT, GL_SHININESS, shininess );
		cout << "Shininess " << shininess << endl; cout.flush();
	}else{
		glColor4fv( diffuse.rgba );
	}
	View::check();
}


//!  Return material name
string Material::getName(){
	return name;
}


};  //  namespace Materials


