
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Maths/CFractal.h"


namespace Maths {


float CFractal::fBm( float *f, float fOctaves ){
	// Initialize locals
	float fValue = 0;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;

	for( i=0; i<m_nDimensions; i++){
		fTemp[i] = f[i];
	}

	// Inner loop of spectral construction, where the fractal is built
	for( i=0; i<fOctaves; i++ ){
		fValue += Noise(fTemp) * m_fExponent[i];
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
	}

	// Take care of remainder in fOctaves
	fOctaves -= (int)fOctaves;
	if( fOctaves > DELTA ){
		fValue += fOctaves * Noise(fTemp) * m_fExponent[i];
	}
	return CLAMP( -0.99999f, 0.99999f, fValue );
}


float CFractal::fBmTest( float *f, float fOctaves ){
	// Initialize locals
	float fValue = 0;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;

	for( i=0; i<m_nDimensions; i++ ){
		fTemp[i] = f[i] * 2;
	}

	//fOctaves *= Abs(Noise(fTemp)) + 1.0f;
	//fOctaves = Clamp(2, 16, fOctaves);

	// Inner loop of spectral construction, where the fractal is built
	for( i=0; i<fOctaves; i++ ){
		fValue += Noise(fTemp) * m_fExponent[i];
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
	}

	// Take care of remainder in fOctaves
	fOctaves -= (int)fOctaves;
	if( fOctaves > DELTA ){
		fValue += fOctaves * Noise(fTemp) * m_fExponent[i];
	}

	if( fValue <= 0.0f ){
		return (float)-pow( -fValue, 0.7f );
	}
	return (float)pow( fValue, 1 + Noise(fTemp) * fValue );
}


float CFractal::Turbulence( float *f, float fOctaves ){
	// Initialize locals
	float fValue = 0;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;

	for( i=0; i<m_nDimensions; i++ ){
		fTemp[i] = f[i];
	}

	// Inner loop of spectral construction, where the fractal is built
	for( i=0; i<fOctaves; i++ ){
		fValue += Abs(Noise(fTemp)) * m_fExponent[i];
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
	}

	// Take care of remainder in fOctaves
	fOctaves -= (int)fOctaves;
	if( fOctaves > DELTA ){
		fValue += fOctaves * Abs(Noise(fTemp) * m_fExponent[i]);
	}
	return CLAMP( -0.99999f, 0.99999f, fValue );
}


float CFractal::Multifractal( float *f, float fOctaves, float fOffset ){
	// Initialize locals
	float fValue = 1;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;

	for( i=0; i<m_nDimensions; i++ ){
		fTemp[i] = f[i];
	}

	// Inner loop of spectral construction, where the fractal is built
	for( i=0; i<fOctaves; i++ ){
		fValue *= Noise(fTemp) * m_fExponent[i] + fOffset;
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
	}

	// Take care of remainder in fOctaves (shouldn't that be a multiply?)
	fOctaves -= (int)fOctaves;
	if( fOctaves > DELTA ){
		fValue *= fOctaves * (Noise(fTemp) * m_fExponent[i] + fOffset);
	}
	return CLAMP( -0.99999f, 0.99999f, fValue );
}


float CFractal::Heterofractal( float *f, float fOctaves, float fOffset ){
	// Initialize locals
	float fValue = Noise(f) + fOffset;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;
	for( i=0; i<m_nDimensions; i++ ){
		fTemp[i] = f[i] * m_fLacunarity;
	}

	// Inner loop of spectral construction, where the fractal is built
	for( i=1; i<fOctaves; i++ ){
		fValue += (Noise(fTemp) + fOffset) * m_fExponent[i] * fValue;
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
	}

	// Take care of remainder in fOctaves
	fOctaves -= (int)fOctaves;
	if( fOctaves > DELTA ){
		fValue += fOctaves * (Noise(fTemp) + fOffset) * m_fExponent[i] * fValue;
	}
	return CLAMP( -0.99999f, 0.99999f, fValue );
}


float CFractal::HybridMultifractal( float *f, float fOctaves, float fOffset, float fGain ){
	// Initialize locals
	float fValue  = (Noise(f) + fOffset) * m_fExponent[0];
	float fWeight = fValue;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;

	for( i=0; i<m_nDimensions; i++ ){
		fTemp[i] = f[i] * m_fLacunarity;
	}

	// Inner loop of spectral construction, where the fractal is built
	for( i=1; i<fOctaves; i++ ){
		if( fWeight > 1 ){
			fWeight = 1;
		}
		float fSignal = (Noise(fTemp) + fOffset) * m_fExponent[i];
		fValue  += fWeight * fSignal;
		fWeight *= fGain * fSignal;
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
	}

	// Take care of remainder in fOctaves
	fOctaves -= (int)fOctaves;
	if( fOctaves > DELTA ){
		if( fWeight > 1 ){
			fWeight = 1;
		}
		float fSignal = (Noise(fTemp) + fOffset) * m_fExponent[i];
		fValue += fOctaves * fWeight * fSignal;
	}
	return CLAMP( -0.99999f, 0.99999f, fValue );
}


float CFractal::RidgedMultifractal( float *f, float fOctaves, float fOffset, float fGain ){
	// Initialize locals
	float fSignal = fOffset - Abs(Noise(f));
	fSignal *= fSignal;
	float fValue = fSignal;
	float fTemp[MAX_DIMENSIONS];
	int   i;
	int   j;

	for( i=0; i<m_nDimensions; i++ ){
		fTemp[i] = f[i];
	}

	// Inner loop of spectral construction, where the fractal is built
	for( i=1; i<fOctaves; i++ ){
		for( j=0; j<m_nDimensions; j++ ){
			fTemp[j] *= m_fLacunarity;
		}
		float fWeight = Clamp( 0, 1, fSignal * fGain );
		fSignal  = fOffset - Abs(Noise(fTemp));
		fSignal *= fSignal;
		fSignal *= fWeight;
		fValue  += fSignal * m_fExponent[i];
	}
	return CLAMP( -0.99999f, 0.99999f, fValue );
}



};  //  namespace Maths


