
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CMatrix
	\ingroup g_maths
	\author  Sean O'Neil
	\brief   4x4 Matrix
	\date    2000

	This class implements a 4x4 matrix. Several functions and operators are
	defined to make working with matrices easier. The values are kept in column-
	major order to make it easier to use with OpenGL. For performance reasons,
	most of the functions assume that all matrices are orthogonal, which means the
	bottom row is [ 0 0 0 1 ]. Since I plan to use the GL_PROJECTION matrix to
	handle the projection matrix, I should never need to use any other kind of
	matrix, and I get a decent performance boost by ignoring the bottom row.

*/


#ifndef TEDDY_MATHS_C_MATRIX_H
#define TEDDY_MATHS_C_MATRIX_H


#include "Maths/CVector.h"
#include <math.h>


#define tanf(a) tan(a)


namespace Maths {


class CVector;
class CQuaternion;


class CMatrix {
public:
	// This class uses column-major order, as used by OpenGL
	// Here are the ways in which the matrix values can be accessed:
	// | f11 f21 f31 f41 |   | f1[0] f1[4] f1[8]  f1[12] |   | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |
	// | f12 f22 f32 f42 |   | f1[1] f1[5] f1[9]  f1[13] |   | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |
	// | f13 f23 f33 f43 | = | f1[2] f1[6] f1[10] f1[14] | = | f2[0][2] f2[1][2] f2[2][2] f2[3][2] |
	// | f14 f24 f34 f44 |   | f1[3] f1[7] f1[11] f1[15] |   | f2[0][3] f2[1][3] f2[2][3] f2[3][3] |
	union {
		struct {
			float f11, f12, f13, f14, f21, f22, f23, f24, f31, f32, f33, f34, f41, f42, f43, f44;
		} A;
		float f1[16];
		float f2[4][4];
	};

	CMatrix(){}
	CMatrix( float        f  ){ *this = f; }
	CMatrix( float       *pf ){ *this = pf; }
	CMatrix( CQuaternion &q  ){ *this = q; }

	// Init functions
	void ZeroMatrix(){
		A.f11 = A.f12 = A.f13 = A.f14 = A.f21 = A.f22 = A.f23 = A.f24 = A.f31 = A.f32 = A.f33 = A.f34 = A.f41 = A.f42 = A.f43 = A.f44 = 0;
	}

	void IdentityMatrix(){
		A.f12 = A.f13 = A.f14 = A.f21 = A.f23 = A.f24 = A.f31 = A.f32 = A.f34 = A.f41 = A.f42 = A.f43 = 0;
		A.f11 = A.f22 = A.f33 = A.f44 = 1;
	}

	operator float*(){
		return f1;
	}

	float   &operator[]( int n        ){ return f1[n]; }
	float   &operator()( int i, int j ){ return f2[i][j]; }

	void     operator= ( float  f       ){ for( register int i=0; i<16; i++ ){ f1[i] = f; }  }
	void     operator= ( float *pf      ){ for( register int i=0; i<16; i++ ){ f1[i] = pf[i]; }  }
	void     operator= ( CQuaternion &q );

	CMatrix  operator* ( CMatrix &m );
	void     operator*=( CMatrix &m ){ *this = *this * m; }
	CVector  operator* ( CVector &v ){ return TransformVector(v); }

	CVector TransformVector( CVector &v ){
		// 9 muls, 9 adds
		// | f11 f21 f31 f41 |   | v.x |   | f11*v.x+f21*v.y+f31*v.z+f41 |
		// | f12 f22 f32 f42 |   | v.y |   | f12*v.x+f22*v.y+f32*v.z+f42 |
		// | f13 f23 f33 f43 | * | v.z | = | f13*v.x+f23*v.y+f33*v.z+f43 |
		// | 0   0   0   1   |   | 1   |   | 1                           |
		return CVector(
			(A.f11*v.x+A.f21*v.y+A.f31*v.z+A.f41),
			(A.f12*v.x+A.f22*v.y+A.f32*v.z+A.f42),
			(A.f13*v.x+A.f23*v.y+A.f33*v.z+A.f43)
		);
	}

	CVector TransformNormal( CVector &v ){
		// 9 muls, 6 adds
		// | f11 f21 f31 f41 |   | v.x |   | f11*v.x+f21*v.y+f31*v.z |
		// | f12 f22 f32 f42 |   | v.y |   | f12*v.x+f22*v.y+f32*v.z |
		// | f13 f23 f33 f43 | * | v.z | = | f13*v.x+f23*v.y+f33*v.z |
		// | 0   0   0   1   |   | 1   |   | 1                       |
		return CVector(
			(A.f11*v.x + A.f21*v.y + A.f31*v.z),
			(A.f12*v.x + A.f22*v.y + A.f32*v.z),
			(A.f13*v.x + A.f23*v.y + A.f33*v.z)
		);
	}

	// Translate functions
	void TranslateMatrix( float x, float y, float z ){
		// | 1  0  0  x |
		// | 0  1  0  y |
		// | 0  0  1  z |
		// | 0  0  0  1 |
		A.f12 = A.f13 = A.f14 = A.f21 = A.f23 = A.f24 = A.f31 = A.f32 = A.f34 = 0;
		A.f11 = A.f22 = A.f33 = A.f44 = 1;
		A.f41 = x;
		A.f42 = y;
		A.f43 = z;
	}

	void TranslateMatrix( float *pf ){
		TranslateMatrix(pf[0], pf[1], pf[2]);
	}

	void Translate( float x, float y, float z ){
		// 9 muls, 9 adds
		// | f11 f21 f31 f41 |   | 1  0  0  x |   | f11 f21 f31 f11*x+f21*y+f31*z+f41 |
		// | f12 f22 f32 f42 |   | 0  1  0  y |   | f12 f22 f32 f12*x+f22*y+f32*z+f42 |
		// | f13 f23 f33 f43 | * | 0  0  1  z | = | f13 f23 f33 f13*x+f23*y+f33*z+f43 |
		// | 0   0   0   1   |   | 0  0  0  1 |   | 0   0   0   1                     |
		A.f41 = A.f11*x + A.f21*y + A.f31*z + A.f41;
		A.f42 = A.f12*x + A.f22*y + A.f32*z + A.f42;
		A.f43 = A.f13*x + A.f23*y + A.f33*z + A.f43;
	}

	void Translate( float *pf ){
		Translate( pf[0], pf[1], pf[2] );
	}

	// Scale functions
	void ScaleMatrix( float x, float y, float z ){
		// | x  0  0  0 |
		// | 0  y  0  0 |
		// | 0  0  z  0 |
		// | 0  0  0  1 |
		A.f12 = A.f13 = A.f14 = A.f21 = A.f23 = A.f24 = A.f31 = A.f32 = A.f34 = A.f41 = A.f42 = A.f43 = 0;
		A.f11 = x;
		A.f22 = y;
		A.f33 = z;
		A.f44 = 1;
	}

	void ScaleMatrix( float *pf ){
		ScaleMatrix(pf[0], pf[1], pf[2]);
	}

	void Scale( float x, float y, float z ){
		// 9 muls
		// | f11 f21 f31 f41 |   | x  0  0  0 |   | f11*x f21*y f31*z f41 |
		// | f12 f22 f32 f42 |   | 0  y  0  0 |   | f12*x f22*y f32*z f42 |
		// | f13 f23 f33 f43 | * | 0  0  z  0 | = | f13*x f23*y f33*z f43 |
		// | 0   0   0   1   |   | 0  0  0  1 |   | 0     0     0     1   |
		A.f11 *= x; A.f21 *= y; A.f31 *= z;
		A.f12 *= x; A.f22 *= y; A.f32 *= z;
		A.f13 *= x; A.f23 *= y; A.f33 *= z;
	}

	void Scale( float *pf ){
		Scale(pf[0], pf[1], pf[2]);
	}

	// Rotate functions
	void RotateXMatrix( float fRadians ){
		// | 1 0    0     0 |
		// | 0 fCos -fSin 0 |
		// | 0 fSin fCos  0 |
		// | 0 0    0     1 |
		A.f12 = A.f13 = A.f14 = A.f21 = A.f24 = A.f31 = A.f34 = A.f41 = A.f42 = A.f43 = 0;
		A.f11 = A.f44 = 1;

		float fCos = cosf(fRadians);
		float fSin = sinf(fRadians);
		A.f22 = A.f33 = fCos;
		A.f23 = fSin;
		A.f32 = -fSin;
	}

	void RotateX( float fRadians ){
		// 12 muls, 6 adds, 2 trig function calls
		// | f11 f21 f31 f41 |   | 1 0    0     0 |   | f11 f21*fCos+f31*fSin f31*fCos-f21*fSin f41 |
		// | f12 f22 f32 f42 |   | 0 fCos -fSin 0 |   | f12 f22*fCos+f32*fSin f32*fCos-f22*fSin f42 |
		// | f13 f23 f33 f43 | * | 0 fSin fCos  0 | = | f13 f23*fCos+f33*fSin f33*fCos-f23*fSin f43 |
		// | 0   0   0   1   |   | 0 0    0     1 |   | 0   0                 0                 1   |
		float fTemp, fCos, fSin;

		fCos  = cosf( fRadians );
		fSin  = sinf( fRadians );

		fTemp = A.f21*fCos + A.f31*fSin;
		A.f31 = A.f31*fCos - A.f21*fSin;
		A.f21 = fTemp;
		fTemp = A.f22*fCos + A.f32*fSin;
		A.f32 = A.f32*fCos - A.f22*fSin;
		A.f22 = fTemp;
		fTemp = A.f23*fCos + A.f33*fSin;
		A.f33 = A.f33*fCos - A.f23*fSin;
		A.f23 = fTemp;
	}

	void RotateYMatrix( float fRadians ){
		// | fCos  0 fSin  0 |
		// | 0     1 0     0 |
		// | -fSin 0 fCos  0 |
		// | 0     0 0     1 |
		A.f12 = A.f14 = A.f21 = A.f23 = A.f24 = A.f32 = A.f34 = A.f41 = A.f42 = A.f43 = 0;
		A.f22 = A.f44 = 1;

		float fCos = cosf( fRadians );
		float fSin = sinf( fRadians );

		A.f11 =  A.f33 = fCos;
		A.f13 = -fSin;
		A.f31 =  fSin;
	}

	void RotateY( float fRadians ){
		// 12 muls, 6 adds, 2 trig function calls
		// | f11 f21 f31 f41 |   | fCos  0 fSin  0 |   | f11*fCos-f31*fSin f21 f11*fSin+f31*fCos f41 |
		// | f12 f22 f32 f42 |   | 0     1 0     0 |   | f12*fCos-f32*fSin f22 f12*fSin+f32*fCos f42 |
		// | f13 f23 f33 f43 | * | -fSin 0 fCos  0 | = | f13*fCos-f33*fSin f23 f13*fSin+f33*fCos f43 |
		// | 0   0   0   1   |   | 0     0 0     1 |   | 0                 0   0                 1   |
		float fTemp, fCos, fSin;

		fCos  = cosf( fRadians );
		fSin  = sinf( fRadians );
		fTemp = A.f11*fCos - A.f31*fSin;
		A.f31 = A.f11*fSin + A.f31*fCos;
		A.f11 = fTemp;
		fTemp = A.f12*fCos - A.f32*fSin;
		A.f32 = A.f12*fSin + A.f32*fCos;
		A.f12 = fTemp;
		fTemp = A.f13*fCos - A.f33*fSin;
		A.f33 = A.f13*fSin + A.f33*fCos;
		A.f13 = fTemp;
	}

	void RotateZMatrix( float fRadians ){
		// | fCos -fSin 0 0 |
		// | fSin fCos  0 0 |
		// | 0    0     1 0 |
		// | 0    0     0 1 |
		A.f13 = A.f14 = A.f23 = A.f24 = A.f31 = A.f32 = A.f34 = A.f41 = A.f42 = A.f43 = 0;
		A.f33 = A.f44 = 1;

		float fCos = cosf (fRadians );
		float fSin = sinf( fRadians );
		A.f11 =  A.f22 = fCos;
		A.f12 =  fSin;
		A.f21 = -fSin;
	}

	void RotateZ( float fRadians ){
		// 12 muls, 6 adds, 2 trig function calls
		// | f11 f21 f31 f41 |   | fCos -fSin 0 0 |   | f11*fCos+f21*fSin f21*fCos-f11*fSin f31 f41 |
		// | f12 f22 f32 f42 |   | fSin fCos  0 0 |   | f12*fCos+f22*fSin f22*fCos-f12*fSin f32 f42 |
		// | f13 f23 f33 f43 | * | 0    0     1 0 | = | f13*fCos+f23*fSin f23*fCos-f13*fSin f33 f43 |
		// | 0   0   0   1   |   | 0    0     0 1 |   | 0                 0                 0   1   |
		float fTemp, fCos, fSin;

		fCos  = cosf( fRadians );
		fSin  = sinf( fRadians );
		fTemp = A.f11*fCos + A.f21*fSin;
		A.f21 = A.f21*fCos - A.f11*fSin;
		A.f11 = fTemp;
		fTemp = A.f12*fCos + A.f22*fSin;
		A.f22 = A.f22*fCos - A.f12*fSin;
		A.f12 = fTemp;
		fTemp = A.f13*fCos + A.f23*fSin;
		A.f23 = A.f23*fCos - A.f13*fSin;
		A.f13 = fTemp;
	}

	void RotateMatrix( CVector &v, float f ){
		// 15 muls, 10 adds, 2 trig function calls
		float fCos = cosf( f );

		CVector vCos = v * (1 - fCos);
		CVector vSin = v * sinf( f );

		A.f14 = A.f24 = A.f34 = A.f41 = A.f42 = A.f43 = 0;
		A.f44 = 1;

		A.f11 = (v.x * vCos.x) + fCos;
		A.f21 = (v.x * vCos.y) - (vSin.z);
		A.f31 = (v.x * vCos.z) + (vSin.y);
		A.f12 = (v.y * vCos.x) + (vSin.z);
		A.f22 = (v.y * vCos.y) + fCos;
		A.f32 = (v.y * vCos.z) - (vSin.x);
		A.f13 = (v.z * vCos.x) - (vSin.y);
		A.f32 = (v.z * vCos.y) + (vSin.x);
		A.f33 = (v.z * vCos.z) + fCos;
	}

	void Rotate( CVector &v, float f ){
		// 51 muls, 37 adds, 2 trig function calls
		CMatrix mat;
		mat.RotateMatrix( v, f );
		*this *= mat;
	}

	void ModelMatrix( CQuaternion &q, CVector &vFrom ){
		*this = q;
		A.f41 = vFrom.x;
		A.f42 = vFrom.y;
		A.f43 = vFrom.z;
	}

	void ViewMatrix( CQuaternion &q, CVector &vFrom ){
		*this = q;
		A.f41 = -(vFrom.x*A.f11 + vFrom.y*A.f21 + vFrom.z*A.f31);
		A.f42 = -(vFrom.x*A.f12 + vFrom.y*A.f22 + vFrom.z*A.f32);
		A.f43 = -(vFrom.x*A.f13 + vFrom.y*A.f23 + vFrom.z*A.f33);
	}

	void ViewMatrix( CVector &vFrom, CVector &vView, CVector &vUp, CVector &vRight ){
		// 9 muls, 9 adds
		A.f11 = vRight.x; A.f21 = vRight.y; A.f31 = vRight.z; A.f41 = -DotProduct(vFrom, vRight);
		A.f12 = vUp.x;    A.f22 = vUp.y;    A.f32 = vUp.z;    A.f42 = -DotProduct(vFrom, vUp);
		A.f13 = -vView.x; A.f23 = -vView.y; A.f33 = -vView.z; A.f43 = -DotProduct(vFrom, -vView);
		A.f14 = 0;        A.f24 = 0;        A.f34 = 0;        A.f44 = 1;
	}

	void ViewMatrix( CVector &vFrom, CVector &vAt, CVector &vUp ){
		CVector vView = vAt - vFrom;
		vView.Normalize();
		CVector vRight = CrossProduct( vView, vUp );
		vRight.Normalize();
		vUp = CrossProduct( vRight, vView );
		vUp.Normalize();
		ViewMatrix( vFrom, vView, vUp, vRight );
	}

	void ProjectionMatrix( float fNear, float fFar, float fFOV, float fAspect ){
		// 2 muls, 3 divs, 2 adds, 1 trig function call
		float h = 1.0f / tanf( DEGTORAD(fFOV * 0.5f) );
		float Q = fFar / (fFar - fNear);
		A.f12 = A.f13 = A.f14 = A.f21 = A.f23 = A.f24 = A.f31 = A.f32 = A.f41 = A.f42 = A.f44 = 0;
		A.f11 = h / fAspect;
		A.f22 = h;
		A.f33 = Q;
		A.f34 = 1;
		A.f43 = -Q*fNear;
	}

	// For orthogonal matrices, I belive this also gives you the inverse.
	void Transpose(){
		float t1;
		float t2;
		SWAP( A.f12, A.f21, t1 );
		SWAP( A.f13, A.f31, t2 );
		SWAP( A.f14, A.f41, t1 );
		SWAP( A.f23, A.f32, t2 );
		SWAP( A.f24, A.f42, t1 );
		SWAP( A.f34, A.f43, t2 );
	}
};


};  //  namespace Maths


#endif  //  TEDDY_MATHS_C_MATRIX_H




