
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Maths/CQuaternion.h"


namespace Maths {


void CQuaternion::operator=( CMatrix &m ){
	// Check the sum of the diagonal
	float tr = m(0, 0) + m(1, 1) + m(2, 2);

	if( tr > 0.0f ){
		// The sum is positive
		// 4 muls, 1 div, 6 adds, 1 trig function call
		float s = sqrtf(tr + 1.0f);
		w = s * 0.5f;
		s = 0.5f / s;
		x = (m(1, 2) - m(2, 1)) * s;
		y = (m(2, 0) - m(0, 2)) * s;
		z = (m(0, 1) - m(1, 0)) * s;
	}else{
		// The sum is negative
		// 4 muls, 1 div, 8 adds, 1 trig function call
		const int nIndex[3] = {1, 2, 0};
		int i, j, k;
		i = 0;
		if(m(1, 1) > m(i, i))
			i = 1;
		if(m(2, 2) > m(i, i))
			i = 2;
		j = nIndex[i];
		k = nIndex[j];

		float s = sqrtf((m(i, i) - (m(j, j) + m(k, k))) + 1.0f);
		(*this)[i] = s * 0.5f;
		if(s != 0.0)
			s = 0.5f / s;
		(*this)[j] = (m(i, j) + m(j, i)) * s;
		(*this)[k] = (m(i, k) + m(k, i)) * s;
		(*this)[3] = (m(j, k) - m(k, j)) * s;
	}
}

CQuaternion CQuaternion::operator*( const CQuaternion &q ){
	// 12 muls, 30 adds
	float E = (x + z)*(q.x + q.y);
	float F = (z - x)*(q.x - q.y);
	float G = (w + y)*(q.w - q.z);
	float H = (w - y)*(q.w + q.z);
	float A = F - E;
	float B = F + E;
	return CQuaternion(
		(w + x)*(q.w + q.x) + (A - G - H) * 0.5f,
		(w - x)*(q.y + q.z) + (B + G - H) * 0.5f,
		(y + z)*(q.w - q.x) + (B - G + H) * 0.5f,
		(z - y)*(q.y - q.z) + (A + G + H) * 0.5f
	);
}


// Spherical linear interpolation between two quaternions
CQuaternion Slerp( const CQuaternion &q1, const CQuaternion &q2, float t ){
	// Calculate the cosine of the angle between the two
	float fScale0, fScale1;
	double dCos = q1.x * q2.x + q1.y * q2.y + q1.z * q2.z + q1.w * q2.w;

	// If the angle is significant, use the spherical interpolation
	if( (1.0 - ABS(dCos)) > DELTA ){
		double dTemp = acos(ABS(dCos));
		double dSin = sin(dTemp);
		fScale0 = (float)(sin((1.0 - t) * dTemp) / dSin);
		fScale1 = (float)(sin(t * dTemp) / dSin);
	}else{  // Else use the cheaper linear interpolation
		fScale0 = 1.0f - t;
		fScale1 = t;
	}
	if( dCos < 0.0 ){
		fScale1 = -fScale1;
	}

	// Return the interpolated result
	return (q1 * fScale0) + (q2 * fScale1);
}


};  //  namespace Maths

