
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CQuaternion
	\ingroup g_maths
	\author  Sean O'Neil
	\date    2000

	This class implements a 4D quaternion. Several functions and operators are
	defined to make working with quaternions easier. Quaternions are often use to
	represent rotations, and have a number of advantages over other constructs.
	Their main disadvantage is that they are unintuitive.
*/


#ifndef TEDDY_MATHS_C_QUATERNION_H
#define TEDDY_MATHS_C_QUATERNION_H


#include "Maths/CVector.h"
#include "Maths/CMatrix.h"
#include <math.h>


namespace Maths {


class CQuaternion : public CVector {
public:
	float w;

public:
	// Constructors
	CQuaternion(){}
	CQuaternion( const CVector &v, const float _w=0 ){
		x = v.x;
		y = v.y;
		z = v.z;
		w = _w;
	}

	CQuaternion( const CQuaternion &q ){
		x = q.x;
		y = q.y;
		z = q.z;
		w = q.w;
	}

	CQuaternion( float _x, float _y, float _z, float _w=0 ){
		x = _x;
		y = _y;
		z = _z;
		w = _w;
	}

	CQuaternion( float f ){
		x = y = z = w = f;
	}

	CQuaternion( float *p ){
		x = p[0];
		y = p[1];
		z = p[2];
		w = p[3];
	}

	// Casting and unary operators
	operator float*(){
		return &x;
	}

	float &operator[]( int n ){
		return (&x)[n];
	}

	CQuaternion operator-(){
		return CQuaternion( -x, -y, -z, -w );
	}

	// Equal and comparison operators
	void operator= ( float              f ){ x = y = z = w = f; }
	void operator= ( float             *p ){ x = p[0]; y = p[1]; z = p[2]; w = p[3]; }
	void operator= ( const CVector     &v ){ x = v.x; y = v.y; z = v.z; w = 0; }
	void operator= ( const CQuaternion &q ){ x = q.x; y = q.y; z = q.z; w = q.w; }
	void operator= ( CMatrix           &m );

	bool operator==( CQuaternion &q ){
		return (Abs(x - q.x) <= DELTA && Abs(y - q.y) <= DELTA && Abs(z - q.z) <= DELTA && Abs(w - q.w) <= DELTA);
	}

	bool operator!=(
		CQuaternion &q ){ return !(*this == q);
	}

	// Arithmetic operators (quaternion and scalar)
	CQuaternion operator+ ( float f )const{ return CQuaternion(x+f, y+f, z+f, w+f); }
	CQuaternion operator- ( float f )const{ return CQuaternion(x-f, y-f, z-f, w-f); }
	CQuaternion operator* ( float f )const{ return CQuaternion(x*f, y*f, z*f, w*f); }
	CQuaternion operator/ ( float f )const{ return CQuaternion(x/f, y/f, z/f, w/f); }
	void        operator+=( float f ){ x+=f; y+=f; z+=f; w+=f; }
	void        operator-=( float f ){ x-=f; y-=f; z-=f; w-=f; }
	void        operator*=( float f ){ x*=f; y*=f; z*=f; w*=f; }
	void        operator/=( float f ){ x/=f; y/=f; z/=f; w/=f; }

	// Arithmetic operators (quaternion and quaternion)
	CQuaternion operator+ ( const CQuaternion &q )const{ return CQuaternion(x+q.x, y+q.y, z+q.z, w+q.w); }
	CQuaternion operator- ( const CQuaternion &q )const{ return CQuaternion(x-q.x, y-q.y, z-q.z, w-q.w); }
	CQuaternion operator* ( const CQuaternion &q );

	void        operator+=( CQuaternion &q ){ x+=q.x; y+=q.y; z+=q.z; w+=q.w; }
	void        operator-=( CQuaternion &q ){ x-=q.x; y-=q.y; z-=q.z; w-=q.w; }
	void        operator*=( CQuaternion &q ){ *this = *this * q; }

	float       MagnitudeSquared()const{ return x*x + y*y + z*z + w*w; }
	float       Magnitude       ()const{ return sqrtf(MagnitudeSquared()); }
	void        Normalize       (){ *this /= Magnitude(); }
	CQuaternion Conjugate       ()const{ return CQuaternion(-x, -y, -z, w); }
	CQuaternion Inverse         ()const{ return Conjugate() / MagnitudeSquared(); }
	CQuaternion UnitInverse     ()const{ return Conjugate(); }
	CVector     RotateVector    ( const CVector &v ){
		return (*this * CQuaternion(v) * UnitInverse());
	}

	void SetAxisAngle( CVector &vAxis, float fAngle ){
		// 4 muls, 2 trig function calls
		fAngle *= 0.5f;
		*this = vAxis * sinf(fAngle);
		w = cosf(fAngle);
	}


	void GetAxisAngle( CVector &vAxis, float &fAngle ){
		// 4 muls, 1 div, 2 trig function calls
		fAngle  = acosf(w);
		vAxis   = *this / sinf(fAngle);
		fAngle *= 2.0f;
	}

	void Rotate( CQuaternion &q ){
		*this = q * *this;
	}

	void Rotate( CVector &vAxis, float fAngle ){
		CQuaternion q;
		q.SetAxisAngle( vAxis, fAngle );
		Rotate( q );
	}

	CVector GetViewAxis(){
		// 6 muls, 7 adds
		float x2 = x + x, y2 = y + y, z2 = z + z;
		float xx = x * x2, xz = x * z2;
		float yy = y * y2, yz = y * z2;
		float wx = w * x2, wy = w * y2;
		return -CVector(xz+wy, yz-wx, 1-(xx+yy));
	}

	CVector GetUpAxis(){
		// 6 muls, 7 adds
		float x2 = x + x, y2 = y + y, z2 = z + z;
		float xx = x * x2, xy = x * y2;
		float yz = y * z2, zz = z * z2;
		float wx = w * x2, wz = w * z2;
		return CVector( xy-wz, 1-(xx+zz), yz+wx );
	}

	CVector GetRightAxis(){
		// 6 muls, 7 adds
		float x2 = x + x, y2 = y + y, z2 = z + z;
		float xy = x * y2, xz = x * z2;
		float yy = y * y2, zz = z * z2;
		float wy = w * y2, wz = w * z2;
		return CVector( 1-(yy+zz), xy+wz, xz-wy );
	}
};


extern CQuaternion Slerp( CQuaternion &q1, CQuaternion &q2, float t );


};  //  namespace Maths


#endif  //  TEDDY_MATHS_C_QUATERNION_H

