
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CVector
	\ingroup g_maths
	\author  Sean O'Neil
	\date    2000

	This class implements a simple 3D vector with x, y, and z coordinates. Several
	functions and operators are defined to make working with vectors easier.
*/


#ifndef TEDDY_MATHS_C_VECTOR_H
#define TEDDY_MATHS_C_VECTOR_H


#include "Maths/NoiseMath.h"


namespace Maths {


class CVector {
public:
	float x, y, z;

public:
	// Constructors
	CVector(){}
	CVector( const float    f ){ x = y = z = f; }
	CVector( const float   *p ){ x = p[0]; y = p[1]; z = p[2]; }
	CVector( const CVector &v ){ x = v.x;  y = v.y;  z = v.z; }
	CVector( const float a, const float b, const float c ){ x = a; y = b; z = c; }

	// Casting, unary, equal, and comparison operators
	operator float *(){ return &x; }
	float   &operator[]( int n ){ return (&x)[n]; }
	CVector  operator- (){ return CVector(-x, -y, -z); }
	void     operator= ( const float    f ){ x = y = z = f; }
	void     operator= ( const float   *p ){ x = p[0]; y = p[1]; z = p[2]; }
	void     operator= ( const CVector &v ){ x = v.x; y = v.y; z = v.z; }
	bool     operator==( const CVector &v ){ return (Abs(x - v.x) <= DELTA && Abs(y - v.y) <= DELTA && Abs(z - v.z) <= DELTA); }
	bool     operator!=( const CVector &v ){ return !(*this == v); }

	// Arithmetic operators (vector with scalar)
	CVector operator+ ( const float f ){ return CVector(x+f, y+f, z+f); }
	CVector operator- ( const float f ){ return CVector(x-f, y-f, z-f); }
	CVector operator* ( const float f ){ return CVector(x*f, y*f, z*f); }
	CVector operator/ ( const float f ){ return CVector(x/f, y/f, z/f); }
	void    operator+=( const float f ){ x += f; y += f; z += f; }
	void    operator-=( const float f ){ x -= f; y -= f; z -= f; }
	void    operator*=( const float f ){ x *= f; y *= f; z *= f; }
	void    operator/=( const float f ){ x /= f; y /= f; z /= f; }

	// Arithmetic operators (vector with vector)
	CVector operator+ ( const CVector &v ){ return CVector(x+v.x, y+v.y, z+v.z); }
	CVector operator- ( const CVector &v ){ return CVector(x-v.x, y-v.y, z-v.z); }
	CVector operator* ( const CVector &v ){ return CVector(x*v.x, y*v.y, z*v.z); }
	CVector operator/ ( const CVector &v ){ return CVector(x/v.x, y/v.y, z/v.z); }
	void    operator+=( const CVector &v ){ x += v.x; y += v.y; z += v.z; }
	void    operator-=( const CVector &v ){ x -= v.x; y -= v.y; z -= v.z; }
	void    operator*=( const CVector &v ){ x *= v.x; y *= v.y; z *= v.z; }
	void    operator/=( const CVector &v ){ x /= v.x; y /= v.y; z /= v.z; }

	// Magnitude members
	float MagnitudeSquared(){ return x*x + y*y + z*z; }
	float Magnitude       (){ return sqrtf(MagnitudeSquared()); }
	void  Normalize       (){ *this /= Magnitude(); }
	float DistanceSquared ( const CVector &v ){ return (*this - v).MagnitudeSquared(); }
	float Distance        ( const CVector &v ){ return (*this - v).Magnitude(); }
};


inline float   DotProduct  ( const CVector &v1, const CVector &v2 ){ return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z; }
inline CVector CrossProduct( const CVector &v1, const CVector &v2 ){ return CVector(v1.y*v2.z - v1.z*v2.y, v1.z*v2.x - v1.x*v2.z, v1.x*v2.y - v1.y*v2.x); }
inline CVector Midpoint    ( CVector &v1, CVector &v2 ){ return CVector((v1 - v2) * 0.5f + v2); }
inline CVector Average     ( CVector &v1, CVector &v2 ){ return CVector((v1 + v2) * 0.5f); }


//!  Returns the normal vector of two vectors (the normalized cross product)
inline CVector Normal( const CVector &v1, const CVector &v2 ){
	CVector v = CrossProduct( v1, v2 );
	v.Normalize();
	return v;
}


//!  Returns the normal vector of a triangle or 3 points on a plane (assumes counterclockwise orientation)
inline CVector Normal( CVector &p1, CVector &p2, CVector &p3 ){
	return Normal( p2-p1, p3-p1 );
}


};  //  namespace Maths


#endif  //  TEDDY_MATHS_C_VECTOR_H


