
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Maths/Matrix.h"
#include "Maths/TranslationMatrix.h"
#include "Maths/RotationMatrix.h"
#include "Maths/Vector.h"
#include "Maths/Vector4.h"
#include "Maths/CrossVector4.h"
#include "Maths/ColumnToVector4.h"


namespace Maths {


//!  Default constructor; does not initialize
Matrix::Matrix(){
}


Vector Matrix::Transform( const Vector &v ){
	Vector t;

    t.r = v.r * A[0][0] + v.s * A[0][1] + v.t * A[0][2];
    t.s = v.r * A[1][0] + v.s * A[1][1] + v.t * A[1][2];
    t.t = v.r * A[2][0] + v.s * A[2][1] + v.t * A[2][2];

	return t;
}


//  Get matrix from quaternion
/*Matrix::Matrix( const Quaternion &q ){
	A[0][0] = 1-2*q.v.s*q.v.s - 2*q.v.t*q.v.t;
	A[0][1] =   2*q.v.r*q.v.s - 2*q.s  *q.v.t;
	A[0][2] =   2*q.v.r*q.v.t + 2*q.s  *q.v.s;
	A[1][0] =   2*q.v.r*q.v.s + 2*q.s  *q.v.t;
	A[1][1] = 1-2*q.v.r*q.v.r - 2*q.v.t*q.v.t;
	A[1][2] =   2*q.v.s*q.v.t - 2*q.s  *q.v.r;
	A[2][0] =   2*q.v.r*q.v.t - 2*q.s  *q.v.s;
	A[2][1] =   2*q.v.s*q.v.t + 2*q.s  *q.v.r;
	A[2][2] = 1-2*q.v.r*q.v.r - 2*q.v.s*q.v.s;

	A[0][3] = A[1][3] = A[2][3] = A[3][0] = A[3][1] = A[3][2] = 0.0;
	A[3][3] = 1.0;
} */



//!  Set column of matrix to contents of Vector4
void Matrix::VectorToColumn( const Vector4 &v, const int col ){
	A[0][col] = v.x;
	A[1][col] = v.y;
	A[2][col] = v.z;
	A[3][col] = v.w;
}


//!  Set column of matrix to contents of Vector
void Matrix::VectorToColumn( const Vector &v, const int col ){
	A[0][col] = v.r;
	A[1][col] = v.s;
	A[2][col] = v.t;
}


//!  Set row of matrix to contents of Vector4
void Matrix::VectorToRow( const Vector4 &v, const int row ){
	A[row][0] = v.x;
	A[row][1] = v.y;
	A[row][2] = v.z;
	A[row][3] = v.w;
}


//!  Set row of matrix to contents of Vector
void Matrix::VectorToRow( const Vector &v, const int row ){
	A[row][0] = v.r;
	A[row][1] = v.s;
	A[row][2] = v.t;
}


//!  Copy matrix from another matrix
Matrix Matrix::Copy( const Matrix &m ){
	const double *f;
	double *t;

	for( t = &A[0][0], f = &m.A[0][0]; f != (double *)m.A+16; t++, f++ ){
		*t = *f;
	}
	return( *this );
}


//!  Invert 4x4 matrix; should work with all invertible matrices
bool Matrix::Invert(){
	Vector4 x, x1;
	Vector4 y, y1;
	Vector4 z, z1;
	Vector4 w, w1;
	double  lx, ly, lz, lw;

    x = ColumnToVector4( *this, 0 );
	y = ColumnToVector4( *this, 1 );
	z = ColumnToVector4( *this, 2 );
    w = ColumnToVector4( *this, 3 );

	x1 = CrossVector4( y, z, w );
    y1 = CrossVector4( z, w, x );
	z1 = CrossVector4( w, x, y );
    w1 = CrossVector4( x, y, z );

	lx = x | x1;
    ly = y | y1;
	lz = z | z1;
    lw = w | w1;

	if( !lx || !ly || !lz || !lw ){
		return( false );
	}

	x1 *= 1.0/lx;
    y1 *= 1.0/ly;
	z1 *= 1.0/lz;
    w1 *= 1.0/lw;

	VectorToRow( x1, 0 );
    VectorToRow( y1, 1 );
	VectorToRow( z1, 2 );
    VectorToRow( w1, 3 );

	return( true );
}


//!  Multiply (concatenate) matrices
Matrix &Matrix::operator*=( const Matrix &m ){
	return *this = *this * m;
}


//!  Multiply (concatenate) matrices
Matrix Matrix::operator*( const Matrix &m ) const {
	Matrix res;
	res.A[0][0] = A[0][0] * m.A[0][0] + A[0][1] * m.A[1][0] + A[0][2] * m.A[2][0] + A[0][3] * m.A[3][0];
    res.A[0][0] = A[0][0] * m.A[0][0] + A[0][1] * m.A[1][0] + A[0][2] * m.A[2][0] + A[0][3] * m.A[3][0];
    res.A[0][1] = A[0][0] * m.A[0][1] + A[0][1] * m.A[1][1] + A[0][2] * m.A[2][1] + A[0][3] * m.A[3][1];
    res.A[0][2] = A[0][0] * m.A[0][2] + A[0][1] * m.A[1][2] + A[0][2] * m.A[2][2] + A[0][3] * m.A[3][2];
    res.A[0][3] = A[0][0] * m.A[0][3] + A[0][1] * m.A[1][3] + A[0][2] * m.A[2][3] + A[0][3] * m.A[3][3];
																							  
    res.A[1][0] = A[1][0] * m.A[0][0] + A[1][1] * m.A[1][0] + A[1][2] * m.A[2][0] + A[1][3] * m.A[3][0];
	res.A[1][1] = A[1][0] * m.A[0][1] + A[1][1] * m.A[1][1] + A[1][2] * m.A[2][1] + A[1][3] * m.A[3][1];
    res.A[1][2] = A[1][0] * m.A[0][2] + A[1][1] * m.A[1][2] + A[1][2] * m.A[2][2] + A[1][3] * m.A[3][2];
	res.A[1][3] = A[1][0] * m.A[0][3] + A[1][1] * m.A[1][3] + A[1][2] * m.A[2][3] + A[1][3] * m.A[3][3];

    res.A[2][0] = A[2][0] * m.A[0][0] + A[2][1] * m.A[1][0] + A[2][2] * m.A[2][0] + A[2][3] * m.A[3][0];
	res.A[2][1] = A[2][0] * m.A[0][1] + A[2][1] * m.A[1][1] + A[2][2] * m.A[2][1] + A[2][3] * m.A[3][1];
	res.A[2][2] = A[2][0] * m.A[0][2] + A[2][1] * m.A[1][2] + A[2][2] * m.A[2][2] + A[2][3] * m.A[3][2];
	res.A[2][3] = A[2][0] * m.A[0][3] + A[2][1] * m.A[1][3] + A[2][2] * m.A[2][3] + A[2][3] * m.A[3][3];

	res.A[3][0] = A[3][0] * m.A[0][0] + A[3][1] * m.A[1][0] + A[3][2] * m.A[2][0] + A[3][3] * m.A[3][0];
	res.A[3][1] = A[3][0] * m.A[0][1] + A[3][1] * m.A[1][1] + A[3][2] * m.A[2][1] + A[3][3] * m.A[3][1];
	res.A[3][2] = A[3][0] * m.A[0][2] + A[3][1] * m.A[1][2] + A[3][2] * m.A[2][2] + A[3][3] * m.A[3][2];
	res.A[3][3] = A[3][0] * m.A[0][3] + A[3][1] * m.A[1][3] + A[3][2] * m.A[2][3] + A[3][3] * m.A[3][3];
	return( res );
}


//!  Concatenate a translation matrix
Matrix &Matrix::ConcatTranslate( const double x, const double y, const double z ){
	return *this *= TranslationMatrix( x, y, z );
}


//!  Concatenate a rotation matrix
Matrix &Matrix::ConcatRotate( const double r, const Vector &orig, const Vector &axis ){
	return *this *= RotationMatrix( r, orig, axis );
}


//!  Concatenate a scaling matrix
Matrix &Matrix::ConcatScale( const double x, const double y, const double z ){
    A[0][0] *= x;
    A[1][0] *= x;
    A[2][0] *= x;
    A[3][0] *= x;
    
    A[0][1] *= y;
    A[1][1] *= y;
    A[2][1] *= y;
    A[3][1] *= y;

    A[0][2] *= z;
    A[1][2] *= z;
    A[2][2] *= z;
    A[3][2] *= z;
	return *this;
}


};  //  namespace Maths

