
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\ingroup g_maths
	\author	 Sean O'Neil
	\date    2000
*/


#ifndef TEDDY_MATHS_NOISE_MATH_H
#define TEDDY_MATHS_NOISE_MATH_H
		
		
#include <math.h>


namespace Maths {
			

#ifndef WIN32	
#define powf(a,b) pow(a,b)
#define logf(a) log(a)
#define expf(a) exp(a)
#define acosf(a) acos(a)
#endif


//  Defines
#ifndef PI
#define PI					3.14159265358979323846      //!<  PI
#endif
#define HALF_PI				1.570796326795f				//!<  PI / 2
#define TWO_PI				6.28318530718f				//!<  PI * 2
#define LOGHALF				-0.6931471805599f			//!<  log(0.5)
#define LOGHALFI			-1.442695040889f			//!<  Inverse of log(0.5)
#define DELTA				1e-6f						//!<  Small number for comparing floating point numbers
#define MAX_DIMENSIONS		4							//!<  Maximum number of dimensions in a noise object
#define MAX_OCTAVES			128							//!<  Maximum # of octaves in an fBm object


//  Macros
#define SQUARE(a)			((a) * (a))
#define FLOOR(a)			((int)(a) - ((a) < 0 && (a) != (int)(a)))
#define CEILING(a)			((int)(a) + ((a) > 0 && (a) != (int)(a)))
#define MIN(a, b)			((a) < (b) ? (a) : (b))
#define MAX(a, b)			((a) > (b) ? (a) : (b))
#define ABS(a)				((a) < 0 ? -(a) : (a))
#define CLAMP(a, b, x)		((x) < (a) ? (a) : ((x) > (b) ? (b) : (x)))
#define LERP(a, b, x)		((a) + (x) * ((b) - (a)))
#define CUBIC(a)			((a) * (a) * (3 - 2*(a)))
#define STEP(a, x)			((x) >= (a))
#define BOXSTEP(a, b, x)	Clamp(0, 1, ((x)-(a))/((b)-(a)))
#define PULSE(a, b, x)		(((x) >= (a)) - ((x) >= (b)))
#define GAMMA(a, g)			powf(a, 1/g)
#define BIAS(a, b)			powf(a, logf(b) * LOGHALFI)
#define EXPOSE(l, k)		(1 - expf(l * k))
#define DEGTORAD(x)			((x) * 0.01745329251994f)
#define RADTODEG(x)			((x) * 57.29577951308f)
#define SWAP(a, b, t)		t = a; a = b; b = t
	

// Inline functions (use instead of macros to avoid performing slow operations twice)
inline float Square ( float a                   ){ return a * a; }
inline int   Floor  ( float a                   ){ return ((int)a - (a < 0 && a != (int)a)); }
inline int   Ceiling( float a                   ){ return ((int)a + (a > 0 && a != (int)a)); }
inline float Min    ( float a, float b          ){ return (a < b ? a : b); }
inline float Max    ( float a, float b          ){ return (a > b ? a : b); }
inline float Abs    ( float a                   ){ return (a < 0 ? -a : a); }
inline float Clamp  ( float a, float b, float x ){ return (x < a ? a : (x > b ? b : x)); }
inline float Lerp   ( float a, float b, float x ){ return a + x * (b - a); }
inline float Cubic  ( float a                   ){ return a * a * (3 - 2*a); }
inline float Step   ( float a, float x          ){ return (float)(x >= a); }
inline float Boxstep( float a, float b, float x ){ return Clamp(0, 1, (x-a)/(b-a)); }
inline float Pulse  ( float a, float b, float x ){ return (float)((x >= a) - (x >= b)); }
inline float Gamma  ( float a, float g          ){ return powf(a, 1/g); }
inline float Bias   ( float a, float b          ){ return powf(a, logf(b) * LOGHALFI); }
inline float Expose ( float l, float k          ){ return (1 - expf(-l * k)); }


inline float Gain( float a, float b ){
	if( a <= DELTA ){
		return 0;
	}
	if( a >= 1-DELTA ){
		return 1;
	}

	register float p = (logf(1 - b) * LOGHALFI);
	if( a < 0.5 ){
		return powf(2 * a, p) * 0.5f;
	}else{
		return 1 - powf(2 * (1 - a), p) * 0.5f;
	}
}


inline float Smoothstep( float a, float b, float x ){
	if( x <= a ){
		return 0;
	}
	if( x >= b ){
		return 1;
	}
	return Cubic( (x - a) / (b - a) );
}


inline float Mod( float a, float b ){
	a -= ((int)(a / b)) * b;
	if( a < 0 ){
		a += b;
	}
	return a;
}


inline void Normalize( float *f, int n ){
	float fMagnitude = 0;
	int   i;
	for( i=0; i<n; i++ ){
		fMagnitude += f[i]*f[i];
	}
	fMagnitude = 1 / sqrtf( fMagnitude );
	for( i=0; i<n; i++ ){
		f[i] *= fMagnitude;
	}
}


};  //  namespace Maths


#endif  //  TEDDY_MATHS_NOISE_MATH_H


