
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Quaternion.h"
#include "Matrix.h"
#include "trigmath.h"


namespace Maths {


//!  Default constructor
Quaternion::Quaternion(){
	W = 1.0;
	X = 0.0;
	Y = 0.0;
	Z = 0.0;
}


//!  Component constructor
Quaternion::Quaternion( const double w, const double x, const double y, const double z ){
	W = w;
	X = x;
	Y = y;
	Z = z;
}


//!  Multiplication
Quaternion operator*( const Quaternion &a, const Quaternion &b ){
	double w;
	double x;
	double y;
	double z;

	w = a.W*b.W - (a.X*b.X + a.Y*b.Y + a.Z*b.Z);
  
	x = a.W*b.X + b.W*a.X + a.Y*b.Z - a.Z*b.Y;
	y = a.W*b.Y + b.W*a.Y + a.Z*b.X - a.X*b.Z;
	z = a.W*b.Z + b.W*a.Z + a.X*b.Y - a.Y*b.X;

	return Quaternion( w, x, y, z );
}


//!  Multiplication
const Quaternion& Quaternion::operator*=( const Quaternion &q ){
	double w = W*q.W - (X*q.X + Y*q.Y + Z*q.Z);

	double x = W*q.X + q.W*X + Y*q.Z - Z*q.Y;
	double y = W*q.Y + q.W*Y + Z*q.X - X*q.Z;
	double z = W*q.Z + q.W*Z + X*q.Y - Y*q.X;

	W = w;
	X = x;
	Y = y;
	Z = z;
	return *this;
}


//!  Conjugate
const Quaternion& Quaternion::operator~(){
	X = -X;
	Y = -Y;
	Z = -Z;
	return *this;
}


//!  Invert
const Quaternion& Quaternion::operator-(){
	double norme = sqrt( W*W + X*X + Y*Y + Z*Z );
	if( norme == 0.0 ){
		norme = 1.0;
	}

	double recip = 1.0 / norme;

	W =  W * recip;
	X = -X * recip;
	Y = -Y * recip;
	Z = -Z * recip;

	return *this;
}


//!  Normalize
const Quaternion& Quaternion::Normalize(){
	double norme = sqrt( W*W + X*X + Y*Y + Z*Z );
	if( norme > DBL_MIN ){
		W = 1.0; 
		X = Y = Z = 0.0;
	}else{
		double recip = 1.0/norme;

		W *= recip;
		X *= recip;
		Y *= recip;
		Z *= recip;
	}
	return *this;
}


//!  Rotation quaternation
const Quaternion& Quaternion::FromAxis( const double Angle, double x, double y, double z ){
	double omega;
	double s;
	double c;
    
	s = sqrt( x*x + y*y + z*z );

	if( fabs(s) > DBL_MIN ){
		c = 1.0/s;
	  
		x *= c;
		y *= c;
		z *= c;

		omega = -0.5f * Angle;
		s = (double)sin( omega );

		X = s*x;
		Y = s*y;
		Z = s*z;
		W = (double)cos( omega );
	}else{
		X = Y = 0.0f;
		Z = 0.0f;
		W = 1.0f;
	}
	Normalize();
	return *this;
}


//!  Quaternation to matrix
void Quaternion::ToMatrix( Matrix &m ) const {
	m.A[0][0] = 1.0 - 2*Y*Y - 2*Z*Z;
	m.A[1][0] =       2*X*Y + 2*W*Z;
	m.A[2][0] =       2*X*Z - 2*W*Y;
	m.A[3][0] = 0.0;

	m.A[0][1] =       2*X*Y - 2*W*Z;
	m.A[1][1] = 1.0 - 2*X*X - 2*Z*Z;
	m.A[2][1] =       2*Y*Z + 2*W*X;
	m.A[3][1] = 0.0;

	m.A[0][2] =       2*X*Z + 2*W*Y;
	m.A[1][2] =       2*Y*Z - 2*W*X;
	m.A[2][2] = 1.0 - 2*X*X - 2*Y*Y;
	m.A[3][2] = 0.0;

	m.A[0][3] = 0.0;
	m.A[1][3] = 0.0;
	m.A[2][3] = 0.0;
	m.A[3][3] = 1.0;
}


//!  Interpolate two quaternations
void Quaternion::Slerp( const Quaternion &a, const Quaternion &b, const double t ){
	double omega;
	double cosom;
	double sinom;
	double sclp;
	double sclq;

	cosom = a.X*b.X + a.Y*b.Y + a.Z*b.Z + a.W*b.W;

	if( (1.0f+cosom) > DBL_EPSILON ){
		if( (1.0f-cosom) > DBL_EPSILON ){
			omega = acos( cosom );
			sinom = sin ( omega );
			sclp  = sin ( (1.0f-t)*omega ) / sinom;
			sclq  = sin ( t*omega ) / sinom;
		}else{
			sclp = 1.0f - t;
			sclq = t;
		}

		X = sclp*a.X + sclq*b.X;
		Y = sclp*a.Y + sclq*b.Y;
		Z = sclp*a.Z + sclq*b.Z;
		W = sclp*a.W + sclq*b.W;
	}else{
		X = -a.Y;
		Y =  a.X;
		Z = -a.W;
		W =  a.Z;

		sclp = sin( (1.0f-t) * M_PI * 0.5 );
		sclq = sin( t * M_PI * 0.5 );

		X = sclp*a.X + sclq*b.X;
		Y = sclp*a.Y + sclq*b.Y;
		Z = sclp*a.Z + sclq*b.Z;
	}
}


const Quaternion& Quaternion::exp(){                               
	double Mul;
	double Length = sqrt( X*X + Y*Y + Z*Z );

	if( Length > 1.0e-4 ){
		Mul = sin( Length )/Length;
	}else{
		Mul = 1.0;
	}

	W = cos( Length );

	X *= Mul;
	Y *= Mul;
	Z *= Mul; 

	return *this;
}


const Quaternion& Quaternion::log(){
	double Length;

	Length = sqrt( X*X + Y*Y + Z*Z );
	Length = atan( Length/W );

	W = 0.0;

	X *= Length;
	Y *= Length;
	Z *= Length;

	return *this;
}


};  //  namespace Maths

