
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\class   Vector
	\ingroup g_maths
	\author  Timo Suoranta
	\brief   Vector class and some methods, see subclasses for more
	\date    1999, 2000
*/


#ifndef TEDDY_MATHS_VECTOR_H
#define TEDDY_MATHS_VECTOR_H


#include "Maths/Vector4.h"
#include <math.h>
#include <float.h>
#include <limits.h>


namespace Maths {


class Vector {
public:
	double r;
	double s;
	double t;

	static Vector NullVector;
	static Vector IdentityX;
	static Vector IdentityY;
	static Vector IdentityZ;

	inline  Vector();
	inline  Vector( const Vector4& v4 );
	inline  Vector( const double r, const double s, const double t );

	virtual void debug() const;

	inline void    Neg        ();
	inline double  Len        () const;
	inline double  Norm       ();
	inline void    GetPerpVect( Vector &hor, Vector &vert ) const;
	inline Vector  operator+  ( const Vector &v ) const;
	inline Vector &operator+= ( const Vector &v );
	inline Vector  operator-  ( const Vector &v ) const;
	inline Vector &operator-= ( const Vector &v );
	inline Vector  operator*  ( const double k ) const;
	inline Vector &operator*= ( const double k );
/*	friend inline Vector  operator*  ( double k, const Vector &v );*/
	inline Vector  operator*  ( const Vector &v ) const;
	inline Vector &operator*= ( const Vector &v );
	inline double  operator|  ( const Vector &v ) const;
	inline Vector  operator/  ( const Vector &v ) const;
	inline Vector &operator/= ( const Vector &v );
	inline Vector  operator/  ( const double k ) const;
	inline Vector &operator/= ( const double k );
	
	inline bool    operator==( const Vector &v ) const;
	inline bool    operator!=( const Vector &v ) const;

};


//!  Default constructor; does not initialize
inline Vector::Vector(){
}


//!  Vector constructor specifying components
inline Vector::Vector( const double r, const double s, const double t ){
	this->r = r;
	this->s = s;
	this->t = t;
}


//!  Vector constructor from Vector4
inline Vector::Vector( const Vector4& v4 ){
	this->r = v4.x;
	this->s = v4.y;
	this->t = v4.z;
}


//!  Negate Vector components
inline void Vector::Neg(){
	r = -r;
	s = -s;
	t = -t;
}


//!  Return length of Vector
inline double Vector::Len() const {
	return( sqrt(r*r+s*s+t*t) );
}


//!  Normalize Vector
inline double Vector::Norm(){
	double len;

	len = sqrt( r*r+s*s+t*t );
	if( len>DBL_MIN ){  //  len==0.0
		r /= len;
		s /= len;
		t /= len;
	}else{
		r = 1.0;
	}

	return( len );
}


//!  Generate a vectors perp. to given vectors
inline void Vector::GetPerpVect( Vector &hor, Vector &vert ) const {
	if( r || t ){
		hor.r = t;
		hor.s = 0.0;
		hor.t = - r;
	    hor.Norm();
	}else{
		hor = Vector::IdentityX;
    }
	vert = (*this) * hor;
}


//!  Add two Vector's
inline Vector Vector::operator+( const Vector &v ) const {
	return( Vector(r+v.r, s+v.s, t+v.t) );
}


//!  Add two Vector's
inline Vector &Vector::operator+=( const Vector &v ){
	r += v.r;
	s += v.s;
	t += v.t;
	return( *this );
}


inline Vector Vector::operator-( const Vector &v ) const {
	return( Vector(r-v.r, s-v.s, t-v.t) );
}


inline Vector &Vector::operator-=( const Vector &v ){
	r -= v.r;
	s -= v.s;
	t -= v.t;
	return( *this );
}


inline Vector Vector::operator*( const double k ) const {
	return( Vector(r*k, s*k, t*k) );
}


inline Vector &Vector::operator*=( const double k ){
	r *= k;
	s *= k;
	t *= k;
	return( *this );
}


inline Vector operator*( double k, const Vector &v ){
	return( Vector(v.r*k, v.s*k, v.t*k ) );
}


//!  CrossProduct
inline Vector Vector::operator*( const Vector &v ) const {
	return Vector((s*v.t)-(t*v.s),(t*v.r)-(r*v.t),(r*v.s)-(s*v.r));
}
	//return( Vector( r*v.r, s*v,s, t*v.t );


inline Vector &Vector::operator*=( const Vector &v ){
	return( *this = (*this) *v );
	/*
	r *= v.r;
	s *= v.s;
	t *= v.t;
	*/
}
					   

/*inline Vector operator*( const Vector &v, double k ){
	return( k * v );
} */


inline double Vector::operator|( const Vector &v ) const {
	return( r*v.r + s*v.s + t*v.t );
}


inline Vector Vector::operator/( const Vector &v ) const {
	return( Vector(r/v.r, s/v.s, t/v.t) );
}


inline Vector &Vector::operator/=( const Vector &v ){
	r /= v.r;
	s /= v.s;
	t /= v.t;
	return( *this );
}


inline Vector Vector::operator/( const double k ) const {
	return( Vector(r/k, s/k, t/k) );
}


inline Vector &Vector::operator/=( const double k ){
	r /= k;
	s /= k;
	t /= k;
	return( *this );
}


inline bool Vector::operator==( const Vector &v ) const {
	if( r == v.r &&
		s == v.s &&
		t == v.t    ){
		return true;
	}else{
		return false;
	}
}

inline bool Vector::operator!=( const Vector &v ) const {
	
	if( r != v.r ||
		s != v.s ||
		t != v.t    ){
		return true;
	}else{
		return false;
	}
}


};  //  namespace Maths


#endif  //  TEDDY_MATHS_VECTOR_H


