
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\class   Vector4
	\ingroup g_maths
	\author  Timo Suoranta
	\brief   Four-vector base class and some methods, see subclasses for more
	\date    1999
*/


#ifndef TEDDY_MATHS_VECTOR4_H
#define TEDDY_MATHS_VECTOR4_H


#include <math.h>
#include <float.h>
#include <limits.h>


namespace Maths {


class Matrix;


class Vector4 {
public:
	double x;
	double y;
	double z;
	double w;

	inline Vector4();
	inline Vector4( const double x, const double y, const double z, const double w );

	inline void     Neg        ();
	inline double   Len        ();
	inline double   Norm       ();

	inline Vector4  operator+  ( const Vector4 &v );
	inline Vector4 &operator+= ( const Vector4 &v );
	inline Vector4  operator-  ( const Vector4 &v );
	inline Vector4 &operator-= ( const Vector4 &v );

	inline Vector4  operator*  ( const double k );
	inline Vector4 &operator*= ( const double k );
	friend
	inline Vector4  operator*  ( double k, const Vector4 &v );
	inline Vector4  operator*  ( const Vector4 &v );
	inline Vector4 &operator*= ( const Vector4 &v );

	inline double   operator|  ( const Vector4 &v );
	inline Vector4  operator/  ( const Vector4 &v );
	inline Vector4 &operator/= ( const Vector4 &v );
	inline Vector4  operator/  ( const double k );
	inline Vector4 &operator/= ( const double k );
	inline bool     AEq        ( const Vector4 &v, const double p );
	inline void     OrtUnit    ( const Vector4 &v, const Vector4 &t );

	inline bool operator==( const Vector4 &v );
	inline bool operator!=( const Vector4 &v );
};


//!  Default Vector4 constructor; does not initialize
inline Vector4::Vector4(){
}


//!  Vector4 constructor specifying components
inline Vector4::Vector4( const double x, const double y, const double z, const double w ){
	this->x = x;
	this->y = y;
	this->z = z;
	this->w = w;
}


//!  Negate Vector4 components
inline void Vector4::Neg(){
	x = -x;
	y = -y;
	z = -z;
	w = -w;
}


//!  Return length of Vector4
inline double Vector4::Len(){
	return( sqrt(x*x+y*y+z*z+w*w) );
}


//!  Normalize Vector4
inline double Vector4::Norm(){
	double len;

	len = sqrt( x*x+y*y+z*z+w*w );
	if( len>DBL_MIN ){  // len==0.0
		x /= len;
		y /= len;
		z /= len;
		w /= len;
	}else{
		x = 1.0;
	}

	return( len );
}


//!  Add two Vector4's
inline Vector4 Vector4::operator+( const Vector4 &v ){
	return( Vector4(x+v.x, y+v.y, z+v.z, w+v.w) );
}


//!  Add two Vector4's
inline Vector4 &Vector4::operator+=( const Vector4 &v ){
	x += v.x;
	y += v.y;
	z += v.z;
	w += v.w;
	return( *this );
}


//!  Substract two Vector4's
inline Vector4 Vector4::operator-( const Vector4 &v ){
	return( Vector4(x-v.x, y-v.y, z-v.z, w-v.w) );
}


//!  Substract two Vector4's
inline Vector4 &Vector4::operator-=( const Vector4 &v ){
	x -= v.x;
	y -= v.y;
	z -= v.z;
	w -= v.w;
	return( *this );
}


//!  Multiply Vector4 with a scalar
inline Vector4 Vector4::operator*( const double k ){
	return( Vector4(x*k, y*k, z*k, w*k) );
}


//!  Multiply Vector4 with a scalar
inline Vector4 &Vector4::operator*=( const double k ){
	x *= k;
	y *= k;
	z *= k;
	w *= k;
	return( *this );
}


//!  Multiply Vector4 by component with components of another Vector4
inline Vector4 Vector4::operator*( const Vector4 &v ){
	return( Vector4(x*v.x, y*v.y, z*v.z, w*v.w) );
}


//!  Multiply Vector4 by component with components of another Vector4
inline Vector4 &Vector4::operator*=( const Vector4 &v ){
	x *= v.x;
	y *= v.y;
	z *= v.z;
	w *= v.w;
	return( *this );
}


//!  Multiply Vector4 with a scalar
inline Vector4 operator*( double k, const Vector4 &v ){
	return( Vector4(v.x*k, v.y*k, v.z*k, v.w*k) );
}


//!  Vector4 dot-product 
inline double Vector4::operator|( const Vector4 &v ){
	return( x*v.x + y*v.y + z*v.z + w*v.w );
}


//!  Divide Vector4 by components with another Vector4
inline Vector4 Vector4::operator/( const Vector4 &v ){
	return( Vector4(x/v.x, y/v.y, z/v.z, w/v.w) );
}


//!  Divide Vector4 by components with another Vector4
inline Vector4 &Vector4::operator/=( const Vector4 &v ){
	x /= v.x;
	y /= v.y;
	z /= v.z;
	w /= v.w;
	return( *this );
}


//!  Divide Vector4 with a scalar
inline Vector4 Vector4::operator/( const double k ){
	return( Vector4(x/k, y/k, z/k, w/k) );
}


//!  Divide Vector4 with a scalar
inline Vector4 &Vector4::operator/=( const double k ){
	x /= k;
	y /= k;
	z /= k;
	w /= k;
	return( *this );
}


//!  Almost Equal operation with component treshold p
inline bool Vector4::AEq( const Vector4 &v, const double p ){
	return(
		( fabs(x-v.x) < p) && 
		( fabs(y-v.y) < p) && 
		( fabs(z-v.z) < p) && 
        ( fabs(w-v.w) < p)
	);
}


//!  Set Vector4 v and t to orthogonal (to this Vector4) unit vectors 
inline void Vector4::OrtUnit( const Vector4 &v, const Vector4 &t ){
	double a;

    a = v.x*t.x + v.y*t.y + v.z*t.z + v.w*t.w;
	x = v.x - a*t.x;
    y = v.y - a*t.y;
	z = v.z - a*t.z;
	w = v.w - a*t.w;
}


//!  Precise Vector4 component equality comparison
inline bool Vector4::operator==( const Vector4 &v ){
	if( x == v.x &&
		y == v.y &&
		z == v.z &&
		w == v.w    ){
		return true;
	}else{
		return false;
	}
}


//!  Precise Vector4 component inequality comparison
inline bool Vector4::operator!=( const Vector4 &v ){
	if( x != v.x ||
		y != v.y ||
		z != v.z ||
		w != v.w    ){
		return true;
	}else{
		return false;
	}
}


};  //  namespace Maths


#endif  //  TEDDY_MATHS_VECTOR4_H


