
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CDiamond
	\ingroup g_models
	\author  Sean O'Neil
	\brief   ROAM diamond
	\date    2000

	This class implements the ROAM diamond. A diamond represents a set of 4
	triangles that can be merged. ROAM's constraints say that you can't merge
	triangles that have themselves been split until they are merged. So unless
	you plan to build the entire tree every frame, you need to keep a list of
	diamonds that can currently be merged. Every split causes existing diamonds
	to be destroyed and new ones to be created and so does every merge, so
	keeping track of them properly is important. Most of the code for that is
	in CSphere.

*/


#ifndef TEDDY_MODELS_C_DIAMOND_H
#define TEDDY_MODELS_C_DIAMOND_H


#include "SysSupport/CMemoryBlock.h"
#include "DataStructures/CLinkedList.h"
#include "DataStructures/CListNode.h"
#include "Maths/CFractal.h"
#include "Maths/CVector.h"
#include "Models/CVertex.h"
#include "Models/CTriangle.h"
#include <string.h>  //  memset
using namespace SysSupport;
using namespace DataStructures;
using namespace Maths;


namespace Models {


#define CDiamondList CLinkedList<CDiamond>


class CDiamond : public CListNode<CDiamond> {
public:
	CTriangle *m_pParent[2];
	CTriangle *m_pChild[2];
	float      m_fMergePriority;

public:
	CDiamond(){
		memset( (char *)this, 0, sizeof(CDiamond) );
	}

	CDiamond( CTriangle *p1, CTriangle *p2, CTriangle *p3, CTriangle *p4 ){
		m_pParent[0] = p1;
		m_pParent[1] = p2;
		m_pChild [0] = p3;
		m_pChild[1] = p4;
		UpdatePriority();
	}

	void UpdatePriority(){
		m_fMergePriority = MAX( m_pParent[0]->m_fSplitPriority, m_pParent[1]->m_fSplitPriority)*2.0f;
	}

	float GetPriority(){
		return m_fMergePriority;
	}

	bool HasTriangle( CTriangle *p ){
		return (m_pParent[0] == p || m_pParent[1] == p || m_pChild[0] == p || m_pChild[1] == p);
	}

	void *operator new( unsigned int n ){
		return m_mbDiamond.GetElement();
	}

	void operator delete( void *p ){
		m_mbDiamond.ReleaseElement( p );
	}

private:
	static CMemoryBlock m_mbDiamond;
};


};  //  namespace Models


#endif  //  TEDDY_MODELS_C_DIAMOND_H

