
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Models/CSphere.h"
#include "Maths/CFractal.h"
#include <stdlib.h>
using namespace Maths;


namespace Models {


#define COLOR_DELTA    0.001f
#define TEMP_DELTA     0.05f
#define SPLIT_PRIORITY 0.00050
#define MERGE_PRIORITY 0.00050


int triangles = -1;
int diamonds  = -1;
int vertices  = -1;


void CSphere::RebuildEdgeList(){
	for( CTriangle *p1 = m_llTriangle.GetHead(); p1->IsInList(); p1 = p1->GetNext() ){
		for( CTriangle *p2 = p1->GetNext(); p2->IsInList(); p2 = p2->GetNext() ){
			bool bMatch = true;

			if( p2->HasVertex(p1->m_pVertex[0]) && p2->HasVertex(p1->m_pVertex[1]) ){
				p1->m_pEdge[0] = p2;
			}else if(p2->HasVertex(p1->m_pVertex[1]) && p2->HasVertex(p1->m_pVertex[2]) ){
				p1->m_pEdge[1] = p2;
			}else if(p2->HasVertex(p1->m_pVertex[2]) && p2->HasVertex(p1->m_pVertex[0]) ){
				p1->m_pEdge[2] = p2;
			}else{
				bMatch = false;
			}

			if( bMatch ){
				if( !p1->HasVertex(p2->m_pVertex[2]) ){
					p2->m_pEdge[0] = p1;
				}else if(!p1->HasVertex(p2->m_pVertex[0]) ){
					p2->m_pEdge[1] = p1;
				}else{
					p2->m_pEdge[2] = p1;
				}
			}

		}
	}
}

void CSphere::BuildCube(){
	#define CUBE_VERTICES			8
	#define CUBE_TRIANGLES			12
	float fCube[CUBE_VERTICES *3] = {-0.57735f, 0.57735f, 0.57735f, 0.57735f, 0.57735f, 0.57735f, 0.57735f, -0.57735f, 0.57735f, -0.57735f, -0.57735f, 0.57735f, -0.57735f, 0.57735f, -0.57735f, 0.57735f, 0.57735f, -0.57735f, 0.57735f, -0.57735f, -0.57735f, -0.57735f, -0.57735f, -0.57735f};
	int   nCube[CUBE_TRIANGLES*3] = { 1, 0, 3, 3, 2, 1, 5, 6, 7, 7, 4, 5, 0, 1, 5, 5, 4, 0, 3, 7, 6, 6, 2, 3, 0, 4, 7, 7, 3, 0, 1, 2, 6, 6, 5, 1};
	int   i;
	int   j;

	m_llVertex  .RemoveAll();
	m_llTriangle.RemoveAll();
	m_llDiamond .RemoveAll();

	CVertex **pVertex = new CVertex *[CUBE_VERTICES];
	for( i=0, j=0; i<CUBE_VERTICES; i++, j+=3 ){
		pVertex[i] = new CVertex( CVector(fCube[j], fCube[j+1], fCube[j+2]) );
		m_llVertex.AddTail( pVertex[i] );
	}
	for( i=0, j=0; i<CUBE_TRIANGLES; i++, j+=3){
		m_llTriangle.AddTail( new CTriangle(pVertex[nCube[j]], pVertex[nCube[j+1]], pVertex[nCube[j+2]]) );
	}
	delete pVertex;
	RebuildEdgeList();
	m_nTriangles = m_llTriangle.GetCount();
}


void CSphere::UpdateTriangleOffset( CTriangle *pTriangle ){
	CVector v = Midpoint(pTriangle->m_pVertex[0]->m_vPosition, pTriangle->m_pVertex[2]->m_vPosition);
	pTriangle->m_vMidpoint = v;
	float fHeight = (float)v.Magnitude();
	v /= fHeight;
	//pTriangle->m_fOffset = m_frTest.Noise(v) * m_fVariance + m_fOffset + m_fRadius - (float)fHeight;
	pTriangle->m_fOffset = m_frTest.fBmTest(v, 8.0f) * m_fVariance + m_fOffset + m_fRadius - (float)fHeight;

	pTriangle->m_bColor = false;
	pTriangle->m_bTemp  = false;
	if( fHeight < m_fRadius+COLOR_DELTA && fHeight + pTriangle->m_fOffset > m_fRadius-COLOR_DELTA ){
		pTriangle->m_bColor = true;
	}
	if( fHeight > m_fRadius-COLOR_DELTA && fHeight + pTriangle->m_fOffset < m_fRadius+COLOR_DELTA ){
		pTriangle->m_bColor = true;
	}
	if( fHeight < m_fRadius-TEMP_DELTA && fHeight + pTriangle->m_fOffset < m_fRadius-TEMP_DELTA ){
		pTriangle->m_bTemp = true;
	}
}


void CSphere::Split( CTriangle *pTriangle, bool bDynamic ){
	// If the opposite triangle is not ready to split with this one, then it must be split first.
	// After that, it should always be ready to split with this one.
	static CTriangle *pOpposite;

	if( pTriangle->m_pEdge[2]->m_pEdge[2] != pTriangle ){
		Split( pTriangle->m_pEdge[2], bDynamic );
	}
	pOpposite = pTriangle->m_pEdge[2];
	ASSERT( pOpposite->m_pEdge[2] == pTriangle );

	// Now split edge 2-0 by getting its midpoint and creating a new vertex
	static CVertex *pNewVertex;
	CVector v = Midpoint( pTriangle->m_pVertex[0]->m_vPosition, pTriangle->m_pVertex[2]->m_vPosition );
	v.Normalize();
	pNewVertex = new CVertex( v );
	pNewVertex->InsertAfter( pTriangle->m_pVertex[0] );
	
	if( bDynamic ){
		float fLattitude = Clamp( 0.01f, 0.99f, Abs(pNewVertex->m_vPosition.y) + 0.5f * m_frTest.fBm(pNewVertex->m_vPosition*4, 8.0f) );
		float fAltitude  = (float)pTriangle->m_vMidpoint.Magnitude() + pTriangle->m_fOffset;
		pNewVertex->m_vPosition *= fAltitude;
		fAltitude -= m_fRadius - m_fMaxVariance;
		fAltitude  = Clamp( 0.01f, 0.99f, (fAltitude * 0.5f) / m_fMaxVariance );
		pNewVertex->m_fTexCoord[0] = fAltitude;
		pNewVertex->m_fTexCoord[1] = fLattitude;
	}else{
		pNewVertex->m_vPosition *= m_fRadius;
	}

	// Now create two new triangles for the split
	static CTriangle *pNewTriangle[2];
	pNewTriangle[0] = new CTriangle;
	pNewTriangle[1] = new CTriangle;

	// Split the first one from pTriangle
	pNewTriangle[0]->InsertAfter( pTriangle );
	pNewTriangle[0]->m_fDistance = pTriangle->m_fDistance;
	pNewTriangle[0]->m_fFacing   = pTriangle->m_fFacing;
	pNewTriangle[0]->m_fAngle    = pTriangle->m_fAngle;
	pNewTriangle[0]->SetVertices( pTriangle->m_pVertex[2], pNewVertex, pTriangle->m_pVertex[1] );
	pTriangle->SetVertices( pTriangle->m_pVertex[1], pNewVertex, pTriangle->m_pVertex[0] );
	pNewTriangle[0]->SetEdges( pOpposite, pTriangle, pTriangle->m_pEdge[1] );
	pTriangle->m_pEdge[1]->ReplaceEdge( pTriangle, pNewTriangle[0] );
	pTriangle->SetEdges( pNewTriangle[0], pNewTriangle[1], pTriangle->m_pEdge[0] );
	if( bDynamic ){
		if( pNewTriangle[0]->m_pEdge[2]->m_pEdge[2] == pNewTriangle[0] ){
			pNewTriangle[0]->CopyPriorityInfo( pNewTriangle[0]->m_pEdge[2] );
		}else{
			UpdateTriangleOffset( pNewTriangle[0] );
			pNewTriangle[0]->UpdatePriority();
		}

		if( pTriangle->m_pEdge[2]->m_pEdge[2] == pTriangle ){
			pTriangle->CopyPriorityInfo( pTriangle->m_pEdge[2] );
		}else{
			UpdateTriangleOffset( pTriangle );
			pTriangle->UpdatePriority();
		}
	}else{
		pNewTriangle[0]->m_fSplitPriority = 0;
		pTriangle->m_fSplitPriority       = 0;
	}


	// Split the second one from pOpposite
	pNewTriangle[1]->InsertAfter( pOpposite );
	pNewTriangle[1]->m_fDistance = pOpposite->m_fDistance;
	pNewTriangle[1]->m_fFacing   = pOpposite->m_fFacing;
	pNewTriangle[1]->m_fAngle    = pOpposite->m_fAngle;
	pNewTriangle[1]->SetVertices( pOpposite->m_pVertex[2], pNewVertex, pOpposite->m_pVertex[1] );
	pOpposite->SetVertices( pOpposite->m_pVertex[1], pNewVertex, pOpposite->m_pVertex[0] );
	pNewTriangle[1]->SetEdges( pTriangle, pOpposite, pOpposite->m_pEdge[1] );
	pOpposite->m_pEdge[1]->ReplaceEdge( pOpposite, pNewTriangle[1] );
	pOpposite->SetEdges( pNewTriangle[1], pNewTriangle[0], pOpposite->m_pEdge[0] );
	if( bDynamic ){
		if( pNewTriangle[1]->m_pEdge[2]->m_pEdge[2] == pNewTriangle[1] ){
			pNewTriangle[1]->CopyPriorityInfo( pNewTriangle[1]->m_pEdge[2] );
		}else{
			UpdateTriangleOffset( pNewTriangle[1] );
			pNewTriangle[1]->UpdatePriority();
		}
		if( pOpposite->m_pEdge[2]->m_pEdge[2] == pOpposite ){
			pOpposite->CopyPriorityInfo( pOpposite->m_pEdge[2] );
		}else{
			UpdateTriangleOffset( pOpposite );
			pOpposite->UpdatePriority();
		}
	}else{
		pNewTriangle[1]->m_fSplitPriority = 0;
		pOpposite->m_fSplitPriority       = 0;
	}

	m_nTriangles += 2;
	pNewVertex->m_vNormal =
		pTriangle->m_vNormal +
		pOpposite->m_vNormal +
		pNewTriangle[0]->m_vNormal +
		pNewTriangle[1]->m_vNormal;
	pNewVertex->m_vNormal.Normalize();

	pNewTriangle[0]->m_pParent = pTriangle;
	pNewTriangle[1]->m_pParent = pOpposite;

	// If either of the split triangles was in the merge list, remove it.
	static CDiamond *pDiamond;
	for( pDiamond=m_llDiamond.GetHead(); pDiamond->IsInList(); pDiamond=pDiamond->GetNext() ){
		if( pDiamond->HasTriangle(pTriangle) || pDiamond->HasTriangle(pOpposite) ){
			pDiamond = pDiamond->GetPrevious();
			delete pDiamond->GetNext();
			m_nDiamonds--;
		}
	}

	// Then add the new diamond to the merge list
	m_llDiamond.AddTail(
		new CDiamond( pTriangle, pOpposite, pNewTriangle[0], pNewTriangle[1] )
	);
	m_nDiamonds++;
}


void CSphere::Merge( CDiamond *pDiamond ){
	int i;
	// First, get the vertex all the triangles share and the parent triangles that the children will be merged into
	CVertex *pVertex = pDiamond->m_pParent[0]->m_pVertex[1];
	ASSERT( pDiamond->m_pParent[1]->m_pVertex[1] == pVertex && pDiamond->m_pChild[0]->m_pVertex[1] == pVertex && pDiamond->m_pChild[1]->m_pVertex[1] == pVertex );

	// Loop twice to merge two child triangles into their parents
	for( i=0; i<2; i++ ){
		// Start by updating the parent's vertices and edges, then deleting the child
		CTriangle *pParent = pDiamond->m_pParent[i];
		CTriangle *pChild  = pDiamond->m_pChild[i];
		pParent->SetVertices( pParent->m_pVertex[2], pParent->m_pVertex[0], pChild->m_pVertex[0] );
		pParent->SetEdges   ( pParent->m_pEdge[2], pChild->m_pEdge[2], pDiamond->m_pParent[1-i] );
		pChild->m_pEdge[2]->ReplaceEdge( pChild, pParent );
		delete pChild;

		if( i == 0 ){
			CVector v = Midpoint( pParent->m_pVertex[0]->m_vPosition, pParent->m_pVertex[2]->m_vPosition );
			pParent->m_vMidpoint = v;
			float fHeight = (float)v.Magnitude();
			pParent->m_fOffset = (float)pVertex->m_vPosition.Magnitude() - fHeight;
			pParent->m_bColor  = false;
			pParent->m_bTemp   = false;
			if( fHeight < m_fRadius+COLOR_DELTA && fHeight + pParent->m_fOffset > m_fRadius-COLOR_DELTA ){
				pParent->m_bColor = true;
			}
			if( fHeight > m_fRadius-COLOR_DELTA && fHeight + pParent->m_fOffset < m_fRadius+COLOR_DELTA ){
				pParent->m_bColor = true;
			}
			if( fHeight < m_fRadius-TEMP_DELTA && fHeight + pParent->m_fOffset < m_fRadius-TEMP_DELTA ){
				pParent->m_bTemp = true;
			}
			pParent->UpdatePriority();
		}else{
			pParent->CopyPriorityInfo( pParent->m_pEdge[2] );
		}

		//pParent->m_fSplitPriority *= 2.0f;

		// Then see if the merged parent forms a new diamond
		CTriangle *p[4];
		p[0] = pParent;
		p[1] = p[0]->m_pEdge[0];
		p[2] = p[1]->m_pEdge[0];
		p[3] = p[2]->m_pEdge[0];
		if( p[3]->m_pEdge[0] == p[0] ){
			// If it does, make sure parents and children match up
			if( p[1]->m_pParent == p[0] && p[3]->m_pParent == p[2] ){
				m_llDiamond.AddTail( new CDiamond(p[0], p[2], p[1], p[3]) );
				m_nDiamonds++;
			}

			if( p[2]->m_pParent == p[1] && p[0]->m_pParent == p[3] ){
				m_llDiamond.AddTail( new CDiamond(p[1], p[3], p[2], p[0]) );
				m_nDiamonds++;
			}
		}
	}
	m_nTriangles -= 2;
	m_nDiamonds--;
	delete pDiamond;
	delete pVertex;
}


bool CSphere::Update( CVector vPosition, CVector vHeading ){
/*	CVector     vPosition = camera->getLoc() - getLoc();
	vPosition             = RotateVector( vPosition );
	CQuaternion q         = *pCamera;

	q.Rotate(*this);
	CVector vHeading = q.GetViewAxis();
	CVector vUp      = q.GetUpAxis();
	CVector vRight   = q.GetRightAxis();
*/
	float   fHeight = (float)vPosition.Magnitude();
	CVector v       = vPosition / fHeight;

	float fAltitude = m_frTest.fBmTest( v, 8.0f ) * m_fVariance + m_fOffset + m_fRadius;
	if( fHeight < fAltitude+0.001 ){
		return true;
	}

	// Update the split prioritities and get the triangle with the greatest priority
	m_nTriangles = 0;
	for( CTriangle *pTriangle=m_llTriangle.GetHead(); pTriangle->IsInList(); pTriangle = pTriangle->GetNext() ){
		pTriangle->UpdatePosition( vPosition, vHeading/*, m_fRadius*/ );
		pTriangle->UpdatePriority();
		m_nTriangles++;
		if( pTriangle->GetPriority() >= SPLIT_PRIORITY ){
			Split( pTriangle );
		}
	}

	// Update the merge priorities and get the diamond with the lowest priority
	CDiamond *pMerge;
	m_nDiamonds = 0;
	for( CDiamond *pDiamond=m_llDiamond.GetHead(); pDiamond->IsInList(); ){
		pMerge = pDiamond;
		pMerge->UpdatePriority();
		pDiamond = pDiamond->GetNext();
		if( pMerge->GetPriority() <= MERGE_PRIORITY ){
			Merge( pMerge );
		}
		m_nDiamonds++;
	}
	m_nVertices = m_llVertex.GetCount();

	triangles = m_nTriangles;
	diamonds  = m_nDiamonds;
	vertices  = m_nVertices;
	return false;
}


void CSphere::Init( int nSeed, float fRadius, float fVariance, float fRoughness ){
	m_frTest.Init( 3, nSeed, 1.0f-fRoughness, 2.25f );
	m_fRadius      = fRadius;
	m_fMaxVariance = fVariance;
	clip_radius = fRadius + fVariance;

	BuildCube();

	CVertex *pVertex;
	srand( nSeed );

	float fMin = (float)1e20, fMax = (float)-1e20;
	for( pVertex=m_llVertex.GetHead(); pVertex->IsInList(); pVertex = pVertex->GetNext() ){
		pVertex->m_fTexCoord[0] = m_frTest.fBmTest(pVertex->m_vPosition, 8.0f);

		if( pVertex->m_fTexCoord[0] < fMin ){
			fMin = pVertex->m_fTexCoord[0];
		}
		if( pVertex->m_fTexCoord[0] > fMax ){
			fMax = pVertex->m_fTexCoord[0];
		}
	}
	m_fVariance = fVariance / 0.75f;
	m_fOffset = -m_fMaxVariance + 0.75f * m_fVariance;

	for( pVertex=m_llVertex.GetHead(); pVertex->IsInList(); pVertex = pVertex->GetNext() ){
		float fLattitude = Clamp( 0.01f, 0.99f, Abs(pVertex->m_vPosition.y) + 0.5f * m_frTest.fBm(pVertex->m_vPosition*4, 8.0f) );
		float fAltitude  = pVertex->m_fTexCoord[0] * m_fVariance + m_fOffset + m_fRadius;
		pVertex->m_vPosition *= fAltitude;
		fAltitude -= m_fRadius - m_fMaxVariance;
		fAltitude  = Clamp( 0.01f, 0.99f, (fAltitude * 0.5f) / m_fMaxVariance );
		pVertex->m_fTexCoord[0] = fAltitude;
		pVertex->m_fTexCoord[1] = fLattitude;
	}

	for( CTriangle *pTriangle=m_llTriangle.GetHead(); pTriangle->IsInList(); pTriangle = pTriangle->GetNext() ){
		UpdateTriangleOffset( pTriangle );
	}
}


/*virtual*/ void CSphere::beginElements(){
	glBegin( GL_TRIANGLES );
}


/*virtual*/ void CSphere::drawElements(){
	for( CTriangle *p = m_llTriangle.GetHead(); p->IsInList(); p = p->GetNext() ){
		p->Draw();
	}
}


/*virtual*/ void CSphere::endElements(){
	glEnd();
}


};  //  namespace Models

