
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CSphere
	\ingroup g_models
	\author  Sean O'Neil
	\date    2000

	This class implements the spherical ROAM algorithm. It keeps lists of the
	vertices, triangles, and diamonds. Because it is currently used to implement
	a fractal-based planet, it also has a member variable of the CFractal class
	and a number of others for defining the planet's basic properties. Its main
	methods are Init(), Update(), and Draw(). Init() initializes the planet
	with a random seed and certain parameters, like radius and roughness.
	Update() updates the priority lists and performs any necessary splits or
	merges in the mesh. Draw() draws the mesh using either smooth or flat shading.
*/


#ifndef TEDDY_MODELS_C_SPHERE_H
#define TEDDY_MODELS_C_SPHERE_H


#include "DataStructures/CLinkedList.h"
#include "DataStructures/CListNode.h"
#include "Maths/Vector.h"
#include "Maths/TransposeMatrix.h"
#include "Maths/CVector.h"
#include "Maths/CFractal.h"
#include "Models/CVertex.h"
#include "Models/CTriangle.h"
#include "Models/CDiamond.h"
#include "Graphics/OpenGL.h"
#include "Scenes/Camera.h"
#include <iostream>
using namespace DataStructures;
using namespace Maths;
using namespace Scenes;
using namespace std;
//using namespace Graphics;


namespace Models {


class CSphere : public Mesh {
public:
	CSphere(){}
	virtual ~CSphere(){}

	int GetVertexCount  (){ return m_nVertices; }
	int GetTriangleCount(){ return m_nTriangles; }
	int GetDiamondCount (){ return m_nDiamonds; }

	void Init  ( int nSeed, float fRadius, float fVariance, float fRoughness );
	bool Update( CVector vPosition, CVector vHeading );

	virtual void update( RenderContext *rc ){
		Camera *c = dynamic_cast<Camera*>(rc);
		if( c==NULL ){
			return;
		}
		Vector c_loc  = c->getLoc();   //c_loc.Neg();
		Vector c_look = c->getLook();  c_look.Neg();

		float f_pos [3]; f_pos [0] = c_loc .r; f_pos [1] = c_loc .s; f_pos [2] = c_loc .t;
		float f_head[3]; f_head[0] = c_look.r; f_head[1] = c_look.s; f_head[2] = c_look.t;
		(void)this->Update( CVector(f_pos), CVector(f_head) );
	}

	virtual void beginElements();
	virtual void drawElements ();
	virtual void endElements  ();

protected:
	CVertexList   m_llVertex;
	CTriangleList m_llTriangle;
	CDiamondList  m_llDiamond;

	int      m_nVertices, m_nTriangles, m_nDiamonds;
	float    m_fRadius;       //!<  The radius of the planet
	float    m_fMaxVariance;  //!<  The maximum variance above or below the radius
	float    m_fVariance;
	float    m_fOffset;
	CFractal m_frTest;        //!<  The fractal object used to determine altitudes for new vertices

	void RebuildEdgeList     ();
	void BuildCube           ();
	void UpdateTriangleOffset( CTriangle *pTriangle );
	void Split               ( CTriangle *pTriangle, bool bDynamic=true );
	void Merge               ( CDiamond *pDiamond );
	void MergeAll(){
		while( !m_llDiamond.IsEmpty() ){
			Merge( m_llDiamond.GetHead() );
		}
	}
};


extern int triangles;
extern int diamonds;
extern int vertices;


};  //  namespace Models


#endif  //  TEDDY_MODELS_C_SPHERE_H


