
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta, Sean O' Neil

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\file
	\class   CTriangle
	\ingroup g_models
	\author  Sean O'Neil
	\brief   ROAM triangle
	\date    2000

	This class implements the ROAM triangle. It is quite large because it
	relies heavily on caching information so it doesn't have to calculate things
	every frame. Some deal with calculating split priority while others deal
	with handling the actual split. Like CVertex, it is derived from CListNode
	and uses CMemoryBlock if USE_ALLOCATOR is defined.
	
	Note that the parent pointer points to the parent of this triangle, but this
	is still not a tree-like structure! I do not keep a binary tree of all
	triangles used to reach the current model, I only keep the triangles that
	are part of the current model. When I split 2 triangles into 4, I modify
	the 2 existing triangles and add 2 new ones, setting the parent pointers of
	the new triangles to point to the existing ones. Later when I merge 4 into
	2, the two child triangles get deleted and the parents get changed back.
	
	One more important thing to note is vertex and edge order. ROAM will only
	split two right triangles that make up a square, so I think of every
	triangle as a right triangle (even though it's not really a right triangle
	because it's on a sphere). My vertices always run counter-clockwise with
	the middle vertex being in the middle. Edges are handled in a similar
	fashion. Edge 0 indicates vertices 0 - 1, edge 1 indicates vertices 1 - 2,
	and edge 2 indicates vertices 2 - 0.
	
	It looks something like this:
	(square)
	1----0 /2   Note: m_pEdge[2] is always the edge that is split, and a
	|   / / |         triangle can be split only if m_pEdge[2]->m_pEdge[2]
	|  / /  |         points back to itself. If it doesn't, its neighbor must
	| / /   |         be split first (recursively).
	2/ 0----1	
*/


#ifndef TEDDY_MODELS_C_TRIANGLE_H
#define TEDDY_MODELS_C_TRIANGLE_H

		
#include <math.h>
#include "SysSupport/CMemoryBlock.h"
#include "DataStructures/CLinkedList.h"
#include "DataStructures/CListNode.h"
#include "Models/CVertex.h"
#include "Maths/CVector.h"
#include "Maths/NoiseMath.h"  //  FIX MAX
#include <string.h>
using namespace SysSupport;
using namespace DataStructures;
using namespace Maths;


namespace Models {


#define CTriangleList CLinkedList<CTriangle>


class CTriangle : public CListNode<CTriangle> {
public:
	CVertex   *m_pVertex[3];      //!<  Pointers to the 3 vertices that make up the triangle
	CVector    m_vNormal;         //!<  The normal of the triangle (used for flat shading)
	CTriangle *m_pEdge[3];        //!<  Pointers to the triangles sharing edges with this one
	CTriangle *m_pParent;         //!<  Pointer to the parent triangle that created this one

	CVector    m_vMidpoint;       //!<  The midpoint of the third edge (the one to split)
	float      m_fOffset;         //!<  The offset in height the midpoint will require when split (used to determine split priority)
									
	bool       m_bColor;          //!<  Set to true if the offset crosses sea level (drastically changing the color of the pixel)
	bool       m_bTemp;           //!<  Set to true if the entire polygon is below sea level
	float      m_fDistance;       //!<  The distance to the viewpoint
	float      m_fFacing;         //!<  The angle the triangle is facing (relative to the camera)
	float      m_fAngle;          //!<  The angle between the camera heading and heading to the midpoint
	float      m_fSplitPriority;  //!<  Priority in the split queue

public:
	CTriangle(){
		memset( (char *)this, 0, sizeof(CTriangle) );
	}

	CTriangle( CVertex *p1, CVertex *p2, CVertex *p3 ){
		memset( (char *)this, 0, sizeof(CTriangle) );
		SetVertices( p1, p2, p3 );
	}

	void CopyPriorityInfo( CTriangle *pTriangle ){
		m_vMidpoint      = pTriangle->m_vMidpoint;
		m_fOffset        = pTriangle->m_fOffset;
		m_bColor         = pTriangle->m_bColor;
		m_bTemp          = pTriangle->m_bTemp;
		m_fSplitPriority = pTriangle->m_fSplitPriority;
	}

	void UpdatePosition( CVector &vPosition, CVector &vHeading/*, float fRadius*/ ){
		m_fDistance = (float)m_vMidpoint.Distance( vPosition );
		CVector v   = m_vMidpoint - vPosition;
		v.Normalize();

		//float fHeight = (float)vPosition.Magnitude() - fRadius;
		/*  TIS clipping removed
		if( m_fDistance > MAX(sqrt(fHeight*fHeight + 2*fHeight*fRadius), 1000) ){
			m_fFacing = -1;
			m_fAngle  = -1;
			return;
		}*/

		m_fFacing = DotProduct(-m_vNormal, v);
		for( int i=0; i<3; i++ ){
			v = m_pVertex[i]->m_vPosition - vPosition;
			v.Normalize();
			float f = DotProduct( vHeading, v );
			if( !i || m_fAngle < f ){
				m_fAngle = f;
			}
			f = (float)vPosition.Distance( m_pVertex[i]->m_vPosition );
			if( m_fDistance > f ){
				m_fDistance = f;
			}
		}
	}

	void UpdatePriority(){
		if( m_fDistance > 1.00f && (m_fAngle < 0.7f || m_fFacing < -0.25f) ){
			m_fSplitPriority = 0.0f;
		}else{
			m_fSplitPriority =
				/*(m_bColor ?
					MIN(25.0f, MAX(1.0f, m_fDistance*10.0)) :
					m_bTemp ?
						0.1f :
						1.0f
				) 
				*  */
				ABS(m_fOffset) / Max(1.00f, m_fDistance*50.0f);
		}
	}

	float GetPriority(){
		return m_fSplitPriority;
	}

	void SetVertices( CVertex *p1, CVertex *p2, CVertex *p3 ){
		m_pVertex[0] = p1;
		m_pVertex[1] = p2;
		m_pVertex[2] = p3;
		ComputeNormal();
	}

	void SetEdges( CTriangle *p1, CTriangle *p2, CTriangle *p3 ){
		m_pEdge[0] = p1;
		m_pEdge[1] = p2;
		m_pEdge[2] = p3;
	}

	void ReplaceEdge( CTriangle *pOld, CTriangle *pNew ){
		if(m_pEdge[0] == pOld){
			m_pEdge[0] = pNew;
		}else if( m_pEdge[1] == pOld ){
			m_pEdge[1] = pNew;
		}else if( m_pEdge[2] == pOld ){
			m_pEdge[2] = pNew;
		}
	}

	void ComputeNormal(){
		m_vNormal = Normal(m_pVertex[0]->m_vPosition, m_pVertex[1]->m_vPosition, m_pVertex[2]->m_vPosition);
	}

	bool HasVertex( CVertex *p ){
		return (p == m_pVertex[0] || p == m_pVertex[1] || p == m_pVertex[2]);
	}

	CVertex *GetOpposite( CVertex *p1, CVertex *p2 ){
		if( m_pVertex[0] != p1 && m_pVertex[0] != p2 ){
			return m_pVertex[0];
		}
		if( m_pVertex[1] != p1 && m_pVertex[1] != p2 ){
			return m_pVertex[1];
		}
		return m_pVertex[2];
	}

	void Draw( GLenum nMode=GL_SMOOTH ){
		//if(m_fFacing < -0.25f || m_fAngle < 0.6f)
		//	return;
		if( nMode != GL_SMOOTH ){
			glNormal3fv( m_vNormal );
		}
		m_pVertex[0]->Draw( nMode );
		m_pVertex[1]->Draw( nMode );
		m_pVertex[2]->Draw( nMode );
	}

	void *operator new( unsigned int n ){
		return m_mbTriangle.GetElement();
	}

	void operator delete( void *p ){
		m_mbTriangle.ReleaseElement( p );
	}

private:
	static CMemoryBlock m_mbTriangle;
};



};  //  namespace Models


#endif  //  TEDDY_MODELS_C_TRIANGLE_H

