
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Models/Mesh.h"
#include "Models/Element.h"
#include "Materials/Material.h"
#include "Maths/IdentityMatrix.h"
#include "Graphics/OpenGL.h"
#include "Graphics/RenderContext.h"
#include "Scenes/Camera.h"
#include <iostream>
using namespace Graphics;
using namespace Materials;
using namespace Scenes;
using namespace std;


namespace Models {


Material *Mesh::default_material = 
	new Material( "Default material", false, GL_FLAT );


//!  Default Constructor
Mesh::Mesh(){
	name            = string("");
	transformations = IdentityMatrix();
	material        = NULL;	
	clip_radius     = 0.0;
}


//!  Constructor with name and default material
Mesh::Mesh( const string &name, Material *material ){
	this->name      = name;
	this->material  = material;
	transformations = IdentityMatrix();
	clip_radius     = 0.0;
}


//!  Insert submesh to Mesh
void Mesh::insert( Mesh *m ){
	submeshes.push_back( m );
}


//!  Insert Element to Mesh
void Mesh::insert( Element *e ){
	elements.push_back( e );
}


/*!
	Update OpenGL displaylist.
	Recursively manages submeshes as well.
	First update submesh (if any) displaylists
	Then put to displaylist: own elements and call submesh (if any) displaylists
*/
/*virtual*/ void Mesh::update( RenderContext *rc ){
	View::check();
	list<Mesh*>::iterator m_it;
	m_it = submeshes.begin();
	while( m_it != submeshes.end() ){
		(*m_it)->update( rc );
		m_it++;
	}

	try{
		beginUpdate( rc );
		drawElements( rc );
		list<Mesh*>::iterator m_it;
		m_it = submeshes.begin();
		while( m_it != submeshes.end() ){
			(*m_it)->draw( rc );
			m_it++;
		}
		endUpdate( rc );
	}catch(...){
		cout << "Mesh::update() error" << endl;
		cout.flush();
	}
	View::check();
}


//!  Mesh Interface - Begin Elements - Pre drawing code
/*virtual*/ void Mesh::beginElements( RenderContext *rc ){
}


//!  Mesh Interface - Draw elements
/*virtual*/ void Mesh::drawElements( RenderContext *rc ){
	if( material!=NULL ){
		material->glApply( rc );
	}

	beginElements( rc );

	list<Element*>::const_iterator e_it = elements.begin();
	while( e_it!=elements.end() ){
		(*e_it)->draw();
		e_it++;
	}

	endElements( rc );

}


//!  Mesh Interface - End Elements - Post drawing code
/*virtual*/ void Mesh::endElements( RenderContext *rc ){
}


//!  Get Mesh material
Material *Mesh::getMaterial() const {
	return material;
}

			   
//!  Set Mesh Material
void Mesh::setMaterial( Material *m ){
	this->material = m;
}


//!  Get Mesh clipping radius
float Mesh::getClipRadius() const {
	return clip_radius;
}

			   
//!  Set Mesh clipping radius
void Mesh::setClipRadius( float r ){
	this->clip_radius = r;
}


//!  Debugging information
void Mesh::debug(){	
	cout << "Mesh " << name.c_str() << " elements " << elements.size() << endl;
	if( material!=NULL ){
		material->debug();
	}

	list<Mesh*>::const_iterator m_it;
	m_it = submeshes.begin();
	while( m_it != submeshes.end() ){
		(*m_it)->debug();
		m_it++;
	}

	cout.flush();
}


//!  Return mesh name
string Mesh::getName(){
	return this->name;
}


};  //  namespace Models


