
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\class   Mesh
	\ingroup g_models
	\author  Timo Suoranta
	\brief   Mesh maintains modeling and rendering data for part of Mesh Model
	\bug     Destructors are missing
	\bug     There is no level of detail support.
	\date    1999, 2000

	Mesh maintains modeling data for an object.
	Mesh is collection of Elements and sub-meshes.
	Each mesh can have individual Material properties.

	The level of detail support should be added.
*/


#ifndef TEDDY_MODELS_MESH_H
#define TEDDY_MODELS_MESH_H


#include "Graphics/RenderContext.h"
#include "Graphics/Renderable.h"
#include "Graphics/OpenGL.h"
#include "Materials/Material.h"
#include "Maths/Matrix.h"
#include "Models/Element.h"
#include <string>
#include <list>
using namespace Graphics;
using namespace Materials;
using namespace Maths;
using namespace std;


namespace Models {


class Mesh : public Renderable {
public:
	Mesh();
    Mesh( const string &name, Material *material = NULL );

	//  Renderable Interface
	virtual void  update       ( RenderContext *rc );

	//  Mesh Interface
	virtual void  beginElements( RenderContext *rc );
	virtual void  drawElements ( RenderContext *rc );
	virtual void  endElements  ( RenderContext *rc );

	void          insert       ( Mesh *c );
	void          insert       ( Element *e );
	void          setClipRadius( float r );
	float         getClipRadius() const;
	void          setMaterial  ( Material *m );
	Material     *getMaterial  () const;
	string        getName      ();
	void          debug        ();

protected:		  
	string           name;              //!<  Name for this mesh
	Material        *material;          //!<  Material property for this Mesh
	Matrix           transformations;   //!<  Transformation matrix specifies location, rotation, scaling...
	list<Mesh*>      submeshes;         //!<  Hierarchial meshes have submeshes
	list<Element*>   elements;          //!<  A single mesh is set of elements
	float            clip_radius;       //!<  For view volume clipping

	static Material *default_material;
};


};  //  namespace Models


#endif  //  TEDDY_MODELS_MESH_H

