
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Graphics/OpenGL.h"
#include "Materials/Material.h"
#include "Maths/ColumnToVector4.h"
#include "Maths/IdentityMatrix.h"
#include "Maths/Quaternion.h"
#include "Maths/RotationMatrix.h"
#include "Maths/RowToVector4.h"
#include "Maths/TransposeMatrix.h"
#include "Maths/TranslationMatrix.h"
#include "Maths/trigmath.h"
#include "Models/Mesh.h"
#include "Models/ModelInstance.h"
#include <iostream>
using namespace Graphics;
using namespace Materials;
using namespace std;


namespace Models {


//!  Get ModelInstance Mesh
Mesh *ModelInstance::getMesh() const {
	return( mesh );
}


/*!
	Set ModelInstance Mesh
	You must update() modelInstance after setMesh()
*/
void ModelInstance::setMesh( Mesh *m ){
	this->mesh = m;
//  FIX
//	update();
}


//!  Create new ModelInstance for Light.
ModelInstance::ModelInstance( const string &name ){
	this->name       = name;
	this->mesh       = NULL;
	this->accessible = true;
	transformations  = IdentityMatrix();
}


//!  Create new ModelInstance for Camera.
ModelInstance::ModelInstance( const string &name, Mesh *mesh ){
	this->name       = name;
	this->mesh       = mesh;
	this->accessible = true;
	transformations  = IdentityMatrix();
}


/*!
	Create new ModelInstance.
	Obtains material from model default or static default
*/
ModelInstance::ModelInstance( const string &name, Mesh *mesh, Vector *offset, Vector *rotation ){
	this->name       = name;
	this->mesh       = mesh;
	this->accessible = true;
	transformations  = TranslationMatrix( offset->r, offset->s, offset->t );
	rotate( rads(rotation->r), Vector(0,0,0), Vector(1,0,0) );
	rotate( rads(rotation->s), Vector(0,0,0), Vector(0,1,0) );
	rotate( rads(rotation->t), Vector(0,0,0), Vector(0,0,1) );
}


//!  ModelInstance update
void ModelInstance::update( RenderContext *rc ){
	View::check();
	mesh->update( rc );
	try{
		beginUpdate( rc );
//		mesh->Renderable::draw( rc );
		mesh->draw( rc );
		endUpdate( rc );
	}catch(...){
		cerr << "ModelInstance::update() error" << endl;
		cerr.flush();
		return;
	}
	View::check();
}


/*!
	Drawing the instance.

	Note that this method shadows Renderable::draw();
	draw() will use default parameter draw( Camera *c = NULL );
*/
/*virtual*/ void ModelInstance::draw( RenderContext *rc ){
//	cout << "ModelInstance::draw() this == " << name << endl;
	View::check();
	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glPushMatrix();
	TransposeMatrix( transformations ).glMult();
	Renderable::draw( rc );
	glPopMatrix();
	glPopAttrib();
	View::check();
}


//!  Return name
const string &ModelInstance::getName() const {
	return name;
}


//!  Move ModelInstance foward
void ModelInstance::foward( const double len ){
	transformations.ConcatTranslate( 0, 0, -len );
}


//!  Translate ModelInstance with (x,y,z) world coordinates
void ModelInstance::translate( const double x, const double y, const double z ){
	transformations = TranslationMatrix( x, y, z ) * transformations;
}


//!  Translate ModelInstance with vector (world coordinates)
void ModelInstance::translate( const Vector &v ){
	transformations = TranslationMatrix(v) * transformations;
}


//!  Rotate ModelInstance andle r around given origo and axis
void ModelInstance::rotate( const double r, const Vector &orig, const Vector &axis ){
	transformations.ConcatRotate( r, orig, axis );
}


//!  Quaternation rotate
void ModelInstance::qrot( const double angle, double x, double y, double z ){
	Quaternion rot_q;
	Matrix     rot_m;

	rot_q.FromAxis( rads(angle), x, y, z);
	rot_q.ToMatrix( rot_m );
	transformations *= rot_m;
}


//!  Scale ModelInstance. Does not fix normals
void ModelInstance::scale( const double k ){
	transformations.ConcatScale( k, k, k );
}


//!  Face target
void ModelInstance::face( const Vector &target ){
	Vector x_axis;
	Vector y_axis;
	Vector z_axis;

	x_axis = Vector(ColumnToVector4( transformations, 0 ));
	z_axis = target - this->getLoc();
	z_axis.Neg();
	z_axis.Norm();
	y_axis = z_axis * x_axis;
	y_axis.Norm();
	x_axis = y_axis * z_axis;
	x_axis.Norm();

	transformations.VectorToColumn( x_axis, 0 );
	transformations.VectorToColumn( y_axis, 1 );
	transformations.VectorToColumn( z_axis, 2 );
}


//!  Return world coordinate location of ModelInstance
Vector ModelInstance::getLoc() const {
	return Vector(ColumnToVector4( transformations, 3 ));
}


//!  Wrapper for mesh->getClipRadius() - can be overridden for those that don't use mesh
/*virtual*/ float ModelInstance::getClipRadius() const {
	return mesh->getClipRadius();
}


//!  Alter ModelInstance pitch
void ModelInstance::pitch( const double pitch ){
	transformations *= RotationMatrix( -rads(pitch), Vector::NullVector, Vector::IdentityX );
}


//!  Alter ModelInstance roll
void ModelInstance::roll( const double roll ){
	transformations *= RotationMatrix( -rads(roll), Vector::NullVector, Vector::IdentityZ );
}


//!  Return lookat vector of ModelInstance
Vector ModelInstance::getLook() const {
	return Vector(ColumnToVector4( transformations, 2 ));  // z
}


//!  Return distance to point
double ModelInstance::distanceTo( const Vector &target ) const {
	Vector delta = Vector(ColumnToVector4( transformations, 3 )) - target;
	return sqrt( delta.r*delta.r + delta.s*delta.s + delta.t*delta.t );
}


//!  Return distance^2 to point
double ModelInstance::distanceSqr( const Vector &target ) const {
	Vector delta = Vector(ColumnToVector4( transformations, 3 )) - target;
	return delta.r*delta.r + delta.s*delta.s + delta.t*delta.t;
}


//!  Transform
/*virtual*/ Vector ModelInstance::transform( const Vector &v ){
	Matrix i = transformations;
	//i.Invert();
	return i.Transform( v );
}


//!  Inverse Transform
/*virtual*/ Vector ModelInstance::invTransform( const Vector &v ){
	Matrix i = transformations;
	i.Invert();
	return i.Transform( v );
}


};  //  namespace Models


