
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "PhysicalComponents/Area.h"
#include "PhysicalComponents/Projection.h"
#include "PhysicalComponents/LayoutConstraint.h"
#include "PhysicalComponents/Style.h"


namespace PhysicalComponents {


//!  Constructor
Area::Area( View *view /*=NULL*/ ){
	constraint   = &LayoutConstraint::default_constraint;
	style        = Style::default_style;
	parent       = NULL;
    viewport[0]  = viewport[1] = viewport[2] = viewport[3] = 0;
	ordering     = pre_self;
	this->view   = view;
}


//!  Destructor
/*virtual*/ Area::~Area(){
	// FIX
}


//!  Renderable interface - Area update
/*virtual*/ void Area::update( RenderContext *rc ){
	//  FIX Think about this
}


//!  Area Layout interface - Set Area ordering
void Area::setOrdering( const e_ordering ordering ){
	this->ordering = ordering;
}


//!  Set parent Area
/*virtual*/ void Area::setParent( Area *parent, View *view ){
	this->parent = parent;
	this->view   = view;

	//  Place Projection areas
	list<Projection*>::iterator p_it = proj_areas.begin();
	while( p_it != proj_areas.end() ){
		(*p_it)->setParent( this, view );
		p_it++;
	}

	//  Place 2D orthogonal projection areas
	if( areas.size() > 0 ){
		list<Area*>::iterator a_it = areas.begin();
		while( a_it != areas.end() ){
			(*a_it)->setParent( this, view );
			a_it++;
		}
	}
}


//!  Area Layout interface - Insert Area to Layer
/*virtual*/ void Area::insert( Area *area ){
	//  Test if area is actually Projection area. If so, put to correct collection
	Projection *proj_area = dynamic_cast<Projection*>(area);
	if( proj_area != NULL ){
		proj_areas.push_back( proj_area );
	}else{
		//  Otherwise put to general collection
		areas.push_back( area );
	}
	area->setParent( this, this->view );
}


//!  Area Layout interface - Layout code
/*virtual*/ void Area::place( int offset_x, int offset_y ){
	const LayoutConstraint &c = *constraint;
	int t_xx;
	int t_yy;
	int p_xx;
	int p_yy;
	int p_x;
	int p_y;

	t_xx = c.local_x_fill_pixels;
	t_yy = c.local_y_fill_pixels;
	if( parent!=NULL ){
		parent->getSize( &p_xx, &p_yy );
		parent->getPos( &p_x, &p_y );
	}else{
		p_xx = p_yy = p_x = p_y = 0;		
	}

	if( c.parent_x_fill_relative >= 0 ){
		t_xx += (int)(c.parent_x_fill_relative * p_xx);
	}
	if( (c.min_x_fill_pixels >= 0) && (t_xx < c.min_x_fill_pixels) ){
		t_xx = c.min_x_fill_pixels;			
	}
	if( (c.max_x_fill_pixels >= 0) && (t_xx > c.max_x_fill_pixels) ){
		t_xx = c.max_x_fill_pixels;
	}
	if( c.parent_y_fill_relative >= 0 ){
		t_yy += (int)(c.parent_y_fill_relative * p_yy);
	}					 
	if( (c.min_y_fill_pixels >= 0) && (t_yy < c.min_y_fill_pixels) ){
		t_yy = c.min_y_fill_pixels;			
	}
	if( (c.max_y_fill_pixels >= 0) && (t_yy > c.max_y_fill_pixels) ){
		t_yy = c.max_y_fill_pixels;
	}
	setSize( t_xx, t_yy );
	setPos(		
		(c.local_x_offset_pixels) +
		(c.local_x_offset_relative * t_xx) +
		(c.parent_x_offset_pixels) +
		(c.parent_x_offset_relative * p_xx) +
		(p_x) + offset_x,
		(c.local_y_offset_pixels) +
		(c.local_y_offset_relative * t_yy) +
		(c.parent_y_offset_pixels) +
		(c.parent_y_offset_relative * p_yy )+
		(p_y) + offset_y
	);

	beginPlace();

	//  Place Projection areas
	list<Projection*>::iterator p_it = proj_areas.begin();
	while( p_it != proj_areas.end() ){
		placeOne( *p_it );
		p_it++;
	}

	//  Place 2D orthogonal projection areas
	if( areas.size() > 0 ){
		list<Area*>::iterator a_it = areas.begin();
		while( a_it != areas.end() ){
			placeOne( *a_it );
			a_it++;
		}
	}

	endPlace();

}


//!  Area Layout Interface - Query minimum accepted size for Area
/*virtual*/ void Area::getMinSize( int *width, int *height ) const {
	*width  = constraint->min_x_fill_pixels;
	*height = constraint->min_y_fill_pixels;
}


/*virtual*/ void Area::placeOne( Area *area ){
	area->place();	
}


//!  Area Layout Interface - Begin placement process for children
/*virtual*/ void Area::beginPlace(){
}


//!  Area Layout Interface - End placement process for children
/*virtual*/ void Area::endPlace(){
}


//!  Area Layout interface - Set Area size
void Area::setSize( const int width, const int height ){
	viewport[2] = viewport[0] + width;
	viewport[3] = viewport[1] + height;
}


//!  Area Layout interface - Get Area size
void Area::getSize( int *width, int *height ) const {
	*width  = viewport[2] - viewport[0];
	*height = viewport[3] - viewport[1];
}


float Area::getRatio() const {
	int width;
	int height;

	getSize( &width, &height );
	if( height == 0 ){
		return 1.0;
	}
	return (float)(width) / (float)(height);
}


View *Area::getView() const {
	return view;
}


GLint *Area::getViewport(){
	return viewport;
}


//!  Area Layout interface - Set Area position
void Area::setPos( const int x, const int y ){
	int width  = viewport[2] - viewport[0];
	int height = viewport[3] - viewport[1];
	viewport[0] = /*x<0?0:*/x;
	viewport[1] = /*y<0?0:*/y;
	viewport[2] = viewport[0] + width;
	viewport[3] = viewport[1] + height;	
}


//!  Area Layout Interface - Get Area position
void Area::getPos( int *x, int *y ) const {
	*x = viewport[0];
	*y = viewport[1];
}


//!  Area Layout Interface - Move area
/*virtual*/ void Area::moveDelta( const int x_delta, const int y_delta ){
	constraint->local_x_offset_pixels += x_delta;
	constraint->local_y_offset_pixels += y_delta;
	place();
}


//!  Area Layout Interface - Resize area
/*virtual*/ void Area::sizeDelta( const int x_delta, const int y_delta ){
	this->constraint->local_x_fill_pixels += x_delta;
	this->constraint->local_y_fill_pixels += y_delta;
	place();
}


//!  Draw area - draw area self component
/*virtual*/ void Area::drawSelf(){
}


//!  Draw area recursively - draw all areas in layer
/*virtual*/ void Area::draw(){
	//  First draw self
	if( ordering == pre_self ){
		drawSelf();
	}

	//  Draw Projection projection areas - each set the projection matrix
	list<Projection*>::iterator p_it = proj_areas.begin();
	while( p_it != proj_areas.end() ){
		//view->begin3d();
		(*p_it)->draw();
		p_it++;
		//view->end3d();
	}

	//  Draw 2D orthogonal projection areas - set projection matrix once
	if( areas.size() > 0 ){
		view->begin2d();
		list<Area*>::iterator a_it = areas.begin();
		while( a_it != areas.end() ){
			(*a_it)->draw();
			a_it++;
		}
		view->end2d();
	}

	if( ordering == post_self ){
		drawSelf();
	}
}


//!  Area Graphics interface - Enter vertex to rendering engine
void Area::vertex2i( const GLint x, const GLint y ) const {
	glVertex2i( viewport[0]+x, viewport[1]+y );
}


//!  Area Graphics interface - Draw filled rectangle - changed polygonmode..
void Area::drawFillRect( const int x1, const int y1, const int x2, const int y2 ){
	view->setPolygonMode( GL_FILL );
	glBegin( GL_QUADS );
	vertex2i( x1, y1 );
	vertex2i( x2, y1 );
	vertex2i( x2, y2 );
	vertex2i( x1, y2 );
	glEnd();
}


//!  Area Graphics interface - Draw non-filled rectangle - changed polygonmode..
void Area::drawRect( const int x1, const int y1, const int x2, const int y2 ){
	view->setPolygonMode( GL_LINE );
	glBegin( GL_QUADS );
	vertex2i( x1,   y1+1 );
	vertex2i( x2-1, y1+1 );
	vertex2i( x2-1, y2 );
	vertex2i( x1,   y2 );
	glEnd();
}


//!  Area Graphics interface - Draw twocolor rectangle
void Area::drawBiColRect( Color &top_left, Color &bottom_right, const int x1, const int y1, const int x2, const int y2 ){
	top_left.glApply();
	glBegin( GL_LINE_STRIP );
	vertex2i( x1, y2   );
	vertex2i( x1, y1+1 );
	vertex2i( x2, y1+1 );
	glEnd();
	bottom_right.glApply();
	glBegin( GL_LINE_STRIP );
	vertex2i( x2-1,  y1+2 );
	vertex2i( x2-1,  y2   );
	vertex2i( x1,    y2   );
	glEnd();
}


//!  Area Graphics interface - Draw string - no formatting
void Area::drawString( Font *font, string &str, const int xp, const int yp ){
	font->drawString( str, viewport[0]+xp, viewport[1]+yp );
}


};  //  namespace PhysicalComponents

