
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


/*!
	\class   Area
	\ingroup g_physical_components
	\author  Timo Suoranta
	\brief   Base class for physical userinterface components
	\date    2000

	Area is the basic base class for all Physical Components.

	Area has two interfaces. It is both RenderContext and
	Renderable. It is RenderContext in the sense that it
	can be drawn to. It is Renderable in sense that there
	must exist some other RenderContext where to draw it.

	Areas can be hierarhial. In the end, 

	Area provides a drawing context. At the moment there is
	no clipping for drawing operations. I might leave it as
	option not to use clipping later, too.

	Area can contain subareas.
	When Area is drawn, all subareas are also drawn.
	Area knows how to layout (place and size) itself,
	based on it's LayoutConstraints (place method).

	setPos() and setSize() methods are protected lowlevel
	methods; to actually set position or size of area you
	must use LayoutConstraint of that Area. At the moment
	it is public member, but that will change later.

	Notice that you should use drawSelf() method for drawing
	in subclasses and leave draw() method as it is. The draw()
	method will take care of rendering children as well.

	Area can also be interactive such that they can be operated
	with pointing device. Area is not interactive by default.
	To make Area interactive, it must be registered to FocusManager.
	See FocusManager for details on this.

	Visual Area properties link font, color and size should be
	stored in Style.
*/


#ifndef TEDDY_PHYSICAL_COMPONENTS_AREA_H
#define TEDDY_PHYSICAL_COMPONENTS_AREA_H


#include "Graphics/Renderable.h"
#include "Graphics/RenderContext.h"
#include "Graphics/View.h"
#include "Graphics/Color.h"
#include "Graphics/Font.h"
#include "Graphics/OpenGL.h"
#include <list>
using namespace Graphics;
using namespace std;


#if defined( WIN32 )
#pragma warning(disable:4786)
#endif


namespace PhysicalComponents {


class Projection;
class LayoutConstraint;
class Style;


class Area : public Renderable, public RenderContext {
public:
	Area( View *view = NULL );
	virtual ~Area();

	enum e_ordering {				   
		pre_self,      //!<  Draw self first, before children
		post_self      //!<  Draw self last, after children
	};
	
	//  Area interface
	virtual void  drawSelf     ();  //!<  const? Will render only self

	//  Area Graphics interface
	void          vertex2i     ( const GLint x, const GLint y ) const;    //!<  Drawing interface
	void          drawString   ( Font *font, string &str, const int xp, const int yp );     // const?
	void          drawFillRect ( const int x1, const int y1, const int x2, const int y2 );  // const?
	void          drawRect     ( const int x1, const int y1, const int x2, const int y2 );  // const?
	void          drawBiColRect( Color &top_left, Color &bottom_right, const int x1, const int y1, const int x2, const int y2 );  // const?

	//  Area Layout interface
	virtual void  insert       ( Area *area );
	virtual void  place        ( int offset_x=0, int offset_y=0 );
	virtual void  getMinSize   ( int *min_width, int *min_height ) const;  //!<  Query minimal accepted size for Area
	void          getSize      ( int *width, int *height ) const;          //!<  Query real size of Area excluding outer spacings
	void          getPos       ( int *x, int *y ) const;                   //!<  Quary real position of Area
	float         getRatio     () const;                                   //!<  Return aspect ratio of Area
	View         *getView      () const;                                   //!<  Return View of Area
	GLint        *getViewport  ();
	virtual void  moveDelta    ( const int x_delta, const int y_delta );   //!<  Move Area
	virtual void  sizeDelta    ( const int x_delta, const int y_delta );   //!<  Resize Area
	void          setOrdering  ( const e_ordering ordering );

	//  Renderable interface
	virtual void  update       ( RenderContext *rc );
	virtual void  draw         ();   //  Does recursive draw

	LayoutConstraint *constraint;   //  FIX protected: ? private: ?
	Style            *style;        //  FIX protected: ? private: ?

protected:
	//  LowLevel Layout
	virtual void beginPlace();
	virtual void endPlace  ();
	virtual void placeOne  ( Area *area );
	virtual void setParent ( Area *parent, View *view = NULL );
	virtual void setSize   ( const int width, const int height );
	virtual void setPos    ( const int x, const int y );

	e_ordering         ordering;      //!<  Draw before or after children
	GLint              viewport[4];   //!<  OpenGL Viewport for this Area
	Area              *parent;        //!<  Parent Area hosting this Area
	View              *view;
	list<Area*>        areas;         //!<  2D children
	list<Projection*>  proj_areas;    //!<  3D children

	void begin2d();
	void end2d  ();
};


};  //  namespace PhysicalComponents


#endif  //  TEDDY_PHYSICAL_COMPONENTS_AREA_H

