
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Graphics/OpenGL.h"
#include "Graphics/View.h"
#include "Maths/ColumnToVector4.h"
#include "Maths/TransposeMatrix.h"
#include "Maths/trigmath.h"
#include "Models/Mesh.h"
#include "PhysicalComponents/Projection.h"
#include "Scenes/Camera.h"
#include "Scenes/Scene.h"
#include "Scenes/SkyBox.h"
#include <iostream>
using namespace Graphics;
using namespace Models;
using namespace PhysicalComponents;
using namespace std;


namespace Scenes {


//!  Constructor for skybox enabled first person camera
Camera::Camera( const string &name, Mesh *skybox_m, Scene *scene ):ModelInstance(name,skybox_m){
	this->scene     = scene;
	this->fov       = 60.0f;
	this->near_clip = 6.0f;
	this->far_clip  = 20000.0f;
}				   


//!  Destructor
/*virtual*/ Camera::~Camera(){
	//  FIX
}



static void Perspective( GLdouble fovy, GLdouble aspect, GLdouble zNear, GLdouble zFar/*, float x0, float x1, float y0, float y1*/ ){
	GLdouble xmin, xmax, ymin, ymax;

	ymax = zNear * tan( fovy * M_PI / 360.0 );
	ymin = -ymax;

	xmin = ymin * aspect;
	xmax = ymax * aspect;

	glFrustum( xmin, xmax, ymin, ymax, zNear, zFar );
}


/*!
	Apply camera transformations to active viewport (Perspective Area)
	Renders skybox

	\warning Assumes valid Projection in proj
*/
void Camera::projectScene( Projection *p ){
	glPushAttrib( GL_ALL_ATTRIB_BITS );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( fov, p->getRatio(), near_clip, far_clip );
	glMatrixMode( GL_MODELVIEW );

	//  Skybox
	Matrix i;
/*	i = transformations;
	i.VectorToColumn( Vector::NullVector, 3 );
	i.Invert();
	TransposeMatrix( i ).glLoad();
	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glDisable( GL_LIGHTING );
	mesh->draw();
	glPopAttrib();*/

	//  Normal camera
	i = transformations;
	i.Invert();

	TransposeMatrix( i ).glLoad();

	updatePlanes();
	scene->draw( this, p );

	glPopAttrib();
}	


/*!
	Apply camera transformations to viewport
	Apply pick matrix for OpenGL selection
	No skybox, no starfield
*/
ModelInstance *Camera::pickInstance( Projection *p, const int x, const int y ){
	int width, height;

	p->getSize( &width, &height );

	Vector c_loc = this->ModelInstance::getLoc();
	scene->translate( c_loc );
	this->translate( -c_loc.r, -c_loc.s, -c_loc.t );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPickMatrix( x, p->getView()->getHeight()-y, 3.0, 3.0, p->getViewport() );
	gluPerspective( fov, p->getRatio(), near_clip, far_clip );
	glMatrixMode( GL_MODELVIEW );
	Matrix i = transformations;
	i.Invert();
	TransposeMatrix( i ).glLoad();

	updatePlanes();
	return scene->pickInstance( this, p );
}	


//!  Set field of vision
void Camera::setFov( const float fov ){
	this->fov = fov;
}


//!  Return field of vision
float Camera::getFov() const {
	return fov;
}


//  matrix and math utility routines and macros


static void matrixConcatenate( float *result, float *ma, float *mb ){
    int i;
    double mb00, mb01, mb02, mb03, mb10, mb11, mb12, mb13, mb20, mb21, mb22, mb23, mb30, mb31, mb32, mb33;
    double mai0, mai1, mai2, mai3;

    mb00 = mb[ 0];  mb01 = mb[ 1];
    mb02 = mb[ 2];  mb03 = mb[ 3];
    mb10 = mb[ 4];  mb11 = mb[ 5];
    mb12 = mb[ 6];  mb13 = mb[ 7];
    mb20 = mb[ 8];  mb21 = mb[ 9];
    mb22 = mb[10];  mb23 = mb[11];
    mb30 = mb[12];  mb31 = mb[13];
    mb32 = mb[14];  mb33 = mb[15];

    for (i = 0; i < 4; i++) {
        mai0 = ma[i*4+0]; mai1 = ma[i*4+1];
	    mai2 = ma[i*4+2]; mai3 = ma[i*4+3];

        result[i*4+0] = mai0 * mb00 + mai1 * mb10 + mai2 * mb20 + mai3 * mb30;
        result[i*4+1] = mai0 * mb01 + mai1 * mb11 + mai2 * mb21 + mai3 * mb31;
        result[i*4+2] = mai0 * mb02 + mai1 * mb12 + mai2 * mb22 + mai3 * mb32;
        result[i*4+3] = mai0 * mb03 + mai1 * mb13 + mai2 * mb23 + mai3 * mb33;
    }
}


/* Calculates the six view volume planes in object coordinate (OC) space.
       
       Algorithm:
       
       A view volume plane in OC is transformed into CC by multiplying it by
       the inverse of the combined ModelView and Projection matrix (M).
       Algebraically, this is written:
              -1
         P   M   = P
          oc        cc
       
       The resulting six view volume planes in CC are:
         [ -1  0  0  1 ]
         [  1  0  0  1 ]
         [  0 -1  0  1 ]
         [  0  1  0  1 ]
         [  0  0 -1  1 ]
         [  0  0  1  1 ]
       
       To transform the CC view volume planes into OC, we simply multiply
       the CC plane equations by the combined ModelView and Projection matrices
       using standard vector-matrix multiplication. Algebraically, this is written:  
         P   M = P
          cc      oc
       
       Since all of the CC plane equation components are 0, 1, or -1, full vector-
       matrix multiplication is overkill. For example, the first element of the
       first OC plane equation is computed as:
         A = -1 * m0 + 0 * m1 + 0 * m2 + 1 * m3
       This simplifies to:
         A = m3 - m0
       
       Other terms simpliofy similarly. In fact, all six plane equations can be
       computed as follows:
         [ m3-m0  m7-m4  m11-m8  m15-m12 ]
         [ m3+m0  m7+m4  m11+m8  m15+m12 ]
         [ m3-m1  m7-m5  m11-m9  m15-m13 ]
         [ m3+m1  m7+m5  m11+m9  m15+m13 ]
         [ m3-m2  m7-m6  m11-m10 m15-m14 ]
         [ m3+m2  m7+m6  m11+m10 m15+m14 ]
     */
void Camera::updatePlanes(){
    GLfloat ocEc[16];
	GLfloat ecCc[16];
	GLfloat ocCc[16];

    glGetFloatv( GL_MODELVIEW_MATRIX,  ocEc );
    glGetFloatv( GL_PROJECTION_MATRIX, ecCc );
    matrixConcatenate( ocCc, ocEc, ecCc );
	//  Optimize above
	
    view_plane[0] = new Plane( ocCc[ 3] - ocCc[ 0], ocCc[ 7] - ocCc[ 4], ocCc[11] - ocCc[ 8], ocCc[15] - ocCc[12] );
    view_plane[1] = new Plane( ocCc[ 3] + ocCc[ 0], ocCc[ 7] + ocCc[ 4], ocCc[11] + ocCc[ 8], ocCc[15] + ocCc[12] );
    view_plane[2] = new Plane( ocCc[ 3] + ocCc[ 1], ocCc[ 7] + ocCc[ 5], ocCc[11] + ocCc[ 9], ocCc[15] + ocCc[13] );
    view_plane[3] = new Plane( ocCc[ 3] - ocCc[ 1], ocCc[ 7] - ocCc[ 5], ocCc[11] - ocCc[ 9], ocCc[15] - ocCc[13] );
    view_plane[4] = new Plane( ocCc[ 3] + ocCc[ 2], ocCc[ 7] + ocCc[ 6], ocCc[11] + ocCc[10], ocCc[15] + ocCc[14] );
	view_plane[5] = new Plane( ocCc[ 3] - ocCc[ 2], ocCc[ 7] - ocCc[ 6], ocCc[11] - ocCc[10], ocCc[15] - ocCc[14] );
}


bool Camera::cull( ModelInstance *mi ){
	int    i;
	Vector loc;
	float  rad;

	loc = mi->getLoc();
	rad = mi->getClipRadius()*2;
	//  For all planes
	for( i=0; i<6; i++ ){
		//  Is object location further from plane than max radius of object?
		if( view_plane[i]->distance(loc) < -rad ){
			return true;
		}
	}
	return false;
}


};  //  Scenes

