
/*
    TEDDY - General graphics application library
    Copyright (C) 1999, 2000  Timo Suoranta

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Timo Suoranta
    tksuoran@cc.helsinki.fi
*/


#include "Graphics/GlFeature.h"
#include "Graphics/OpenGL.h"
#include "Graphics/View.h"
#include "Materials/Light.h"
#include "Materials/Material.h"
#include "Scenes/Camera.h"
#include "Scenes/Scene.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <map>
using namespace Graphics;
using namespace Materials;
using namespace std;


#ifdef WIN32
#define isnan _isnan
#else
#define isnan __isnan
#endif


namespace Scenes {


int Scene::culled = 0;


//!  Constructor
Scene::Scene( const string &name ){
	this->name = name;
	this->sort = false;
}


//!  Add new ModelInstance to Scene
void Scene::addInstance( ModelInstance *i ){
	instances.push_back( i );
}


//!  Add new Light to Scene
void Scene::addLight( Light *l ){
	lights.push_back( l );
}


//!  Update ModelInstances Meshes in scene - required when lighting is changed for example
void Scene::update( Projection *p ){
	list<ModelInstance*>::iterator i_it = instances.begin();
	while( i_it != instances.end() ){
		(*i_it)->update( p );
		i_it++;
	}
}


//!  Set sorting on/off
void Scene::setSort( const bool state ){
	sort = state;
}


/*!
	Reposition objects in scene for new camera position.
	This is a little dirty hack used to keep origo with camera.
*/
void Scene::translate( const Vector &c_pos ){
	list<Light*>::iterator         l_it;
	list<ModelInstance*>::iterator i_it;

	//  Translate lights 
	l_it = lights.begin();
	while( l_it != lights.end() ){
		(*l_it)->translate( -c_pos.r, -c_pos.s, -c_pos.t );
		l_it++;
	}

	//  Translate objects
	i_it = instances.begin();
	while( i_it != instances.end() ){
		(*i_it)->translate( -c_pos.r, -c_pos.s, -c_pos.t );
		i_it++;
	}

}


//!  Return instance list
list<ModelInstance*> &Scene::getInstances(){
	return instances;
}


//!  Draw scene to OpenGL viewport
void Scene::draw( Camera *c, Projection *p ){
	list<Light*>::iterator         l_it;
	list<ModelInstance*>::iterator i_it;

	culled = 0;

	//  Apply lights 
	if( GlFeature::lighting.isMasterEnabled() ){
		l_it = lights.begin();
		while( l_it != lights.end() ){
			(*l_it)->applyLight();
			l_it++;
		}
	}

	//  Render objects
	if( sort ){
		/*
			FIX NOTICE: This is very bad way to do sorting!

			If several objects have same distance, only one
			of them is actually drawn!

			Also, sorting (those objects which are close enough,
			or all object) should happen by faces, not by whole objects.
		*/

		//  Calc ranges
		map<double, ModelInstance*, greater<double> > sorted_instances;
		i_it = instances.begin();
		while( i_it != instances.end() ){
			double r = (*i_it)->distanceSqr( Vector(0,0,0)/*c_pos*/ );
			if( isnan(r) != 0 ){
				r = DBL_MAX;
			}
			sorted_instances.insert( map<double, ModelInstance*, greater<double> >::value_type( r, *i_it ) );
			i_it++;
		}
		map<double, ModelInstance*, greater<double> >::iterator si_it;
		si_it = sorted_instances.begin();
		while( si_it != sorted_instances.end() ){
			//(*si_it).second->draw( c );
			if( c->cull( (*si_it).second ) ){
				culled++;
			}else{
				(*si_it).second->draw( p );
			}
			si_it++;
		}
	}else{
		i_it = instances.begin();
		while( i_it != instances.end() ){
			double r2 = (*i_it)->distanceSqr( Vector::NullVector );
			/*  FIX Think about this  */
			if( r2 < 10000000.0*10000000.0 ){
				if( c->cull( *i_it ) ){
					culled++;
					//cout << "Culled " << (*i_it)->getName() << endl;
				}else{
					(*i_it)->draw( p );
				}
			}
			i_it++;
		}
	}
}


//!  Return ModelInstance close to given view coordinates  3
ModelInstance *Scene::pickInstance( Camera *c, Projection *p ){
	ModelInstance                  *mi_lookup[1024];
	ModelInstance                  *pick = NULL;
	GLuint                          hits[1024];
	list<ModelInstance*>::iterator  i_it;

	glPushAttrib( GL_ALL_ATTRIB_BITS );

	glSelectBuffer( 1024, hits );
	//  OpenGL select mode does not need lighting, culling is used
	(void)glRenderMode( GL_SELECT );

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glEnable( GL_CULL_FACE );

	glInitNames();
	glPushName( 0 );

	
	//  Render objects
	int mi = 1;
	i_it = instances.begin();

	if( sort ){
		//  Calc ranges
		map<double, ModelInstance*, greater<double> > sorted_instances;
		//i_it = instances.begin();
		while( i_it != instances.end() ){
			double r = (*i_it)->distanceSqr( Vector(0,0,0)/*c_pos*/ );
			if( isnan(r) != 0 ){
				r = DBL_MAX;
			}
			sorted_instances.insert( map<double, ModelInstance*, greater<double> >::value_type( r, *i_it ) );
			i_it++;
		}
		map<double, ModelInstance*, greater<double> >::iterator si_it;
		si_it = sorted_instances.begin();
		while( si_it != sorted_instances.end() ){
			mi_lookup[mi] = (*si_it).second;
			glLoadName( mi++ );
//			(*si_it).second->draw( c );
			/*if( c->cull( (*si_it).second ) ){
				culled++;
			}else{ */
				(*si_it).second->draw( p );
			//}
			si_it++;
		}
	}else{
		//i_it = instances.begin();		
		while( i_it != instances.end() ){
			mi_lookup[mi] = *i_it;
			glLoadName( mi++ );
//			(*i_it).second->draw( c );
			/*if( c->cull( *i_it ) ){
				culled++;
			}else{*/
				(*i_it)->draw( p );
			//}
			i_it++;
		}
	}
	GLuint  num_hits = glRenderMode( GL_RENDER );
	GLuint *ptr      = hits;
	GLuint  names;
	GLuint  z_min;
	GLuint  z_max;
	GLuint  hit_name;
	GLuint  nearest = 0xffffffff;
	for( GLuint i=0; i<num_hits; i++ ){
		names = *ptr++;
		z_min = *ptr++;
		z_max = *ptr++;
		for( GLuint j=0; j<names; j++ ){
			hit_name = *ptr++;
			if( z_min<nearest ){
				nearest = z_min;
				pick    = mi_lookup[hit_name];
			}
		}		
	}
	glPopAttrib();
	return pick;
}


};  //  namespace Scenes

