-------------------------------------------------------------------------------------
-- RTGExport.ms
-- This utility exports geometry from Max 3.0 into the Alias text-based RTG v1.8 format.
-- The RTG (RealTimeGames) format is not very well supported, but it is so simple that
-- writing support for it is easy.
-------------------------------------------------------------------------------------
utility RTGExport "Alias RTG v1.8 Export"
(
	-- Define variables that are visible to all functions in the utility
	local ostream, tabs = ""

	-- Define the GUI interface
	group "Options"
	(
		checkbox cb_exportSelOnly "Export Selected Only"
		checkbox cb_outputVertNorms "Output Vertex Normals" checked:true
		checkbox cb_outputVertColors "Output Vertex Colors" checked:true
		checkbox cb_outputPolyNorms "Output Polygon Normals" checked:true
		checkbox cb_outputTexCoords "Output Texture Coords" checked:true
		checkbox cb_outputHierarchy "Output Hierarchy" checked:true
	)
	button btn_export "Save As..." width:100

	-- Now define the functions that the utility uses

	-------------------------------------------------------------------------------------
	-- This function exports a geometry object to the RTG file.
	function ExportMesh meshObj name =
	(
		-- Output the object header
		Format "\n\nOBJECT_START % v%" name meshObj.numVerts to:ostream
		if cb_outputVertNorms.checked then
		Format " n%" meshObj.numVerts to:ostream
		if cb_outputTexCoords.checked then
		Format " t%" meshObj.numtverts to:ostream
		Format " p%\n\n" meshObj.numFaces to:ostream

		-- Vertex output is in local space
		Format "VERTEX local\n" to:ostream
		if meshObj.numVerts > 0 then
		(
			for i = 1 to meshObj.numVerts do
			(
				vert = ((GetVert meshObj i)-meshObj.pos)
				Format "% % % %\n" (i-1) vert.x vert.y vert.z to:ostream
			)
		)
		
		-- Vertex normals
		if meshObj.numVerts > 0 and cb_outputVertNorms.checked then
		(
			Format "\nNORMAL\n" to:ostream
			for i = 1 to meshObj.numVerts do
			(
				normal = GetNormal meshObj i
				Format "% % % %\n" (i-1) normal.x normal.y normal.z to:ostream
			)
		)
		
		-- Texture coords
		if meshObj.numTVerts > 0 and cb_outputTexCoords.checked then
		(
			Format "\nTEXCOORD\n" to:ostream
			for i = 1 to meshObj.numTVerts do
			(
				uvw = GetTVert meshObj i
				Format "% % % %\n" (i-1) uvw.x uvw.y uvw.z to:ostream
			)
		)
		
		-- Polygon
		Format "\nPOLYGON\n" to:ostream
		if meshObj.NumFaces > 0 then
		(
			for i = 1 to meshObj.numFaces do
			(
				poly = GetFace meshObj i
				Format "% 3 v % % %" (i-1) (poly.x as integer -1) (poly.y as integer -1) (poly.z as integer -1) to:ostream
				-- Specify vert normals
				if cb_outputVertNorms.checked then
					Format " n % % %" (poly.x as integer -1) (poly.y as integer -1) (poly.z as integer -1) to:ostream
				-- Tex coord
				if cb_outputTexCoords.checked and meshObj.numTverts > 0 then
				(
					tvert = GetTVFace meshObj i
					Format " t % % %" (tvert.x as integer -1) (tvert.y as integer -1) (tvert.z as integer -1) to:ostream
				)
				-- Face normal
				if cb_outputPolyNorms.checked then
				(
					normal = GetFaceNormal meshObj i
					Format " N % % %" normal.x normal.y normal.z to:ostream
				)
				-- Texture ref
				Format " T 0\n" to:ostream
			)
		)
		
		Format "\nOBJECT_END %\n" name to:ostream
	)

	-------------------------------------------------------------------------------------
	-- This function is called once per node in the scene.
	-- A node in Max may be all sorts of things. We are only interested in geometry.
	function ExportNode node =
	(
		-- Create node and export class specific data
		if SuperClassOf node == GeometryClass and ClassOf node == Editable_mesh then
			ExportMesh node node.name
		else if SuperClassOf node == GeometryClass then
		(
			-- Build a mesh out of this object and save it
			local temp = copy node
			convertToMesh temp
			ExportMesh temp node.name
			delete temp
		)
		else	-- Not geometry.. could be a camera, light, etc.
			return false

		return true
	)

	-------------------------------------------------------------------------------------
	-- This function recurses down the node hierarchy calling ExportNode for each node.
	function RecursiveExportNode node =
	(
		if (ExportNode node) == false then
			return false

		-- Recurse children before writing this node
		for child in node.children do
			RecursiveExportNode child
	)
	-------------------------------------------------------------------------------------
	-- Output the RTG file header.
	function OutputHeader =
	(
		Format "OUTPUT_VERT_NORMS " to:ostream
		if cb_outputVertNorms.checked then ; Format "on\n" to:ostream
		else Format "off\n" to:ostream
		Format "OUTPUT_VERT_COLORS " to:ostream
		if cb_outputVertColors.checked then ; Format "on\n" to:ostream
		else Format "off\n" to:ostream
		Format "OUTPUT_POLY_NORMS " to:ostream
		if cb_outputPolyNorms.checked then ; Format "on\n" to:ostream
		else Format "off\n" to:ostream
		Format "OUTPUT_TEX_COORDS " to:ostream
		if cb_outputTexCoords.checked then ; Format "on\n" to:ostream
		else Format "off\n" to:ostream
		Format "OUTPUT_HIERARCHY " to:ostream
		if cb_outputHierarchy.checked then ; Format "on\n" to:ostream
		else Format "off\n" to:ostream

		Format "SHOW_INDEX_COUNTERS on\n" to:ostream
		Format "TEXTURE_MODE per_shader\n" to:ostream
	)
	-------------------------------------------------------------------------------------
	-- A helper function to write indentation to the file.
	function doTabs depth =
	(
		for x = 1 to depth do
		(
			Format "\t" to:ostream
		)
	)
	-------------------------------------------------------------------------------------
	-- A function to output the hierarchical relationship of geometric objects.
	function RecursiveWriteHierarchy node depth =
	(
		-- Write this node
		if SuperClassOf node == GeometryClass then
		(
			-- Geometry info
			doTabs (depth as integer -1)
			Format "% G %\n" depth node.name to:ostream
			doTabs (depth+1)
			Format "tran: % % %\n" node.pos.x node.pos.y node.pos.z to:ostream
			doTabs (depth+1)
			rot = node.rotation as eulerangles
			Format "rot: % % %\n" rot.x rot.y rot.z to:ostream
			doTabs (depth+1)
			Format "scal: % % %\n" node.scale.x node.scale.y node.scale.z to:ostream
			doTabs (depth+1)
			Format "sPiv: % % %\n" node.pivot.x node.pivot.y node.pivot.z to:ostream
			doTabs (depth+1)
			Format "rPiv: % % %\n" node.pivot.x node.pivot.y node.pivot.z to:ostream
		)
		-- Pivot info
		-- Recurse to children
	--	for children in node.children do
	--		RecursiveWriteHierarchy children
	)
	-------------------------------------------------------------------------------------
	-- The highest level function called to export an RTG file.
	function ExportRTG =
	(
		-- Write the RTG header
		Format "HEADER_TITLE\t\tMax 3.0 Real Times Game Output\n" to:ostream
		Format "HEADER_VERSION\t\tv1.8\n" to:ostream
		Format ("HEADER_DATE \t\t" + localtime + "\n") to:ostream
		Format "\n\n" to:ostream
		
		if cb_exportSelOnly.checked then
		(
			Format "NUMBER_OF_OBJECTS = %\n" selection.count to:ostream
			OutputHeader()
	
			for node in selection do
				ExportNode node
		)
		else
		(
			Format "NUMBER_OF_OBJECTS = %\n" rootnode.children.count to:ostream
			OutputHeader()
			
			if cb_outputHierarchy.checked then
			(
				Format "\nHIERARCHY_LIST HXP 1 top_level\n\n" to:ostream

				for node in rootnode.children do
					RecursiveWriteHierarchy node 0

				Format "END_HIERARCHY_LIST\n\n" to:ostream
			)
			
			for node in rootnode.children do
				RecursiveExportNode node
		)
	)
	-------------------------------------------------------------------------------------
	-- Open an prepare a file handle for writing.
	function GetSaveFileStream =
	(
		fname = GetSaveFileName types:"Alias RTG (*.rtg)|*.rtg|All Files(*.*)|*.*|"
		if fname == undefined then
			return undefined

		ostream = CreateFile fname
		if ostream == undefined then
		(
			MessageBox "Couldn't open file for writing !"
			return undefined
		)

		return ostream
	)

	-------------------------------------------------------------------------------------
	-- This is the function called when the user activates the utility by pressing on
	-- the export button. It opens the file and calls the export routine.
	on btn_export pressed do
	(
		ostream = GetSaveFileStream()
		if ostream != undefined then
		(
			ExportRTG()
			close ostream
		)
	)

) -- End RTGExport


