//
//   Copyright (c) 1999-2000 Intel Corporation.
//
// Permition is granted to use, copy, distribute and prepare derivative works 
// of this library for any purpose and without fee, provided, that the above 
// copyright notice and this statement appear in all copies.  
// Intel makes no representations about the suitability of this software for 
// any purpose, and specifically disclaims all warranties. 
// See LEGAL.TXT for all the legal information.
//

#include "PIIIBezier.h"
#include <fvec.h>
#include "util.h"


CBezierPatch *CPIIIBezierSpline2::CreatePatch(DWORD *patchIndices) {
	return new CPIIIBezierPatch2(this,patchIndices);
}

void CPIIIBezierSpline2::PrepearSpline(CScene *scene) {
	CPIIIBezierSpline::PrepearSpline(scene);
	DWORD i;
	for (i=0;i<GetCtrlPtsCount();i++) {
		Vector4 *projCtrlPts = GetProjCtrlPts();
		Vector4 *worldCtrlPts = GetViewCtrlPts();
		m_p3ProjCtrlPts[i].x = F32vec4(projCtrlPts[i].x);
		m_p3ProjCtrlPts[i].y = F32vec4(projCtrlPts[i].y);
		m_p3ProjCtrlPts[i].z = F32vec4(projCtrlPts[i].z);
		m_p3ProjCtrlPts[i].w = F32vec4(projCtrlPts[i].w);
		m_p3WorldCtrlPts[i].x = F32vec4(worldCtrlPts[i].x);
		m_p3WorldCtrlPts[i].y = F32vec4(worldCtrlPts[i].y);
		m_p3WorldCtrlPts[i].z = F32vec4(worldCtrlPts[i].z);
	}
}

void	CPIIIBezierSpline2::CreateSplineFromFile(const char *filename) {
	CPIIIBezierSpline::CreateSplineFromFile(filename);
	m_p3ProjCtrlPts = (PIIIVector4 *)alignedAlloc(sizeof(PIIIVector4) * GetCtrlPtsCount());
	m_p3WorldCtrlPts = (PIIIVector3 *)alignedAlloc(sizeof(PIIIVector3) * GetCtrlPtsCount());
}


CPIIIBezierSpline2::CPIIIBezierSpline2(CPIIIBezierTessalation2 *tesl) : CPIIIBezierSpline(tesl) {
	m_p3ProjCtrlPts = NULL;
	m_p3WorldCtrlPts = NULL;
}

CPIIIBezierSpline2::~CPIIIBezierSpline2() {
	alignedFree(m_p3ProjCtrlPts);
	alignedFree(m_p3WorldCtrlPts);
}




CPIIIBezierPatch2::CPIIIBezierPatch2(CBezierSpline *spline,DWORD *indices) : CPIIIBezierPatch(spline,indices) 
{
	DWORD i,j;
	CPIIIBezierSpline2 *p3spl2 = (CPIIIBezierSpline2 *)spline;
	for (i=0;i<4;i++)
		for (j=0;j<4;j++) {
			DWORD idx = m_indices[i][j];
			m_p3ProjCtrlPtsOffs[i][j] = idx * sizeof(PIIIVector4);
			m_p3WorldCtrlPtsOffs[i][j] = idx * sizeof(PIIIVector3);
		}
}



bool	CPIIIBezierPatch2::EvalSurface(LPDIRECT3DVERTEXBUFFER7 vb,CScene *scene,LPWORD &indices,DWORD &indicesCnt) {
	static F32vec4 _255_ = F32vec4(255.0f);
	D3DTLVERTEX *vtx;
	DWORD i,j,k,t;
	DWORD teslLevel = m_spline->GetTeslLevel();
	CBezierTessalation::TessLevel *tessLevel = m_spline->GetBezierTessalation()->GetTessalationLevel(teslLevel);
	DWORD coeffsSkip = tessLevel->m_teslValuesSkip;

	F32vec4 *coeffs;
	
	coeffs = (F32vec4 *)(tessLevel->m_teslValues);

	DWORD lightState = scene->GetLightingState();

	vb->Lock(DDLOCK_WAIT,(LPVOID *)&vtx,NULL);
	
	
	PIIIVector3 scale,translate;
	Vector3 scaling = scene->GetScreenScaling();
	Vector3 trans = scene->GetScreenTranslation();
	scale.x = F32vec4(scaling.x);
	scale.y = F32vec4(scaling.y);
	scale.z = F32vec4(scaling.z);
	translate.x = F32vec4(trans.x);
	translate.y = F32vec4(trans.y);
	translate.z = F32vec4(trans.z);



	PIIIVector4 *projBase = (((CPIIIBezierSpline2 *)m_spline)->m_p3ProjCtrlPts);
	PIIIVector3 *worldBase = (((CPIIIBezierSpline2 *)m_spline)->m_p3WorldCtrlPts);
	DWORD nIter = (teslLevel*teslLevel + 3) >> 2;
	for (i=0;i<nIter;i++) {
		PIIIVector4 vertex;
		PIIIVector3 normk,normt,wpos;
		
		_mm_prefetch((char*)&vtx[4].sx,_MM_HINT_NTA);
		_mm_prefetch((char*)&vtx[5].sx,_MM_HINT_NTA);
		_mm_prefetch((char*)&vtx[6].sx,_MM_HINT_NTA);
		_mm_prefetch((char*)&vtx[7].sx,_MM_HINT_NTA);

		F32vec4 _ZERO_= _mm_setzero_ps();
		vertex.x = _ZERO_;
		vertex.y = _ZERO_;
		vertex.z = _ZERO_;
		vertex.w = _ZERO_;
		normk.x = _ZERO_;
		normk.y = _ZERO_;
		normk.z = _ZERO_;
		normt.x = _ZERO_;
		normt.y = _ZERO_;
		normt.z = _ZERO_;
		wpos.x = _ZERO_;
		wpos.y = _ZERO_;
		wpos.z = _ZERO_;
		for (k=0;k<4;k++) {
			for (t=0;t<4;t++) {
				F32vec4 coeff = coeffs[k] * coeffs[t + 8];
				DWORD idx = m_indices[t][k];
				PIIIVector4 *pt = &projBase[idx];
				vertex.x += coeff*pt->x;
				vertex.y += coeff*pt->y;
				vertex.z += coeff*pt->z;
				vertex.w += coeff*pt->w;
				if (lightState) {
					PIIIVector3 *wpt = &worldBase[idx];

					wpos.x += coeff*wpt->x;
					wpos.y += coeff*wpt->y;
					wpos.z += coeff*wpt->z;

					coeff = coeffs[k+4] * coeffs[t + 8];
					normt.x += coeff * wpt->x;
					normt.y += coeff * wpt->y;
					normt.z += coeff * wpt->z;

					coeff = coeffs[k] * coeffs[t + 12];
					normk.x += coeff * wpt->x;
					normk.y += coeff * wpt->y;
					normk.z += coeff * wpt->z;
				}
			}
		}

		coeffs += 16;
		
		F32vec4 rhw = rcp(vertex.w);
		F32vec4 sx = vertex.x*scale.x*rhw + translate.x;
		F32vec4 sy = vertex.y*scale.y*rhw + translate.y;
		F32vec4 sz = vertex.z*scale.z*rhw + translate.z;

		_MM_TRANSPOSE4_PS(sx,sy,sz,rhw);
	
		storeu(&vtx[0].sx,sx);
		storeu(&vtx[1].sx,sy);
		storeu(&vtx[2].sx,sz);
		storeu(&vtx[3].sx,rhw);

		if (lightState) {
			PIIIVector3 norm;
			PIIIVector3 color;
			PIIIVector3 dir;
			F32vec4 dot;

			color.x = _mm_setzero_ps();
			color.y = _mm_setzero_ps();
			color.z = _mm_setzero_ps();


			norm.x = normk.y*normt.z - normk.z*normt.y;
			norm.y = normk.z*normt.x - normk.x*normt.z;
			norm.z = normk.x*normt.y - normk.y*normt.x;

			F32vec4 size = rsqrt(norm.x*norm.x + norm.y*norm.y + norm.z*norm.z);
			F32vec4 dir_size;

			for (k = 1,t = 0;t< 8; t++,k <<= 1) {
				if (lightState & k) {
					const CPIIIBezierSpline::PIIILight& light = ((CPIIIBezierSpline2 *)m_spline)->m_lights[t];
					if (light.type == CScene::Light::DIRECTIONAL_LIGHT) {
						dir = light.norm;
						dir_size = F32vec4(1.0f);
					} else {
						dir.x = wpos.x-light.pos.x;
						dir.y = wpos.y-light.pos.y;
						dir.z = wpos.z-light.pos.z;
						dir_size = rsqrt(dir.x*dir.x + dir.y*dir.y + dir.z*dir.z);
					}
					dot = dir.x*norm.x + dir.y*norm.y + dir.z*norm.z;
					
					if (move_mask(dot) != 0x0f) {
						dot = simd_max(dot,_mm_setzero_ps());
						dot *= size * dir_size;
						color.x += dot*light.color.x;
						color.y += dot*light.color.y;
						color.z += dot*light.color.z;
					}
				}
			}

			F32vec4 r = simd_min(color.x*_255_,_255_);
			F32vec4 g = simd_min(color.y*_255_,_255_);
			F32vec4 b = simd_min(color.z*_255_,_255_);

			DWORD Color[4];

			__m64 dr,dg,db,*outCol = (__m64 *)Color;

			// conversion of lower two
			dr = _mm_cvtt_ps2pi(r); 
			dg = _mm_cvtt_ps2pi(g); 
			db = _mm_cvtt_ps2pi(b); 
			// packing
			dr  = _m_psllqi(dr,16);	 
			dg =  _m_psllqi(dg,8);	 
//			outCol[0]  = _m_por(_m_por(db,_m_por(dr,dg)),0xff000000ff000000);	
			outCol[0]  = (_m_por(db,_m_por(dr,dg)));	
			// conversion of upper two
			dr = _mm_cvtt_ps2pi(_mm_shuffle_ps(r,r,0xe)); 
			dg = _mm_cvtt_ps2pi(_mm_shuffle_ps(g,g,0xe)); 
			db = _mm_cvtt_ps2pi(_mm_shuffle_ps(b,b,0xe)); 
			// packing
			dr  = _m_psllqi(dr,16);	 
			dg =  _m_psllqi(dg,8);	 
			//outCol[1]  = _m_por(_m_por(db,_m_por(dr,dg)),0xff000000ff000000);	
			outCol[1] = (_m_por(db,_m_por(dr,dg)));

			vtx[0].color = Color[0];
			vtx[1].color = Color[1];
			vtx[2].color = Color[2];
			vtx[3].color = Color[3];

		} else {
			vtx[0].color = 0xffffffff;
			vtx[1].color = 0xffffffff;
			vtx[2].color = 0xffffffff;
			vtx[3].color = 0xffffffff;
		}

		vtx[0].specular = 0x0;
		vtx[1].specular = 0x0;
		vtx[2].specular = 0x0;
		vtx[3].specular = 0x0;
		vtx[0].tu = 0;
		vtx[1].tu = 0;
		vtx[2].tu = 0;
		vtx[3].tu = 0;
		vtx[0].tv = 0;
		vtx[1].tv = 0;
		vtx[2].tv = 0;
		vtx[3].tv = 0;
		vtx += 4;
	}

	_m_empty();
	indices = tessLevel->m_indices;
	indicesCnt = tessLevel->m_indicesCount;
	vb->Unlock();
	return true;
}


