#include "common.h"
#include "planet.h"
#include "scene.h"
#include "utils.h"
#include <math.h>


/*  
Transform the scene light array to the pipeline 
local light array by transforming the light data to object coordinates
*/

static void Init(CScene *scene,SLight *NewLight,CPlanet *planet) {
	int i;
	D3DMATRIX iWorld = planet->GetIWorldMatrix();
	SVector   *lv,*ov;
	SLight OrigLight;
	float w,x,y,z;
	SColor mat_col = planet->GetMaterialColor();

	for (i=0;i<scene->GetLightCount();i++) {
		
		// copy scene light to pipeline pipe.
		NewLight[i] = OrigLight = scene->GetLight(i);
		
		switch (OrigLight.type) {
		case POINT_LIGHT:			//recalc the position
			lv = &NewLight[i].pos;
			ov = &OrigLight.pos;
			break;

		case DIR_LIGHT:				//recalc the direction
			lv = &NewLight[i].dir;
			ov = &OrigLight.dir;
			break;
		}


		// Reverse transform - World to Object spaces.

		w =	ov->x * iWorld._14 + 
			ov->y * iWorld._24 +
			ov->z * iWorld._34 +
			iWorld._44;

		w = 1 / w;

		x = (ov->x * iWorld._11 + 
			 ov->y * iWorld._21 +
			 ov->z * iWorld._31 +
			 iWorld._41) * w;

		y = (ov->x * iWorld._12 +
	    	 ov->y * iWorld._22 +
			 ov->z * iWorld._32 +
			 iWorld._42) * w;

		z = (ov->x * iWorld._13 + 
			 ov->y * iWorld._23 +
			 ov->z * iWorld._33 +
			 iWorld._43) * w;

		*lv = SVector(x,y,z);


		// pre-multiply light color by material color.

		NewLight[i].col.r = OrigLight.col.r * mat_col.r;
		NewLight[i].col.g = OrigLight.col.g * mat_col.g;
		NewLight[i].col.b = OrigLight.col.b * mat_col.b;


		// directional light vector have to be normalized.

		if (OrigLight.type == DIR_LIGHT)
			*lv = VectorNormalize(*lv);
	}

}




void CPlanet::BasicPipeline(CScene *scene) {
	D3DTLVERTEX *out = m_Out;
	SVertex		*in  = m_AOSData;
	int i,j;
	float x,y,z,w,we;
	float lw,len,magnitude,dot;
	D3DVECTOR	tlp,dir;
	D3DMATRIX	x_mat;
	float r,g,b;
	DWORD Color;
	DWORD nLights = scene->GetLightCount();
	SLight	lights[MAX_LIGHTS];
	float mag;


	//init local light structure.
	Init(scene,lights,this);

	//calc effective (combined) matrix.
	x_mat = MatrixMult(scene->GetCompositeMatrix(),m_World);

	for (i=0;i<m_nVerts;i++) {

		//transform
		w = in->x * x_mat._14 + 
			in->y * x_mat._24 +
			in->z * x_mat._34 +
			x_mat._44;

		we = 1 / w;

		x = (in->x * x_mat._11 + 
			 in->y * x_mat._21 +
			 in->z * x_mat._31 +
			 x_mat._41) * we;

		y = (in->x * x_mat._12 + 
			 in->y * x_mat._22 +
			 in->z * x_mat._32 +
			 x_mat._42) * we;

		z = (in->x * x_mat._13 + 
			 in->y * x_mat._23 +
			 in->z * x_mat._33 +
			 x_mat._43) * we;


		//initial color values.
		r = m_Ambient.r;
		g = m_Ambient.g;
		b = m_Ambient.b;

		for (j=0;j<nLights;j++) {
	
			//calc dot according to light type.
			switch (lights[j].type) {
			case POINT_LIGHT:
				dir.x = lights[j].pos.x - in->x;
				dir.y = lights[j].pos.y - in->y;
				dir.z = lights[j].pos.z - in->z;

				mag = 1.0f/sqrt(dir.x*dir.x + dir.y*dir.y + dir.z*dir.z);
				
				dot = (in->nx * dir.x +
					   in->ny * dir.y + 
					   in->nz * dir.z ) * mag;
				break;
			case DIR_LIGHT:
				dot = in->nx * lights[j].dir.x + 
					  in->ny * lights[j].dir.y + 
					  in->nz * lights[j].dir.z;
				break;
			}

			//apply the light color (if any)
			if (dot > 0) {
				r += dot * lights[j].col.r;
				g += dot * lights[j].col.g;
				b += dot * lights[j].col.b;
			}
		}

		//scale colors from [0..1] to [0..255]
		r *= 255.0f;
		g *= 255.0f;
		b *= 255.0f;
		
		//trim colors if over 255.
		if (r > 255.0f)
			r=255.0f;
		if (g > 255.0f)
			g=255.0f;
		if (b > 255.0f)
			b=255.0f;
		
		//pack to ARGB format (A = 0).
		Color = ((int) r << 16) + ((int)g << 8) + ((int)b);

		//write to output as TLVERTEX.
		out->sx = x;
		out->sy = y;
		out->sz = z;
		out->rhw = we;
		out->color = Color;
		out->specular = 0;
		out->tu = in->tu;	//texture coordinates copy.
		out->tv = in->tv;

		in++;
		out++;
	}
}

