#include "common.h"
#include "planet.h"
#include "scene.h"
#include <math.h>

void CPlanet::GenerateSphere( int nDivisions)
{
	const int   nVertices = (2 * nDivisions * nDivisions) + 2;
	const int   nIndices = 12 * nDivisions * nDivisions;
	const float fRadius = 0.5f;


	SVertex *pVertices;
	WORD    *pIndices;
	int     i, j;
	float   dj = (float)(pi/(nDivisions+1.0f));
	float	di = (float)(pi/nDivisions);
	int	    v;		// vertex offset
	int     ind;	// indices offset

	// allocate the needed memory
	pVertices = (SVertex *)malloc(sizeof(D3DVERTEX) * nVertices);
	pIndices = (WORD *)malloc(sizeof(WORD) * nIndices);

	if(pVertices && pIndices)
	{
		// generate the sphere data, note the random texture coords
		// vertices 0 and 1 are the north and south poles
		pVertices[0] = SVertex(SVector(0.0f,  fRadius, 0.0f), VectorNormalize(SVector(0.0f,  1.0f, 0.0f)), 0.0f, 0.0f);
		pVertices[1] = SVertex(SVector(0.0f, -fRadius, 0.0f), VectorNormalize(SVector(0.0f, -1.0f, 0.0f)), 1.0f, 1.0f);

		for (j=0; j<nDivisions; j++) 
		{
			for (i=0; i<nDivisions*2; i++) 
			{
				SVector	p;
				SVector	n;
				float	u, v;

				// calculate points and normal vectors
				n.y=p.y = (float) cos((j+1) * dj);
				n.x=p.x = (float) sin(i * di) * (float) sin((j+1) * dj);
				n.z=p.z = (float) cos(i * di) * (float) sin((j+1) * dj);

				// apply radius
				p.y *= fRadius;
				p.x *= fRadius;
				p.z *= fRadius;

				// apply normal bumpping.
				n.x += 0.6f*(rand01() - 0.5f);
				n.y += 0.6f*(rand01() - 0.5f);
				n.z += 0.6f*(rand01() - 0.5f);

				n = VectorNormalize(n);

				//calc texture coordinates
				u = ((float)i/nDivisions) / 2.0f;
				v = 1.0f - (float)j/nDivisions;

				pVertices[2+i+j*nDivisions*2] = SVertex(p, n, u, v);
			}
		}

		// now generate the triangle indices
		// strip around north pole first
		for (i=0; i<nDivisions*2; i++) 
		{
			pIndices[3*i] = (WORD)0;
			pIndices[3*i+1] = (WORD)(i+2);
			pIndices[3*i+2] = (WORD)(i+3);

			if (i==nDivisions*2-1)
				pIndices[3*i+2] = 2;
		}

		// now all the middle strips
		for (j=0; j<nDivisions-1; j++) 
		{
			v = 2+j*nDivisions*2;
			ind = 3*nDivisions*2 + j*6*nDivisions*2;

			for (i=0; i<nDivisions*2; i++) 
			{
				pIndices[6*i+ind] = (WORD)(v + i);
				pIndices[6*i+2+ind] = (WORD)(v + i + 1);
				pIndices[6*i+1+ind] = (WORD)(v + i + nDivisions * 2);

				pIndices[6*i+ind+3] = (WORD)(v + i + nDivisions * 2);
				pIndices[6*i+2+ind+3] = (WORD)(v + i + 1);
				pIndices[6*i+1+ind+3] = (WORD)(v + i + nDivisions * 2 + 1);

				if (i==nDivisions*2-1) 
				{
					pIndices[6*i+2+ind] = (WORD)(v + i + 1 - 2 * nDivisions);
					pIndices[6*i+2+ind+3] = (WORD)(v + i + 1 - 2 * nDivisions);
					pIndices[6*i+1+ind+3] = (WORD)(v + i + nDivisions * 2 + 1 - 2 * nDivisions);
				}
			}
		}

		// finally strip around south pole
		v = nVertices - nDivisions * 2;
		ind = nIndices - 3 * nDivisions * 2;

		for (i=0; i<nDivisions*2; i++) 
		{
			pIndices[3*i+ind] = (WORD)1;
			pIndices[3*i+1+ind] = (WORD)(v + i + 1);
			pIndices[3*i+2+ind] = (WORD)(v + i);

			if (i==nDivisions*2-1)
				pIndices[3*i+1+ind] = (WORD)v;
		}

		m_AOSData = pVertices;
		m_nVerts = nVertices;
		m_Indices = pIndices;
		m_nIndices = nIndices;
		m_OutUnalignedMem = malloc(sizeof(D3DTLVERTEX)*m_nVerts + 31);
		m_Out = (D3DTLVERTEX *)(((DWORD)(m_OutUnalignedMem) + 31) & 0xffffffe0);
	}
	return;
}




void CPlanet::FillSOAData() {
	int i,j,k;
	DWORD nIter = (m_nVerts + 3)/4;
	
	m_SOAUnalignedMem = malloc(sizeof(SSimdVertex) * nIter + 31);
	m_SOAData = (SSimdVertex *)(((unsigned)(m_SOAUnalignedMem) + 31) & 0xfffffe0);
	
	if (m_SOAData == NULL)
		return;
	
	for (i=0,k=0;i<nIter;i++,k+=4) {
		for (j=0;j<4;j++) {
			m_SOAData[i].x[j] = m_AOSData[k + j].x;
			m_SOAData[i].y[j] = m_AOSData[k + j].y;
			m_SOAData[i].z[j] = m_AOSData[k + j].z;
			m_SOAData[i].nx[j] = m_AOSData[k + j].nx;
			m_SOAData[i].ny[j] = m_AOSData[k + j].ny;
			m_SOAData[i].nz[j] = m_AOSData[k + j].nz;
			m_SOAData[i].tex[j].u = m_AOSData[k + j].tu;
			m_SOAData[i].tex[j].v = m_AOSData[k + j].tv;
		}
	}
	return;
}


void  CPlanet::SetWorldMatrix(D3DMATRIX& mat) {
	m_World = mat;
	m_Recalc = TRUE;
}

const D3DMATRIX& CPlanet::GetIWorldMatrix() {
	if (m_Recalc) {
		m_iWorld = MatrixInverse(m_World);
		m_Recalc = FALSE;
	}
	return m_iWorld;
}



CPlanet::CPlanet(int LOD) : m_Ambient(0,0,0){
	m_Recalc = FALSE;
	m_iWorld = m_World = IdentityMatrix();
	GenerateSphere(LOD);
	FillSOAData();
}



CPlanet::~CPlanet() {
	free(m_AOSData);
	free(m_Indices);
	free(m_OutUnalignedMem);
	free(m_SOAUnalignedMem);
}



void CPlanet::Render(LPDIRECT3DDEVICE2 lpd3dDev) {
	HRESULT hr;
	hr = lpd3dDev->DrawIndexedPrimitive (D3DPT_TRIANGLELIST, 
									  D3DVT_TLVERTEX, 
									  (LPVOID)GetOutputBuffer(), 
									  GetVertexCount(), 
									  GetIndices(), 
									  GetIndexCount(), 
									  D3DDP_DONOTUPDATEEXTENTS);
	VALIDATE(hr);
}
