#include "common.h"
#include "planet.h"
#include "scene.h"



inline F32vec4 Upper2ToLower2(const F32vec4& vec) {
	return _mm_shuffle_ps(vec,vec,0xe);
}


void ExpandMatrix(const D3DMATRIX &mat,SSimdMatrix& smat) {
	smat._11 = F32vec4(mat._11);
	smat._12 = F32vec4(mat._12);
	smat._13 = F32vec4(mat._13);
	smat._14 = F32vec4(mat._14);
	smat._21 = F32vec4(mat._21);
	smat._22 = F32vec4(mat._22);
	smat._23 = F32vec4(mat._23);
	smat._24 = F32vec4(mat._24);
	smat._31 = F32vec4(mat._31);
	smat._32 = F32vec4(mat._32);
	smat._33 = F32vec4(mat._33);
	smat._34 = F32vec4(mat._34);
	smat._41 = F32vec4(mat._41);
	smat._42 = F32vec4(mat._42);
	smat._43 = F32vec4(mat._43);
	smat._44 = F32vec4(mat._44);
}

void ExpandVector(const SVector &vec,SSimdVector& svec) {
	svec.x = F32vec4(vec.x);
	svec.y = F32vec4(vec.y);
	svec.z = F32vec4(vec.z);
}

void ExpandLight(const SLight &light,SSimdLight& slight) {
	ExpandVector(light.pos,slight.pos);
	ExpandVector(light.dir,slight.dir);

	slight.col.r = F32vec4(light.col.r);
	slight.col.g = F32vec4(light.col.g);
	slight.col.b = F32vec4(light.col.b);
	slight.type = light.type;
}

static void Init(CScene *scene,SSimdLight *NewLight,CPlanet *planet) {
	// calculate effective light params
	int i;
	D3DMATRIX iWorld = planet->GetIWorldMatrix();
	SVector   *lv,*ov;
	SLight OrigLight,TempLight;
	float w,x,y,z;
	SColor mat_col = planet->GetMaterialColor();

	for (i=0;i<scene->GetLightCount();i++) {
		TempLight = OrigLight = scene->GetLight(i);
		switch (OrigLight.type) {
		case POINT_LIGHT:
			lv = &TempLight.pos;
			ov = &OrigLight.pos;
			break;

		case DIR_LIGHT:
			lv = &TempLight.dir;
			ov = &OrigLight.dir;
			break;
		}

		w =	ov->x * iWorld._14 + 
			ov->y * iWorld._24 +
			ov->z * iWorld._34 +
			iWorld._44;

		w = 1 / w;

		x = (ov->x * iWorld._11 + 
			 ov->y * iWorld._21 +
			 ov->z * iWorld._31 +
			 iWorld._41) * w;

		y = (ov->x * iWorld._12 +
	    	 ov->y * iWorld._22 +
			 ov->z * iWorld._32 +
			 iWorld._42) * w;

		z = (ov->x * iWorld._13 + 
			 ov->y * iWorld._23 +
			 ov->z * iWorld._33 +
			 iWorld._43) * w;

		*lv = SVector(x,y,z);

		TempLight.col.r = OrigLight.col.r * mat_col.r;
		TempLight.col.g = OrigLight.col.g * mat_col.g;
		TempLight.col.b = OrigLight.col.b * mat_col.b;


		if (OrigLight.type == DIR_LIGHT)
			*lv = VectorNormalize(*lv);

		ExpandLight(TempLight,NewLight[i]);
	}

}



void CPlanet::SimdPipeline(CScene *scene) {
	//expanded constants.
	F32vec4 _255_ = 255.0f;
	F32vec4 _ZERO_ = 0.0f;

	//expanded ambient colors
	F32vec4 amb_r = m_Ambient.r;
	F32vec4 amb_g = m_Ambient.g;
	F32vec4 amb_b = m_Ambient.b;

	//local vars
	F32vec4 x,y,z,w,we;
	F32vec4 r,g,b;
	F32vec4 mag,dot;
	SSimdVector	dir;

	//expanded transformation matrix.
	SSimdMatrix	x_mat;

	//expanded & transformed lights data
	SSimdLight  lights[MAX_LIGHTS];

	//temp. storage for packed color values (stored as 2 MMX double DWORDS)
	Is32vec2 Color[2];
	DWORD nLights = scene->GetLightCount();


	D3DTLVERTEX *out = m_Out;
	SSimdVertex	*in  = m_SOAData;
	SSimdVertex *nextVtx = in + 1;
	int i,j;

	int nIter = (m_nVerts + 3)/4;
	
	//calculate transformed lights (into object space) & expand the lights.
	Init(scene,lights,this);

	//calculate & expand transform matrix.
	ExpandMatrix(MatrixMult(scene->GetCompositeMatrix(),m_World),x_mat);


	for (i=0;i<nIter;i++) {

		//transformation
		w = in->x * x_mat._14 + 
			in->y * x_mat._24 +
			in->z * x_mat._34 +
			x_mat._44;

		we = rcp_nr(w);

		_mm_prefetch((char *)&nextVtx->x,_MM_HINT_NTA);

		x = (in->x * x_mat._11 + 
			 in->y * x_mat._21 +
			 in->z * x_mat._31 +
			 x_mat._41) * we;

		_mm_prefetch((char *)&nextVtx->z,_MM_HINT_NTA);

		y = (in->x * x_mat._12 + 
			 in->y * x_mat._22 +
			 in->z * x_mat._32 +
			 x_mat._42) * we;

		_mm_prefetch((char *)&nextVtx->ny,_MM_HINT_NTA);

		z = (in->x * x_mat._13 + 
			 in->y * x_mat._23 +
			 in->z * x_mat._33 +
			 x_mat._43) * we;

		
		//set default light values to ambient colors.
		r = amb_r;
		g = amb_g;
		b = amb_b;

		_mm_prefetch((char *)&nextVtx->tex[0].u,_MM_HINT_NTA);

		//calculate each light donation
		for (j=0;j<nLights;j++) {
			SSimdVector dir;
	
			switch (lights[j].type) {
			case POINT_LIGHT:
				//light to vertex direction.
				dir.x = lights[j].pos.x - in->x;
				dir.y = lights[j].pos.y - in->y;
				dir.z = lights[j].pos.z - in->z;
				
				//calc normalization factor (mag = ~1/sqrt(||dir||))
				mag = rsqrt(dir.x*dir.x + dir.y*dir.y + dir.z*dir.z);
				
				//calculate dot product & normalize
				dot = (in->nx * dir.x + 
					   in->ny * dir.y + 
					   in->nz * dir.z ) * mag;
				break;
			case DIR_LIGHT:
				//calculate dot product (already normalized)
				dot = in->nx * lights[j].dir.x + 
					  in->ny * lights[j].dir.y + 
					  in->nz * lights[j].dir.z;
				break;
			default:
				dot = _ZERO_;
			}

			//skip light if all four dots are negative
			if (move_mask(dot) == 0x0f) 
				continue;

			//trim negative dots to zero.
			dot = simd_max(dot,_ZERO_);

			// add light donation to vertex color.
			r += dot * lights[j].col.r;
			g += dot * lights[j].col.g;
			b += dot * lights[j].col.b;
		}

		//scale colors from [0..1] to [0..255]
		r *= _255_;
		g *= _255_;
		b *= _255_;
		
		//trim to 255.0 if above 255.0 
		r = simd_min(r,_255_);
		g = simd_min(g,_255_);
		b = simd_min(b,_255_);

		
		// converting & packing 2 lower colors to ARGB stored in MMX 2 DWORDS format
		Color[0]  = (F32vec4ToIs32vec2(r) << 16) | 
					(F32vec4ToIs32vec2(g) << 8)  | 
					(F32vec4ToIs32vec2(b));	

		// converting & packing 2 upper colors to ARGB stored in MMX 2 DWORDS format
		Color[1]  = (F32vec4ToIs32vec2(Upper2ToLower2(r)) << 16) | 
					(F32vec4ToIs32vec2(Upper2ToLower2(g)) << 8)  | 
					(F32vec4ToIs32vec2(Upper2ToLower2(b)));	
		
		
		//transposing the position coordinates back to AOS.
	    _MM_TRANSPOSE4_PS(x,y,z,we);
		

		//since output buffer is aligned. we can write each four float in one inst.
		*(F32vec4 *)(&out[0].sx) = x;
		*(F32vec4 *)(&out[1].sx) = y;
		*(F32vec4 *)(&out[2].sx) = z;
		*(F32vec4 *)(&out[3].sx) = we;

		//writing the 4 diffuse colors elements
		out[0].color = Color[0][0];
		out[1].color = Color[0][1];
		out[2].color = Color[1][0];
		out[3].color = Color[1][1];

		//writing specular color elemnts
		out[0].specular = 0;
		out[1].specular = 0;
		out[2].specular = 0;
		out[3].specular = 0;

		//textures are in AOS write each tex. pair with one MMX inst.
		*(__m64 *)(&out[0].tu) = *(__m64 *)(&in->tex[0].u);
		*(__m64 *)(&out[1].tu) = *(__m64 *)(&in->tex[1].u);
		*(__m64 *)(&out[2].tu) = *(__m64 *)(&in->tex[2].u);
		*(__m64 *)(&out[3].tu) = *(__m64 *)(&in->tex[3].u);
		

		//increment pointers.
		in = nextVtx;
		nextVtx++;
		out+=4;
	}

	//back from MMX mode to floating point (emms instr).
	_m_empty();
}

